#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include "kernel/kernel.h"
#include "geomlib/geomlib.h"
#include "export/export.h"
#include <string.h>
#include <ctype.h>
#include "import/import.h"
#include "lib/garr.h"
#include "font/font.h"



static FILE *ps_file;

#define OUT(...) fprintf(ps_file, __VA_ARGS__)

/* Margin sizes in mm */
#define HORIZ_MARGIN	10
#define VERT_MARGIN	10


#define GARR_PREFIX(x) psfont_##x
#define GARR_TYPE byte
#define GARR_ZERO
#include "lib/garr.h"

garr psfonts;
static int max_fntid;

static int exclude_fonts;
static int type_eps;
static int page_cnt;


/*
 * Definition of postscript macros (abbreviations)
 */
static void define_macros(void)
{
	if (type_eps) {
		OUT("%%%%BeginProcSet: VrrEPSProcSet\n");
		OUT("/VrrEPSProcSetDict dup 40 dict def load begin\n");
	}
	else
		OUT("%%%%BeginProcSet: VrrProcSet\n");

	OUT("/np {newpath} bind def\n");
	OUT("/lt {lineto} bind def\n");
	OUT("/mt {moveto} bind def\n");
	OUT("/fi {fill} bind def\n");
	OUT("/gs {gsave} bind def\n");
	OUT("/gr {grestore} bind def\n");
	OUT("/sc {scale} bind def\n");
	OUT("/st {stroke} bind def\n");
	OUT("/ct {curveto} bind def\n");
	OUT("/cp {closepath} bind def\n");
	OUT("/a {arc} bind def\n");				// counterclockwise
	OUT("/an {arcn} bind def\n");				// clockwise
	OUT("/fsd {findfont exch scalefont} bind def\n");	// find & scale font
	OUT("/sms {setfont moveto show} bind def\n");		// set current font and draw string
	OUT("/mkf {makefont} bind def\n");			// apply transformation to font
	OUT("/tr {translate} bind def\n");
	OUT("/rot {rotate} bind def\n");
	OUT("/slw {setlinewidth} bind def\n");
	OUT("/sclr {setrgbcolor} bind def\n");
	if (type_eps)
		OUT("end\n");
	OUT("%%%%EndProcSet\n");
}

static void store_fontfile(int fntid)
{
	int type = font_get_type(fntid);
	DBG("Discovered font '%s'\n", font_filename(fntid));

	if (!*psfont_get(psfonts, fntid)) {
		const char *psnm = font_get_ps_name(fntid);

		max_fntid = MAX(max_fntid, fntid);
		if (exclude_fonts) {
			DBG("Excluding font %s\n", psnm);
			OUT("%%%%IncludeResource: font %s\n", psnm);
		}
		else {
			DBG("Dumping it\n");
#ifdef MAGIC_PSNAMES
			OUT("%%%%BeginResource: font F%d\n", fntid);
			OUT("/vrrfontkey {/F%d} def\n", fntid);
#else
			if (psnm)
				OUT("%%%%BeginResource: font %s\n", psnm);
			else
				OUT("%%%%BeginResource: font F%d\n", fntid);
#endif

			if (type == FONT_TYPE_TYPE1) {
				if (font_pfb_to_pfa(fntid, ps_file)<0)
					err("Error converting T1 font %d from PFB to PFA\n", fntid);
			}
			else if (type == FONT_TYPE_TRUETYPE) {
				if (font_tt_to_type42(fntid, psnm, ps_file)<0)
					err("Error converting TT font '%s'\n", font_filename(fntid));
			}
			else
				err("Unsupported font type!\n");

			OUT("%%%%EndResource\n");
		}
		*psfont_get(psfonts, fntid) = 1;
	}
	else
		DBG("Ha, already got it!\n");
}

static void store_fonts(struct go *go)
{
	if (go->type == GOT_TEXT) {
		int fntid = ((struct go_text *) go)->font_id;

		store_fontfile(fntid);
	}
	if (go->type == GOT_TEX_TEXT) {
		uns i;
		struct go_tex_text *t = (struct go_tex_text*)go;

		for (i=0; i < t->process->glyphs_cnt; i++) {
			struct tex_glyph tg = *(struct tex_glyph *) garr_get(t->process->glyphs, i);

			if (!tg.is_rule)
				store_fontfile(tg.font_id);
		}
	}
}

static void file_header(struct obj *obj, int llx, int lly, int urx, int ury)
{
	if (type_eps)
		OUT("%%!PS-Adobe-3.0 EPSF-3.0\n");
	else
		OUT("%%!PS-Adobe-3.0\n");

	OUT("%%%%Title: %s\n", obj_lookup_name(obj));
	OUT("%%%%Creator: VRR  (http://vrr.ucw.cz/)\n");
	time_t tm = time(NULL);
	OUT("%%%%CreationDate: %s", ctime(&tm));
	OUT("%%%%Pages: %d\n", page_cnt);
	OUT("%%%%BoundingBox: %d %d %d %d\n", llx, lly, urx, ury);
	OUT("%%%%DocumentData: Clean7Bit\n");
	OUT("%%%%PageOrder: Ascend\n");

	OUT("%%%%DocumentSuppliedResources: (atend)\n");
	if (exclude_fonts)
		OUT("%%%%DocumentNeedeResources: (atend)\n");
	OUT("%%%%EndComments\n");

	OUT("%%%%BeginProlog\n");
	if (type_eps)
		OUT("save\n");
	define_macros();
	OUT("%%%%EndProlog\n");

	OUT("%%%%BeginSetup\n");

#ifdef MAGIC_PSNAMES
	/* Black magic follows */
	OUT("5 dict begin\n");
	OUT("/vrrend {end} bind def\n");
	OUT("/vrreexec {eexec} bind def\n");
	OUT("/eexec {/FontName where {begin /FontName vrrfontkey def vrrend} if vrreexec} def\n");
	OUT("/end {currentdict /FontName known {/FontName vrrfontkey def} if vrrend} def\n");
#endif

	if (obj->type == OT_DOCUMENT) {
		struct obj_tlo *tlo;
		struct go *go;
		struct obj_doc *doc = (struct obj_doc *) obj;

		WALK_DOC_TLO(doc, tlo)
			WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, go)
				store_fonts(go);
	}
	else if (obj->type == OT_TLO) {
		struct go *go;
		struct obj_tlo *tlo = (struct obj_tlo *) obj;

		WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, go)
			store_fonts(go);
	}
	else
		BUG();

#ifdef MAGIC_PSNAMES
	OUT("vrrend\n");
#else
	for (int i=1; i <= max_fntid; i++)
		if (*psfont_get(psfonts, i)) {
			const char *psnm = font_get_ps_name(i);
			if (psnm)
				OUT("/F%d /%s def\n", i, psnm);
			else
				OUT("/F%d /F%d def\n", i, i);
		}
#endif

	OUT("%%%%EndSetup\n");
}

static int set_color(struct go *go)
{
	uns clr = prop_lookup_uns(go2o(go), string_stroke_color, 0);

	if (clr) {
		uns r, g, b, a;

		uns2rgba_uns(clr, &r, &g, &b, &a);
		OUT("%f %f %f sclr\n", r/255.0, g/255.0, b/255.0);
	}
	/* Always successful. FIXME: should detect undefined color */
	return 1;
}

static int set_fillcolor(struct go *go)
{
	uns clr = prop_lookup_uns(go2o(go), string_fill_color, 0);

	if (clr) {
		uns r, g, b, a;

		uns2rgba_uns(clr, &r, &g, &b, &a);
		OUT("%f %f %f sclr\n", r/255.0, g/255.0, b/255.0);
	}
	/* Always successful. FIXME: should detect undefined color */
	return 1;
}

static void set_linewidth(struct go *go)
{
	real r = prop_lookup_real(go2o(go), string_stroke_width, -1);

	if (r >= 0)
		OUT("%f slw\n", r);
}


static void export_elliptic_arc(struct go_elarc *gea)
{
	if (isinf(gea->geom.b_radius / gea->geom.a_radius))
		return;
	OUT("gs\n");
	OUT("%f %f tr\n%f rot\n", gea->geom.center.x, gea->geom.center.y, gea->geom.rotation * 180 / R_PI);
	OUT("1 %f sc\n", gea->geom.b_radius/gea->geom.a_radius );

	set_linewidth(&gea->go);
	if (set_fillcolor(&gea->go))
		OUT("np 0 0 %f %f %f a fi\n", gea->geom.a_radius, gea->geom.start * 180 / R_PI, (gea->geom.start + gea->geom.dif) * 180 / R_PI);
	if (set_color(&gea->go))
		OUT("np 0 0 %f %f %f a st\n", gea->geom.a_radius, gea->geom.start * 180 / R_PI, (gea->geom.start + gea->geom.dif) * 180 / R_PI);
	OUT("gr\n");
}

static void export_segment(struct go_segment *gs)
{
	OUT("gs\n");
	set_color(&gs->go);
	set_linewidth(&gs->go);
	OUT("np %f %f mt %f %f lt st\n", gs->geom.a.x, gs->geom.a.y, gs->geom.b.x, gs->geom.b.y);
	OUT("gr\n");
}

static void export_point(struct go_point *gp)
{
	OUT("gs\n");
	set_color(&gp->go);
	OUT("np %f %f 0.1 0 360 a fi\n", gp->geom.point.x, gp->geom.point.y);
	OUT("gr\n");
}

#define HEX(b) ((b)<10 ? (b)+'0' : (b)-10+'a')
static void string_encode(const char *s)
{
	const char *o = s;

	OUT("<");
	while (*s) {
		OUT("%c%c", HEX(*s >> 4), HEX(*s & 15));
		s++;
		if (s-o > 32 && *s)
			OUT("\n");
	}
	OUT(">");
}

static void export_text(struct go_text *t)
{
	struct geom_transform tr;
	struct geom_point pt;

	OUT("gs\n");
	set_color(&t->go);
	string_encode(t->text->text);
	go_text_get_transformation(t, &tr);
	go_text_get_visualisation_point(t, &pt);

	if (t->transformation->flags & GEOM_TRANSFORM_IDENTITY)
#ifdef MAGIC_PSNAMES
		OUT(" %f %f %f /F%d fsd sms\n", pt.x, pt.y, t->font_size, t->font_id);
#else
		OUT(" %f %f %f F%d fsd sms\n", pt.x, pt.y, t->font_size, t->font_id);
#endif
	else {
#ifdef MAGIC_PSNAMES
		OUT(" %f %f %f /F%d fsd [%f %f %f %f 0 0] mkf sms\n", pt.x, pt.y,
			t->font_size, t->font_id, tr.coef[0][0], tr.coef[1][0], tr.coef[0][1], tr.coef[1][1]);
#else
		OUT(" %f %f %f F%d fsd [%f %f %f %f 0 0] mkf sms\n", pt.x, pt.y,
			t->font_size, t->font_id, tr.coef[0][0], tr.coef[1][0], tr.coef[0][1], tr.coef[1][1]);
#endif
	OUT("gr\n");
	}
}

static void export_tex_text(struct go_tex_text *t)
{
	uns i;
	struct geom_transform tr;
	struct geom_point pt;

	go_tex_text_get_transformation(t, &tr);
	go_tex_text_get_visualisation_point(t, &pt);

	OUT("gs\n");
	set_color(&t->go);
	OUT("[%f %f %f %f %f %f] concat\n", tr.coef[0][0], tr.coef[1][0], tr.coef[0][1], tr.coef[1][1], pt.x, pt.y);

	for (i=0; i < t->process->glyphs_cnt; i++) {
		struct tex_glyph tg = *(struct tex_glyph *) garr_get(t->process->glyphs, i);

		if (tg.is_rule) {
			OUT("np %f %f mt %f %f lt %f %f lt %f %f lt cp fi\n",
					tg.xpos+tg.bbox.ll.x, tg.ypos+tg.bbox.ll.y,
					tg.xpos+tg.bbox.ur.x, tg.ypos+tg.bbox.ll.y,
					tg.xpos+tg.bbox.ur.x, tg.ypos+tg.bbox.ur.y,
					tg.xpos+tg.bbox.ll.x, tg.ypos+tg.bbox.ur.y);
		}
		else {
#ifdef MAGIC_PSNAMES
			OUT("<%c%c> %f %f %f /F%d fsd sms\n", HEX((tg.code & 0xf0) >> 4), HEX(tg.code & 0x0f),
					tg.xpos, tg.ypos, tg.font_size, tg.font_id);
#else
			OUT("<%c%c> %f %f %f F%d fsd sms\n", HEX((tg.code & 0xf0) >> 4), HEX(tg.code & 0x0f),
					tg.xpos, tg.ypos, tg.font_size, tg.font_id);
#endif
		}
	}
	OUT("gr\n");
}

static void export_bezier(struct go_bezier *gobj)
{
	struct geom_bezier gb = gobj->geom;
	struct geom_approx approx;
	struct geom_approx_item *item;
	struct geom_bezier *b;

	OUT("gs\n");
	set_linewidth(&gobj->go);

	geom_approx_init(&approx);
	if (geom_approx_bezier(&approx, &gb, GEOM_APPROX_SHAPE | GEOM_APPROX_CUBIC_NONRATIONAL) < 0) {
		geom_approx_destroy(&approx);
		trans_fail("Error occured during bezier curve approximation.\n");
	}
	if (set_fillcolor(&gobj->go)) {
		OUT("np %f %f mt ", gb.pt[0].x, gb.pt[0].y);
		for (uns i = 0; i < approx.count; i++) {
			item = approx.items + i;
			b = &(item->data.bezier.bezier);
			OUT("%f %f %f %f %f %f ct\n", b->pt[1].x, b->pt[1].y,
					b->pt[2].x, b->pt[2].y, b->pt[3].x, b->pt[3].y);
		}
		OUT("fi\n");
	}
	if (set_color(&gobj->go)) {
		OUT("np %f %f mt ", gb.pt[0].x, gb.pt[0].y);
		for (uns i = 0; i < approx.count; i++) {
			item = approx.items + i;
			b = &(item->data.bezier.bezier);
			OUT("%f %f %f %f %f %f ct\n", b->pt[1].x, b->pt[1].y,
					b->pt[2].x, b->pt[2].y, b->pt[3].x, b->pt[3].y);
		}
		OUT("st\n");
	}
	geom_approx_destroy(&approx);
	OUT("gr\n");
}

static void export_decorator(struct go_decorator_point *dp)
{
	/* FIXME! */
}

static void export_tlo(struct obj_tlo *tlo)
{
	struct go *gobj;

	WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, gobj)
	{
		switch(gobj->type)
		{
			case GOT_ELARC:
				export_elliptic_arc((struct go_elarc *) gobj);
				break;
			case GOT_SEGMENT:
				export_segment((struct go_segment *) gobj);
				break;
			case GOT_POINT:
				export_point((struct go_point *) gobj);
				break;
			case GOT_TEXT:
				export_text((struct go_text *) gobj);
				break;
			case GOT_TEX_TEXT:
				export_tex_text((struct go_tex_text *) gobj);
				break;
			case GOT_BEZIER:
				export_bezier((struct go_bezier *) gobj);
				break;
			case GOT_DECORATOR_POINT:
				export_decorator((struct go_decorator_point *) gobj);
				break;
			case GOT_GROUP:
				/* FIXME: implement these */
				break;
			case GOT_PARAMETRIC_POINT:
			case GOT_INTERSECTION:
				/* do nothing */
				break;
			default:
				BUG();
		}
	}
}

static void page_header(struct obj_tlo *tlo, int ord, real dx1, real dy1, real mul, real dx2, real dy2)
{
	const char *name = obj_lookup_name((struct obj *) tlo);
	int i = 0;
	char tmp[strlen(name) + 1];

	while (name[i]) {
		if (!isalnum(name[i]))
			tmp[i] = '_';
		else
			tmp[i] = name[i];
		i++;
	}
	tmp[i] = 0;

	OUT("\n%%%%Page: %s %d\n%%%%BeginPageSetup\n", tmp, ord);
	if (type_eps)
		OUT("VrrEPSProcSetDict begin\n");
	else
		OUT("save\n");

	/* Set the default linewidth to 0.2 mm. */
	OUT("0.2 slw\n");

	OUT("%f %f tr %f %f sc %f %f tr\n", dx1, dy1, mul, mul, dx2, dy2);

	OUT("%%%%EndPageSetup\n");
}

static void page_footer(void)
{
	if (!type_eps)
		OUT("restore ");
	OUT("showpage\n\n");
}

static void export_page(struct obj_tlo *tlo, int ord, real dx, real dy, int fit_to_page, real page_width, real page_height)
{
	struct geom_rectangle bbox;
	real mul;

	bbox = *tlo_get_expand_bbox(tlo);
	bbox.ur.x *= MM2POINT;
	bbox.ur.y *= MM2POINT;
	bbox.ll.x *= MM2POINT;
	bbox.ll.y *= MM2POINT;

	if (fit_to_page) {
		dx = - (bbox.ur.x + bbox.ll.x)/2;
		dy = - (bbox.ur.y + bbox.ll.y)/2;
		real w = bbox.ur.x - bbox.ll.x + HORIZ_MARGIN;
		real h = bbox.ur.y - bbox.ll.y + VERT_MARGIN;

		if (w/page_width > h/page_height)
			mul = MM2POINT * page_width / w;
		else
			mul = MM2POINT * page_height / h;

		page_header(tlo, ord, dx, dy, mul, page_width/2/mul, page_height/2/mul);
	}
	else
		page_header(tlo, ord, dx, dy, MM2POINT, 0, 0);

	export_tlo(tlo);
	page_footer();
}

static void file_footer(void)
{
	int i, len;

	OUT("%%%%Trailer\n");
	if (type_eps)
		OUT("%%%%DocumentSuppliedResources: procset VrrEPSProcSet\n");
	else
		OUT("%%%%DocumentSuppliedResources: procset VrrProcSet\n");

	int first = 1;
	for (i = 1, len = 0; i <= max_fntid; i++) {
		if (*psfont_get(psfonts, i)) {
			if (len > 70) {
				OUT("\n%%%%+ font");
				len = 0;
			}
#ifdef MAGIC_PSNAMES
			len += OUT(" F%d", i);
#else
			const char *nm = font_get_ps_name(i);
			if (first) {
				if (exclude_fonts)
					OUT("%%%%DocumentNeededResources: font");
				else
					OUT("%%%%+ font");
				first = 0;
			}
			if (nm)
				len += OUT(" %s", nm);
			else
				len += OUT(" F%d", i);
#endif
			}
	}
	if (type_eps) {
#ifdef MAGIC_PSNAMES
		OUT("\nvrrend restore");
#else
		OUT("\nend restore");
#endif
	}
	OUT("\n%%%%EOF\n");
}

void export_ps(struct obj_doc *doc, FILE *f, real page_width, real page_height, int fit_to_page, int without_fonts)
{
	struct obj_tlo *tlo;
	struct geom_rectangle bbox;

	ps_file = f;
	psfonts = psfont_new(256);
	max_fntid = 0;
	type_eps = 0;
	page_cnt = 0;
	exclude_fonts = without_fonts;

	geom_rectangle_set_empty(&bbox);
	WALK_DOC_TLO(doc, tlo)
	{
		expand_ref(tlo);
		geom_rectangle_merge(tlo_get_expand_bbox(tlo), &bbox, &bbox);
		page_cnt++;
	}

	bbox.ll.x *= MM2POINT;
	bbox.ll.y *= MM2POINT;
	bbox.ur.x *= MM2POINT;
	bbox.ur.y *= MM2POINT;
	page_width *= MM2POINT;
	page_height *= MM2POINT;

	if (!page_width)
		page_width = bbox.ur.x - bbox.ll.x + 1;
	if (!page_height)
		page_height = bbox.ur.y - bbox.ur.y + 1;

	/* Center the image */
	real dx = page_width/2 - (bbox.ur.x + bbox.ll.x)/2;
	real dy = page_height/2 - (bbox.ur.y + bbox.ll.y)/2;

	bbox.ll.x += dx;
	bbox.ll.y += dy;
	bbox.ur.x += dx;
	bbox.ur.y += dy;

	if (fit_to_page)
		file_header((struct obj *) doc, HORIZ_MARGIN * MM2POINT, VERT_MARGIN * MM2POINT,
				page_width - HORIZ_MARGIN * MM2POINT, page_height - VERT_MARGIN * MM2POINT);
	else
		file_header((struct obj *) doc, bbox.ll.x, bbox.ll.y, bbox.ur.x, bbox.ur.y);

	/* Export all document pages */
	page_cnt = 0;
	WALK_DOC_TLO(doc, tlo) {
		page_cnt++;
		export_page(tlo, page_cnt, dx, dy, fit_to_page, page_width, page_height);
		expand_unref(tlo);
	}

	file_footer();
	psfont_destroy(psfonts);
}

void export_eps(struct obj_tlo *tlo, FILE *f, int without_fonts)
{
	const struct geom_rectangle *bbox;

	ps_file = f;
	psfonts = psfont_new(256);
	max_fntid = 0;
	type_eps = 1;
	exclude_fonts = without_fonts;

	expand_ref(tlo);
	bbox = tlo_get_expand_bbox(tlo);

	real h = (bbox->ur.y - bbox->ll.y) * MM2POINT;
	real w = (bbox->ur.x - bbox->ll.x) * MM2POINT;

	page_cnt = 1;
	file_header((struct obj *) tlo, 0, 0, w, h);

	export_page(tlo, 1, -bbox->ll.x * MM2POINT, -bbox->ll.y * MM2POINT, 0, 0, 0);
	expand_unref(tlo);

	file_footer();
	psfont_destroy(psfonts);
}
