#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <stdio.h>
#include <libxml/xmlwriter.h>
#include "geomlib/geomlib.h"
#include "kernel/kernel.h"
#include "font/font.h"
#include "export/export.h"
#include "import/import.h"

static xmlTextWriterPtr ptr;

#define START_DTD_ELEMENT if(xmlTextWriterWriteDTD(ptr, "svg", "-//W3C//DTD SVG 1.1//EN", "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd", NULL) < 0) { DBG("Error while exporting at line %d\n", __LINE__); return; }
#define START_ELEMENT(x) if(xmlTextWriterStartElement(ptr, x) < 0) { DBG("Error while exporting at line %d\n", __LINE__); return; }
#define END_ELEMENT if(xmlTextWriterEndElement(ptr) < 0) { DBG("Error while exporting at line %d\n", __LINE__); return; }
#define ADD_STRING(str) if(xmlTextWriterWriteString(ptr, str) < 0) { DBG("Error while exporting at line %d\n", __LINE__); return; }
#define ADD_ATTRIBUTE(x, y, z) if(xmlTextWriterWriteFormatAttribute(ptr, x, y, z) < 0) { DBG("Error at line %d\n", __LINE__); return; }
#define ADD_OPTIONAL_ATTRIBUTE(x, y, z) if(xmlTextWriterWriteFormatAttribute(ptr, x, y, z) < 0) { DBG("Error at line %d\n", __LINE__); return; }
#define TEST_RESULT(x) if(x < 0) { DBG("Error at line %d\n", __LINE__); return; }

#define TRANSFORMX(x) ((x - xoffset) * MM2POINT)
#define TRANSFORMY(y) ((height - y + yoffset) * MM2POINT)
#define TRANSFORMLEN(x) (x * MM2POINT)

#define OPACITY_FROM_ALPHA(x) (((double)x)/255.0)

static char tmp[1000];
static int *res;
static real xoffset, yoffset, width, height;

static void export_style (struct go * go)
{
	uns stroke_color = prop_lookup_uns (go2o(go), string_stroke_color, 0);
	uns fill_color = prop_lookup_uns (go2o(go), string_fill_color, 0);
	uns sr, sg, sb, sa, fr, fg, fb, fa;
	uns2rgba_uns (stroke_color, &sr, &sg, &sb, &sa);
	uns2rgba_uns (fill_color, &fr, &fg, &fb, &fa);

	sprintf (tmp, "opacity:1.0;stroke:rgb(%u,%u,%u);stroke-opacity:%.2f;"
		"fill:rgb(%u,%u,%u);fill-opacity:%.2f;stroke-width:%.2f",
		sr, sg, sb, OPACITY_FROM_ALPHA(sa), fr, fg, fb, OPACITY_FROM_ALPHA(fa),
		TRANSFORMLEN( (double) prop_lookup_real (go2o(go), string_stroke_width, 1.0)));
	
	ADD_ATTRIBUTE ("style", "%s", tmp);	
}

static void export_svg_bezier_path (struct go *go)
{
	struct geom_approx approx;
	struct geom_approx_item *item;
	struct geom_bezier *b;
	char *str;

	geom_approx_init (&approx);

	DBG("Geom approximation\n");
	if (go_approx (&approx, go, GEOM_APPROX_SHAPE | GEOM_APPROX_CUBIC_NONRATIONAL))
	{
	   	str = alloca (approx.count * 72 + 128);
		sprintf(str, "M%f %f",
				TRANSFORMX(approx.items[0].data.bezier.bezier.pt[0].x),
				TRANSFORMY(approx.items[0].data.bezier.bezier.pt[0].y));
		for (uns i = 0; i < approx.count; i++)
		{
			item = approx.items + i;
			b = &(item->data.bezier.bezier);
			sprintf(tmp, " C%10f,%10f %10f,%10f %10f,%10f",
					TRANSFORMX(b->pt[1].x), TRANSFORMY(b->pt[1].y),
					TRANSFORMX(b->pt[2].x), TRANSFORMY(b->pt[2].y),
					TRANSFORMX(b->pt[3].x), TRANSFORMY(b->pt[3].y));
			DBG("data: %s\n", tmp);
			str = strcat (str, tmp);
		}
		START_ELEMENT("path");
		ADD_ATTRIBUTE("d", "%s", str);
		export_style(go);
		END_ELEMENT;
	}
	geom_approx_destroy(&approx);
}

static void export_point (struct go_point *go)
{
	START_ELEMENT("circle");
	ADD_ATTRIBUTE("cx", "%f", TRANSFORMX(go->geom.point.x));
	ADD_ATTRIBUTE("cy", "%f", TRANSFORMY(go->geom.point.y));
	ADD_ATTRIBUTE("r", "%d", 1);
	export_style(&go->go);
	END_ELEMENT;
}

static void export_line (real x1, real y1, real x2, real y2, struct go *go)
{
	START_ELEMENT("line");
	ADD_ATTRIBUTE("x1", "%f", TRANSFORMX(x1));
	ADD_ATTRIBUTE("y1", "%f", TRANSFORMY(y1));
	ADD_ATTRIBUTE("x2", "%f", TRANSFORMX(x2));
	ADD_ATTRIBUTE("y2", "%f", TRANSFORMY(y2));
	if (go)
		export_style(go);
	END_ELEMENT;
}

static void export_segment (struct go_segment *go)
{
	export_line (go->geom.a.x, go->geom.a.y, go->geom.b.x, go->geom.b.y, (struct go *) go);
}

static void export_elliptic_arc (struct go *go)
{
	DBG("Exporting elliptic arc\n");
	export_svg_bezier_path (go);
}

static void export_rectangle (real x, real y, real width, real height)
{
	//char tmp[128];

	START_ELEMENT("rect");
	ADD_ATTRIBUTE("x", "%f", TRANSFORMX(x));
	ADD_ATTRIBUTE("y", "%f", TRANSFORMY(y));
	ADD_ATTRIBUTE("width", "%f", TRANSFORMLEN(width));
	ADD_ATTRIBUTE("height", "%f", TRANSFORMLEN(height));
	//sprintf(tmp, "rgb(%d,%d,%d)", r, g, b);
	//ADD_OPTIONAL_ATTRIBUTE("fill", "%s", tmp);
	//ADD_OPTIONAL_ATTRIBUTE("stroke", "%s", tmp);
	//FIXME text zatim nema barvu
	//a takhle se tam stejne vyexportovala barva predchoziho objektu
	END_ELEMENT;
}

static void export_text (struct go_text *t)
{
	const char *fontname;
	char transf[1024];
	struct geom_transform *tr;

	DBG("Exporting text\n");

	START_ELEMENT("text");
	ADD_ATTRIBUTE("x", "%f", TRANSFORMX(t->geom.point.x));
	ADD_ATTRIBUTE("y", "%f", TRANSFORMY(t->geom.point.y));
	fontname = font_get_ps_name (t->font_id);
	DBG("fontname: %s, fontsize: %f\n", fontname, t->font_size);
	if (fontname)
		ADD_ATTRIBUTE("font-family", "%s", fontname);
	ADD_ATTRIBUTE("font-size", "%d", t->font_size);
	tr = t->transformation;
	sprintf(transf, "%f %f %f %f %f %f",
				tr->coef[0][0], tr->coef[1][0], tr->coef[0][1],
				tr->coef[1][1], tr->coef[0][2], tr->coef[1][2]);
	ADD_ATTRIBUTE("transform", "matrix(%s)", transf);
	ADD_STRING(t->text->text);
	END_ELEMENT;
}

static void export_tex_text (struct go_tex_text *t)
{
	uns i;
	char transf[1024];

	for (i = 0; i < t->process->glyphs_cnt; i++)
	{
		struct tex_glyph *glyph = (struct tex_glyph*) garr_get (t->process->glyphs, i);
		struct geom_transform *tr = t->transformation;
		real x, y;
		struct geom_point pos;

		pos.x = glyph->xpos;
		pos.y = glyph->ypos;
			
		if (geom_point_transform(&pos, tr, &pos) == GEOM_ERR_NUMERIC)
			continue;	// numeric error
		x = t->geom.point.x + pos.x;
		y = t->geom.point.y + pos.y;
		sprintf(transf, "%f %f %f %f %f %f",
				tr->coef[0][0], tr->coef[1][0], tr->coef[0][1],
				tr->coef[1][1], tr->coef[0][2], tr->coef[1][2]);

		if (glyph->is_rule)
		{
			struct geom_point p1, p2;

			p1 = glyph->bbox.ll;
			p2 = glyph->bbox.ur;
			DBG("hbox: %f, %f - %f, %f\n", glyph->bbox.ll.x, glyph->bbox.ll.y, glyph->bbox.ur.x, glyph->bbox.ur.y);

			geom_point_transform (&p1, tr, &p1);
			geom_point_transform (&p2, tr, &p2);

			export_rectangle (x + p1.x, y + p1.y, p2.x - p1.x, p2.y - p1.y);
		}
		else	// character
		{
			char str[2];
			struct geom_point p = glyph->bbox.ll;
			struct font_info fi;

			START_ELEMENT("text");
			ADD_ATTRIBUTE("x", "%f", TRANSFORMX(x + p.x));
			ADD_ATTRIBUTE("y", "%f", TRANSFORMY(y + p.y));
			
			if (font_info (glyph->font_id, &fi))
				ADD_ATTRIBUTE("font-family", "%s", fi.family);
			ADD_ATTRIBUTE("font-size", "%d", glyph->font_size);
			ADD_ATTRIBUTE("transform", "matrix(%s)", transf);
			str[1] = 0;
			str[0] = glyph->code;
			ADD_STRING(str);
			END_ELEMENT;
		}
	}
}

static void export_bezier (struct go *go)
{
	DBG("Exporting bezier\n");
	export_svg_bezier_path (go);
}

#if 0
static void export_path (struct go_path *go UNUSED)
{
}
#endif

static void export_objects (struct obj_tlo *tlo)
{
	struct go *gobj;

	WALK_GO_RECURSIVE_WITHOUT_ROOT(&tlo->tlg->group.go, gobj)
	{
		switch (gobj->type)
		{
			case GOT_POINT:
				export_point ((struct go_point *) gobj);
				break;
			case GOT_SEGMENT:
				export_segment ((struct go_segment *) gobj);
				break;
			case GOT_ELARC:
				export_elliptic_arc (gobj);
				break;
			case GOT_TEXT:
				export_text ((struct go_text*)gobj);
				break;
			case GOT_TEX_TEXT:
				export_tex_text ((struct go_tex_text*)gobj);
				break;
			case GOT_BEZIER:
				export_bezier (gobj);
				break;
// FIXME: Not implemented yet.
#if 0
			case GOT_PATH:
				export_path ((struct go_path *) gobj);
				break;
			case GOT_PARAMETRIC_POINT:
			case GOT_INTERSECTION:
				break; 	// do not export
			case GOT_INSTANCE:
			case GOT_SCRIPT:
			case GOT_GROUP:
				break;	//NIY
#endif
			default:
				DBG("Bad go type!!!\n");
		}
	}
}

static void export_tlo(struct obj_tlo *tlo, const char *filename)
{
	ptr = xmlNewTextWriterFilename(filename, 0);
	if(!ptr)
	{
		DBG("Couldn't create text writer.\n");
		*res = 0;
		return;
	}
	TEST_RESULT(xmlTextWriterStartDocument(ptr, NULL, NULL, NULL));
	TEST_RESULT(xmlTextWriterSetIndent(ptr, 1));
	TEST_RESULT(xmlTextWriterSetIndentString(ptr, "\t"));
	START_DTD_ELEMENT;
	START_ELEMENT("svg");
	ADD_ATTRIBUTE("width", "%d", (int) (width * MM2POINT));
	ADD_ATTRIBUTE("height", "%d", (int) (height * MM2POINT));
	ADD_ATTRIBUTE("version", "%s", "1.1");
	ADD_ATTRIBUTE("xmlns", "%s", "http://www.w3.org/2000/svg");
	START_ELEMENT("desc");
	ADD_STRING("SVG file exported by VRR");
	END_ELEMENT;
	export_objects(tlo);
	END_ELEMENT;
	TEST_RESULT(xmlTextWriterEndDocument(ptr));
	TEST_RESULT(xmlTextWriterFlush(ptr));
	xmlFreeTextWriter(ptr);
}

void export_svg (struct obj_tlo *tlo, const char *filename, int *result)
{
	const struct geom_rectangle *bbox;

	res = result;
	expand_ref(tlo);
	bbox = tlo_get_expand_bbox (tlo);

	width = bbox->ur.x - bbox->ll.x;
	height = bbox->ur.y - bbox->ll.y;

	xoffset = bbox->ll.x;
	yoffset = bbox->ll.y;

	export_tlo(tlo, filename);
	expand_unref(tlo);
	*result = 0;
}
