#define GEOM_SOURCE
#define UNSTABLE
#define NO_MESSAGES

#include "geomlib/base.h"
#include "geomlib/bezier.h"
#include "geomlib/path.h"
#include "geomlib/approx.h"


#define GEOM_APPROX_INITIAL_SIZE  16
#define GEOM_APPROX_MAX_DEPTH     8


/* geom_approx_init */

void geom_approx_init(struct geom_approx *approx)
{
	DBG("Initializing geom_approx structure at %p.\n", approx);

	PASSERT(approx);

	approx->size = GEOM_APPROX_INITIAL_SIZE;
	approx->count = 0;
	approx->right = 0;
	approx->items = (struct geom_approx_item *)xmalloc(approx->size * sizeof(struct geom_approx_item));
}


/* geom_approx_destroy */

void geom_approx_destroy(struct geom_approx *approx)
{
	DBG("Destroying geom_approx structure at %p.\n", approx);

	PASSERT(approx);

	xfree(approx->items);
}


/* geom_approx_realloc */

static void geom_approx_realloc(struct geom_approx *approx)
{
	DBG("Reallocating approximation buffer.\n");

	approx->size *= 2;
	approx->items = (struct geom_approx_item *)xrealloc(approx->items, approx->size * sizeof(struct geom_approx_item));
}


/* geom_approx_alloc */

static inline struct geom_approx_item *geom_approx_alloc(struct geom_approx *approx)
{
	if (approx->size == approx->count)
		geom_approx_realloc(approx);

	return &approx->items[approx->count++];
}


/* geom_approx_clear */

void geom_approx_clear(struct geom_approx *approx)
{
	PASSERT(approx);

	approx->count = 0;
	approx->right = 0;
}


/* geom_approx_add_bezier */

int geom_approx_add_bezier(struct geom_approx *approx, struct geom_bezier *bezier, uns flags, double len)
{
	struct geom_approx_item *item;

#if 0
	GEOM_APPROX_ASSERT(approx);
	GEOM_BEZIER_ASSERT(bezier);
	BASSERT(isfinite(len));
	switch (bezier->degree) {
		case GEOM_BEZIER_LINEAR:
			if (bezier->flags & GEOM_BEZIER_NONRATIONAL)
				BASSERT(flags == GEOM_APPROX_LINEAR_NONRATIONAL);
			else
				BASSERT(flags == GEOM_APPROX_LINEAR_RATIONAL);
			break;

		case GEOM_BEZIER_QUADRATIC:
			if (bezier->flags & GEOM_BEZIER_NONRATIONAL)
				BASSERT(flags == GEOM_APPROX_QUADRATIC_NONRATIONAL);
			else
				BASSERT(flags == GEOM_APPROX_QUADRATIC_RATIONAL);
			break;

		case GEOM_BEZIER_CUBIC:
			if (bezier->flags & GEOM_BEZIER_NONRATIONAL)
				BASSERT(flags == GEOM_APPROX_CUBIC_NONRATIONAL);
			else
				BASSERT(flags == GEOM_APPROX_CUBIC_RATIONAL);
			break;

		default:
			BUG();
	}
#endif

	item = geom_approx_alloc(approx);
	item->type = flags;
	item->left = approx->right;
	approx->right = item->right = item->left + len;
	GEOM_INSTANCE_CREATE2(&item->data.bezier.bezier, bezier, geom_bezier_copy, bezier);

	if (!isfinite(approx->right)) {
		DBG("Parametrization overflow.\n");
		return -1;
	}

	return 0;
}


/* geom_approx_nonrational_enough */

static int geom_approx_nonrational_enough(struct geom_bezier *bezier)
{
	int i;
	double d;

	d = 0;
	for (i = 0; i < BEZIER_DEGREE(bezier); i++) {
		d += SQR(bezier->pt[i].w - bezier->pt[i + 1].w);
	}

	return !isfinite(d) || d < 0.1;
}


/* geom_approx_set_nonrational */

static void geom_approx_set_nonrational(struct geom_bezier *bezier)
{
	uns i;

	INSTANCE_FLAGS(bezier) |= GEOM_BEZIER_NONRATIONAL;

	for (i = 0; i <= BEZIER_DEGREE(bezier); i++)
		bezier->pt[i].w = 1;
}


/* geom_approx_bezier */

int geom_approx_bezier(struct geom_approx *approx, struct geom_bezier *bezier, uns flags)
{
	uns i;
	struct geom_bezier b1;
	double len;

	PASSERT(approx);
	ITEM_ASSERT(bezier, bezier);

	GEOM_INSTANCE_CREATE2(&b1, bezier, geom_bezier_copy, bezier); // FIXME: no destructor is called...

	if (flags & GEOM_APPROX_TYPE == GEOM_APPROX_LENGTH) {
		len = geom_bezier_alength(&b1);
		if (!isfinite(len))
			return -1;
	}
	else
		len = 1;

	switch (BEZIER_DEGREE(bezier)) {
		case GEOM_BEZIER_LINEAR:

			/* nonrational linear curve */
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL) {

				if (flags & GEOM_APPROX_LINEAR) {
					DBG("Inserting nonrational linear curve.\n");
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_LINEAR_NONRATIONAL, len);
				}

				else if (flags & GEOM_APPROX_QUADRATIC) {
					DBG("Converting nonrational linear curve to nonrational quadratic curve.\n");
					geom_bezier_elevate(&b1, GEOM_BEZIER_QUADRATIC, &b1);
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_QUADRATIC_NONRATIONAL, len);
				}

				else if (flags & GEOM_APPROX_CUBIC) {
					DBG("Converting nonrational linear curve to nonrational cubic curve.\n");
					geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_NONRATIONAL, len);
				}

				else {
					DBG("Could not approximate nonrational linear curve.\n");
					return -1;
				}
			}

			/* rational linear curve */
			else {

				if (flags & GEOM_APPROX_TYPE == GEOM_APPROX_TIME) {

					if (flags & GEOM_APPROX_LINEAR_RATIONAL) {
						DBG("Inserting rational linear curve.\n");
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_LINEAR_RATIONAL, len);
					}

					else if (flags & GEOM_APPROX_QUADRATIC_RATIONAL) {
						DBG("Converting rational linear curve to rational quadratic curve.\n");
						geom_bezier_elevate(&b1, GEOM_BEZIER_QUADRATIC, &b1);
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_QUADRATIC_RATIONAL, len);
					}

					else if (flags & GEOM_APPROX_CUBIC_RATIONAL) {
						DBG("Converting rational linear curve to rational cubic curve.\n");
						geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_RATIONAL, len);
					}

					else if (flags & GEOM_APPROX_BEZIER) {
						if (flags & GEOM_APPROX_CUBIC_NONRATIONAL) {
							DBG("Approximating rational linear curve with nonrational cubic curves.\n");
							i = GEOM_BEZIER_CUBIC;
						}
						else if (flags & GEOM_APPROX_QUADRATIC_NONRATIONAL) {
							DBG("Approximating rational linear curve with nonrational quadratic curves.\n");
							i = GEOM_BEZIER_QUADRATIC;
						}
						else {
							DBG("Approximating rational linear curve with nonrational linear curves.\n");
							i = GEOM_BEZIER_LINEAR;
						}
						geom_bezier_elevate(&b1, i, &b1);
						GEOM_BEZIER_SUBDIVIDE_BEGIN(sub_rl_nc, &b1, part, deep, left, right, GEOM_APPROX_MAX_DEPTH, geom_approx_nonrational_enough(part))
						{
							geom_approx_set_nonrational(part);
							if (geom_approx_add_bezier(approx, part, GEOM_APPROX_CUBIC_NONRATIONAL, right - left) < 0)
								return -1;
						}
						GEOM_BEZIER_SUBDIVIDE_END(sub_rl_nc);
						return 0;
					}

					else {
						DBG("Cound not approximate rational linear curve.\n");
						return -1;
					}
				}

				else {

					INSTANCE_FLAGS(&b1) |= GEOM_BEZIER_NONRATIONAL;
					b1.pt[0].w = 1;
					b1.pt[1].w = 1;

					if (flags & GEOM_APPROX_LINEAR) {
						DBG("Converting rational linear curve to nonrational linear curve.\n");
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_LINEAR_NONRATIONAL, len);
					}

					else if (flags & GEOM_APPROX_QUADRATIC) {
						DBG("Converting rational linear curve to nonrational quadratic curve.\n");
						geom_bezier_elevate(&b1, GEOM_BEZIER_QUADRATIC, &b1);
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_QUADRATIC_NONRATIONAL, len);
					}

					else if (flags & GEOM_APPROX_CUBIC) {
						DBG("Converting rational linear curve to nonrational cubic curve.\n");
						geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
						return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_NONRATIONAL, len);
					}

					else {
						DBG("Cound not approximate rational linear curve.\n");
						return -1;
					}
				}
			}

		case GEOM_BEZIER_QUADRATIC:

			/* nonrational quadratic curve */
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL) {

				if (flags & GEOM_APPROX_QUADRATIC) {
					DBG("Inserting nonrational quadratic curve.\n");
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_QUADRATIC_NONRATIONAL, len);
				}

				else if (flags & GEOM_APPROX_CUBIC) {
					DBG("Converting nonrational quadratic curve to nonrational cubic curve.\n");
					geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_NONRATIONAL, len);
				}

				else {
					DBG("Could not approximate nonrational quadratic curve.\n");
					return -1;
				}
			}

			/* rational quadratic curve */
			else {

				if (flags & GEOM_APPROX_QUADRATIC_RATIONAL) {
					DBG("Inserting rational quadratic curve.\n");
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_QUADRATIC_RATIONAL, len);
				}

				else if (flags & GEOM_APPROX_CUBIC_RATIONAL) {
					DBG("Converting rational quadratic curve to rational cubic curve.\n");
					geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_RATIONAL, len);
				}

				else if (flags & GEOM_APPROX_CUBIC_NONRATIONAL) {
					DBG("Approximating rational quadratic curve with nonrational cubic curves.\n");
					geom_bezier_elevate(&b1, GEOM_BEZIER_CUBIC, &b1);
					GEOM_BEZIER_SUBDIVIDE_BEGIN(sub_rq_nc, &b1, part, deep, left, right, GEOM_APPROX_MAX_DEPTH, geom_approx_nonrational_enough(part))
					{
						geom_approx_set_nonrational(part);
						if (geom_approx_add_bezier(approx, part, GEOM_APPROX_CUBIC_NONRATIONAL, right - left) < 0)
							return -1;
					}
					GEOM_BEZIER_SUBDIVIDE_END(sub_rq_nc);
					return 0;
				}

				else {
					DBG("Could not approximate rational quadratic curve.\n");
					return -1;
				}
			}

		case GEOM_BEZIER_CUBIC:

			/* nonrational cubic curve */
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL) {

				if (flags & GEOM_APPROX_CUBIC) {
					DBG("Inserting nonrational cubic curve.\n");
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_NONRATIONAL, len);
				}

				else {
					DBG("Could not approximate nonrational cubic curve.\n");
					return -1;
				}
			}

			/* rational cubic curve */
			else {

				if (flags & GEOM_APPROX_CUBIC_RATIONAL) {
					DBG("Inserting rational cubic curve.\n");
					return geom_approx_add_bezier(approx, &b1, GEOM_APPROX_CUBIC_RATIONAL, len);
				}

				else if (flags & GEOM_APPROX_CUBIC_NONRATIONAL) {
					DBG("Approximating rational cubic curve with nonrational cubic curves.\n");
					GEOM_BEZIER_SUBDIVIDE_BEGIN(sub_rc_nc, &b1, part, deep, left, right, GEOM_APPROX_MAX_DEPTH, geom_approx_nonrational_enough(part))
					{
						geom_approx_set_nonrational(part);
						if (geom_approx_add_bezier(approx, part, GEOM_APPROX_CUBIC_NONRATIONAL, right - left) < 0)
							return -1;
					}
					GEOM_BEZIER_SUBDIVIDE_END(sub_rc_nc);
					return 0;
				}

				else {
					DBG("Could not approximate rational cubic curve.\n");
					return -1;
				}
			}

			return 0;

		default:
			BUG();
	}
}


/* geom_approx_path */

int geom_approx_path(struct geom_approx *approx, const struct geom_path *path, uns flags)
{
	struct geom_bezier *bezier;

	GEOM_GROUP_WALK((void *)path, bezier) {
		if (geom_approx_bezier(approx, bezier, flags) < 0)
			return -1;
	}

	return 0;
}



