/**
 * @file geomlib/base.h
 *
 * Geomlib - Base definitions and includes in Geomlib.
 */

#ifndef _VRR_GEOMLIB_BASE_H_
#define _VRR_GEOMLIB_BASE_H_


/* Geomlib source keywords */

#if defined GEOM_SOURCE && defined ALL_MESSAGES
#define GEOM_DEBUG_SLOW
#undef GEOM_DEBUG_IN
#undef GEOM_DEBUG_OUT
#undef GEOM_DEBUG_TRACE
#undef GEOM_DEBUG_FUNC
#undef GEOM_DEBUG_ERR
#undef GEOM_DEBUG_RET
#define GEOM_DEBUG_IN 1000
#define GEOM_DEBUG_OUT 1000
#define GEOM_DEBUG_TRACE 1000
#define GEOM_DEBUG_FUNC 1000
#define GEOM_DEBUG_ERR 1000
#define GEOM_DEBUG_RET 1000
#endif

#if defined GEOM_SILENT || (defined GEOM_SOURCE && defined NO_MESSAGES)
#undef GEOM_DEBUG_IN
#undef GEOM_DEBUG_OUT
#undef GEOM_DEBUG_TRACE
#undef GEOM_DEBUG_FUNC
#undef GEOM_DEBUG_ERR
#undef GEOM_DEBUG_RET

#endif


/* Global includes */

#include "lib/lib.h"
#include <math.h>
#include <float.h>
#include <string.h>
#include <stdlib.h>
#include "lib/clists.h"
#include "lib/cclists.h"
#include "lib/slists.h"
#include "lib/avltree.h"
#include "lib/garr.h"


/** If defined, enables slow debugging routines in Geomlib sources.  */
#define GEOM_DEBUG_SLOW

/** If nonzero, turns on input parameters dumping in entire Geomlib to given call depth.  */
#ifndef GEOM_DEBUG_IN
#define GEOM_DEBUG_IN 0
#endif

/** If nonzero, turns on output dumping in entire Geomlib to given call depth. */
#ifndef GEOM_DEBUG_OUT
#define GEOM_DEBUG_OUT 0
#endif

/** If nonzero, turns on trace messages in entire Geomlib to given call depth.  */
#ifndef GEOM_DEBUG_TRACE
#define GEOM_DEBUG_TRACE 0
#endif

/** If nonzero, turns on reporting of called functions in entire Geomlib to given call depth.  */
#ifndef GEOM_DEBUG_FUNC
#define GEOM_DEBUG_FUNC 0
#endif

/** If nonzero, turns on error reporting in entire Geomlib to given call depth.  */
#ifndef GEOM_DEBUG_ERR
#define GEOM_DEBUG_ERR 1000
#endif

/** If nonzero, turns on reporting of all functions results in Geomlib to given call depth.  */
#ifndef GEOM_DEBUG_RET
#define GEOM_DEBUG_RET 0
#endif


/* Solve debug dependencies */

#ifndef GEOM_DEBUG
#undef GEOM_DEBUG_SLOW
#undef GEOM_DEBUG_IN
#undef GEOM_DEBUG_OUT
#undef GEOM_DEBUG_TRACE
#undef GEOM_DEBUG_FUNC
#undef GEOM_DEBUG_ERR
#undef GEOM_DEBUG_RET
#define GEOM_DEBUG_IN 0
#define GEOM_DEBUG_OUT 0
#define GEOM_DEBUG_TRACE 0
#define GEOM_DEBUG_FUNC 0
#define GEOM_DEBUG_ERR 0
#define GEOM_DEBUG_RET 0
#endif


/* Messages
 * - GEOM_MSG(format, args) - debug message, controled by GEOM_DEBUG_TRACE macro
 * - GEOM_TRACE(format, args) - debug message with source position, controled by GEOM_DEBUG_TRACE macro
 * - GEOM_WPT - message with source position, controled by GEOM_DEBUG_TRACE macro
 */

#ifdef __GNUC__
#define GEOM_POS __FUNCTION__,__FILE__,__LINE__
#define GEOM_POS_DEF const char UNUSED *func, const char UNUSED *file, int UNUSED line
#define GEOM_POS_COPY func,file,line
#define GEOM_POS_FORMAT "[%s:%s:%d]"
#else
#define GEOM_POS __FILE__,__LINE__
#define GEOM_POS_DEF const char UNUSED *file, int UNUSED line
#define GEOM_POS_COPY file,line
#define GEOM_POS_FORMAT "[%s:%d]"
#endif
#ifdef GEOM_DEBUG
#define GEOM_MSG(...) do{if(geom_func_depth<=GEOM_DEBUG_TRACE)msg(__VA_ARGS__);}while(0)
#define GEOM_TRACE(...) do{if(geom_func_depth<=GEOM_DEBUG_TRACE){msg(GEOM_POS_FORMAT ": ",GEOM_POS);msg(__VA_ARGS__);msg("\n");}}while(0)
#else
#define GEOM_MSG(...) do{}while(0)
#define GEOM_TRACE(...) do{}while(0)
#endif
#define GEOM_WPT GEOM_TRACE("Waypoint.")


/* Inline dumping routines
 * - GEOM_DUMP(format, args) - dumping message, controlled by recursively entered blocks (see below); messages are separated by commas
 * - GEOM_DUMP_ENTER(on, format, args) - enters dumping block; if on is 1, dumping in block is enabled; prints header message
 * - GEOM_DUMP_GROUP - enters subgroup surrounded by braces
 * - GEOM_DUMP_LEAVE - leaves block or subgroup, inserts line break after enabled block
 * */

extern int geom_dump_depth;
extern int geom_dump_on;
extern uns geom_dump_flags;
extern int geom_dump_separator;
void geom_dump_group(uns flags);
void geom_dump_leave(void);
void geom_dump_sep(const char *c);
#define GEOM_DUMP_ENTER(f,...) do{if((f&1)&&!geom_dump_on){geom_dump_on=geom_dump_depth+1;geom_dump_separator=0;msg(__VA_ARGS__);}geom_dump_group(f);}while(0)
#define GEOM_DUMP_GROUP do{geom_dump_group(0);}while(0)
#define GEOM_DUMP_LEAVE do{geom_dump_leave();}while(0)
#define GEOM_DUMP(...) do{if(geom_dump_on){geom_dump_separator=1;msg(__VA_ARGS__);}}while(0)
#define GEOM_DUMP_SEP do{geom_dump_sep(", ");}while(0)
#define GEOM_SDUMP(...) do{GEOM_DUMP_SEP;GEOM_DUMP(__VA_ARGS__);}while(0)
#define GEOM_DUMP_PTR(p) if(!PTEST(p))GEOM_SDUMP("0");else
#define GEOM_DUMP_INLINE(x) do{if(!geom_dump_on){geom_dump_on=geom_dump_depth+1;geom_dump_separator=0;}geom_dump_group(2);{x;}GEOM_DUMP_LEAVE;}while(0)

/* Assertions
 * - GEOM_TEST(value) - untyped boolean test (optimized for true value)
 * - GEOM_PTEST(value) - assigned void pointer test
 * - GEOM_RTEST(value) - finite real test
 * - GEOM_FTEST(value) - finite double test
 * - GEOM_ETEST(value) - geomlib error code test (true if >=0)
 * - GEOM_RTEST_MINMAX(value, min, max) - real range test
 * - GEOM_FTEST_MINMAX(value, min, max) - double range test
 * - GEOM_xASSERT(value) - GEOM_xTEST assertion, cannot be used in expression, ignored in debug mode
 * - GEOM_xASSERTE(value) - GEOM_xTEST assertion, cannot be used in expression, value evaluated exactly once
 * - GEOM_xASSERTX(value) - GEOM_xTEST assertion, can be used in expression, value evaluated exactly once
 * */

#define GEOM_TEST(x) (likely((x)?1:0))
#define GEOM_PTEST(x) (likely((x) != NULL))
#define GEOM_RTEST(x) (likely(risfinite(x)))
#define GEOM_FTEST(x) (likely(isfinite((double)(x))))
#define GEOM_ETEST(x) (likely((int)(x) >= 0))
static inline int geom_rtest_minmax(real x, real min, real max) { return likely(x >= min && x <= max); }
static inline int geom_ftest_minmax(double x, double min, double max) { return likely(x >= min && x <= max); }
#define GEOM_RTEST_MINMAX(x,min,max) (likely(geom_rtest_minmax(x,min,max)))
#define GEOM_FTEST_MINMAX(x,min,max) (likely(geom_ftest_minmax(x,min,max)))

#define GEOM_ASSERT_CALL(m) m,GEOM_POS
#define GEOM_ASSERT_DEF const char UNUSED *assert_msg,GEOM_POS_DEF
#define GEOM_ASSERT_COPY assert_msg,GEOM_POS_COPY
#define GEOM_ASSERT_FORMAT "Geomlib assertion `%s' failed at "GEOM_POS_FORMAT
int geom_b_assert(GEOM_ASSERT_DEF, int value);
const void *geom_p_assert(GEOM_ASSERT_DEF, const void *value);
real geom_r_assert(GEOM_ASSERT_DEF, real value);
double geom_f_assert(GEOM_ASSERT_DEF, double value);
int geom_e_assert(GEOM_ASSERT_DEF, int value);
real geom_r_minmax_assert(GEOM_ASSERT_DEF, real value, real min, real max);
double geom_f_minmax_assert(GEOM_ASSERT_DEF, double value, double min, double max);

#ifdef GEOM_DEBUG
#define GEOM_ASSERT(n,v) do{geom_##n##_assert(GEOM_ASSERT_CALL("geom_"#n"_assert("#v")"),v);}while(0)
#define GEOM_ASSERTX(n,v) (geom_##n##_assert(GEOM_ASSERT_CALL("geom_"#n"_assert("#v")"),v))
#define GEOM_ASSERT2(n,v,...) do{geom_##n##_assert(GEOM_ASSERT_CALL("geom_"#n"_assert("#v", ...)"),v,__VA_ARGS__);}while(0)
#define GEOM_ASSERTX2(n,v,...) (geom_##n##_assert(GEOM_ASSERT_CALL("geom_"#n"_assert("#v", ...)"),v,__VA_ARGS__))
#else
#define GEOM_ASSERT(n,v) do{}while(0)
#define GEOM_ASSERTX(n,v) (v)
#define GEOM_ASSERT2(n,v,...) do{}while(0)
#define GEOM_ASSERTX2(n,v,...) (v)
#endif
#define GEOM_BASSERT(v) GEOM_ASSERT(b,(v)?1:0)
#define GEOM_BASSERTX(v) GEOM_ASSERTX(b,(v)?1:0)
#define GEOM_PASSERT(v) GEOM_ASSERT(p,(void *)(v))
#define GEOM_PASSERTX(v) GEOM_ASSERTX(p,(void *)(v))


/* Functions debugging routines
 * - GEOM_FUNC - enters function, prints with debug message (controlled by GEOM_DEBUG_FUNC), function must be ended with GEOM_*RET
 * - GEOM_*RET - leaves function, prints debug messages (controlled by GEOM_DEBUG_RET and GEOM_DEBUG_ERR)
 * - GEOM_IN{,_ENTER,_LEAVE} - input dumping, evaluated only in debug mode (controlled by GEOM_DEBUG_IN)
 * - GEOM_OUT(,_ENTER,_LEAVE) - output dumping, evaluated only in debug mode (controlled by GEOM_DEBUG_OUT)
 * - GEOM_RET_ERR(suffix) - leaves function with GEOM_ERR_suffix result
 * - GEOM_ERR_NUM - leaves function with GEOM_ERR_NUMERIC result
 * - GEOM_ERR_NAN - leaves function with NAN result
 * */

#ifdef GEOM_DEBUG

extern int geom_func_depth;
void geom_func(GEOM_POS_DEF, int max_depth);
void geom_ret(GEOM_POS_DEF, int max_depth);
int geom_eret(GEOM_POS_DEF, int max_depth, int max_err_depth, int value);
real geom_rret(GEOM_POS_DEF, int max_depth, int max_err_depth, real value);
double geom_fret(GEOM_POS_DEF, int max_depth, int max_err_depth, double value);
void *geom_pret(GEOM_POS_DEF, int max_depth, int max_err_depth, void *value);
#define GEOM_FUNC do{geom_func(GEOM_POS,GEOM_DEBUG_FUNC);}while(0)
#define GEOM_RET do{geom_ret(GEOM_POS,GEOM_DEBUG_RET);return;}while(0)
#define GEOM_VRET(v) do{geom_ret(GEOM_POS,GEOM_DEBUG_RET);return(v);}while(0)
#define GEOM_ERET(v) do{return geom_eret(GEOM_POS,GEOM_DEBUG_RET,GEOM_DEBUG_ERR,v);}while(0)
#define GEOM_RRET(v) do{return geom_rret(GEOM_POS,GEOM_DEBUG_RET,GEOM_DEBUG_ERR,v);}while(0)
#define GEOM_FRET(v) do{return geom_fret(GEOM_POS,GEOM_DEBUG_RET,GEOM_DEBUG_ERR,v);}while(0)
#define GEOM_PRET(v) do{return geom_pret(GEOM_POS,GEOM_DEBUG_RET,GEOM_DEBUG_ERR,v);}while(0)
#define GEOM_IN_ENTER do{GEOM_DUMP_ENTER(geom_func_depth<=GEOM_DEBUG_IN,"Input: ");
#define GEOM_IN_LEAVE GEOM_DUMP_LEAVE;}while(0)
#define GEOM_IN(x) do{GEOM_IN_ENTER{x}GEOM_IN_LEAVE;}while(0)
#define GEOM_OUT_ENTER do{GEOM_DUMP_ENTER(geom_func_depth<=GEOM_DEBUG_OUT,"Output: ");
#define GEOM_OUT_LEAVE GEOM_DUMP_LEAVE;}while(0)
#define GEOM_OUT(x) do{GEOM_OUT_ENTER{x}GEOM_OUT_LEAVE;}while(0)

#else

#define GEOM_FUNC do{}while(0)
#define GEOM_RET do{return;}while(0)
#define GEOM_VRET(v) do{return(v);}while(0)
#define GEOM_ERET(v) do{return(v);}while(0)
#define GEOM_RRET(v) do{return(v);}while(0)
#define GEOM_FRET(v) do{return(v);}while(0)
#define GEOM_PRET(v) do{return(v);}while(0)
#define GEOM_IN_ENTER do{if(0){
#define GEOM_IN_LEAVE }}while(0)
#define GEOM_IN(x) do{}while(0)
#define GEOM_OUT_ENTER do{if(0){
#define GEOM_OUT_LEAVE }}while(0)
#define GEOM_OUT(x) do{}while(0)

#endif


/* Shortcuts in Geomlib sources */

#ifdef GEOM_SOURCE

#define MSG(...) GEOM_DBG(__VA_ARGS__)
#define TRACE(...) GEOM_TRACE(__VA_ARGS__)
#define WPT GEOM_WPT

#define DUMP_ENTER(in,...) GEOM_DUMP_ENTER(in,__VA_ARGS__)
#define DUMP_GROUP GEOM_DUMP_GROUP
#define DUMP_LEAVE GEOM_DUMP_LEAVE
#define DUMP(...) GEOM_DUMP(__VA_ARGS__)
#define DUMP_SEP GEOM_DUMP_SEP
#define SDUMP(...) GEOM_SDUMP(__VA_ARGS__)
#define DUMP_PTR(p) GEOM_DUMP_PTR(p)

#define TEST(x) GEOM_TEST(x)
#define PTEST(x) GEOM_PTEST(x)
#define RTEST(x) GEOM_RTEST(x)
#define FTEST(x) GEOM_FTEST(x)
#define ETEST(x) GEOM_ETEST(x)
#define RTEST_MINMAX(x,min,max) GEOM_RTEST_MINMAX(x,min,max)
#define FTEST_MINMAX(x,min,max) GEOM_FTEST_MINMAX(x,min,max)

//#undef ASSERT
#define GASSERT(n,v) GEOM_ASSERT(n,v)
#define GASSERTX(n,v) GEOM_ASSERTX(n,v)
#define GASSERT2(n,v,...) GEOM_ASSERT2(n,v,__VA_ARGS__)
#define GASSERTX2(n,v,...) GEOM_ASSERTX2(n,v,__VA_ARGS__)
#define BASSERT(v) GEOM_BASSERT(v)
#define BASSERTX(v) GEOM_BASSERTX(v)
#define PASSERT(v) GEOM_PASSERT(v)
#define PASSERTX(v) GEOM_PASSERTX(v)
#define RASSERT(v) GEOM_ASSERT(r,v)
#define RASSERTX(v) GEOM_ASSERTX(r,v)
#define FASSERT(v) GEOM_ASSERT(f,v)
#define FASSERTX(v) GEOM_ASSERTX(f,v)
#define EASSERT(v) GEOM_ASSERT(e,v)
#define EASSERTX(v) GEOM_ASSERTX(e,v)

#define FUNC GEOM_FUNC
#define RET GEOM_RET
#define VRET(v) GEOM_VRET(v)
#define ERET(v) GEOM_ERET(v)
#define RRET(v) GEOM_RRET(v)
#define FRET(v) GEOM_FRET(v)
#define PRET(v) GEOM_PRET(v)
#define IN_ENTER GEOM_IN_ENTER
#define IN_LEAVE GEOM_IN_LEAVE
#define IN(x) GEOM_IN(x)
#define OUT_ENTER GEOM_OUT_ENTER
#define OUT_LEAVE GEOM_OUT_LEAVE
#define OUT(x) GEOM_OUT(x)

#define ERR(s) GEOM_ERR_##s
#define RET_ERR(s) ERET(ERR(s))
#define RET_NUM RET_ERR(NUMERIC)
#define RET_NAN RRET(NAN)
#define RET_ERRNO ERET(geom_errno)
#define TEST_ERR(v,s) do{if(!TEST(v))RET_ERR(s);}while(0)
#define TEST_NUM(v) do{if(!TEST(v))RET_NUM;}while(0)
#define TEST_NAN(v) do{if(!TEST(v))RET_NAN;}while(0)
#define TEST_ERRNO(v) do{if(!TEST(v))RET_ERRNO;}while(0)
#define ETEST_RET(v) do{int _etest_ret=(v);if(!ETEST(_etest_ret))ERET(_etest_ret);}while(0)
#define ETEST_NAN(v) do{if(!ETEST(v))RET_NAN;}while(0)
#define RTEST_NUM(v) do{if(!RTEST(v))RET_NUM;}while(0)
#define RTEST_NAN(v) do{if(!RTEST(v))RET_NAN;}while(0)
#define RTEST_ERRNO(v) do{if(!RTEST(v))RET_ERRNO;}while(0)
#define FTEST_NUM(v) do{if(!FTEST(v))RET_NUM;}while(0)
#define FTEST_NAN(v) do{if(!FTEST(v))RET_NAN;}while(0)
#define FTEST_ERRNO(v) do{if(!FTEST(v))RET_ERRNO;}while(0)
#define PTEST_ERR(v,s) do{if(!PTEST(v))RET_ERR(s);}while(0)
#define PTEST_NUM(v) do{if(!PTEST(v))RET_NUM;}while(0)
#define PTEST_NAN(v) do{if(!PTEST(v))RET_NAN;}while(0)
#define PTEST_ERRNO(v) do{if(!PTEST(v))RET_ERRNO;}while(0)


#endif /* GEOM_SOURCE */


/* Global Geomlib includes */

#include "geomlib/class.h"
#include "geomlib/types.h"
#include "geomlib/numeric.h"
#include "geomlib/err.h"


/* Library initializaction and finalization routines */

void geomlib_init(void);
void geomlib_finish(void);

#endif /* _VRR_GEOMLIB_BASE_H_ */

