#define GEOM_SOURCE
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"
#include "geomlib/transform.h"
#include "geomlib/bezier.h"


/* geom_bezier_valid */

int geom_bezier_valid(struct geom_bezier *bezier)
{
	uns i, d;
	if (!PTEST(bezier))
		return 0;
	if (!geom_instance_is(bezier, CLASS(bezier)))
		return 0;
	d = GEOM_BEZIER_DEGREE(bezier);
	if (d < GEOM_BEZIER_LINEAR || d > GEOM_BEZIER_CUBIC)
		return 0;
	for (i = 0; i <= d; i++)
		if (!geom_point_w_valid(&bezier->pt[i]))
			return 0;
	return 1;
}


/* geom_bezier_subdump */

void geom_bezier_subdump(struct geom_bezier *bezier, uns flags)
{
	uns i;
	geom_item_subdump((struct geom_item *)bezier, flags);
	SDUMP("flags=0x%x", INSTANCE_FLAGS(bezier));
	if (GEOM_BEZIER_DEGREE(bezier) <= GEOM_BEZIER_CUBIC)
		for (i = 0; i <= GEOM_BEZIER_DEGREE(bezier); i++) {
			SDUMP("pt[%d]=", i);
			geom_point_w_dump(&bezier->pt[i], 0);
		}
}


/*----------------------------------------------*
 *                                              *
 *               TRANSFORMATIONS                *
 *                                              *
 *----------------------------------------------*/


/* geom_bezier_transform */

int geom_bezier_transform(struct geom_bezier *bezier, struct geom_transform *transform, struct geom_bezier *result)
{
	struct geom_point pt[4];
	uns i;

	FUNC;
	ITEM_ASSERT(bezier, bezier);
	GASSERT(transform, transform);
	INSTANCE_IS_ASSERT(bezier, result);

	if (result != bezier)
		geom_bezier_copy(result, bezier);
	if (!(transform->flags & GEOM_TRANSFORM_IDENTITY)) {
		if (result != bezier) {
			for (i = 0; i <= BEZIER_DEGREE(bezier); i++) {
				RTEST_NUM(result->pt[i].x = transform->coef[0][0] * bezier->pt[i].x +
					transform->coef[0][1] * bezier->pt[i].y + transform->coef[0][2]);
				RTEST_NUM(result->pt[i].y = transform->coef[1][0] * bezier->pt[i].x +
					transform->coef[1][1] * bezier->pt[i].y + transform->coef[1][2]);
			}
		}
		else {
			for (i = 0; i <= BEZIER_DEGREE(bezier); i++) {
				RTEST_NUM(pt[i].x = transform->coef[0][0] * bezier->pt[i].x +
					transform->coef[0][1] * bezier->pt[i].y + transform->coef[0][2]);
				RTEST_NUM(pt[i].y = transform->coef[1][0] * bezier->pt[i].x +
					transform->coef[1][1] * bezier->pt[i].y + transform->coef[1][2]);
			}
			for (i = 0; i <= BEZIER_DEGREE(bezier); i++) {
				result->pt[i].x = pt[i].x;
				result->pt[i].y = pt[i].y;
			}
		}
		geom_bezier_after_change(result);
	}

	ITEM_ASSERT(bezier, result);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/* geom_bezier_point_at_time */

int geom_bezier_point_at_time(struct geom_bezier *bezier, real time, struct geom_point *result)
{
	ITEM_ASSERT(bezier, bezier);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		return geom_bezier_point_start(bezier, result);

	else if (time >= 1)
		return geom_bezier_point_end(bezier, result);

	else {
		struct geom_bezier right;
		GEOM_INSTANCE_INIT(&right, bezier);
		geom_bezier_split_at(bezier, time, NULL, &right);
		result->x = right.pt[0].x;
		result->y = right.pt[0].y;
		GEOM_INSTANCE_DESTROY2(&right, bezier);
		return 0;
	}
}


/* geom_bezier_split_middle
 * Function splits a bezier curve at parameter 0.5. */

int geom_bezier_split_middle(struct geom_bezier *b, struct geom_bezier *l, struct geom_bezier *r)
{
	return geom_bezier_split_at(b, 0.5, l, r);
#if 0
	int d, i, j;
	double u, v;
	struct geom_point_w p[4][4];

	BASSERT(geom_bezier_valid(b));


	/* copy control points */
	d = b->degree;
	for (i = 0; i <= d; i++)
		p[d][i] = b->pt[i];

	/* nonrational curve */
	if (b->flags & GEOM_BEZIER_NONRATIONAL) {
		/* fast version */
		for (j = d; j > 0; j--)
			for (i = 0; i < j; i++) {
				p[j - 1][i].x = 0.5 * (p[j][i].x + p[j][i + 1].x);
				p[j - 1][i].y = 0.5 * (p[j][i].y + p[j][i + 1].y);
				p[j - 1][i].w = b->pt[0].w;
			}

		/* overflow-safe version */
		if (unlikely(!isfinite(p[0][0].x) || !isfinite(p[0][0].y)))
			for (j = d; j > 0; j--)
				for (i = 0; i < j; i++) {
					if (unlikely(isinf(p[j - 1][i].x = 0.5 * p[j][i].x + 0.5 * p[j][i + 1].x)))
						p[j - 1][i].x = p[j][i].x;
					if (unlikely(isinf(p[j - 1][i].y = 0.5 * p[j][i].y + 0.5 * p[j][i + 1].y)))
						p[j - 1][i].y = p[j][i].y;
				}

	/* rational curve */
	} else {
		/* fast version */
		for (j = d; j > 0; j--)
			for (i = 0; i < j; i++) {
				p[j - 1][i].w = 0.5 * (p[j][i].w + p[j][i + 1].w);
				u = 0.5 * p[j][i + 1].w / p[j - 1][i].w;
				p[j - 1][i].x = p[j][i].x + u * (p[j][i + 1].x - p[j][i].x);
				p[j - 1][i].y = p[j][i].y + u * (p[j][i + 1].y - p[j][i].y);
			}

		/* overflow-safe version  */
		if (unlikely(!isfinite(p[0][0].x) || !isfinite(p[0][0].y) || !isfinite(p[0][0].w))) {
			for (j = d; j > 0; j--)
				for (i = 0; i < j; i++) {
					p[j - 1][i].w = 0.5 * p[j][i].w + 0.5 * p[j][i + 1].w;
					if (unlikely(isinf(p[j - 1][i].w) || p[j - 1][i].w == 0))
						p[j - 1][i].w = p[j][i].w;
					if (unlikely((u = 0.5 * p[j][i + 1].w / p[j - 1][i].w) > 1))
						u = 1;
					v = 1 - u;
					if (unlikely(isinf(p[j - 1][i].x = v * p[j][i].x + u * p[j][i + 1].x)))
						p[j - 1][i].x = p[j][i].x;
					if (unlikely(isinf(p[j - 1][i].y = v * p[j][i].y + u * p[j][i + 1].y)))
						p[j - 1][i].y = p[j][i].y;
				}
		}
	}

	/* store results */
	if (l) {
		l->degree = d;
		l->flags = b->flags & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
		for (i = 0; i <= d; i++)
			l->pt[i] = p[d - i][0];
	}
	if (r) {
		r->degree = d;
		r->flags = b->flags & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
		for (i = 0; i <= d; i++)
			r->pt[i] = p[i][i];
	}

	return 0;
#endif
}


/* geom_bezier_split_at
 * Function splits a bezier curve in a given parameter t. */

int geom_bezier_split_at(struct geom_bezier *b, real t, struct geom_bezier *l, struct geom_bezier *r)
{
	int d, i, j;
	double s, u, v;
	struct geom_point_w p[4][4];

#ifdef GEOM_DEBUG
	ITEM_ASSERT(bezier, b);
	RASSERT(t);
	if (l) INSTANCE_IS_ASSERT(bezier, l);
	if (r) INSTANCE_IS_ASSERT(bezier, r);
#endif

	d = BEZIER_DEGREE(b);

	/* special cases */
	if (t <= 0) {
		if (r)
			geom_bezier_copy(r, b);
		if (l) {
			INSTANCE_FLAGS(l) = INSTANCE_FLAGS(b) & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
			for (i = 0; i <= d; i++)
				l->pt[i] = b->pt[0];
		}
		return 0;
	} else if (t >= 1) {
		if (l)
			geom_bezier_copy(l, b);
		if (r) {
			INSTANCE_FLAGS(r) = INSTANCE_FLAGS(b) & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
			for (i = 0; i <= d; i++)
				r->pt[i] = b->pt[d];
		}
		return 0;
	} /*else if (t == 0.5)
		return geom_bezier_split_middle(b, l, r);*/

	/* copy control points */
	for (i = 0; i <= d; i++)
		p[d][i] = b->pt[i];

	/* nonrational curve */
	if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL) {
		/* fast version */
		for (j = d; j > 0; j--)
			for (i = 0; i < j; i++) {
				p[j - 1][i].x = p[j][i].x + t * (p[j][i + 1].x - p[j][i].x);
				p[j - 1][i].y = p[j][i].y + t * (p[j][i + 1].y - p[j][i].y);
				p[j - 1][i].w = b->pt[0].w;
			}

		/* overflow-safe version */
		if (unlikely(!isfinite(p[0][0].x) || !isfinite(p[0][0].y))) {
			s = 1 - t;
			for (j = d; j > 0; j--)
				for (i = 0; i < j; i++) {
					if (unlikely(isinf(p[j - 1][i].x = s * p[j][i].x + t * p[j][i + 1].x)))
						p[j - 1][i].x = p[j][i].x;
					if (unlikely(isinf(p[j - 1][i].y = s * p[j][i].y + t * p[j][i + 1].y)))
						p[j - 1][i].y = p[j][i].y;
				}
		}

	/* rational curve */
	} else {
		/* fast version */
		for (j = d; j > 0; j--)
			for (i = 0; i < j; i++) {
				p[j - 1][i].w = p[j][i].w + t * (p[j][i + 1].w - p[j][i].w);
				u = t * p[j][i + 1].w / p[j - 1][i].w;
				p[j - 1][i].x = p[j][i].x + u * (p[j][i + 1].x - p[j][i].x);
				p[j - 1][i].y = p[j][i].y + u * (p[j][i + 1].y - p[j][i].y);
			}

		/* overflow-safe version  */
		if (unlikely(!isfinite(p[0][0].x) || !isfinite(p[0][0].y) || !isfinite(p[0][0].w))) {
			s = 1 - t;
			for (j = d; j > 0; j--)
				for (i = 0; i < j; i++) {
					p[j - 1][i].w = s * p[j][i].w + t * p[j][i + 1].w;
					if (unlikely(isinf(p[j - 1][i].w) || p[j - 1][i].w == 0))
						p[j - 1][i].w = p[j][i].w;
					if (unlikely((u = 0.5 * p[j][i + 1].w / p[j - 1][i].w) > 1))
						u = 1;
					v = 1 - u;
					if (unlikely(isinf(p[j - 1][i].x = v * p[j][i].x + u * p[j][i + 1].x)))
						p[j - 1][i].x = p[j][i].x;
					if (unlikely(isinf(p[j - 1][i].y = v * p[j][i].y + u * p[j][i + 1].y)))
						p[j - 1][i].y = p[j][i].y;
				}
		}
	}

	/* copy results */
	if (l) {
		INSTANCE_FLAGS(l) = INSTANCE_FLAGS(b) & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
		for (i = 0; i <= d; i++)
			l->pt[i] = p[d - i][0];
		ITEM_ASSERT(bezier, l);
	}
	if (r) {
		INSTANCE_FLAGS(r) = INSTANCE_FLAGS(b) & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
		for (i = 0; i <= d; i++)
			r->pt[i] = p[i][i];
		ITEM_ASSERT(bezier, r);
	}

	return 0;
}

/* geom_bezier_reverse
 * Function returns bezier curve with reversed parametrization. */

// FIXME

int geom_bezier_reverse(struct geom_bezier *b, struct geom_bezier *r)
{
	int i, j;
	struct geom_point_w s;

	if (b == r)
		for (i = 0, j = GEOM_BEZIER_DEGREE(b); i < j; i++, j--) {
			s = r->pt[i];
			r->pt[i] = r->pt[j];
			r->pt[j] = s;
		}

	else {
		GEOM_INSTANCE_FLAGS(r) = GEOM_INSTANCE_FLAGS(b) & ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
		GEOM_BEZIER_DEGREE(r) = GEOM_BEZIER_DEGREE(b);
		for (i = 0; i <= GEOM_BEZIER_DEGREE(b); i++)
			r->pt[i] = b->pt[GEOM_BEZIER_DEGREE(b) - i];
	}

	return 0;
}

/* geom_bezier_part
 * Function returns subinterval of bezier curve. */

// FIXME

int geom_bezier_part(struct geom_bezier *b, double t, double s, struct geom_bezier *result)
{
	int reverse;
	double u;

	/* align parameters to unit interval */
	if (t < 0)
		t = 0;
	else if (t > 1)
		t = 1;
	if (s < 0)
		s = 0;
	else if (s > 1)
		s = 1;

	/* sort parameters */
	if (t > s) {
		u = t;
		t = s;
		s = u;
		reverse = 1;
	} else
		reverse = 0;

	/* two cases to avoid numeric errors */
	if (s >= 0.5) {
		u = t / s;
		geom_bezier_split_at(b, s, result, NULL);
		geom_bezier_split_at(result, u, NULL, result);
	} else {
		u = (s - t) / (1 - t);
		geom_bezier_split_at(b, t, NULL, result);
		geom_bezier_split_at(result, u, result, NULL);
	}

	/* reverse if it was t > s */
	if (reverse)
		return geom_bezier_reverse(result, result);
	else
		return 0;
}

/* geom_bezier_elevate
 * Function to increase bezier curve degree. */

int geom_bezier_elevate(struct geom_bezier *b, uns degree, struct geom_bezier *r)
{
	int i;
	double t, s;

	ITEM_ASSERT(bezier, b);
	BASSERT(degree <= GEOM_BEZIER_CUBIC);
	BASSERT(BEZIER_DEGREE(b) <= degree);
	INSTANCE_IS_ASSERT(bezier, r);

	if (b != r)
		geom_bezier_copy(r, b);

	/* elevate in single degree steps */
	while (BEZIER_DEGREE(r) < degree) {
		r->pt[BEZIER_DEGREE(r) + 1] = r->pt[BEZIER_DEGREE(r)];
		for (i = BEZIER_DEGREE(r)++; i > 0; i--) {
			t = (double)i / BEZIER_DEGREE(r);
			s = 1 - t;

			/* rational curve */
			if (!(INSTANCE_FLAGS(r) & GEOM_BEZIER_NONRATIONAL)) {
				r->pt[i].w = s * r->pt[i].w + t * r->pt[i - 1].w;
				if (unlikely(isinf(r->pt[i].w) || r->pt[i].w == 0))
					r->pt[i].w = r->pt[i - 1].w;
				if (unlikely((t = t * r->pt[i - 1].w / r->pt[i].w) > 1))
					t = 1;
				s = 1 - t;
			}
			if (unlikely(isinf(r->pt[i].x = s * r->pt[i].x + t * r->pt[i - 1].x)))
				r->pt[i].x = r->pt[i - 1].x;
			if (unlikely(isinf(r->pt[i].y = s * r->pt[i].y + t * r->pt[i - 1].y)))
				r->pt[i].y = r->pt[i - 1].y;
		}
	}

	ITEM_ASSERT(bezier, r);
	return 0;
}


/* */

static void add_point(double **buffer, int *buffer_size, int *points, double x, double y)
{
	int op = (*points)++;
	op *= 2;
	if (*buffer_size <= op) {
		if (*buffer_size == 0)
			*buffer_size = 16;
		else
			*buffer_size *= 2;

		*buffer = xrealloc(*buffer, *buffer_size * sizeof(double));
	}
	(*buffer)[op] = x;
	(*buffer)[op + 1] = y;
}

static double line_point_square_dist(struct geom_point_w *l1, struct geom_point_w *l2, struct geom_point_w *p)
{
	double dx = l2->x - l1->x;
	double dy = l2->y - l1->y;
	double t;

	/* FIXME: signal error? */
	if ((dx == 0) && (dy == 0))
	{
	}

	t = ((p->x - l1->x) * dx + (p->y - l1->y) * dy) / (dx * dx + dy * dy);

	{
		double ox = l1->x + t * dx - p->x;
		double oy = l1->y + t * dy - p->y;
		return (ox * ox + oy * oy);
	}
}

/* compute error of approximation that bezier curve with line from startpoint
 to endpoint. Should i use some better metric? */

static double compute_square_delta(struct geom_bezier *b)
{
	double d1, d2;

	switch (BEZIER_DEGREE(b)) {
	case GEOM_BEZIER_LINEAR:	/* linear curve */
		return (0);
		break;
	case GEOM_BEZIER_QUADRATIC:	/* quadratic curve */
		d1 = line_point_square_dist(b->pt, b->pt + 2, b->pt + 1);
		return (d1);
		break;
	default:	/* cubic curve */
		d1 = line_point_square_dist(b->pt, b->pt + 2, b->pt + 1);
		d2 = line_point_square_dist(b->pt, b->pt + 3, b->pt + 2);
		return (d1 < d2 ? d2 : d1);
		break;
	}
}

int geom_bezier_approximate(struct geom_bezier *b, double delta, double **buffer, int *buffer_size, int *points)
{
	struct geom_bezier temp[3];
	struct geom_bezier *b1, *b2, *bt;
	int false_size = 0;
	uns i;

	for (i = 0; i < 3; i++)
		GEOM_INSTANCE_INIT(temp + i, bezier);
	b1 = temp + 1;
	b2 = temp + 2;

	*points = 0;
	if (buffer_size == NULL)
		buffer_size = &false_size;

	add_point(buffer, buffer_size, points, b->pt[0].x, b->pt[0].y);

	if (BEZIER_DEGREE(b) != GEOM_BEZIER_LINEAR) {
		double step = 0.2;
		double subdelta = 0.3 * delta;
		double ad;

		int overshot = 0;
		int iteration = 0;
		int raising = 0;

		double delta2 = delta * delta;
		double subdelta2 = subdelta * subdelta;

		*b1 = *b;
		while (step < 1) {
			geom_bezier_split_at(b1, step, temp, b2);
			ad = compute_square_delta(temp);
			if (ad > delta2) {
				overshot++;
				step /= 2;
			} else {
				add_point(buffer, buffer_size, points, b2->pt[0].x, b2->pt[0].y);

				bt = b1;
				b1 = b2;
				b2 = b1;

				iteration++;
				step *= (1 / (1 - step));
				if (ad < subdelta2) {
					raising++;
					step *= 1.4;
				}
			}
		}
		DBG("geom_bezier_approximate: (%f) Iters: %d, overs: %d, ratio: %d/100, raising: %d, ratio: %d/100\n", delta,
		       iteration, overshot, (overshot * 100) / (overshot + iteration), raising, (raising * 100) / iteration);
	}

	add_point(buffer, buffer_size, points, b->pt[BEZIER_DEGREE(b)].x, b->pt[BEZIER_DEGREE(b)].y);
	for (i = 0; i < 3; i++)
		GEOM_INSTANCE_DESTROY2(&temp[i], bezier);
	return (0);
}

