#ifndef _VRR_GEOMLIB_BEZIER_H_
#define _VRR_GEOMLIB_BEZIER_H_


/**
 * @file geomlib/bezier.h
 *
 * Geomlib - Rational Bezier curves.
 */


#include "geomlib/curve.h"
#include "geomlib/fpath.h"


/*----------------------------------------------*
 *                                              *
 *                BEZIER CLASS                  *
 *                                              *
 *----------------------------------------------*/


#define geom_bezier_VMT    geom_curve_VMT
#define geom_bezier_INIT   geom_curve_INIT \
	\
	GEOM_CURVE_TIME_EQUALS_BTIME \
	\
	.valid = (int (*)(const void *)) &geom_bezier_valid, \
	.subdump = (void (*)(const void *, uns)) &geom_bezier_subdump, \
	\
	.after_change = (void (*)(void *)) &geom_bezier_after_change, \
	\
	.alength = (real (*)(void *)) &geom_bezier_alength, \
	.time_to_atime = (real (*)(void *, real)) &geom_bezier_time_to_atime, \
	.times_to_atimes = (int (*)(void *, uns, real *, real *)) &geom_bezier_times_to_atimes, \
	.atime_to_time = (real (*)(void *, real)) &geom_bezier_atime_to_time, \
	.atimes_to_times = (int (*)(void *, uns, real *, real *)) &geom_bezier_atimes_to_times, \
	\
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_bezier_expansion_append, \
	.to_bezier = (int (*)(void *, struct geom_bezier *)) &geom_bezier_to_bezier, \
	\
	.transform = (int (*)(void *, struct geom_transform *, void *)) &geom_bezier_transform, \
	\
	.point_start = (int (*)(void *, struct geom_point *)) &geom_bezier_point_start, \
	.point_end = (int (*)(void *, struct geom_point *)) &geom_bezier_point_end, \
	.point_at_time = (int (*)(void *, real, struct geom_point *)) &geom_bezier_point_at_time, \
	\
	.derivation_start = (int (*)(void *, struct geom_vector *)) &geom_bezier_derivation_start, \
	.derivation_end = (int (*)(void *, struct geom_vector *)) &geom_bezier_derivation_end, \
	.derivation_at_time = (int (*)(void *, real, struct geom_vector *)) &geom_bezier_derivation_at_time, \
	\
	.bbox = (int (*)(void *, struct geom_rectangle *)) &geom_bezier_bbox, \
	.nearest_to_point = (int (*)(void *, struct geom_point *, uns, struct geom_nearest *)) &geom_bezier_nearest_to_point, \

GEOM_CLASS_HEAD(bezier, curve);


/*----------------------------------------------*
 *                                              *
 *                  CONSTANTS                   *
 *                                              *
 *----------------------------------------------*/


#define GEOM_BEZIER_LINEAR     1
#define GEOM_BEZIER_QUADRATIC  2
#define GEOM_BEZIER_CUBIC      3

#define GEOM_BEZIER_NONRATIONAL   0x100
#define GEOM_BEZIER_BBOX_VALID    0x200
#define GEOM_BEZIER_ALENGTH_VALID 0x400

#define GEOM_BEZIER_DEGREE(b) (*(u8 *)&GEOM_INSTANCE_FLAGS(b))


/*----------------------------------------------*
 *                                              *
 *               INITIALIZATION                 *
 *                                              *
 *----------------------------------------------*/


/**
 * Default Bezier curve constructor.
 * Resulting structure is invalid curve and must be set by geom_bezier_set_x or destroyed.
 */
static inline void geom_bezier_create(struct geom_bezier *bezier)
{ geom_curve_create(&bezier->curve); }


/** Destructor. */
static inline void geom_bezier_destroy(struct geom_bezier *bezier)
{ geom_curve_destroy(&bezier->curve); }


/**
 * Sets nonrational Bezier curve from two control points.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 */
static inline void geom_bezier_set_linear_nonrational(struct geom_bezier *bezier, struct geom_point *pt0, struct geom_point *pt1)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_LINEAR | GEOM_BEZIER_NONRATIONAL;
	*(struct geom_point *)&bezier->pt[0] = *GEOM_ASSERTX(point, pt0);
	*(struct geom_point *)&bezier->pt[1] = *GEOM_ASSERTX(point, pt1);
	bezier->pt[0].w = bezier->pt[1].w = 1;
	geom_curve_after_change(&bezier->curve);
}


/**
 * Sets rational Bezier curve from two control points and weights.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 */
static inline void geom_bezier_set_linear_rational(struct geom_bezier *bezier, struct geom_point_w *pt0, struct geom_point_w *pt1)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_LINEAR;
	bezier->pt[0] = *GEOM_ASSERTX(point_w, pt0);
	bezier->pt[1] = *GEOM_ASSERTX(point_w, pt1);
	geom_curve_after_change(&bezier->curve);
}


/**
 * Sets nonrational Bezier curve from three control points.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 * @ai  pt2     Third control point.
 */
static inline void geom_bezier_set_quadratic_nonrational(struct geom_bezier *bezier, struct geom_point *pt0, struct geom_point *pt1, struct geom_point *pt2)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_QUADRATIC | GEOM_BEZIER_NONRATIONAL;
	*(struct geom_point *)&bezier->pt[0] = *GEOM_ASSERTX(point, pt0);
	*(struct geom_point *)&bezier->pt[1] = *GEOM_ASSERTX(point, pt1);
	*(struct geom_point *)&bezier->pt[2] = *GEOM_ASSERTX(point, pt2);
	bezier->pt[0].w = bezier->pt[1].w = bezier->pt[2].w = 1;
	geom_curve_after_change(&bezier->curve);
}


/**
 * Sets rational Bezier curve from three control points.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 * @ai  pt2     Third control point.
 */
static inline void geom_bezier_set_quadratic_rational(struct geom_bezier *bezier, struct geom_point_w *pt0, struct geom_point_w *pt1, struct geom_point_w *pt2)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_QUADRATIC;
	bezier->pt[0] = *GEOM_ASSERTX(point_w, pt0);
	bezier->pt[1] = *GEOM_ASSERTX(point_w, pt1);
	bezier->pt[2] = *GEOM_ASSERTX(point_w, pt2);
	geom_curve_after_change(&bezier->curve);
}


/**
 * Sets nonrational Bezier curve from four control points.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 * @ai  pt2     Third control point.
 * @ai  pt3     Fouth control point.
 */
static inline void geom_bezier_set_cubic_nonrational(struct geom_bezier *bezier, struct geom_point *pt0, struct geom_point *pt1, struct geom_point *pt2, struct geom_point *pt3)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_CUBIC | GEOM_BEZIER_NONRATIONAL;
	*(struct geom_point *)&bezier->pt[0] = *GEOM_ASSERTX(point, pt0);
	*(struct geom_point *)&bezier->pt[1] = *GEOM_ASSERTX(point, pt1);
	*(struct geom_point *)&bezier->pt[2] = *GEOM_ASSERTX(point, pt2);
	*(struct geom_point *)&bezier->pt[3] = *GEOM_ASSERTX(point, pt3);
	bezier->pt[0].w = bezier->pt[1].w = bezier->pt[2].w = bezier->pt[3].w = 1;
	geom_curve_after_change(&bezier->curve);
}


/**
 * Sets rational Bezier curve from four control points.
 * Can be used as constructor.
 *
 * @ao  bezier  Bezier instance.
 * @ai  pt0     First control point.
 * @ai  pt1     Second control point.
 * @ai  pt2     Third control point.
 * @ai  pt3     Fouth control point.
 */
static inline void geom_bezier_set_cubic_rational(struct geom_bezier *bezier, struct geom_point_w *pt0, struct geom_point_w *pt1, struct geom_point_w *pt2, struct geom_point_w *pt3)
{
	GEOM_INSTANCE_FLAGS(GEOM_INSTANCE_IS_ASSERTX(bezier, bezier)) = GEOM_BEZIER_CUBIC;
	bezier->pt[0] = *GEOM_ASSERTX(point_w, pt0);
	bezier->pt[1] = *GEOM_ASSERTX(point_w, pt1);
	bezier->pt[2] = *GEOM_ASSERTX(point_w, pt2);
	bezier->pt[3] = *GEOM_ASSERTX(point_w, pt3);
	geom_curve_after_change(&bezier->curve);
}


/*----------------------------------------------*
 *                                              *
 *           OVERRIDED VIRTUAL METHODS          *
 *                                              *
 *----------------------------------------------*/

int geom_bezier_valid(struct geom_bezier *bezier);
void geom_bezier_subdump(struct geom_bezier *bezier, uns flags);

static inline void geom_bezier_after_change(struct geom_bezier *bezier)
{
	GEOM_INSTANCE_FLAGS(bezier) &= ~(GEOM_BEZIER_BBOX_VALID | GEOM_BEZIER_ALENGTH_VALID);
	geom_curve_after_change(&bezier->curve);
}

static inline void geom_bezier_copy(struct geom_bezier *bezier, struct geom_bezier *src)
{ geom_curve_copy(&bezier->curve, &src->curve); }


/* Parametrizations */

real geom_bezier_alength(struct geom_bezier *bezier);

real geom_bezier_time_to_atime(struct geom_bezier *bezier, real time);
int geom_bezier_times_to_atimes(struct geom_bezier *bezier, uns count, real *times, real *atimes);
real geom_bezier_atime_to_time(struct geom_bezier *bezier, real atime);
int geom_bezier_atimes_to_times(struct geom_bezier *bezier, uns count, real *atimes, real *times);


/* Conversions */

static inline int geom_bezier_expansion_append(struct geom_bezier *bezier, struct geom_fpath *expansion)
{ void *p; GEOM_FPATH_APPEND_NEW_CREATE2(p, expansion, bezier, geom_bezier_copy, bezier); return 0; }

static inline int geom_bezier_to_bezier(struct geom_bezier *bezier, struct geom_bezier *result)
{ geom_bezier_copy(result, bezier); return 0; }


/* Transformations */

int geom_bezier_transform(struct geom_bezier *bezier, struct geom_transform *transform, struct geom_bezier *result);


/* Geometric properties */

static inline int geom_bezier_point_start(struct geom_bezier *bezier, struct geom_point *result)
{ result->x = bezier->pt[0].x; result->y = bezier->pt[0].y; return 0; }

static inline int geom_bezier_point_end(struct geom_bezier *bezier, struct geom_point *result)
{ result->x = bezier->pt[GEOM_BEZIER_DEGREE(bezier)].x; result->y = bezier->pt[GEOM_BEZIER_DEGREE(bezier)].y; return 0; }

int geom_bezier_point_at_time(struct geom_bezier *bezier, real time, struct geom_point *result);

int geom_bezier_derivation_start(struct geom_bezier *bezier, struct geom_vector *result);
int geom_bezier_derivation_end(struct geom_bezier *bezier, struct geom_vector *result);
int geom_bezier_derivation_at_time(struct geom_bezier *bezier, real time, struct geom_vector *result);

int geom_bezier_bbox(struct geom_bezier *bezier, struct geom_rectangle *bbox);
int geom_bezier_nearest_to_point(struct geom_bezier *bezier, struct geom_point *point, uns flags, struct geom_nearest *nearest);


/*----------------------------------------------*
 *                                              *
 *                   SUBDIVISION                *
 *                                              *
 *----------------------------------------------*/


#define GEOM_BEZIER_SUBDIVISION_MAX 25

int geom_bezier_split_middle(struct geom_bezier *b, struct geom_bezier *left, struct geom_bezier *right);

struct geom_bezier_subdivision {
	struct geom_bezier *bezier, stack_bezier[GEOM_BEZIER_SUBDIVISION_MAX + 1];
	uns depth, stack_top, init_top, stack_depth[GEOM_BEZIER_SUBDIVISION_MAX + 1];
	real left, right, stack_time[GEOM_BEZIER_SUBDIVISION_MAX + 2];
};

static inline void geom_bezier_subdivision_init(struct geom_bezier_subdivision *s, struct geom_bezier *b)
{
	GEOM_INSTANCE_CREATE2(&s->stack_bezier[0], bezier, geom_bezier_copy, b);
	s->init_top = 0; s->stack_top = 1; s->stack_depth[0] = 0; s->stack_time[1] = 0; s->stack_time[0] = 1;
}

static inline void geom_bezier_subdivision_destroy(struct geom_bezier_subdivision *s)
{
	for (uns i = 0; i <= s->init_top; i++) geom_bezier_destroy(&s->stack_bezier[i]);
}

static inline struct geom_bezier *geom_bezier_subdivision_next(struct geom_bezier_subdivision *s)
{
	if (unlikely(s->stack_top-- == 0)) return NULL;
	s->bezier = &s->stack_bezier[s->stack_top]; s->depth = s->stack_depth[s->stack_top];
	s->left = s->stack_time[s->stack_top + 1]; s->right = s->stack_time[s->stack_top];
	return s->bezier;
}

static inline void geom_bezier_subdivision_split(struct geom_bezier_subdivision *s)
{
	if (unlikely(s->stack_top == s->init_top)) GEOM_INSTANCE_INIT(&s->stack_bezier[++s->init_top], bezier);
	geom_bezier_split_middle(&s->stack_bezier[s->stack_top], &s->stack_bezier[s->stack_top + 1], &s->stack_bezier[s->stack_top]);
	s->stack_depth[s->stack_top + 1] = ++s->stack_depth[s->stack_top];
	s->stack_time[s->stack_top + 1] = fmidrngr(s->stack_time[s->stack_top], s->stack_time[s->stack_top + 2] = s->stack_time[s->stack_top + 1]);
	s->stack_top += 2;
}

real geom_bezier_nonrational_error(struct geom_bezier *bezier);


/*----------------------------------------------*
 *                                              *
 *                 INTERSECTIONS                *
 *                                              *
 *----------------------------------------------*/


/**
 * Finds intersections of two given Bezier curves.
 *
 * Intersections (geom_intersection structures) are added to current
 * position of a growing array structure.
 *
 * @ai  bezier1  First Bezier curve.
 * @ai  bezier2  Second Bezier curve.
 * @ai  flags    Flags.
 * @aio result   Growing array, where to store found intersections.
 *
 * @rv  >=0               Number of found intersections.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 *
 * @todo Better code.
 */
int geom_beziers_intersections(struct geom_bezier *bezier1, struct geom_bezier *bezier2, uns flags, struct garr *result);






/* FIXME: santiaguv interface */

static inline void geom_bezier_fill_linear_nonrational(struct geom_bezier *bezier, real x0, real y0, real x1, real y1)
{
	geom_bezier_set_linear_nonrational(bezier, &(struct geom_point){.x=x0, .y=y0}, &(struct geom_point){.x=x1,.y=y1});
}




/* FIXME: CHTELO BY PROBRAT... */

/*----------------------------------------------*
 *                                              *
 *            RECURSIVE SUBDIVISION             *
 *                                              *
 *----------------------------------------------*/


/**
 * Macro implements recursive subdivision of Bezier curve.
 *
 * Given condition statement is evaluated in each node of subdivision tree
 * to decide whether subdivide or not. Main code is executed only for leaf nodes.
 * Description of current node is stored in varriables depth, left, right and part.
 *
 * Usage:
 * \code
 * struct geom_bezier *bezier = ???;
 *
 * GEOM_BEZIER_SUBDIVIDE_BEGIN(walk1, bezier, part, depth, left, right, 3, 1) {
 *
 * 	msg("Depth of subdivision = %d, current interval = (%f, %f)\n", depth, left, right);
 *	geom_bezier_dump(part);
 *
 * } GEOM_BEZIER_SUBDIVIDE_END(walk1);
 * \endcode
 *
 * @ai  name       Unique prefix in current function (used to generate internal identifiers).
 * @ai  bezier     Pointer to bezier curve to subdivide.
 * @ao  part       Current bezier subinterval.
 * @ao  depth      Current subdivision depth (0 for entire curve).
 * @ao  left       Current left TIME subinterval limit.
 * @ao  right      Current right TIME subinterval limit.
 * @ai  max_depth  Maximum subdivision depth (unsigned int).
 * @ai  test       Condition - nonzero value cause subdivision of current curve.
 */

#define GEOM_BEZIER_SUBDIVIDE_BEGIN(name, bez, part, depth, left, right, max_depth, test) \
	do{ struct geom_bezier *part, name##_s[(max_depth) + 2]; \
		uns depth, name##_t, name##_d[(max_depth) + 2]; \
		real left, right, name##_l[(max_depth) + 2]; \
		GEOM_ITEM_ASSERT(bezier, bez); \
		GEOM_INSTANCE_CREATE2(&name##_s[1], bezier, geom_bezier_copy, bez); \
		name##_t = 1; \
		name##_d[1] = 0; name##_l[1] = 0; name##_l[0] = 1; \
		while (name##_t) { \
			depth = name##_d[name##_t]; \
			part = &name##_s[name##_t]; \
			left = name##_l[name##_t]; \
			right = name##_l[name##_t - 1]; \
			if (depth < max_depth && !(test)) { \
				GEOM_INSTANCE_INIT(&name##_s[name##_t + 1], bezier); \
				geom_bezier_split_middle(&name##_s[name##_t], \
					&name##_s[name##_t + 1], &name##_s[name##_t]); \
				name##_d[name##_t + 1] = ++name##_d[name##_t]; \
				if (!isfinite(name##_l[name##_t] = 0.5 * (name##_l[name##_t - 1] + \
					(name##_l[name##_t + 1] = name##_l[name##_t])))) \
					name##_l[name##_t] = name##_l[name##_t - 1]; \
				name##_t++; } \
			else { GEOM_INSTANCE_DESTROY2(&name##_s[name##_t], bezier); \
				name##_t--; \

#define GEOM_BEZIER_SUBDIVIDE_END(name) }} name##_break:; if (0) goto name##_break; }while(0)
#define GEOM_BEZIER_SUBDIVIDE_BREAK(name) goto name##_break;


/*----------------------------------------------*
 *                                              *
 *                                              *
 *                                              *
 *----------------------------------------------*/



/** Flag used in geom_bezier_direction_times. */
#define GEOM_BEZIER_DIRECTION_NOBACK 0x00000100


#define GEOM_BEZIER_EXTREME_MIN      0x00000001
#define GEOM_BEZIER_EXTREME_MAX      0x00000002
#define GEOM_BEZIER_EXTREME_GLOBAL   0x00000004
#define GEOM_BEZIER_EXTREME_LOCAL    0x00000000
#define GEOM_BEZIER_EXTREME_LIMITS   0x00000008
#define GEOM_BEZIER_EXTREME_ALL      (GEOM_BEZIER_EXTREME_MIN | GEOM_BEZIER_EXTREME_MAX)

/**
	\brief Function finds parameters, where the curve is parallel to a given vector.
		If the whole curve is parallel to v or v is zero vector, no values are returned.
		Otherwise, there are at most 1, 3, 2 or 5 solutions for quadratic nonrational,
		quadratic rational, cubic nonrational or cubic rational curves respectively.
		We can also set flag GEOM_SOLVE_LEFT_ONLY (return only left most solution, if any) and/or
		GEOM_BEZIER_DIRECTION_NOBACK (skip solutions with negative derivation in the vector direction).

	\param b pointer to the bezier curve
	\param v pointer to the vector
	\param flags OR combination of GEOM_SOLVE_LEFT_ONLY and GEOM_BEZIER_DIRECTION_NOBACK
	\param result buffer to store result parameters (needed size in brief description, 1 with flag GEOM_SOLVE_LEFT_ONLY)

	\return 0 always
*/
int geom_bezier_direction_times(struct geom_bezier *b, struct geom_point *v, uns flags, real result[]);


/**
    \brief Function splits a bezier curve at parameter 0.5 and returns its two parts.

    \param b pointer to the bezier curve
    \param left buffer for the left curve part (interval <0, 0.5>)
    \param right buffer for the right curve part (interval <0.5, 1>)

	\return 0 always
*/


/**
    \brief Function splits a bezier curve at given parameter t and returns its two parts.
		Parameters outside the range <0, 1> are aligned to this interval.

    \param b pointer to the bezier curve
    \param t curve parameter
    \param left buffer for the left curve part (interval <0, t>)
    \param right buffer for the right curve part (interval <t, 1>)

	\return 0 always
*/
int geom_bezier_split_at(struct geom_bezier *b, real t, struct geom_bezier *left, struct geom_bezier *right);


/**
    \brief Function returns the same bezier curve, but with reversed parametrization.

    \param b pointer to the bezier curve
    \param result buffer for the reversed curve

    \return 0 always
*/
int geom_bezier_reverse(struct geom_bezier *b, struct geom_bezier *result);


/**
    \brief Function returns part of bezier curve given by an interval <t, s>.
		If t > s, parametrization of resulting path is reversed.
		Parameters outside the range <0, 1> are aligned to this interval.

    \param b pointer to the bezier curve
    \param t curve parameter
    \param s curve parameter
    \param result buffer for the result curve (subinterval <t, s> of original curve)

	\return 0 always
*/
int geom_bezier_part(struct geom_bezier *b, double t, double s, struct geom_bezier *result);


/**
	\brief Function increases bezier curve degree to degree. Resulting curve has the same
		parametrization as the original.

	\param b pointer to the bezier curve
	\param degree wanted degree, b->degree <= degree <= GEOM_BEZIER_CUBIC
	\param result buffer for the resulting curve

	\return 0 always
*/
int geom_bezier_elevate(struct geom_bezier *b, uns degree, struct geom_bezier *result);


/**
	\brief Function returns bounding box of given bezier curve control points.
		Usually, this region is larger than the exact curve bounding box, but
		is easier to compute.

	\param b pointer to the bezier curve
	\param bbox buffer for bounding box

	\return 0 always
*/
int geom_bezier_control_box(struct geom_bezier *b, struct geom_rectangle *bbox);


/**
    \brief Function returns parameter of a given point very close to curve.

    \param b pointer to the bezier curve
    \param p pointer to the point

    \return curve parameter (NAN for numeric error)
*/
double geom_bezier_time_of(struct geom_bezier *b, struct geom_point *p);


/**
	\brief Function returns sorted list of parameters, where the curve crosses given line.
		If the whole curve is on the line or v is zero vector, no intersections are returned,
		otherwise, there are at most curve degree intersections.
		We can also set flags GEOM_SOLVE_LEFT_ONLY (only left most parameter returned, if any)
		or GEOM_SOLVE_MULTIPLICITY (higher degree intersections are returned as more equal values).

	\param b pointer to the bezier curve
	\param a a point on the line
	\param v a vector in the line
	\param flags OR combination of GEOM_SOLVE_LEFT_ONLY and GEOM_SOLVE_MULTIPLICITY
	\param result buffer to store result parameters (of curve degree size or 1 with flag GEOM_SOLVE_LEFT_ONLY)

	\return number of intersections found (-1 for numeric overflow)
*/
//int geom_bezier_line_intersection_times(struct geom_bezier *b, struct geom_point *a, struct geom_point *v, uns flags, double result[]);


/**
	\brief Function returns sorted list of parameters, where has the curve given x coordinate.
		If the whole curve is parallel to y axis, no intersections are returned,
		otherwise, there are at most curve degree intersections.
		We can also set flags GEOM_SOLVE_LEFT_ONLY (only left most parameter returned, if any)
		or GEOM_SOLVE_MULTIPLICITY (higher degree intersections are returned as more equal values).

	\param b pointer to the bezier curve
	\param x x coordinate
	\param flags OR combination of GEOM_SOLVE_LEFT_ONLY and GEOM_SOLVE_MULTIPLICITY
	\param result buffer to store result parameters (of curve degree size or 1 with flag GEOM_SOLVE_LEFT_ONLY)

	\return number of intersections found (-1 for numeric overflow)
*/
/*static inline int geom_bezier_x_intersection_times(struct geom_bezier *b, real x, uns flags, double result[])
{
	struct geom_point point; point.x = x; point.y = 0;
	return geom_bezier_line_intersection_times(b, &point, &geom_point_up, flags, result);
}*/


/**
	\brief Function returns sorted list of parameters, where has the curve given y coordinate.
		If the whole curve is parallel to x axis, no intersections are returned,
		otherwise, there are at most curve degree intersections.
		We can also set flags GEOM_SOLVE_LEFT_ONLY (only left most parameter returned, if any)
		or GEOM_SOLVE_MULTIPLICITY (higher degree intersections are returned as more equal values).

	\param b pointer to the bezier curve
	\param y y coordinate
	\param flags OR combination of GEOM_SOLVE_LEFT_ONLY and GEOM_SOLVE_MULTIPLICITY
	\param result buffer to store result parameters (of curve degree size or 1 with flag GEOM_SOLVE_LEFT_ONLY)

	\return number of intersections found (-1 for numeric overflow)
*/
/*static inline int geom_bezier_y_intersection_times(struct geom_bezier *b, const real y, uns flags, double result[])
{
	struct geom_point point; point.x = 0; point.y = y;
	return geom_bezier_line_intersection_times(b, &point, &geom_point_right, flags, result);
}*/


/*
	TODO: jeste si porad nejsem jisty, co to ma vracet pro vicenasobne pruseciky,
		tak	az to dodelam, sepisu dokumentaci... Pavel
*/
int geom_bezier_intersection_times(struct geom_bezier *b1, struct geom_bezier *b2, uns flags, double result1[], double result2[]);


/**
	\brief Function finds extreme points on bezier curve. Used metric is distance from given center point.
		It is possible to find global (flag GEOM_BEZIER_EXTREME_GLOBAL) or all local extremes (GEOM_BEZIER_EXTREME_LOCAL).
		With GEOM_BEZIER_EXTREME_MIN set, result contains global/local minimum(s),
		with GEOM_BEZIER_EXTREME_MAX maximum(s) and with GEOM_BEZIER_EXTREME_ALL both minimum(s) and maximum(s).


	\param b pointer to the bezier curve
	\param p pointer to center point
	\param flags special flags (see function brief description)
	\param result buffer for resulting parameters

	\return number of found extreme points (-1 for numeric overflow)
*/
int geom_bezier_center_extreme_times(struct geom_bezier *b, struct geom_point *p, uns flags, double result[]);


/**
	\brief Function finds extreme points on bezier curve. Used metric is distance from given center point.
		It is possible to find global (flag GEOM_BEZIER_EXTREME_GLOBAL) or all local extremes (GEOM_BEZIER_EXTREME_LOCAL).
		With GEOM_BEZIER_EXTREME_MIN set, result contains global/local minimum(s),
		with GEOM_BEZIER_EXTREME_MAX maximum(s) and with GEOM_BEZIER_EXTREME_ALL both minimum(s) and maximum(s).

	\param b pointer to the bezier curve
	\param p pointer to center point
	\param flags special flags (see function brief description)
	\param result buffer for resulting points

	\return number of found extreme points (-1 for numeric overflow)
*/
int geom_bezier_center_extreme_points(struct geom_bezier *b, struct geom_point *p, uns flags, struct geom_point result[]);


/**
	\brief Function finds extreme points on bezier curve. Used metric is distance in given direction vector.
		It is possible to find global (flag GEOM_BEZIER_EXTREME_GLOBAL) or all local extremes (GEOM_BEZIER_EXTREME_LOCAL).
		With GEOM_BEZIER_EXTREME_MIN set, result contains global/local minimum(s),
		with GEOM_BEZIER_EXTREME_MAX maximum(s) and with GEOM_BEZIER_EXTREME_ALL both minimum(s) and maximum(s).

	\param b pointer to the bezier curve
	\param v pointer to direction vector
	\param flags special flags (see function brief description)
	\param result buffer for resulting parameters

	\return number of found extreme points (-1 for numeric overflow)
*/
int geom_bezier_direction_extreme_times(const struct geom_bezier *b, const struct geom_point *v, uns flags, double result[]);


/**
	\brief Function finds extreme points on bezier curve. Used metric is distance in given direction vector.
		It is possible to find global (flag GEOM_BEZIER_EXTREME_GLOBAL) or all local extremes (GEOM_BEZIER_EXTREME_LOCAL).
		With GEOM_BEZIER_EXTREME_MIN set, result contains global/local minimum(s),
		with GEOM_BEZIER_EXTREME_MAX maximum(s) and with GEOM_BEZIER_EXTREME_ALL both minimum(s) and maximum(s).

	\param b pointer to the bezier curve
	\param v pointer to direction vector
	\param flags special flags (see function brief description)
	\param result buffer for resulting points

	\return number of found extreme points (-1 for numeric overflow)
*/
int geom_bezier_direction_extreme_points(struct geom_bezier *b, const struct geom_point *v, uns flags, struct geom_point result[]);


/**
	\brief Function returns approximation of bezier curve as a sequence of points
		(returned in buffer as x, y pairs) with error limited by delta

	\param b pointer to the curve
	\param delta maximal error
	\param buffer result buffer, if *NULL, buffer is newly allocated
	\param buffer_size buffer size (in doubles), if NULL, buffer is newly allocated and size not returned
	\param points number of computed points

	\retval 0 success
	\retval -1 error occured
*/

int geom_bezier_approximate(struct geom_bezier *b, double delta, double **buffer, int *buffer_size, int *points);


/* GEOM_BEZIER_PARAM_TO_CUBIC_NONRATIONAL */

static inline int geom_bezier_cubic_nonrational_enough(struct geom_bezier *b)
{
	return (fabs(b->pt[1].w - b->pt[0].w) + fabs(b->pt[2].w - b->pt[1].w) +
			fabs(b->pt[3].w - b->pt[2].w) < 0.1);
}

#define GEOM_BEZIER_PARAM_TO_CUBIC_NONRATIONAL_BEGIN(_name, _b, _left, _right, _part) \
	{ struct geom_bezier _name##_cub; \
		geom_bezier_elevate((_b), GEOM_BEZIER_CUBIC, &_name##_cub); \
		GEOM_BEZIER_SUBDIVIDE_BEGIN(_name##_sub, &_name##_cub, _part, _name##_deep, _left, _right, 4, \
			geom_bezier_cubic_nonrational_enough(_part)) { \
			_part->pt[0].w = _part->pt[1].w = _part->pt[2].w = _part->pt[3].w = 1; \
			_part->flags |= GEOM_BEZIER_NONRATIONAL;

#define GEOM_BEZIER_PARAM_TO_CUBIC_NONRATIONAL_END(_name) \
	} GEOM_BEZIER_SUBDIVIDE_END(_name##_sub); } \

int geom_bezier_to_cubic_nonrational(struct geom_bezier *b);


/* GEOM_BEZIER_SHAPE_TO_CUBIC_NONRATIONAL */

#define GEOM_BEZIER_SHAPE_TO_CUBIC_NONRATIONAL_BEGIN(_name, _b, _left, _right, _part) \
	GEOM_BEZIER_PARAM_TO_CUBIC_NONRATIONAL_BEGIN(_name, _b, _left, _right, _part)

#define GEOM_BEZIER_SHAPE_TO_CUBIC_NONRATIONAL_END(_name) \
	GEOM_BEZIER_PARAM_TO_CUBIC_NONRATIONAL_END(_name)



#ifdef GEOM_SOURCE
#define BEZIER_DEGREE(b) GEOM_BEZIER_DEGREE(b)
#endif


#endif

