#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/bezier.h"


/*----------------------------------------------*
 *                                              *
 *               PARAMETRIZATIONS               *
 *                                              *
 *----------------------------------------------*/

/* FIXME:
 * - neracionalni kvadraticke bezierky by se daly pocitat vzoreckem (delka useku paraboly)
 * - racionalni mozna nakou jednodussi aproximaci (delka elipsy/hyperboly... vzorecek neexistuje)
 * - pro neracionalni kubicke vzorecek neexistuje
 * - chybi kontrola epsilonove chyby na konci intervalu... atime / alength, atime <= alength + epsilon
 * */


/* subdivision options */

static real geom_bezier_arc_length_tolerance = 1e-3; /* subdivision tolerance, lower value = slower/better results (0 <= tolerance <= 1) */


/**
 * Returns length of chord of given bezier curve.
 */
inline static real geom_bezier_chord(struct geom_bezier *b)
{
	return hypot(b->pt[0].x - b->pt[BEZIER_DEGREE(b)].x, b->pt[0].y - b->pt[BEZIER_DEGREE(b)].y);
}


/**
 * Returns length of control polyline of given bezier curve.
 */
inline static real geom_bezier_polyline(struct geom_bezier *b)
{
	real r = 0;
	for (uns i = 0; i < BEZIER_DEGREE(b); i++)
		r += hypot(b->pt[i].x - b->pt[i + 1].x, b->pt[i].y - b->pt[i + 1].y);
	return r;
}


/**
 * Returns nonzero value if bezier curve is very near to straight line (chord and polyline same length).
 */
inline static int geom_bezier_straight_enough(real chord, real poly)
{
	return (poly == 0 || chord / poly >= 1 - geom_bezier_arc_length_tolerance) ? 1 : 0;
}


/**
 * Returns nonzero valie if bezier curve is very near to a linear function of parameter.
 */
static int geom_bezier_linear_enough(struct geom_bezier *b)
{
	int i;
	double dx, dy, t, s, u, v;

	/* weights must be almost equal */
	u = 1 - geom_bezier_arc_length_tolerance;
	v = 1 + geom_bezier_arc_length_tolerance;
	if (!(INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL)) {
		t = 1 / b->pt[0].w;
		if (!isfinite(t))
			return 0;
		for (i = 1; i <= BEZIER_DEGREE(b); i++) {
			s = t * b->pt[i].w;
			if (!isfinite(s) || s < u || s > v)
				return 0;
		}
	}

	/* control polyline parts must be almost equal */
	dx = b->pt[0].x - b->pt[1].x;
	dy = b->pt[0].y - b->pt[1].y;
	t = dx * dx + dy * dy;
	if (!isfinite(t))
		return 0;
	else if (t == 0)
		for (i = 2; i <= BEZIER_DEGREE(b); i++) {
			if (b->pt[i].x != b->pt[0].x || b->pt[i].y != b->pt[0].y)
				return 0;
		}
	else
	{
		t = 1 / t;
		for (i = 1; i < BEZIER_DEGREE(b); i++) {
			dx = b->pt[i].x - b->pt[i + 1].x;
			dy = b->pt[i].y - b->pt[i + 1].y;
			s = t * (dx * dx + dy * dy);
			if (!isfinite(s) || s < u || s > v)
				return 0;
		}
	}

	return 1;
}


/* GEOM_BEZIER_ALENGTH_SUBDIVIDE */
/* FIXME: co de, tak pocitat v doublech */

#define MAX_DEPTH 10
#define GEOM_BEZIER_ALENGTH_SUBDIVIDE_BEGIN(name, bez) \
	do{ struct geom_bezier *part, stack_b[MAX_DEPTH + 2]; \
		uns top1, stack_d[MAX_DEPTH + 2]; \
		real left_time, right_time, left_atime, right_atime, poly, chord, alength = 0, stack_l[MAX_DEPTH + 2]; \
		ITEM_ASSERT(bezier, bez); \
		GEOM_INSTANCE_CREATE2(&stack_b[1], bezier, geom_bezier_copy, bez); \
		top1 = 1; \
		stack_d[1] = 0; stack_l[1] = 0; stack_l[0] = 1; \
		while (top1) { \
			part = &stack_b[top1]; \
			if (!isfinite(chord = geom_bezier_chord(part))) goto name##_error; \
			if (stack_d[top1] < MAX_DEPTH && !(poly = geom_bezier_polyline(part), \
						isfinite(poly) && geom_bezier_straight_enough(chord, poly))) { \
				GEOM_INSTANCE_INIT(&stack_b[top1 + 1], bezier); \
				geom_bezier_split_middle(&stack_b[top1], &stack_b[top1 + 1], &stack_b[top1]); \
				stack_d[top1 + 1] = ++stack_d[top1]; \
				if (!isfinite(stack_l[top1] = 0.5 * (stack_l[top1 - 1] + (stack_l[top1 + 1] = stack_l[top1])))) \
					stack_l[top1] = stack_l[top1 - 1]; \
				top1++; } \
			else { \
				left_time = stack_l[top1]; right_time = stack_l[top1 - 1]; \
				if (!isfinite(right_atime = alength = (left_atime = alength) + chord)) goto name##_error; \
				GEOM_INSTANCE_DESTROY2(&stack_b[top1], bezier); \
				top1--;

#define GEOM_BEZIER_ALENGTH_SUBDIVIDE_END(name) }} if (0) { goto name##_error; name##_error: alength = NAN; } \
	bezier->alength = alength; INSTANCE_FLAGS(bezier) |= GEOM_BEZIER_ALENGTH_VALID; }while(0)


/* GEOM_BEZIER_ATIME_SUBDIVIDE */

#define GEOM_BEZIER_ATIME_SUBDIVIDE_BEGIN(name, bez, test) \
	do { uns top2; \
		real chord1, chord2, coef, stack_a[MAX_DEPTH + 2]; \
		GEOM_BEZIER_ALENGTH_SUBDIVIDE_BEGIN(name, bez) { \
			stack_a[top1] = right_atime; stack_a[top2 = top1 + 1] = left_atime; \
			while (top2 > top1) { \
				part = &stack_b[top2]; \
				left_time = stack_l[top2]; right_time = stack_l[top2 - 1]; \
				left_atime = stack_a[top2]; right_atime = stack_a[top2 - 1]; \
				if (stack_d[top2] < MAX_DEPTH && !((test) && geom_bezier_linear_enough(part))) { \
					GEOM_INSTANCE_INIT(&stack_b[top2 + 1], bezier); \
					geom_bezier_split_middle(&stack_b[top2], &stack_b[top2 + 1], &stack_b[top2]); \
					stack_d[top2 + 1] = ++stack_d[top2]; \
					if (!isfinite(stack_l[top2] = 0.5 * (stack_l[top2 - 1] + (stack_l[top2 + 1] = stack_l[top2])))) \
						stack_l[top2] = stack_l[top2 - 1]; \
					chord1 = geom_bezier_chord(&stack_b[top2 + 1]); \
					chord2 = geom_bezier_chord(&stack_b[top2]); \
					coef = chord2 / (chord1 + chord2); if (!isfinite(coef)) coef = 0.5; \
					if (!isfinite(stack_a[top2] = stack_a[top2 - 1] + coef * (-stack_a[top2 - 1] +\
						(stack_a[top2 + 1] = stack_a[top2])))) stack_a[top2] = stack_a[top2 - 1]; \
					top2++; } \
				else { GEOM_INSTANCE_DESTROY2(&stack_b[top2], bezier); \
					top2--;

#define GEOM_BEZIER_ATIME_SUBDIVIDE_END(name) }}} GEOM_BEZIER_ALENGTH_SUBDIVIDE_END(name); }while(0)


/* geom_bezier_update_linear_alength */

static inline void geom_bezier_update_linear_alength(struct geom_bezier *bezier)
{
	bezier->alength = hypot(
		(double)bezier->pt[0].x - (double)bezier->pt[1].x,
		(double)bezier->pt[0].y - (double)bezier->pt[1].y);
	INSTANCE_FLAGS(bezier) |= GEOM_BEZIER_ALENGTH_VALID;
}


/* geom_bezier_alength */

real geom_bezier_alength(struct geom_bezier *bezier)
{
	FUNC;
	ITEM_ASSERT(bezier, bezier);

	if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID))
		switch (BEZIER_DEGREE(bezier)) {
			case GEOM_BEZIER_LINEAR:
				geom_bezier_update_linear_alength(bezier);
				break;

			default:
				GEOM_BEZIER_ALENGTH_SUBDIVIDE_BEGIN(sub1, bezier) {
				} GEOM_BEZIER_ALENGTH_SUBDIVIDE_END(sub1);
				break;
		}

	RRET(bezier->alength);
}


/* geom_bezier_time_to_atime */

real geom_bezier_time_to_atime(struct geom_bezier *bezier, real time)
{
	FUNC;
	ITEM_ASSERT(bezier, bezier);
	RASSERT(time);

	if (time <= 0)
		RRET(0);

	switch (BEZIER_DEGREE(bezier)) {
		case GEOM_BEZIER_LINEAR:
			if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID))
				geom_bezier_update_linear_alength(bezier);
			if (time >= 1)
				RRET(bezier->alength);
			else if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)
				RRET(time * bezier->alength);
			else
				RRET(bezier->alength *
					(((double)time * bezier->pt[1].w) / (bezier->pt[0].w + time * ((double)bezier->pt[1].w - bezier->pt[0].w))));

		default: ;
			real atime = 0; // assign not necessary
			int found = 0;
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID && (time >= 1 || !RTEST(bezier->alength)))
				RRET(bezier->alength);
			GEOM_BEZIER_ATIME_SUBDIVIDE_BEGIN(sub1, bezier, !found && time < right_time) {
				if (!found && time < right_time) {
					found = 1;
					if (!RTEST(atime = left_atime + (right_atime - left_atime) * (time - left_time) / (right_time - left_time)))
						atime = left_atime;
				}
			} GEOM_BEZIER_ATIME_SUBDIVIDE_END(sub1);
			RRET((RTEST(bezier->alength) && found) ? atime : bezier->alength);
	}
}


/* geom_bezier_times_to_atimes */

int geom_bezier_times_to_atimes(struct geom_bezier *bezier, uns count, real *times, real *atimes)
{
	uns l, r, i;
	real t, s;

	FUNC;
	ITEM_ASSERT(bezier, bezier);
	GEOM_REALS_SORTED_ASSERT(count, times);
	PASSERT(atimes);

	l = 0;
	r = count;
	while (l < r && times[l] <= 0)
		atimes[l++] = 0;
	if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID) {
		RTEST_NUM(bezier->alength);
		while (l < r && times[r - 1] >= 1)
			atimes[--r] = bezier->alength;
	}
	if (l == r)
		goto EXIT;

	switch (BEZIER_DEGREE(bezier)) {
		case GEOM_BEZIER_LINEAR:
			if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID))
				geom_bezier_update_linear_alength(bezier);
			RTEST_NUM(bezier->alength);
			while (l < r && times[r - 1] >= 1)
				atimes[--r] = bezier->alength;
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)
				for (i = l; i < r; i++)
					RASSERTX(atimes[i] = times[i] * bezier->alength);
			else
				for (i = l; i < r; i++)
					RTEST_NUM(atimes[i] = bezier->alength *
						(((double)times[i] * bezier->pt[1].w) / (bezier->pt[0].w + times[i] * ((double)bezier->pt[1].w - bezier->pt[0].w))));
			break;

		default: ;
			i = l;
			GEOM_BEZIER_ATIME_SUBDIVIDE_BEGIN(sub1, bezier, i < r && times[i] < right_time) {
				for (; i < r && times[i] < right_time; i++)
					if (!RTEST(atimes[i] = left_atime + (right_atime - left_atime) * (times[i] - left_time) / (right_time - left_time)))
						atimes[i] = left_atime;
			} GEOM_BEZIER_ATIME_SUBDIVIDE_END(sub1);
			RTEST_NUM(bezier->alength);
			while (i < r)
				atimes[--r] = bezier->alength;
			break;
	}
	for (t = 0, i = l; i < r; i++) {
		s = likely(atimes[i] <= bezier->alength) ? atimes[i] : bezier->alength;
		if (likely(s >= t)) t = s;
		atimes[i] = t;
	}
EXIT: ;
	GEOM_REALS_SORTED_ASSERT(count, atimes);
	ERET(0);
}


/* geom_bezier_atime_to_time */

real geom_bezier_atime_to_time(struct geom_bezier *bezier, real atime)
{
	FUNC;
	ITEM_ASSERT(bezier, bezier);
	RASSERT(atime);

	if (atime <= 0)
		RRET(0);

	switch (BEZIER_DEGREE(bezier)) {
		case GEOM_BEZIER_LINEAR:
			if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID))
				geom_bezier_update_linear_alength(bezier);
			if (atime >= bezier->alength)
				RRET(1);
			else if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)
				RRET(atime / bezier->alength);
			else {
				double s = atime / bezier->alength;
				RRET((s * bezier->pt[0].w) / (bezier->pt[1].w + s * ((double)bezier->pt[0].w - bezier->pt[1].w)));
			}

		default: ;
			real time = 0; // assign not necessary
			int found = 0;
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID && (atime >= bezier->alength || !RTEST(bezier->alength)))
				RRET(1);
			GEOM_BEZIER_ATIME_SUBDIVIDE_BEGIN(sub1, bezier, !found && atime < right_atime) {
				if (!found && atime < right_atime) {
					found = 1;
					if (!RTEST(time = left_time + (right_time - left_time) * (atime - left_atime) / (right_atime - left_atime)))
						time = left_time;
				}
			} GEOM_BEZIER_ATIME_SUBDIVIDE_END(sub1);
			if (!RTEST(bezier->alength))
				RET_NAN;
			else
				RRET(found ? time : 1);
	}
}


/* geom_bezier_atimes_to_times */

int geom_bezier_atimes_to_times(struct geom_bezier *bezier, uns count, real atimes[], real times[])
{
	uns l, r, i;
	real t, s;

	FUNC;
	ITEM_ASSERT(bezier, bezier);
	GEOM_REALS_SORTED_ASSERT(count, atimes);
	PASSERT(times);

	l = 0;
	r = count;
	while (l < r && atimes[l] <= 0)
		times[l++] = 0;
	if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID) {
		RTEST_NUM(bezier->alength);
		while (l < r && atimes[r - 1] >= bezier->alength)
			times[--r] = 1;
	}
	if (l == r)
		goto EXIT;

	switch (BEZIER_DEGREE(bezier)) {
		case GEOM_BEZIER_LINEAR:
			if (!(INSTANCE_FLAGS(bezier) & GEOM_BEZIER_ALENGTH_VALID))
				geom_bezier_update_linear_alength(bezier);
			RTEST_NUM(bezier->alength);
			while (l < r && atimes[r - 1] >= bezier->alength)
				times[--r] = 1;
			if (INSTANCE_FLAGS(bezier) & GEOM_BEZIER_NONRATIONAL)
				for (i = l; i < r; i++)
					RASSERTX(times[i] = atimes[i] / bezier->alength);
			else
				for (i = l; i < r; i++) {
					double s = atimes[i] / bezier->alength;
					RTEST_NUM(times[i] = (s * bezier->pt[0].w) /
						(bezier->pt[1].w + s * ((double)bezier->pt[0].w - bezier->pt[1].w)));
				}
			break;

		default: ;
			i = l;
			GEOM_BEZIER_ATIME_SUBDIVIDE_BEGIN(sub1, bezier, i < r && atimes[i] < right_atime) {
				for (; i < r && atimes[i] < right_atime; i++)
					if (!RTEST(times[i] = left_time + (right_time - left_time) * (atimes[i] - left_atime) / (right_atime - left_atime)))
						times[i] = left_time;
			} GEOM_BEZIER_ATIME_SUBDIVIDE_END(sub1);
			RTEST_NUM(bezier->alength);
			while (i < r)
				times[--r] = 1;
			break;
	}
	for (t = 0, i = l; i < r; i++) {
		s = likely(times[i] <= 1) ? times[i] : 1;
		if (likely(s >= t)) t = s;
		times[i] = t;
	}
EXIT:
	GEOM_REALS_SORTED_ASSERT(count, times);
	ERET(0);
}


