#define GEOM_SOURCE
#include "geomlib/polynomial.h"
#include "geomlib/bernstein.h"
#include "geomlib/bezier.h"


/* geom_bezier_time_of
 * Function returns parameter of point very close to given curve. */

double geom_bezier_time_of(struct geom_bezier *b, struct geom_point *p)
{
	struct geom_nearest nearest;
	if (geom_bezier_nearest_to_point(b, p, GEOM_NEAREST_TIME, &nearest) < 0)
		return 0;
	return nearest.time;
}


/* geom_bezier_center_extreme_times  */

int geom_bezier_center_extreme_times(struct geom_bezier *b, struct geom_point *p, uns flags, double result[])
{
	double x[4], y[4], w[4], a[70], c[9], v[10];
	int n, i, j, k, m, t[10];
	struct geom_point q;

	ITEM_ASSERT(bezier, b);
	GASSERT(point, p);
	PASSERT(result);

	/* compute bernstein coefficients */

	if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL || BEZIER_DEGREE(b) == GEOM_BEZIER_LINEAR)
		for (i = 0; i <= BEZIER_DEGREE(b); i++) {
			x[i] = (b->pt[i].x - p->x);
			y[i] = (b->pt[i].y - p->y);
	} else
		for (i = 0; i <= BEZIER_DEGREE(b); i++) {
			w[i] = b->pt[i].w;
			x[i] = (b->pt[i].x - p->x) * w[i];
			y[i] = (b->pt[i].y - p->y) * w[i];
		}

	switch (BEZIER_DEGREE(b)) {
	case GEOM_BEZIER_LINEAR:	/* linear curve */

		a[0] = SQR(x[0]) + SQR(y[0]);
		a[1] = x[0] * x[1] + y[0] * y[1];
		a[2] = SQR(x[1]) + SQR(y[1]);

		if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL) {
			n = 1;

			c[0] = a[1] - a[0];
			c[1] = a[2] - a[1];
		} else {
			n = 2;

			c[0] = (a[1] - a[0]) * b->pt[0].w;
			c[2] = (a[2] - a[1]) * b->pt[1].w;
			c[1] = 1.0 / 2 * (c[0] + c[2]);
		}
		break;

	case GEOM_BEZIER_QUADRATIC:	/* quadratic curve */

		a[0] = SQR(x[0]) + SQR(y[0]);
		a[1] = x[0] * x[1] + y[0] * y[1];
		a[2] = 2.0 * (SQR(x[1]) + SQR(y[1])) + (x[0] * x[2] + y[0] * y[2]);
		a[3] = x[1] * x[2] + y[1] * y[2];
		a[4] = SQR(x[2]) + SQR(y[2]);

		if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL) {
			n = 3;

			a[2] = 1.0 / 3 * a[2];

			c[0] = a[1] - a[0];
			c[1] = a[2] - a[1];
			c[2] = a[3] - a[2];
			c[3] = a[4] - a[3];
		} else {
			n = 5;

			c[0] = 5.0 * (a[1] * w[0] - a[0] * w[1]);
			c[1] = a[1] * (w[0] - 2 * w[1]) - a[0] * (w[1] + w[2]) + a[2] * w[0];
			c[2] = -1.0 / 2 * a[0] * w[2] - a[1] * (w[1] + 3.0 / 2 * w[2]) + (1.0 / 2 * a[2] + 3.0 / 2 * a[3]) * w[0];
			c[3] = 1.0 / 2 * a[4] * w[0] + a[3] * (w[1] + 3.0 / 2 * w[0]) - (1.0 / 2 * a[2] + 3.0 / 2 * a[1]) * w[2];
			c[4] = a[3] * (2.0 * w[1] - w[2]) + a[4] * (w[0] + w[1]) - a[2] * w[2];
			c[5] = 5.0 * (a[4] * w[1] - a[3] * w[2]);
		}
		break;

	default: /* cubic curve */

		if (INSTANCE_FLAGS(b) & GEOM_BEZIER_NONRATIONAL) {
			n = 5;

			a[0] = SQR(x[0]) + SQR(y[0]);
			a[1] = x[0] * x[1] + y[0] * y[1];
			a[2] = 3.0 / 5 * (SQR(x[1]) + SQR(y[1])) + 2.0 / 5 * (x[0] * x[2] + y[0] * y[2]);
			a[3] = 1.0 / 10 * (x[0] * x[3] + y[0] * y[3]) + 9.0 / 10 * (x[1] * x[2] + y[1] * y[2]);
			a[4] = 3.0 / 5 * (SQR(x[2]) + SQR(y[2])) + 2.0 / 5 * (x[1] * x[3] + y[1] * y[3]);
			a[5] = x[2] * x[3] + y[2] * y[3];
			a[6] = SQR(x[3]) + SQR(y[3]);

			c[0] = a[1] - a[0];
			c[1] = a[2] - a[1];
			c[2] = a[3] - a[2];
			c[3] = a[4] - a[3];
			c[4] = a[5] - a[4];
			c[5] = a[6] - a[5];
		} else {
			n = 8;

			/* we need to solve polynome
			   x*(w*dx-x*dw)+y*(w*dy-y*dw) = 0 */

			if (geom_bernstein_derivate(3, x, a + 0) < 0 ||	/* dx d2 */
			    geom_bernstein_derivate(3, y, a + 3) < 0 ||	/* dy d2 */
			    geom_bernstein_derivate(3, w, a + 6) < 0 ||	/* dw d2 */
			    geom_bernstein_mul(3, x, 2, a + 6, a + 9) < 0 ||	/* x*dw d5 */
			    geom_bernstein_mul(3, y, 2, a + 6, a + 15) < 0 ||	/* y*dw d5 */
			    geom_bernstein_mul(3, w, 2, a + 0, a + 21) < 0 ||	/* w*dx d5 */
			    geom_bernstein_mul(3, w, 2, a + 3, a + 27) < 0 ||	/* w*dy d5 */
			    geom_bernstein_sub(5, a + 21, a + 9, a + 33) < 0 ||	/* w*dx-x*dw d5 */
			    geom_bernstein_sub(5, a + 27, a + 15, a + 39) < 0 ||	/* w*dy-y*dw d5 */
			    geom_bernstein_mul(5, a + 33, 3, x, a + 45) < 0 ||	/* x*(w*dx-x*dw) d8 */
			    geom_bernstein_mul(5, a + 39, 3, y, a + 54) < 0 ||	/* y*(w*dy-y*dw) d8 */
			    geom_bernstein_add(8, a + 45, a + 54, c) < 0)
				return -1;
		}
		break;
	}

	/* test for numeric errors in bernstein basis */

	for (i = 0; i <= n; i++)
		if (!isfinite(c[i]))
			return -1;

	/* find bernstain roots - candidates for local extremes */

	n = geom_bernstein_solve(n, c, GEOM_SOLVE_UNIT_INTERVAL, a + 1) + 1;

	/* compute distances in found candidates and interval limits */
	a[0] = 0;
	a[n] = 1;
	for (i = 0; i <= n; i++)
	{
		geom_bezier_point_at_time(b, a[i], &q);
		v[i] = SQR(q.x - p->x) + SQR(q.y - p->y);
		if (!isfinite(v[i]))
			return -1;
	}

	m = 0;

	/* we need global extremes */
	if (flags & GEOM_BEZIER_EXTREME_GLOBAL)
	{
		j = k = 0;
		for (i = 1; i <= n; i++)
		{
			if (v[i] > v[j])
				j = i;
			if (v[i] < v[k])
				k = i;
		}
		if (flags & GEOM_BEZIER_EXTREME_MIN)
			result[m++] = a[k];
		if (flags & GEOM_BEZIER_EXTREME_MAX)
			result[m++] = a[j];
	}

	/* we need local extremes */
	else
	{
		for (i = 0; i <= n; i++)
		{
			if ((i == 0 || v[i] <= v[i - 1]) && (i == n || v[i] <= v[i + 1]))
				t[i] = -1;
			else if ((i == 0 || v[i] >= v[i - 1]) && (i == n || v[i] >= v[i + 1]))
				t[i] = 1;
			else
				t[i] = 0;
		}
		if (flags & GEOM_BEZIER_EXTREME_LIMITS)
		{
			j = 0;
			k = n;
		}
		else
		{
			j = 1;
			k = n - 1;
		}
		for (i = j; i <= k; i++)
			if (t[i] < 0 && flags & GEOM_BEZIER_EXTREME_MIN)
				result[m++] = a[i];
			else if (t[i] > 0 && flags & GEOM_BEZIER_EXTREME_MAX)
				result[m++] = a[i];
	}

	return m;
}


/* geom_bezier_center_extreme_points */

int geom_bezier_center_extreme_points(struct geom_bezier *b, struct geom_point *p, uns flags, struct geom_point result[])
{
	int i, n;
	double t[8];

	if ((n = geom_bezier_center_extreme_times(b, p, flags, t)) < 0)
		return -1;

	for (i = 0; i < n; i++)
		geom_bezier_point_at_time(b, t[i], &result[i]);

	return n;
}


/* geom_bezier_nearest_to_point */

int geom_bezier_nearest_to_point(struct geom_bezier *bezier, struct geom_point *point, uns flags, struct geom_nearest *nearest)
{
	double t;
	real rt;

	ITEM_ASSERT(bezier, bezier);
	GASSERT(point, point);
	PASSERT(nearest);

	if (geom_bezier_center_extreme_times(bezier, point, GEOM_BEZIER_EXTREME_GLOBAL | GEOM_BEZIER_EXTREME_MIN, &t) < 0 || !isfinite(rt = t))
		return GEOM_ERR_NUMERIC;
	else {
		nearest->time = rt;
		if (flags & (GEOM_NEAREST_POINT | GEOM_NEAREST_DIST))
			geom_bezier_point_at_time(bezier, rt, &nearest->point);
		if (flags & GEOM_NEAREST_DIST)
			if (!isfinite(nearest->dist = SQR(nearest->point.x - point->x) + SQR(nearest->point.y - point->y)))
				return GEOM_ERR_NUMERIC;
		return 0;
	}
}


