#define GEOM_SOURCE
#include "geomlib/cache.h"
#include "geomlib/fpath.h"
#include "lib/hashfunc.h"
#include "lib/cache.h"

/* Caching in Geomlib can be completely turned off */
#ifndef GEOM_USE_NO_CACHE

/* A tiny local cache useful for debugging */
#ifdef GEOM_USE_TINY_CACHE
#define GEOM_USE_LOCAL_CACHE
#endif

/* Switch between using global Vrr cache and local Geomlib cache */
#ifdef GEOM_USE_LOCAL_CACHE
#define CACHE (&geom_local_cache)
static struct cache geom_local_cache;
#else
#define CACHE (&global_cache)
#endif

/* Geomlib cache node header */
struct hash_node { void *key; };

/* Hash table generator */
#define HASH_NODE struct hash_node
#define HASH_PREFIX(x) hash_##x
#define HASH_KEY_ATOMIC key
#define HASH_ATOMIC_TYPE void *
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_INSERT
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void hash_free(void *p UNUSED) { }
#define HASH_GIVE_HASHFN
static inline uns hash_hash(void *p) { return hash_int((intptr_t)p); }
#define HASH_CONSERVE_SPACE
#include "lib/sht.h"

/* Geomlib cache node header */
struct node {
	CACHE_NODE;
	u8 flags;
	u16 ref_count;
	struct hash_bucket b;
};

/* flags values */
#define NODE_WANT_FREE 0x1


/*------------------------------------------------------------------------*
 *                                                                        *
 *                         BEZIER EXPANSIONS CACHING                      *
 *                                                                        *
 *------------------------------------------------------------------------*/

 // FIXME: generator for this type of cache-hashing mecahism may be useful in vrrlib

#define EXPANSION_USER (&geom_expansion_user)
static struct cache_user geom_expansion_user;

/* Bezier expansion node */

struct expansion_node {
	struct node n;
	struct geom_expansion e;
	int errno;
};

#define EXPANSION_KEY(item) (item)
#define EXPANSION_IS_VALID(item) (GEOM_ITEM_FLAGS(item) & GEOM_ITEM_EXPANSION_VALID)
#define EXPANSION_SET_VALID(item) do{GEOM_ITEM_FLAGS(item) |= GEOM_ITEM_EXPANSION_VALID;}while(0)

static inline uns expansion_size(struct expansion_node *n)
{ return sizeof(struct expansion_node) + n->e.path.total_size; }

static inline void expansion_free(struct expansion_node *n)
{ geom_fpath_destroy(&n->e.path); xfree(n); }

static inline struct expansion_node *expansion_find(void *key)
{ // FIXME: can be changed when caching more structures via same pointer
	struct hash_node *h = hash_find(key);
	return h ? SKIP_BACK(struct expansion_node, n.b.n, h) : NULL;
}

void geom_expansion_unref(struct geom_expansion *expansion)
{
	struct expansion_node *n = SKIP_BACK(struct expansion_node, e, expansion);
	if (--n->n.ref_count == 0) {
		if (n->n.flags & NODE_WANT_FREE) {
			TRACE("Expansion of %p unrefed and released (displayed pointer may be invalid)", n->n.b.n.key);
			expansion_free(n);
		}
		else {
			TRACE("Expansion of %p unrefed and cached", n->n.b.n.key);
			cache_used_inc(CACHE, expansion_size(n));
			cache_add(CACHE, EXPANSION_USER, (struct cache_node *)n);
		}
	}
	else
		TRACE("Number of refs to expansion of %p decreased to %d", n->n.b.n.key, n->n.ref_count);
}

struct geom_expansion *geom_item_expand(void *item)
{
	void *key = EXPANSION_KEY(item);
	struct expansion_node *n;
	int errno;
	if (n = expansion_find(key)) {
		if (EXPANSION_IS_VALID(item)) {
			if (n->n.ref_count++ == 0) {
				TRACE("Expansion of %p found in cache", item);
				cache_remove(CACHE, (struct cache_node *)n);
				cache_used_dec(CACHE, expansion_size(n));
			}
			else
				TRACE("Expansion of %p already in use, increasing refcount to %d", item, n->n.ref_count);
			errno = n->errno;
			if (!ETEST(errno)) {
				TRACE("... but only errno %d fetched", errno);
				geom_expansion_unref(&n->e);
				geom_errno = errno;
				return NULL;
			}
			else
				return &n->e;
		}

		if (n->n.ref_count == 0) {
			TRACE("Using previously hashed structure to recompute expansion of %p", n->n.b.n.key);
			cache_remove(CACHE, (struct cache_node *)n);
			cache_used_dec(CACHE, expansion_size(n));
			geom_fpath_clear(&n->e.path);
			goto REUSE;
		}
		else {
			TRACE("Cannot reuse old hashed structure for %p expansion because of %d references, allocating a new one", n->n.b.n.key, n->n.ref_count);
			hash_remove(&n->n.b.n);
			n->n.flags |= NODE_WANT_FREE;
		}
	}
	n = xmalloc_zero(sizeof(struct expansion_node));
	hash_insert(key, &n->n.b);
	GEOM_INSTANCE_ZERO_CREATE(&n->e.path, fpath, geom_fpath_create);
REUSE:
	n->n.ref_count = 1;
	EXPANSION_SET_VALID(item);
	errno = geom_item_expansion_append(item, &n->e.path);
	if (ETEST(errno)) {
		geom_fpath_finish(&n->e.path);
		TRACE("Expansion of %p computed and hashed (size=%d)", item, expansion_size(n));
	}
	else {
		TRACE("Expansion of %p failed, errno %d hashed", item, errno);
		geom_fpath_clear(&n->e.path);
		n->errno = errno;
		geom_expansion_unref(&n->e);
		geom_errno = errno;
		return NULL;
	}
	return &n->e;
}

void geom_item_expansion_release(void *item)
{
	struct expansion_node *n;
	if (n = expansion_find(EXPANSION_KEY(item))) {
		hash_remove(&n->n.b.n);
		if (n->n.ref_count == 0) {
			TRACE("Expansion of %p released", n->n.b.n.key);
			cache_remove(CACHE, (struct cache_node *)n);
			cache_used_dec(CACHE, expansion_size(n));
			expansion_free(n);
		}
		else {
			TRACE("Expansion of %p with %d references will be released as soon as possible.", n->n.b.n.key, n->n.ref_count);
			n->n.flags |= NODE_WANT_FREE;
		}
	}
}

static void expansion_callback(struct expansion_node *n)
{
	TRACE("Cache callback - releasing expansion of %p", n->n.b.n.key);
	hash_remove(&n->n.b.n);
	cache_used_dec(CACHE, expansion_size(n));
	expansion_free(n);
}

/*------------------------------------------------------------------------*
 *                                                                        *
 *                      CACHE CALLBACK AND INITIALIZATION                 *
 *                                                                        *
 *------------------------------------------------------------------------*/


/* Geomlib cache initialization. */

void geom_cache_init(void)
{
	TRACE("Initializing Geomlib cache");
#ifdef GEOM_USE_LOCAL_CACHE
	TRACE("Using Geomlib local cache");
	cache_init(CACHE);
#ifdef GEOM_USE_TINY_CACHE
	cache_set_limits(CACHE, 512 * 1024, 0, 3);
#else
	cache_set_limits(CACHE, 5 * 1024 * 1024, 2, 50000);
#endif
#endif
	hash_init();
	cache_user_init(EXPANSION_USER, CACHE, (void (*)(struct cache_node *)) &expansion_callback);
}


/* Geomlib cache finalization. */

void geom_cache_finish(void)
{
	TRACE("Finalizing Geomlib cache");
	cache_user_destroy(EXPANSION_USER, CACHE);
#ifdef GEOM_DEBUG
	uns i = 0;
	HASH_FOR_ALL(hash, node) { if (node) {}; i++; } HASH_END_FOR;
	if (i)
		die("There are %d left items in Geomlib hash table... propably missing destructors or calls to geom_item_expansion_release().\n", i);
#endif
	hash_cleanup();
#ifdef GEOM_USE_LOCAL_CACHE
	cache_destroy(CACHE);
#endif
}


/*------------------------------------------------------------------------*
 *                                                                        *
 *                             IF CACHE IS TURNED OFF                     *
 *                                                                        *
 *------------------------------------------------------------------------*/

#else /* GEOM_USE_NO_CACHE */

struct geom_expansion *geom_item_expand(void *item)
{
	struct geom_expansion *e = xmalloc_zero(sizeof(struct geom_expansion));
	FUNC;
	INSTANCE_IS_ASSERT(item, item);
	GEOM_INSTANCE_ZERO_CREATE(&e->path, fpath, geom_fpath_create);
	if (ETEST(geom_errno = geom_item_expansion_append(item, &e->path)))
		PRET(e);
	geom_fpath_destroy(&e->path);
	xfree(e);
	PRET(NULL);
}

void geom_expansion_unref(struct geom_expansion *expansion)
{
	FUNC;
	PASSERT(expansion);
	geom_fpath_destroy(&expansion->path);
	xfree(expansion);
	RET;
}

#endif /* GEOM_USE_NO_CACHE */
