#ifndef _VRR_GEOMLIB_CONIC_H_
#define _VRR_GEOMLIB_CONIC_H_


enum geom_conic_arc_type
{
	GEOM_CONIC_ARC_START_ENTIRE,
	GEOM_CONIC_ARC_START_DIF,
	GEOM_CONIC_ARC_POINT_ENTIRE,
	GEOM_CONIC_ARC_POINT_DIF,
	GEOM_CONIC_ARC_CCW,
	GEOM_CONIC_ARC_CW,
	GEOM_CONIC_ARC_SMALLER,
	GEOM_CONIC_ARC_BIGGER,
	GEOM_CONIC_ARC_MIDDLE,
	GEOM_CONIC_ARC_OPPOSITE,
	GEOM_CONIC_ARC_ENTIRE,
	GEOM_CONIC_ARC_MAX,
	GEOM_CONIC_ARC_0P = GEOM_CONIC_ARC_START_DIF,
	GEOM_CONIC_ARC_1P = GEOM_CONIC_ARC_POINT_DIF,
	GEOM_CONIC_ARC_2P = GEOM_CONIC_ARC_BIGGER,
	GEOM_CONIC_ARC_3P = GEOM_CONIC_ARC_OPPOSITE,
};


/**
	\file geomlib/conic.h

	\brief Handling conic curves and their sections.
*/

#include "lib/lib.h"
#include "geomlib/types.h"
#include "geomlib/numeric.h"
#include "geomlib/vector.h"


/*-------------------------------------------------------------------------------------------*/


/**
	\brief Makes assertions to test validity of geom_parabolic_arc structure.

	\param  a  Pointer to geom_parabolic_arc structure.
*/
#define GEOM_PARABOLIC_ARC_ASSERT(a) do{ \
		GEOM_BASSERT((a) != NULL); \
		GEOM_POINT_ASSERT(&(a)->center); \
		GEOM_BASSERT(isfinite((a)->rotation)); \
		GEOM_BASSERT(isfinite((a)->distance)); \
		GEOM_BASSERT((a)->distance >= 0); \
		GEOM_BASSERT(isfinite((a)->start)); \
		GEOM_BASSERT(isfinite((a)->end)); \
	}while(0)


/**
	\brief Test validity of geom_parabolic_arc structure.

	\param   arc      Pointer to geom_parabolic_arc structure.

	\retval  nonzero  Structure is valid.
	\retval  zero     Structure is invalid.
*/
int geom_parabolic_arc_valid(const struct geom_parabolic_arc *arc);


/**
	\brief Dumps geom_parabolic_arc structure data.

	\param  arc  Pointer to geom_parabolic_arc structure (could has invalid data).
*/
void geom_parabolic_arc_dump(const struct geom_parabolic_arc *arc);


/**
	\brief Initialize parabolic arc.

	\param  arc       Resulting parabolic arc.
	\param  center    Center point.
	\param  rotation  Rotation around the center point (CCW in radians).
	\param  distance  Distance of center point to parabola focus and directrix (distance >= 0).
	\param  start     Parameter of parabolic arc starting point.
	\param  end       Parameter of parabolic arc ending point.
*/
void geom_parabolic_arc_set(struct geom_parabolic_arc *arc, const struct geom_point *center, const real rotation, const real distance, const real start, const real end);


/**
	\brief Initialize "unit" parabolic arc.

	\param  arc       Resulting parabola.
	\param  center    Center point.
	\param  rotation  Rotation around the center point (CCW in radians).
	\param  distance  Distance of center point to parabola focus and directrix (distance >= 0).
*/
static inline void geom_parabolic_arc_set_unit(struct geom_parabolic_arc *arc, const struct geom_point *center, const real rotation, const real distance)
{ geom_parabolic_arc_set(arc, center, rotation, distance, -1, 1); }


/**
	\brief Normalize parabolic arc structure. Rotation angle is aligned to [0, 2 * PI].

	\param  arc  Parabolic arc to normalize.

	\retval  0  Always.
*/
int geom_parabolic_arc_normalize(struct geom_parabolic_arc *arc);


/**
	\brief Sets parabolic arc endpoints. Given points should be near the parabola.

	\param  arc    Parabolic arc structure to change.
	\param  start  Starting point.
	\param  end    Ending point.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_set_endpoints(struct geom_parabolic_arc *arc, const struct geom_point *start, const struct geom_point *end);


/**
	\brief Returns point on parabola for given parameter (see parabola in parametric form).

	\param  arc    Parabolic arc.
	\param  time   Finite parameter.
	\param  point  Resulting point.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_point_at(const struct geom_parabolic_arc *arc, const real time, struct geom_point *point);


/**
	\brief Compute focus point of given parabola.

	\param   arc    Parabolic arc.
	\param   focus  Resulting parabola focus.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_focus(const struct geom_parabolic_arc *arc, struct geom_point *focus);


/**
	\brief Compute directrix of given parabola.

	\param   arc     Parabolic arc.
	\param   point   Resulting point on parabola directrix (nearest to focus).
	\param   vector  Line direction vector (unit, parabola is to the left of this vector).

	\retval  0  Success (always, if point = NULL).
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_directrix(const struct geom_parabolic_arc *arc, struct geom_point *point, struct geom_vector *vector);


/**
	\brief Compute focal parameter of given parabola (distance of focus to directrix).

	\param  arc  Parabolic arc.

	\return      Focal parameter (infinite for numeric overflow).
*/
static inline real geom_parabolic_arc_focal_parameter(const struct geom_parabolic_arc *arc)
{ return 2 * arc->distance; }


/**
	\brief Compute latus rectum of given parabola.

	\param  arc  Parabolic arc.

	\return      Latus rectum (infinite for numeric overflow).
*/
static inline real geom_parabolic_arc_latus_rectum(const struct geom_parabolic_arc *arc)
{ return 4 * arc->distance; }


/**
	\brief Converts parabola to form: Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.

	\param   arc   Parabola to convert.
	\param   poly  Resulting coefficients (A, B, C, D, E, F).

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_to_poly(const struct geom_parabolic_arc *arc, real *poly);


/**
	\brief Converts parabolic arc to single bezier curve.

	\param   arc     Parabolic arc to convert.
	\param   bezier  Resulting bezier curve.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_to_bezier(const struct geom_parabolic_arc *arc, struct geom_bezier *bezier);


/**
	\brief Converts parabolic arc to single bezier curve and adds it to end of given path.

	\param   arc   Parabolic arc to convert.
	\param   path  Resulting path.

	\retval  1  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_parabolic_arc_to_path(const struct geom_parabolic_arc *arc, struct geom_path *path);



/*-------------------------------------------------------------------------------------------*/


/**
	\brief Makes assertions to test validity of geom_hyperbolic_arc structure.

	\param  a  Pointer to geom_hyperbolic_arc structure.
*/
#define GEOM_HYPERBOLIC_ARC_ASSERT(a) do{ \
		GEOM_BASSERT((a) != NULL); \
		GEOM_POINT_ASSERT(&(a)->center); \
		GEOM_BASSERT(isfinite((a)->rotation)); \
		GEOM_BASSERT(isfinite((a)->major)); \
		GEOM_BASSERT((a)->major >= 0); \
		GEOM_BASSERT(isfinite((a)->minor)); \
		GEOM_BASSERT((a)->minor >= 0); \
		GEOM_BASSERT(isfinite((a)->start)); \
		GEOM_BASSERT(isfinite((a)->end)); \
	}while(0)


/**
	\brief Test validity of geom_hyperbolic_arc structure.

	\param   arc      Pointer to geom_hyperbolic_arc structure.

	\retval  nonzero  Structure is valid.
	\retval  zero     Structure is invalid.
*/
int geom_hyperbolic_arc_valid(const struct geom_hyperbolic_arc *arc);


/**
	\brief Dumps geom_hyperbolic_arc structure data.

	\param  arc  Pointer to geom_hyperbolic_arc structure (could has invalid data).
*/
void geom_hyperbolic_arc_dump(const struct geom_hyperbolic_arc *arc);


/**
	\brief Initialize hyperbolic arc.

	\param  arc       Resulting hyperbolic arc.
	\param  center    Center point.
	\param  rotation  Rotation around the center point (CCW in radians).
	\param  major     Semimajor axis (major >= 0).
	\param  minor     Semiminor axis (minor >= 0).
	\param  start     Parameter of hyperbolic arc starting point (-1 for entire hyperbola).
	\param  end       Parameter of hyperbolic arc ending point (+1 for entire hyperbola).
*/
void geom_hyperbolic_arc_set(struct geom_hyperbolic_arc *arc, const struct geom_point *center, const real rotation, const real major, const real minor, const real start, const real end);


/**
	\brief Initialize "unit" hyperbolic arc.

	\param  arc       Resulting hyperbola.
	\param  center    Center point.
	\param  rotation  Rotation around the center point (CCW in radians).
	\param  major     Semimajor axis (major >= 0).
	\param  minor     Semiminor axis (minor >= 0).
*/
static inline void geom_hyperbolic_arc_set_unit(struct geom_hyperbolic_arc *arc, const struct geom_point *center, const real rotation, const real major, const real minor)
{ geom_hyperbolic_arc_set(arc, center, rotation, major, minor, -1, 1); }


/**
	\brief Normalize hyperbolic arc structure. Rotation angle is aligned to [0, 2 * PI].

	\param  arc  Hyperbolic arc to normalize.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_normalize(struct geom_hyperbolic_arc *arc);


/**
	\brief Sets hyperbolic arc endpoints. Given points should be near one branch of hyperbola.

	\param  arc    Hyperbolic arc structure to change.
	\param  start  Starting point.
	\param  end    Ending point.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_set_endpoints(struct geom_hyperbolic_arc *arc, const struct geom_point *start, const struct geom_point *end);


/**
	\brief Returns point on hyperbola for given parameter (see hyperbola in parametric form).

	\param  arc    Hyperbola.
	\param  time   Finite parameter.
	\param  point  Resulting point.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_point_at(const struct geom_hyperbolic_arc *arc, const real time, struct geom_point *point);


/**
	\brief Converts hyperbola to form: Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.

	\param   arc   Hyperbola to convert.
	\param   poly  Resulting coefficients (A, B, C, D, E, F).

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_to_poly(const struct geom_hyperbolic_arc *arc, real *poly);


/**
	\brief Converts hyperbolic arc to single bezier curve.

	\param   arc     Hyperbolic arc to convert.
	\param   bezier  Resulting bezier curve.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_to_bezier(const struct geom_hyperbolic_arc *arc, struct geom_bezier *bezier);


/**
	\brief Converts hyperbolic arc to single bezier curve and adds it to end of given path.

	\param   arc   Hyperbolic arc to convert.
	\param   path  Resulting path.

	\retval  1   Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_hyperbolic_arc_to_path(const struct geom_hyperbolic_arc *arc, struct geom_path *path);


/*-------------------------------------------------------------------------------------------*/


/**
	\brief Makes assertions to test validity of geom_conic structure.

	\param  a  Pointer to geom_conic structure.
*/
#define GEOM_CONIC_ASSERT(a) do{ \
		GEOM_BASSERT((a) != NULL); \
		switch ((a)->type) { \
			case GEOM_CONIC_PARABOLIC_ARC: GEOM_PARABOLIC_ARC_ASSERT(&(a)->data.parabolic_arc); break; \
			case GEOM_CONIC_ARC: GEOM_ARC_ASSERT(&(a)->data.arc); break; \
			case GEOM_CONIC_ELLIPTIC_ARC: GEOM_ELLIPTIC_ARC_ASSERT(&(a)->data.elliptic_arc); break; \
			case GEOM_CONIC_HYPERBOLIC_ARC: GEOM_HYPERBOLIC_ARC_ASSERT(&(a)->data.hyperbolic_arc); break; \
			default: GEOM_BASSERT(0); \
		} \
	}while(0)


/**
	\brief Test validity of geom_conic structure.

	\param   conic    Pointer to geom_conic structure.

	\retval  nonzero  Structure is valid.
	\retval  zero     Structure is invalid.
*/
int geom_conic_valid(struct geom_conic *conic);


/**
	\brief Dumps geom_conic structure data.

	\param  conic  Pointer to geom_conic structure (could has invalid data).
*/
void geom_conic_dump(struct geom_conic *conic);


/**
	\brief Converts conic to form: Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.

	\param   arc   Conic to convert.
	\param   poly  Resulting coefficients (A, B, C, D, E, F).

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_conic_to_poly(struct geom_conic *conic, real *poly);


/**
	\brief Converts conic to single bezier curve.

	\param   arc     Conic to convert.
	\param   bezier  Resulting bezier curve.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC         Numeric error.
	\retval  GEOM_ERR_CONIC_TOO_LONG  Arc or elliptic arc angle too big.
*/
int geom_conic_to_bezier(struct geom_conic *conic, struct geom_bezier *bezier);


/**
	\brief Converts conic to list of bezier curves and adds them to end of given path.

	\param   arc   Conic to convert.
	\param   path  Resulting path.

	\retval  >0  Success, number of beziers added.
	\retval  GEOM_ERR_NUMERIC         Numeric error.
	\retval  GEOM_ERR_CONIC_TOO_LONG  Arc or elliptic arc angle too big.
*/
int geom_conic_to_path(struct geom_conic *conic, struct geom_path *path);


/**
	\brief Convers form AX^2 + Bxy + Cy^2 + Dx + Ey + F = 0 to conic curve.

	\param   conic  Resulting conic.
	\param   poly   Coefficients to convert (A, B, C, D, E, F).

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC            Numeric error.
	\retval  GEOM_ERR_CONIC_DEGENERATED  Degenerated case of conic.
*/
int geom_conic_from_poly(struct geom_conic *conic, const real *poly);


/**
	\brief Find conic containing 5 given points.

	\param  conic   Resulting conic.
	\param  points  Array of 5 points.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC            Numeric error.
	\retval  GEOM_ERR_CONIC_DEGENERATED  Degenerated case of conic.
	\retval  GEOM_ERR_CONIC_AMBIGUOUS    More solutions.
*/
int geom_conic_from_5_points(struct geom_conic *conic, const struct geom_point *points);


/**
	\brief Find conic containing 4 given points and with axis in given angle.

	\param  conic     Resulting Conic.
	\param  rotation  Angle of one axis (periode = PI / 2).
	\param  points    Array of 4 points.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC            Numeric error.
	\retval  GEOM_ERR_CONIC_DEGENERATED  Degenerated case of conic.
	\retval  GEOM_ERR_CONIC_AMBIGUOUS    More solutions.
*/
int geom_conic_from_rotation_and_4_points(struct geom_conic *conic, const real rotation, const struct geom_point *points);


/**
	\brief Finds conic for 3 given points and 2 tangential angles in first 2 of them.

	\param  conic   Resulting Conic.
	\param  angles  Array of 2 angles.
	\param  points  Array of 3 points.

	\retval  0  Success.
	\retval  GEOM_ERR_NUMERIC            Numeric error.
	\retval  GEOM_ERR_CONIC_DEGENERATED  Degenerated case of conic.
	\retval  GEOM_ERR_CONIC_AMBIGUOUS    More solutions.
*/
int geom_conic_from_2_tangents_and_point(struct geom_conic *conic, const real *angles, const struct geom_point *points);


#endif

