#define GEOM_SOURCE
#include "geomlib/vector.h"
#include "geomlib/transform.h"
#include "geomlib/fpath.h"
#include "geomlib/bezier.h"
#include "geomlib/conic.h"
#include "geomlib/circle.h"
#include "geomlib/ellipse.h"



/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


/* geom_elliptic_arc_set */

void geom_elliptic_arc_set(struct geom_elliptic_arc *arc, struct geom_point *center, real rotation, real a_radius, real b_radius, real start, real dif)
{
	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, center);
	RASSERT(rotation);
	GASSERT2(r_minmax, a_radius, 0, REAL_MAX);
	GASSERT2(r_minmax, b_radius, 0, REAL_MAX);
	RASSERT(start);
	RASSERT(dif);

	arc->center = *center;
	arc->rotation = rotation;
	arc->a_radius = a_radius;
	arc->b_radius = b_radius;
	arc->start = start;
	arc->dif = dif;

	EASSERTX(geom_elliptic_arc_normalize(arc));
	ITEM_ASSERT(elliptic_arc, arc);
	RET;
}


/**
 * Converts coordinates in ellipse to unit circle angle.
 *
 * @ai  a_radius  First radius of the ellipse.
 * @ai  b_radius  Second radius of the ellipse.
 * @ai  dx        Finite X coordinate.
 * @ai  dy        Finite Y coordinate.
 *
 * @rv  [-PI,PI]  Angle in unit circle.
 */
static real geom_elliptic_arc_vector_to_angle(real a_radius, real b_radius, double dx, double dy)
{
	GASSERT2(r_minmax, a_radius, 0, REAL_MAX);
	GASSERT2(r_minmax, b_radius, 0, REAL_MAX);
	FASSERT(dx);
	FASSERT(dy);

	if (a_radius == 0) {
		if (b_radius == 0)
			return 0;
		else if (dy >= b_radius)
			return R_PI2;
		else if (dy <= -b_radius)
			return -R_PI2;
		else
			return asin(dy / b_radius);
	}
	else if (b_radius == 0) {
		if (dx >= a_radius)
			return 0;
		else if (dx <= -a_radius)
			return R_PI;
		else
			return acos(dx / a_radius);
	}
	else
		return nansetr(atan2(a_radius * dy, b_radius * dx), 0);
}


/* geom_elliptic_arc_set_endpoints */

int geom_elliptic_arc_set_endpoints(struct geom_elliptic_arc *arc, struct geom_point *start, struct geom_point *end, struct geom_point *mid, real st, real dif, uns flags)
{
	double cosa, sina, cx, cy, dx, dy, t1, t2, t3;
	int swap;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);

	if (flags == GEOM_CONIC_ARC_ENTIRE) {
		arc->start = 0;
		arc->dif = (dif >= 0) ? R_2PI : -R_2PI;
		goto EXIT;
	}
	else if (flags == GEOM_CONIC_ARC_START_ENTIRE) {
		arc->start = fmodfloorr(st, R_2PI);
		arc->dif = (dif >= 0) ? R_2PI : -R_2PI;
		goto EXIT;
	}
	else if (flags == GEOM_CONIC_ARC_START_DIF) {
		arc->start = fmodfloorr(st, R_2PI);
		arc->dif = fminmaxr(dif, -R_2PI, R_2PI);
		goto EXIT;
	}

	GASSERT(point, start);

	cosa = cos(arc->rotation);
	sina = sin(arc->rotation);
	cx = start->x - arc->center.x;
	cy = start->y - arc->center.y;
	FTEST_NUM(dx = cosa * cx + sina * cy);
	FTEST_NUM(dy = cosa * cy - sina * cx);
	t1 = geom_elliptic_arc_vector_to_angle(arc->a_radius, arc->b_radius, dx, dy);

	if (flags == GEOM_CONIC_ARC_POINT_DIF) {
		GASSERT(r,dif);
		arc->start = fmodfloorr(t1, R_2PI);
		arc->dif = fminmaxr(dif, -R_2PI, R_2PI);
		goto EXIT;
	}
	else if (flags == GEOM_CONIC_ARC_POINT_ENTIRE) {
		GASSERT(r,dif);
		arc->start = fmodfloorr(t1, R_2PI);
		arc->dif = (dif >= 0) ? R_2PI : -R_2PI;
		goto EXIT;
	}

	GASSERT(point, end);

	cx = end->x - arc->center.x;
	cy = end->y - arc->center.y;
	FTEST_NUM(dx = cosa * cx + sina * cy);
	FTEST_NUM(dy = cosa * cy - sina * cx);
	t2 = geom_elliptic_arc_vector_to_angle(arc->a_radius, arc->b_radius, dx, dy);

	switch (flags & ~1) {
		case GEOM_CONIC_ARC_CCW:
			swap = t1 > t2;
			break;

		case GEOM_CONIC_ARC_SMALLER:
			swap = fabs(t2 - t1) > M_PI;
			break;

		case GEOM_CONIC_ARC_MIDDLE:
			GASSERT(point, mid);

			cx = mid->x - arc->center.x;
			cy = mid->y - arc->center.y;
			FTEST_NUM(dx = cosa * cx + sina * cy);
			FTEST_NUM(dy = cosa * cy - sina * cx);
			t3 = geom_elliptic_arc_vector_to_angle(arc->a_radius, arc->b_radius, dx, dy);

			swap = (t1 < t3) ^ (t3 < t2);
			break;

		default:
			BUG();
	}

	if (swap ^ (flags & 1)) {
		if (t1 < t2)
			t1 += 2 * M_PI;
		else
			t2 += 2 * M_PI;
	}

	RASSERTX(arc->start = fmodfloorr(t1, R_2PI));
	RASSERTX(arc->dif = fminmaxr(t2 - t1, -R_2PI, R_2PI));

EXIT:
	ITEM_ASSERT(elliptic_arc, arc);
	ERET(0);
}


/* geom_elliptic_arc_from_foci_major */

int geom_elliptic_arc_from_foci_major(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2, real major)
{
	struct geom_point center;
	double dx, dy, d, e;
	real rotation, minor;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, focus1);
	GASSERT(point, focus2);
	GASSERT2(r_minmax, major, 0, REAL_MAX);

	dx = focus2->x - focus1->x;
	dy = focus2->y - focus1->y;
	rotation = nansetr(atan2r(dy, dx), 0);
	d = SQR(dx) + SQR(dy);
	e = SQR((double)major) - d;
	if (e < 0) {
		minor = 0;
		RTEST_NUM(major = sqrt(d));
	}
	else
		RTEST_NUM(minor = 0.5 * sqrt(e));
	major *= 0.5;
	center.x = fmidfinr(focus1->x, focus2->x);
	center.y = fmidfinr(focus1->y, focus2->y);

	geom_elliptic_arc_set_closed(arc, &center, rotation, major, minor);
	ERET(0);
}


/* geom_elliptic_arc_from_foci_point */

int geom_elliptic_arc_from_foci_point(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2, struct geom_point *point)
{
	real major;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, focus1);
	GASSERT(point, focus2);
	GASSERT(point, point);

	RTEST_NUM(major =
		hypot(point->x - focus1->x, point->y - focus1->y) +
		hypot(point->x - focus2->x, point->y - focus2->y));

	ETEST_RET(geom_elliptic_arc_from_foci_major(arc, focus1, focus2, major));
	ERET(0);
}


/* geom_elliptic_arc_from_3_points_smallest */

int geom_elliptic_arc_from_3_points_smallest(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3)
{
	struct geom_transform transform;
	double x, y;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, point1);
	GASSERT(point, point2);
	GASSERT(point, point3);

	transform.flags = 0;
	RTEST_NUM(transform.coef[0][0] = sqrtr(3) * (point3->x - point2->x));
	RTEST_NUM(transform.coef[1][0] = sqrtr(3) * (point3->y - point2->y));
	RTEST_NUM(transform.coef[0][1] = fma(2, point1->x, -(x = point2->x + point3->x)));
	RTEST_NUM(transform.coef[1][1] = fma(2, point1->y, -(y = point2->y + point3->y)));
	RTEST_NUM(transform.coef[0][2] = (1.0 / 3) * (point1->x + x));
	RTEST_NUM(transform.coef[1][2] = (1.0 / 3) * (point1->y + y));
	GASSERT(transform, &transform);

	geom_elliptic_arc_set_closed(arc, &geom_point_zero, 0, (1.0 / 3), (1.0 / 3));
	ETEST_RET(geom_elliptic_arc_transform(arc, &transform, arc));
	ERET(0);
}


/* geom_elliptic_arc_from_3_points_circle */

int geom_elliptic_arc_from_3_points_circle(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3)
{
	int err;
	struct geom_arc a;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, point1);
	GASSERT(point, point2);
	GASSERT(point, point3);

	GEOM_INSTANCE_INIT(&a, arc);
	if (ETEST(err = geom_arc_from_3_points(&a, point1, point2, point3)))
		EASSERTX(geom_elliptic_arc_from_arc(arc, &a));
	GEOM_INSTANCE_DESTROY2(&a, arc);

	ERET(err);
}



/* geom_elliptic_arc_from_3_points_rotation_eccentricity */

int geom_elliptic_arc_from_3_points_rotation_eccentricity(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3, real rotation, real eccentricity)
{
	double cos1, sin1, cos2, sin2, coef, rcoef, x, y;
	struct geom_point tmp1, tmp2, tmp3;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, point1);
	GASSERT(point, point2);
	GASSERT(point, point3);
	RASSERT(rotation);
	RASSERT(eccentricity);

	if (IS_ZERO(eccentricity))
		ETEST_RET(geom_elliptic_arc_from_3_points_circle(arc, point1, point2, point3));

	else {
		coef = sqrt(fminmax(1 - SQR(eccentricity), 0, 1));
		RTEST_NUM(rcoef = 1 / coef); /* FIXME: unstable near eccentricity = 1  */

		cos1 = cos(rotation);
		sin1 = sin(rotation);
		cos2 = cos1 * rcoef;
		sin2 = -sin1 * rcoef;
		RTEST_NUM(tmp1.x = point1->x * cos1 + point1->y * sin1);
		RTEST_NUM(tmp1.y = point1->x * sin2 + point1->y * cos2);
		RTEST_NUM(tmp2.x = point2->x * cos1 + point2->y * sin1);
		RTEST_NUM(tmp2.y = point2->x * sin2 + point2->y * cos2);
		RTEST_NUM(tmp3.x = point3->x * cos1 + point3->y * sin1);
		RTEST_NUM(tmp3.y = point3->x * sin2 + point3->y * cos2);
		ETEST_RET(geom_elliptic_arc_from_3_points_circle(arc, &tmp1, &tmp2, &tmp3));

		x = arc->center.x;
		y = arc->center.y * coef;
		RTEST_NUM(arc->center.x = x * cos1 - y * sin1);
		RTEST_NUM(arc->center.y = x * sin1 + y * cos1);
		arc->b_radius = MIN(arc->a_radius, coef * arc->b_radius);
		arc->rotation = /*fmodfloorr(rotation, R_PI)*/ fminmaxr(rotation, 0, R_PI);

		ITEM_ASSERT(elliptic_arc, arc);
	}

	ERET(0);
}


/* geom_elliptic_arc_from_center_point_circle */

int geom_elliptic_arc_from_center_point_circle(struct geom_elliptic_arc *arc, struct geom_point *center, struct geom_point *point)
{
	int err;
	struct geom_arc a;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, center);
	GASSERT(point, point);

	GEOM_INSTANCE_INIT(&a, arc);
	if (ETEST(err = geom_arc_from_center_point(&a, center, point)))
		geom_elliptic_arc_from_arc(arc, &a);
	GEOM_INSTANCE_DESTROY2(&a, arc);

	ERET(err);
}


/* geom_elliptic_arc_from_center_point_rotation_eccentricity */
// FIXME: pujde pouzit mnohem jednodussi vzorecek

int geom_elliptic_arc_from_center_point_rotation_eccentricity(struct geom_elliptic_arc *arc, struct geom_point *center, struct geom_point *point, real rotation, real eccentricity)
{
	double cos1, sin1, cos2, sin2, coef, rcoef, x, y;
	struct geom_point tmp1, tmp2;

	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, center);
	GASSERT(point, point);
	RASSERT(rotation);
	RASSERT(eccentricity);

	if (IS_ZERO(eccentricity))
		ETEST_RET(geom_elliptic_arc_from_center_point_circle(arc, center, point));

	else {
		coef = sqrt(fminmax(1 - SQR(eccentricity), 0, 1));
		RTEST_NUM(rcoef = 1 / coef); /* FIXME: unstable near eccentricity = 1  */

		cos1 = cos(rotation);
		sin1 = sin(rotation);
		cos2 = cos1 * rcoef;
		sin2 = -sin1 * rcoef;
		RTEST_NUM(tmp1.x = center->x * cos1 + center->y * sin1);
		RTEST_NUM(tmp1.y = center->x * sin2 + center->y * cos2);
		RTEST_NUM(tmp2.x = point->x * cos1 + point->y * sin1);
		RTEST_NUM(tmp2.y = point->x * sin2 + point->y * cos2);
		ETEST_RET(geom_elliptic_arc_from_center_point_circle(arc, &tmp1, &tmp2));

		x = arc->center.x;
		y = arc->center.y * coef;
		RTEST_NUM(arc->center.x = x * cos1 - y * sin1);
		RTEST_NUM(arc->center.y = x * sin1 + y * cos1);
		arc->b_radius = MIN(arc->a_radius, coef * arc->b_radius);
		arc->rotation = /*fmodfloorr(rotation, R_PI)*/ fminmaxr(rotation, 0, R_PI);

		ITEM_ASSERT(elliptic_arc, arc);
	}


	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             PARAMETRIZATIONS                 *
 *                                              *
 *----------------------------------------------*/

/* FIXME:
 * - dopsat prevody mezi TIME<->ATIME
 *   - pro kruznice jednoduche
 *   - pro elipsy by sla pouzit naka rychlejsi aproximace eliptickeho integralu nez
 *     pres bezierky */

/* geom_elliptic_arc_time_to_btime */

real geom_elliptic_arc_time_to_btime(struct geom_elliptic_arc *arc, struct geom_expansion *expansion, real time)
{
	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	GASSERT(expansion, expansion);
	RASSERT(time);

	if (time <= 0)
		RRET(0);
	if (time >= 1)
		RRET(geom_group_count(&expansion->path));
	else {
		double t = time * geom_group_count(&expansion->path), index = floor(t), btime, alpha, beta, abs_beta, s;
		if (!geom_elliptic_arc_is_closed(arc))
			alpha = 0.25 * fabs(arc->dif) / geom_group_count(&expansion->path);
		else
			alpha = 0.5 * M_PI / 3;
		beta = 2 * (t - index - 0.5) * alpha;
		abs_beta = fabs(beta);
		t = sin(alpha + abs_beta);
		s = sin(alpha - abs_beta);
		t += s;
		s *= 2;
		if (s <= 0)
			btime = beta < 0 ? index : index + 1;
		else if (s >= t)
			btime = index + 0.5;
		else {
			t = 0.5 * s / t;
			btime = index + (beta < 0 ? t : 1 - t);
		}
		RRET(btime);
	}
}


/* geom_elliptic_arc_btime_to_time */

real geom_elliptic_arc_btime_to_time(struct geom_elliptic_arc *arc, struct geom_expansion *expansion, real btime)
{
	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	GASSERT(expansion, expansion);
	RASSERT(btime);

	if (btime <= 0)
		RRET(0);
	else if (btime >= geom_group_count(&expansion->path))
		RRET(1);
	else {
		double index = floorr(btime), t = btime - index, alpha, half, beta, y, x;
		if (!geom_elliptic_arc_is_closed(arc)) {
			alpha = arc->dif / geom_group_count(&expansion->path);
			half = 0.5 * alpha;
			y = sin(half) * (0.5 - t);
			t = t - SQR(t);
			x = cos(half) * (0.5 - t) + t;
			beta = atan2(y, x);
			RRET((index + 0.5 - beta / alpha) / geom_group_count(&expansion->path));
		}
		else {
			y = sqrt(3) * (0.5 - t);
			x = 0.5 + t - SQR(t);
			beta = atan2(y, x);
			RRET((index + 0.5 - beta / (2 * M_PI / 3)) / 3);
		}
	}
}


/*----------------------------------------------*
 *                                              *
 *                  CONVERSION                  *
 *                                              *
 *----------------------------------------------*/


/* geom_elliptic_arc_expansion_append */

int geom_elliptic_arc_expansion_append(struct geom_elliptic_arc *arc, struct geom_fpath *expansion)
{
	uns i, n;
	double end, dif, abs_dif, mid, w, a, b, x, y, cos1, sin1, acos, asin, bcos, bsin, acosbsin1, acosbsin2, asinbcos1, asinbcos2;
	real x1, y1;
	struct geom_point_w pt0, pt1, pt2;
	struct geom_bezier *bezier;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	ITEM_ASSERT(fpath, expansion);

	pt0.w = pt2.w = 1;

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	if (geom_elliptic_arc_is_closed(arc)) {
		pt1.w = 0.5;
		acos = arc->a_radius * cos1;
		asin = arc->a_radius * sin1;
		b = sqrt(3) * arc->b_radius;
		bcos = b * cos1;
		bsin = b * sin1;
		acosbsin1 = acos + bsin;
		acosbsin2 = acos - bsin;
		asinbcos1 = asin + bcos;
		asinbcos2 = asin - bcos;
		RTEST_NUM(x1 = acos + arc->center.x);
		RTEST_NUM(y1 = asin + arc->center.y);

		pt0.x = x1;
		pt0.y = y1;
		RTEST_NUM(pt1.x = acosbsin2 + arc->center.x);
		RTEST_NUM(pt1.y = asinbcos1 + arc->center.y);
		RTEST_NUM(pt2.x = -0.5 * acosbsin1 + arc->center.x);
		RTEST_NUM(pt2.y = -0.5 * asinbcos2 + arc->center.y);
		GEOM_FPATH_APPEND_NEW_CREATE2(bezier, expansion, bezier, geom_bezier_set_quadratic_rational, &pt0, &pt1, &pt2);

		pt0 = pt2;
		RTEST_NUM(pt1.x = -2 * acos + arc->center.x);
		RTEST_NUM(pt1.y = -2 * asin + arc->center.y);
		RTEST_NUM(pt2.x = -0.5 * acosbsin2 + arc->center.x);
		RTEST_NUM(pt2.y = -0.5 * asinbcos1 + arc->center.y);
		GEOM_FPATH_APPEND_NEW_CREATE2(bezier, expansion, bezier, geom_bezier_set_quadratic_rational, &pt0, &pt1, &pt2);

		pt0 = pt2;
		RTEST_NUM(pt1.x = acosbsin1 + arc->center.x);
		RTEST_NUM(pt1.y = asinbcos2 + arc->center.y);
		pt2.x = x1;
		pt2.y = y1;
		GEOM_FPATH_APPEND_NEW_CREATE2(bezier, expansion, bezier, geom_bezier_set_quadratic_rational, &pt0, &pt1, &pt2);
	}

	else {
		dif = arc->dif;
		abs_dif = fabs(dif);
		n = 1 + (int)(abs_dif * (4.0 / (3 * M_PI)));
		dif /= n;
		end = arc->start;
		mid = end - 0.5 * dif;

		pt1.w = w = cos(0.5 * dif);
		TEST_NUM(pt1.w > 0);
		a = arc->a_radius / w;
		b = arc->b_radius / w;

		x = cos(arc->start) * arc->a_radius;
		y = sin(arc->start) * arc->b_radius;
		RTEST_NUM(pt2.x = x * cos1 - y * sin1 + arc->center.x);
		RTEST_NUM(pt2.y = x * sin1 + y * cos1 + arc->center.y);

		for (i = 1; i <= n; i++) {
			end += dif;
			mid += dif;

			pt0 = pt2;
			x = cos(mid) * a;
			y = sin(mid) * b;
			RTEST_NUM(pt1.x = x * cos1 - y * sin1 + arc->center.x);
			RTEST_NUM(pt1.y = x * sin1 + y * cos1 + arc->center.y);
			x = cos(end) * arc->a_radius;
			y = sin(end) * arc->b_radius;
			RTEST_NUM(pt2.x = x * cos1 - y * sin1 + arc->center.x);
			RTEST_NUM(pt2.y = x * sin1 + y * cos1 + arc->center.y);
			GEOM_FPATH_APPEND_NEW_CREATE2(bezier, expansion, bezier, geom_bezier_set_quadratic_rational, &pt0, &pt1, &pt2);
		}
	}

	ERET(0);
}


/* geom_elliptic_arc_to_bezier */

int geom_elliptic_arc_to_bezier(struct geom_elliptic_arc *arc, struct geom_bezier *bezier)
{
	double w, mid, dif, cos1, sin1, x, y;
	struct geom_point_w pt0, pt1, pt2;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	INSTANCE_IS_ASSERT(bezier, bezier);

	dif = fabs(arc->dif);
	TEST_ERR(dif < M_PI, CONIC_TOO_LONG);

	pt0.w = pt2.w = 1;

	FASSERTX(mid = arc->start + 0.5 * arc->dif);
	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	x = cos(arc->start) * arc->a_radius;
	y = sin(arc->start) * arc->b_radius;
	RTEST_NUM(pt0.x = x * cos1 - y * sin1 + arc->center.x);
	RTEST_NUM(pt0.y = x * sin1 + y * cos1 + arc->center.y);

	x = cos(arc->start + arc->dif) * arc->a_radius;
	y = sin(arc->start + arc->dif) * arc->b_radius;
	RTEST_NUM(pt2.x = x * cos1 - y * sin1 + arc->center.x);
	RTEST_NUM(pt2.y = x * sin1 + y * cos1 + arc->center.y);

	pt1.w = w = cos(0.5 * dif);
	TEST_NUM(pt1.w > 0);
	x = cos(mid) * arc->a_radius / w;
	y = sin(mid) * arc->b_radius / w;
	RTEST_NUM(pt1.x = x * cos1 - y * sin1 + arc->center.x);
	RTEST_NUM(pt1.y = x * sin1 + y * cos1 + arc->center.y);

	geom_bezier_set_quadratic_rational(bezier, &pt0, &pt1, &pt2);

	ERET(0);
}


/* geom_elliptic_arc_from_arc */

int geom_elliptic_arc_from_arc(struct geom_elliptic_arc *elliptic_arc, struct geom_arc *arc)
{
	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, elliptic_arc);
	ITEM_ASSERT(arc, arc);

	elliptic_arc->center = arc->center;
	elliptic_arc->rotation = 0;
	elliptic_arc->a_radius = elliptic_arc->b_radius = arc->radius;
	elliptic_arc->start = arc->start;
	RASSERTX(elliptic_arc->dif = fminmaxr(arc->end - arc->start, -R_2PI, R_2PI));
	EASSERTX(geom_elliptic_arc_normalize(elliptic_arc));

	ERET(0);
}


/* geom_elliptic_arc_to_poly */

int geom_elliptic_arc_to_poly(struct geom_elliptic_arc *arc, real poly[6])
{
	double a2, b2, cos1, sin1, cos2, sin2, rotx, roty;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(poly);

	a2 = SQR(arc->a_radius);
	b2 = SQR(arc->b_radius);
	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	cos2 = SQR(cos1);
	sin2 = SQR(sin1);
	rotx = cos1 * arc->center.x + sin1 * arc->center.y;
	roty = sin1 * arc->center.x - cos1 * arc->center.y;

	poly[0] = b2 * cos2 + a2 * sin2;
	poly[1] = 2 * sin1 * cos1 * (b2 - a2);
	poly[2] = b2 * sin2 + a2 * cos2;
	poly[3] = -2 * (b2 * cos1 * rotx + a2 * sin1 * roty);
	poly[4] = -2 * (b2 * sin1 * rotx - a2 * cos1 * roty);
	poly[5] = b2 * (SQR(rotx) - a2) + a2 * SQR(roty);
	for (uns i = 0; i < 6; i++)
		RTEST_NUM(poly[i]);

	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *                TRANSFORMATIONS               *
 *                                              *
 *----------------------------------------------*/


/* geom_elliptic_arc_transform */

int geom_elliptic_arc_transform(struct geom_elliptic_arc *arc, struct geom_transform *transform, struct geom_elliptic_arc *result)
{
	/* FIXME:
	 * - predelat interni vypocty na double, zjednodusit pomoci maker v numeric.h
	 * - rychlejsi ortogonalni transformovani... hlavne podobnost */

	struct geom_point center;
	real rotation, a_radius, b_radius, start, dif;
	real cos1, sin1, cos2, sin2, sincos, x, y, tx, ty, a, b, c, det, abdet, c1, c2, c3, c4, c12, c34;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	GASSERT(transform, transform);
	INSTANCE_IS_ASSERT(elliptic_arc, result);

	if (transform->flags & GEOM_TRANSFORM_IDENTITY)
		geom_elliptic_arc_copy(arc, result);

	ETEST_RET(geom_point_transform(&arc->center, transform, &center));
	rotation = arc->rotation;
	a_radius = arc->a_radius;
	b_radius = arc->b_radius;
	start = arc->start;
	dif = arc->dif;

	/* Special case: a_radius = b_radius = 0 */
	if (IS_ZERO(a_radius) && IS_ZERO(b_radius)) {
RETURN_POINT:
		rotation = a_radius = b_radius = 0;
		if (!geom_elliptic_arc_is_closed(arc))
			start = dif = 0;
		goto RETURN;
	}

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	/* Special case: a_radius = 0, b_radius > 0 */
	if (a_radius == 0) {
		x = -transform->coef[0][0] * sin1 + transform->coef[0][1] * cos1;
		y = -transform->coef[1][0] * sin1 + transform->coef[1][1] * cos1;
		RTEST_NUM(b_radius *= hypot(x, y));
		if (IS_ZERO(b_radius))
			goto RETURN_POINT;
		RTEST_NUM(rotation = atan2(y, x) - 0.5 * M_PI);
		goto RETURN;
	}

	/* Special case: a_radius > 0, b_radius = 0 */
	if (b_radius == 0) {
		x = transform->coef[0][0] * cos1 + transform->coef[0][1] * sin1;
		y = transform->coef[1][0] * cos1 + transform->coef[1][1] * sin1;
		RTEST_NUM(a_radius *= hypot(x, y));
		if (IS_ZERO(a_radius))
			goto RETURN_POINT;
		RTEST_NUM(rotation = atan2(y, x));
		goto RETURN;
	}

	/* Compute T' = Scale(1 / a_radius, 1 / b_radius) * Rotate(rotation) * T^-1
	 * and store coefficients of (det(T) * a_radius * b_radius) * T' */
	RTEST_NUM(det = transform->coef[0][0] * transform->coef[1][1] - transform->coef[0][1] * transform->coef[1][0]);

	RTEST_NUM(abdet = fabs(arc->a_radius * arc->b_radius * det));

	if (IS_ZERO(det) || IS_ZERO(abdet))
		goto SINGULAR;
	c1 = arc->b_radius * (transform->coef[1][1] * cos1 - transform->coef[1][0] * sin1);
	c2 = arc->b_radius * (-transform->coef[0][1] * cos1 + transform->coef[0][0] * sin1);
	c3 = arc->a_radius * (-transform->coef[1][1] * sin1 - transform->coef[1][0] * cos1);
	c4 = arc->a_radius * (transform->coef[0][1] * sin1 + transform->coef[0][0] * cos1);

	/* Transformed ellipse:
	 *   0 = (T'(x, y)_x)^2 + (T'(x, y)_y)^2 - 1
	 *   0 = A * x^2 + B * xy + C * y^2 + F */
	a = SQR(c1) + SQR(c3);
	RTEST_NUM(b = 2 * (c1 * c2 + c3 * c4));
	c = SQR(c2) + SQR(c4);

	/* Rotate to clear xy coefficient */
	if (IS_ZERO(b))
		rotation = 0;
	else {
		rotation = 0.5 * atan2(b, a - c);
		cos1 = cos(rotation);
		sin1 = sin(rotation);
		cos2 = SQR(cos1);
		sin2 = SQR(sin1);
		sincos = b * sin1 * cos1;
		b = a;
		a = b * cos2 + sincos + c * sin2;
		c = b * sin2 - sincos + c * cos2;
	}

	/* Compute ellipse radii from A and C, use special code for near-singular transformation */
	RTEST_NUM(a);
	RTEST_NUM(c);

	if (IS_ZERO(a) || IS_ZERO(c))
		goto SINGULAR;
	RTEST_NUM(a = sqrt(fabs(a)));
	RTEST_NUM(c = sqrt(fabs(c)));
	a_radius = abdet / a;
	b_radius = abdet / c;

	if (!isfinite(a_radius) || !isfinite(b_radius))
		goto SINGULAR;

	goto START_END;

SINGULAR:
	TRACE("singular case");

	/* Compute T' coefficients */
	c1 = arc->a_radius * (transform->coef[0][0] * cos1 + transform->coef[0][1] * sin1);
	c2 = arc->b_radius * (-transform->coef[0][0] * sin1 + transform->coef[0][1] * cos1);
	c3 = arc->a_radius * (transform->coef[1][0] * cos1 + transform->coef[1][1] * sin1);
	c4 = arc->b_radius * (-transform->coef[1][0] * sin1 + transform->coef[1][1] * cos1);

	/* Find extreme point of {T' * (cos t, sin t)} */
	RTEST_NUM(c12 = SQR(c1) + SQR(c2));
	RTEST_NUM(c34 = SQR(c3) + SQR(c4));

	/* T' image of unit circle is single point (0, 0) */
	if (IS_ZERO(c12) && IS_ZERO(c34))
		goto RETURN_POINT;

	/* T' image of unit circle is segment (x, y)--(-x, -y) */
	else {
		if (c12 >= c34)
			y = c34 / (x = sqrt(c12));
		else
			x = c12 / (y = sqrt(c34));
		RTEST_NUM(a_radius = hypot(x, y));
		b_radius = 0;
		RTEST_NUM(rotation = atan2(y, x));
		goto START_END;
	}

START_END:
	/* Convert starting and ending angle */
	if (!geom_elliptic_arc_is_closed(arc)) {
		cos1 = cos(arc->rotation);
		sin1 = sin(arc->rotation);
		cos2 = cos(rotation);
		sin2 = sin(rotation);

		tx = arc->a_radius * cos(arc->start);
		ty = arc->b_radius * sin(arc->start);
		x = tx * cos1 - ty * sin1;
		y = tx * sin1 + ty * cos1;
		tx = transform->coef[0][0] * x + transform->coef[0][1] * y;
		ty = transform->coef[1][0] * x + transform->coef[1][1] * y;
		RTEST_NUM(x = tx * cos2 + ty * sin2);
		RTEST_NUM(y = -tx * sin2 + ty * cos2);
		start = geom_elliptic_arc_vector_to_angle(a_radius, b_radius, x, y);

		if (det < 0)
			dif = -dif;
		/*real end = arc->start + arc->dif;
		tx = arc->a_radius * cos(end);
		ty = arc->b_radius * sin(end);
		x = tx * cos1 - ty * sin1;
		y = tx * sin1 + ty * cos1;
		tx = transform->coef[0][0] * x + transform->coef[0][1] * y;
		ty = transform->coef[1][0] * x + transform->coef[1][1] * y;
		RTEST_NUM(x = tx * cos2 + ty * sin2);
		RTEST_NUM(y = -tx * sin2 + ty * cos2);
		end = geom_elliptic_arc_vector_to_angle(a_radius, b_radius, x, y);
		RTEST_NUM(b = arc->dif);
		if ((det < 0) ^ (b > 0))
			end = 2 * M_PI * round((fabs(b) - end) / (2 * M_PI)) + end;
		else
			end = -2 * M_PI * round((fabs(b) + end) / (2 * M_PI)) + end;
		RTEST_NUM(dif = end - start);*/
	}

RETURN:
	result->center = center;
	result->rotation = rotation;
	result->a_radius = a_radius;
	result->b_radius = b_radius;
	result->start = start;
	result->dif = dif;
	EASSERTX(geom_elliptic_arc_normalize(result));
	ERET(0);
}


/* geom_elliptic_arc_normalize */

int geom_elliptic_arc_normalize(struct geom_elliptic_arc *arc)
{
	FUNC;
	INSTANCE_IS_ASSERT(elliptic_arc, arc);
	GASSERT(point, &arc->center);
	RASSERT(arc->rotation);
	GASSERT2(r_minmax, arc->a_radius, 0, REAL_MAX);
	GASSERT2(r_minmax, arc->b_radius, 0, REAL_MAX);
	RASSERT(arc->start);
	RASSERT(arc->dif);

	/* dif = [-2 * PI, 2 * PI] */
	arc->dif = fminmaxr(arc->dif, -R_2PI, R_2PI);

	/* a_radius >= b_radius */
	if (arc->a_radius < arc->b_radius) {
		RSWAP(arc->a_radius, arc->b_radius);
		arc->rotation += R_PI2;
		arc->start -= R_PI2;
	}

	/* rotation = [0, PI] */
	arc->rotation = fmodfloorr(nfinsetr(arc->rotation, 0), R_2PI);
	if (arc->rotation > R_PI) {
		arc->rotation = fminmaxr(arc->rotation - R_PI, 0, R_PI);
		arc->start += R_PI;
	}

	/* start = [0, 2 * PI] */
	arc->start = fmodfloorr(nfinsetr(arc->start, 0), R_2PI);

	ITEM_ASSERT(elliptic_arc, arc);
	ERET(0);
}


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/* geom_elliptic_arc_point_start */

int geom_elliptic_arc_point_start(struct geom_elliptic_arc *arc, struct geom_point *result)
{
	double x, y, c, s;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(result);

	c = cos(arc->rotation);
	s = sin(arc->rotation);
	if (geom_elliptic_arc_is_closed(arc)) {
		result->x = arc->a_radius * c + arc->center.x;
		result->y = arc->a_radius * s + arc->center.y;
	}
	else {
		x = arc->a_radius * cos(arc->start);
		y = arc->b_radius * sin(arc->start);
		result->x = c * x - s * y + arc->center.x;
		result->y = c * y + s * x + arc->center.y;
	}
	RTEST_NUM(result->x);
	RTEST_NUM(result->y);
	ERET(0);
}


/* geom_elliptic_arc_point_end */

int geom_elliptic_arc_point_end(struct geom_elliptic_arc *arc, struct geom_point *result)
{
	double x, y, c, s, end;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(result);

	c = cos(arc->rotation);
	s = sin(arc->rotation);
	if (geom_elliptic_arc_is_closed(arc)) {
		result->x = arc->a_radius * c + arc->center.x;
		result->y = arc->a_radius * s + arc->center.y;
	}
	else {
		FASSERTX(end = (double)arc->start + arc->dif);
		x = arc->a_radius * cos(end);
		y = arc->b_radius * sin(end);
		result->x = c * x - s * y + arc->center.x;
		result->y = c * y + s * x + arc->center.y;
	}
	RTEST_NUM(result->x);
	RTEST_NUM(result->y);
	ERET(0);
}


/* geom_elliptic_arc_points */

int geom_elliptic_arc_points(struct geom_elliptic_arc *arc, struct geom_point *start, struct geom_point *end)
{
	double x, y, c, s, e;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(start);
	PASSERT(end);

	c = cos(arc->rotation);
	s = sin(arc->rotation);
	if (geom_elliptic_arc_is_closed(arc)) {
		start->x = end->x = arc->a_radius * c + arc->center.x;
		start->y = end->y = arc->a_radius * s + arc->center.y;
	}
	else {
		x = arc->a_radius * cos(arc->start);
		y = arc->b_radius * sin(arc->start);
		start->x = c * x - s * y + arc->center.x;
		start->y = c * y + s * x + arc->center.y;

		FASSERTX(e = (double)arc->start + arc->dif);
		x = arc->a_radius * cos(e);
		y = arc->b_radius * sin(e);
		end->x = c * x - s * y + arc->center.x;
		end->y = c * y + s * x + arc->center.y;

		RTEST_NUM(end->x);
		RTEST_NUM(end->y);
	}
	RTEST_NUM(start->x);
	RTEST_NUM(start->y);
	ERET(0);
}


/* geom_elliptic_arc_point_at_time */

int geom_elliptic_arc_point_at_time(struct geom_elliptic_arc *arc, real time, struct geom_point *result)
{
	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		ETEST_RET(geom_elliptic_arc_point_start(arc, result));
	else if (time >= 1)
		ETEST_RET(geom_elliptic_arc_point_end(arc, result));
	else {
		double x, y, c, s, t;
		c = cos(arc->rotation);
		s = sin(arc->rotation);
		if (geom_elliptic_arc_is_closed(arc))
			t = (2 * M_PI) * time;
		else
			FASSERTX( t = arc->start + (double)time * arc->dif);
		x = arc->a_radius * cos(t);
		y = arc->b_radius * sin(t);
		RTEST_NUM(result->x = c * x - s * y + arc->center.x);
		RTEST_NUM(result->y = c * y + s * x + arc->center.y);
	}
	ERET(0);
}


/* geom_elliptic_arc_derivation_start */

int geom_elliptic_arc_derivation_start(struct geom_elliptic_arc *arc, struct geom_vector *result)
{
	double x, y, c, s;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(result);

	c = cos(arc->rotation);
	s = sin(arc->rotation);
	if (geom_elliptic_arc_is_closed(arc)) {
		result->dx = arc->b_radius * -s;
		result->dy = arc->b_radius * c;
	}
	else {
		x = arc->a_radius * -sin(arc->start);
		y = arc->b_radius * cos(arc->start);
		result->dx = c * x - s * y;
		result->dy = c * y + s * x;
	}
	RTEST_NUM(result->dx);
	RTEST_NUM(result->dy);
	ERET(0);
}


/* geom_elliptic_arc_derivation_end */

int geom_elliptic_arc_derivation_end(struct geom_elliptic_arc *arc, struct geom_vector *result)
{
	double x, y, c, s, end;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(result);

	c = cos(arc->rotation);
	s = sin(arc->rotation);
	if (geom_elliptic_arc_is_closed(arc)) {
		result->dx = arc->b_radius * -s;
		result->dy = arc->b_radius * c;
	}
	else {
		FASSERTX(end = arc->start + arc->dif);
		x = arc->a_radius * -sin(end);
		y = arc->b_radius * cos(end);
		result->dx = c * x - s * y;
		result->dy = c * y + s * x;
	}
	RTEST_NUM(result->dx);
	RTEST_NUM(result->dy);
	ERET(0);
}


/* geom_elliptic_arc_derivation_at_time */

int geom_elliptic_arc_derivation_at_time(struct geom_elliptic_arc *arc, real time, struct geom_vector *result)
{
	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	RASSERT(time);
	PASSERT(result);

	if (time <= 0)
		ETEST_RET(geom_elliptic_arc_derivation_start(arc, result));
	else if (time >= 1)
		ETEST_RET(geom_elliptic_arc_derivation_end(arc, result));
	else {
		double x, y, c, s, t;
		c = cos(arc->rotation);
		s = sin(arc->rotation);
		if (geom_elliptic_arc_is_closed(arc))
			t = (2 * M_PI) * time;
		else
			FASSERTX(t = arc->start + (double)time * arc->dif);
		x = arc->a_radius * -sin(t);
		y = arc->b_radius * cos(t);
		RTEST_NUM(result->dx = c * x - s * y);
		RTEST_NUM(result->dy = c * y + s * x);
	}
	ERET(0);
}


/* geom_elliptic_arc_foci */

int geom_elliptic_arc_foci(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2)
{
	double cos1, sin1, c, x, y;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(focus1);
	PASSERT(focus2);

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	c = sqrt(fabs(SQR(arc->a_radius) - SQR(arc->b_radius)));

	if (arc->a_radius >= arc->b_radius) {
		x = cos1 * c;
		y = sin1 * c;
	}
	else {
		x = -sin1 * c;
		y = cos1 * c;
	}

	if (focus1) {
		RTEST_NUM(focus1->x = arc->center.x + x);
		RTEST_NUM(focus1->y = arc->center.y + y);
	}

	if (focus2) {
		RTEST_NUM(focus2->x = arc->center.x - x);
		RTEST_NUM(focus2->y = arc->center.y - y);
	}

	ERET(0);
}


/* geom_elliptic_arc_directrices */

int geom_elliptic_arc_directrices(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_vector *vector)
{
	double cos1, sin1, c, x, y;

	FUNC;
	ITEM_ASSERT(elliptic_arc, arc);
	PASSERT(point1);
	PASSERT(point2);
	PASSERT(vector);

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);

	if (arc->a_radius >= arc->b_radius) {
		c = SQR(arc->b_radius);
		c = c / sqrt(SQR(arc->a_radius) - c);
		x = cos1 * c;
		y = sin1 * c;
		if (vector) {
			vector->dx = -sin1;
			vector->dy = cos1;
		}
	}

	else {
		c = SQR(arc->a_radius);
		c = c / sqrt(SQR(arc->b_radius) - c);
		x = -sin1 * c;
		y = cos1 * c;
		if (vector) {
			vector->dx = -cos1;
			vector->dy = -sin1;
		}
	}

	if (point1) {
		RTEST_NUM(point1->x = arc->center.x + x);
		RTEST_NUM(point1->y = arc->center.x + y);
	}

	if (point2) {
		RTEST_NUM(point2->x = arc->center.x - x);
		RTEST_NUM(point2->y = arc->center.x - y);
	}

	ERET(0);
}


/* geom_elliptic_arc_point_error
 * FIXME: neco jednodussiho a stabilnejsiho pro b_radius =~ 0 */

real geom_elliptic_arc_point_error(struct geom_elliptic_arc *arc, struct geom_point *point)
{
	double x1, y1, x2, y2, cos1, sin1, ab, e;

	FUNC;

	cos1 = cos(arc->rotation);
	sin1 = sin(arc->rotation);
	x1 = point->x - arc->center.x;
	y1 = point->y - arc->center.y;
	x2 = (x1 * cos1 + y1 * sin1) * arc->b_radius;
	y2 = (-x1 * sin1 + y1 * cos1) * arc->a_radius;
	ab = SQR(arc->b_radius * arc->a_radius);
	e = (SQR(x2) + SQR(y2) - ab) / MAX(1, ab);

	RRET(e);
}


