#ifndef _VRR_GEOMLIB_ELLIPSE_H_
#define _VRR_GEOMLIB_ELLIPSE_H_


/**
 * @file geomlib/ellipse.h
 *
 * Geomlib - Ellipses and elliptic arcs.
 *
 * @sa \ref geomlib_curves_elliptic_arc "Ellipses and elliptic arcs"
 */


#include "geomlib/curve.h"
#include "geomlib/vector.h"


#define geom_elliptic_arc_VMT    geom_curve_VMT
#define geom_elliptic_arc_INIT   geom_curve_INIT \
	\
	.time_to_btime = (real (*)(void *, struct geom_expansion *, real)) &geom_elliptic_arc_time_to_btime, \
	.btime_to_time = (real (*)(void *, struct geom_expansion *, real)) &geom_elliptic_arc_btime_to_time, \
	\
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_elliptic_arc_expansion_append, \
	.to_bezier = (int (*)(void *, struct geom_bezier *)) &geom_elliptic_arc_to_bezier, \

GEOM_CLASS_HEAD(elliptic_arc, curve);


/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


static inline void geom_elliptic_arc_create(struct geom_elliptic_arc *arc)
{ geom_curve_create(&arc->curve); }

static inline void geom_elliptic_arc_destroy(struct geom_elliptic_arc *arc)
{ geom_curve_destroy(&arc->curve); }

static inline void geom_elliptic_arc_after_change(struct geom_elliptic_arc *arc)
{ geom_curve_after_change(&arc->curve); }

static inline void geom_elliptic_arc_copy(struct geom_elliptic_arc *arc, struct geom_elliptic_arc *src)
{ geom_item_copy(arc, src); }


/**
 * Sets elliptic arc from given center point, rotation, radii and endpoints angles.
 *
 * @ao  arc       Resulting elliptic arc (initialized).
 * @ai  center    Center point.
 * @ai  rotation  CCW rotation around the center point.
 * @ai  a_radius  First radius (>=0).
 * @ai  b_radius  Second radius (>=0).
 * @ai  start     Angle of starting point in unit circle.
 * @ai  dif       CCW angles difference in unit circle.
 *
 * @rv  zero  Always.
 */
void geom_elliptic_arc_set(struct geom_elliptic_arc *arc, struct geom_point *center, real rotation, real a_radius, real b_radius, real start, real dif);


/**
 * Sets entire ellipse from given center point, rotation and radii.
 *
 * @ao  arc       Resulting elliptic arc (initialized).
 * @ai  center    Center point.
 * @ai  rotation  CCW rotation around the center point.
 * @ai  a_radius  First radius (>=0).
 * @ai  b_radius  Second radius (>=0).
 *
 * @rv  zero  Always.
 */
static inline void geom_elliptic_arc_set_closed(struct geom_elliptic_arc *arc, struct geom_point *center, real rotation, real a_radius, real b_radius)
{ geom_elliptic_arc_set(arc, center, rotation, a_radius, b_radius, 0, 2 * M_PI); }


/**
 * Sets new endpoints to ellipse or elliptic arc.
 *
 * @aio  arc    Ellipse or elliptic arc.
 * @ai   start  Point very near to ellipse (or NULL).
 * @ai   end    Point very near to ellipse (or NULL).
 * @ai   mid    Point very near to ellipse (or NULL).
 * @ai   dif    Angle difference, if needed.
 * @ai   flags  Flags describing the arc to select (see GEOM_CONIC_ARC_...).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_set_endpoints(struct geom_elliptic_arc *arc, struct geom_point *start, struct geom_point *end, struct geom_point *mid, real st, real dif, uns flags);


/**
 * Sets the ellipse from 2 foci and major axis.
 *
 * If axis is too short, resulting ellipse is singular.
 *
 * @ao  arc     Resulting ellipse (initialized).
 * @ai  focus1  First focus.
 * @ai  focus2  Second focus.
 * @ai  major   Length of major axis (>=0).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_foci_major(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2, real major);


/**
 * Sets the ellipse from 2 foci going through given point.
 *
 * @ao  arc     Resulting ellipse (initialized).
 * @ai  focus1  First focus.
 * @ai  focus2  Second focus.
 * @ai  point   Point on resulting ellipse.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_foci_point(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2, struct geom_point *point);


/**
 * Sets the ellipse with smallest area going through 3 given points.
 *
 * @ao  arc     Resulting ellipse (initialized).
 * @ai  point1  First point.
 * @ai  point2  Second point.
 * @ai  point3  Third point.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_3_points_smallest(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3);


/**
 * Sets the ellipse to be circle going through 3 given points.
 *
 * @ao  arc           Resulting ellipse (initialized).
 * @ai  point1        First point.
 * @ai  point2        Second point.
 * @ai  point3        Third point.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_3_points_circle(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3);


/**
 * Sets the ellipse from 3 points, rotation and eccentricity.
 *
 * @ao  arc           Resulting ellipse (initialized).
 * @ai  point1        First point.
 * @ai  point2        Second point.
 * @ai  point3        Third point.
 * @ai  rotation      Rotation of major axis.
 * @ai  eccentricity  Ellipse eccentricity.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_3_points_rotation_eccentricity(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_point *point3, real rotation, real eccentricity);


/**
 * Sets the ellipse to be circle with given center and one point.
 *
 * @ao  arc           Resulting ellipse (initialized).
 * @ai  center        Center.
 * @ai  point         Point.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_center_point_circle(struct geom_elliptic_arc *arc, struct geom_point *center, struct geom_point *point);


/**
 * Sets the ellipse with given center, one point, rotation and eccentricity.
 *
 * @ao  arc           Resulting ellipse (initialized).
 * @ai  center        Center.
 * @ai  point         Point.
 * @ai  rotation      Rotation of major axis.
 * @ai  eccentricity  Ellipse eccentricity.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_from_center_point_rotation_eccentricity(struct geom_elliptic_arc *arc, struct geom_point *center, struct geom_point *point, real rotation, real eccentricity);


/*----------------------------------------------*
 *                                              *
 *           OVERRIDED VIRTUAL METHODS          *
 *                                              *
 *----------------------------------------------*/


/* Parametrizations */

real geom_elliptic_arc_time_to_btime(struct geom_elliptic_arc *arc, struct geom_expansion *expansion, real time);
real geom_elliptic_arc_btime_to_time(struct geom_elliptic_arc *arc, struct geom_expansion *expansion, real btime);


/* Conversions */

int geom_elliptic_arc_expansion_append(struct geom_elliptic_arc *arc, struct geom_fpath *path);
int geom_elliptic_arc_to_bezier(struct geom_elliptic_arc *arc, struct geom_bezier *bezier);


/*----------------------------------------------*
 *                                              *
 *                  NEW METHODS                 *
 *                                              *
 *----------------------------------------------*/


/**
 * Converts circular arc to elliptic arc.
 *
 * @ai  elliptic_arc  Resulting elliptic arc.
 * @ao  arc           Circular arc to convert.
 *
 * @rv  zero  Always.
 */
int geom_elliptic_arc_from_arc(struct geom_elliptic_arc *elliptic_arc, struct geom_arc *arc);


/**
 * Converts ellipse to form Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0.
 *
 * @ai  arc   Elliptic arc to convert.
 * @ao  poly  Resulting coefficents.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_to_poly(struct geom_elliptic_arc *arc, real poly[6]);


/*----------------------------------------------*
 *                                              *
 *                TRANSFORMATIONS               *
 *                                              *
 *----------------------------------------------*/


/**
 * Apply affine transformation to given elliptic arc.
 *
 * @ai  arc        Elliptic arc to transform.
 * @ai  transform  Affine transformation.
 * @ao  result     Resulting elliptic arc (possibly result = arc).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_transform(struct geom_elliptic_arc *arc, struct geom_transform *transform, struct geom_elliptic_arc *result);


/**
 * Normalize elliptic arc structure.
 * Input:
 * - rotation, a_radius, b_radius, start and dif must be finite
 * Output:
 * - rotation in [0, R_PI]
 * - a_radius >= b_radius
 * - start in [0, R_2PI]
 * - dif in [-R_2PI, R_2PI]
 *
 * @aio  arc  Elliptic arc to normalize.
 *
 * @rv  zero  Always.
 */
int geom_elliptic_arc_normalize(struct geom_elliptic_arc *arc);


/*----------------------------------------------*
 *                                              *
 *             GEOMETRIC PROPERTIES             *
 *                                              *
 *----------------------------------------------*/


/**
 * Returns starting point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ai  result  Resulting starting point.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_point_start(struct geom_elliptic_arc *arc, struct geom_point *result);


/**
 * Returns ending point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ai  result  Resulting ending point.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_point_end(struct geom_elliptic_arc *arc, struct geom_point *result);


/**
 * Returns starting and ending points of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ai  start   Resulting starting point.
 * @ai  end     Resulting ending point.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_points(struct geom_elliptic_arc *arc, struct geom_point *start, struct geom_point *end);


/**
 * Returns point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ai  time    TIME parameter.
 * @ao  result  Resulting point.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_point_at_time(struct geom_elliptic_arc *arc, real time, struct geom_point *result);


/**
 * Returns first derivation in starting point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ao  result  Resulting derivation vector.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_derivation_start(struct geom_elliptic_arc *arc, struct geom_vector *result);


/**
 * Returns first derivation in ending point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ao  result  Resulting derivation vector.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_derivation_end(struct geom_elliptic_arc *arc, struct geom_vector *result);


/**
 * Returns first derivation in point of given elliptic arc.
 *
 * @ai  arc     Elliptic arc.
 * @ai  time    TIME parameter.
 * @ao  result  Resulting derivation vector.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_elliptic_arc_derivation_at_time(struct geom_elliptic_arc *arc, real time, struct geom_vector *result);


/**
 * Returns whether the structure contains entire ellipse.
 *
 * @ai  arc  Ellipse or elliptic arc.
 *
 * @rv  nonzero  Ellipse.
 * @rv  zero     Elliptic arc.
 */
static inline int geom_elliptic_arc_is_closed(struct geom_elliptic_arc UNUSED *arc)
{ return 0; }


/**
 * Returns whether the elliptic arc is singular (zero radius or radii).
 *
 * @ai  arc  Elliptic arc.
 *
 * @rv  nonzero  Elliptic arc is singular.
 * @rv  zero     Elliptic arc is not singular.
 */
static inline int geom_elliptic_arc_is_singular(struct geom_elliptic_arc *arc)
{ return arc->b_radius == 0; }


/**
 * Returns semimajor axis of given ellipse.
 *
 * @ai  arc  Ellipse.
 *
 * @rv  finite  Semimajor axis.
 */
static inline real geom_elliptic_arc_semimajor(struct geom_elliptic_arc *arc)
{ return arc->a_radius; }


/**
 * Returns semiminor axis of given ellipse.
 *
 * @ai  arc  Ellipse.
 *
 * @rv  finite  Semiminor axis.
 */
static inline real geom_elliptic_arc_semiminor(struct geom_elliptic_arc *arc)
{ return arc->b_radius; }


/**
 * Returns eccentricity of given ellipse.
 *
 * @ai  arc  Ellipse.
 *
 * @rv  finite     Eccentricity (in interval [0, 1]).
 * @rv  otherwise  Numeric error.
 */
static inline real geom_elliptic_arc_eccentricity(struct geom_elliptic_arc *arc)
{ return sqrt(1 - SQR(geom_elliptic_arc_semiminor(arc) / (geom_elliptic_arc_semimajor(arc)))); }


/**
 * Returns focal parameter of given ellipse (distance of focus to directrix).
 *
 * @ai  arc  Ellipse.
 *
 * @rv  finite     Focal parameter (>=0).
 * @rv  otherwise  Numeric error.
 */
static inline real geom_elliptic_arc_focal_parameter(struct geom_elliptic_arc *arc)
{ real a2 = SQR(geom_elliptic_arc_semiminor(arc));
	return a2 / sqrt(SQR(geom_elliptic_arc_semimajor(arc)) - a2); }


/**
 * Returns foci of given ellipse.
 *
 * @ai  arc     Ellipse.
 * @ai  focus1  Rotated right (possibly NULL).
 * @ai  focus2  Rotated left focus (possibly NULL).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_foci(struct geom_elliptic_arc *arc, struct geom_point *focus1, struct geom_point *focus2);


/**
 * Returns directrices of given ellipse.
 *
 * @ai  arc  Ellipse.
 * @ai  point1  Rotated point on right directrix (possibly NULL).
 * @ai  point2  Rotated point on left directrix (possibly NULL).
 * @ai  vertex  Direction vector of directrices, ellipse is to the left of vector from point1 (possibly NULL).
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_elliptic_arc_directrices(struct geom_elliptic_arc *arc, struct geom_point *point1, struct geom_point *point2, struct geom_vector *vector);


/**
 * Returns value near zero for points near the ellipse.
 *
 * @ai  arc    Ellipse.
 * @ai  point  Point to test.
 *
 * @rv  ~zero       Point is near the ellipse.
 * @rv  ~nonzero    Point is far from the ellipse.
 * @rv  not_finite  Numeric error.
 */
real geom_elliptic_arc_point_error(struct geom_elliptic_arc *arc, struct geom_point *point);


#endif

