#ifndef _VRR_GEOMLIB_FPATH_H_
#define _VRR_GEOMLIB_FPATH_H_

/**
 * @file geomlib/fpath.h
 *
 * Geomlib - Fast paths.
 *
 * These paths are optimized for tail-construction.
 * Grouping is not allowed (conflicts with planned optimalizations).
 *
 * Example:
 *
 *   struct geom_fpath path;
 *   void *ptr;
 *
 *   GEOM_INSTANCE_CREATE(&path, fpath, geom_fpath_create);
 *
 *   GEOM_FPATH_APPEND_NEW_CREATE2(ptr, &path, bezier, ...bezier constructor, ...constructor parameters));
 *   GEOM_FPATH_APPEND_NEW_CREATE2(ptr, &path, bezier, ...bezier constructor, ...constructor parameters));
 *   geom_fpath_finish(fpath);
 *
 *   ...path routines are allowed here
 *
 *   GEOM_INSTANCE_DESTROY2(&path, fpath);
 */

#include "geomlib/path.h"


/*----------------------------------------------*
 *                                              *
 *                BEZIER CLASS                  *
 *                                              *
 *----------------------------------------------*/


#define geom_fpath_VMT    geom_path_VMT
#define geom_fpath_INIT   geom_path_INIT \
	\
	.destroy = (void (*)(void *)) &geom_fpath_destroy, \
	.copy = (void (*)(void *, void *)) &geom_fpath_copy, \

GEOM_CLASS_HEAD(fpath, path);


static inline void geom_fpath_create(struct geom_fpath *path)
{ geom_path_create(&path->path); }

void geom_fpath_clear(struct geom_fpath *path);
void geom_fpath_destroy(struct geom_fpath *path);
void geom_fpath_copy(struct geom_fpath *path, struct geom_path *src);


/**
 * Allocates buffer of a given size.
 * This buffer is automatically freed on path destruction.
 *
 * @ai  path  Path pointer.
 * @ai  size  Number of bytes to allocate (must be greater than zero).
 *
 * @r  Pointer to newly allocated buffer. Buffer is filled by zeros.
 */
void *geom_fpath_alloc(struct geom_fpath *path, uns size);


/**
 * Appends given curve to tail.
 *
 * @ai  path   Path pointer.
 * @ai  curve  Curve pointer.
 */
static inline void geom_fpath_append(struct geom_fpath *path, void *curve)
{ geom_group_add_tail(&path->path.group, curve); }


/**
 * Allocate new curve instance on internal path allocator.
 * This curved is automatically destroyed with path.
 *
 * @ai  path    Path pointer.
 * @ai  pclass  Curve class.
 *
 * @r  Pointer to newly allocated curve instance.
 */
static inline void *geom_fpath_append_new(struct geom_fpath *path, struct geom_class *pclass)
{
	void *curve = geom_fpath_alloc(path, geom_instance_size(pclass));
	geom_instance_zero_init(curve, pclass);
	geom_fpath_append(path, curve);
	return curve;
}


static inline void *geom_fpath_append_dup(struct geom_fpath *path, void *curve)
{
	void *copy = geom_fpath_append_new(path, geom_instance_class(curve));
	geom_item_copy(copy, curve);
	return copy;
}


#define GEOM_FPATH_APPEND_NEW(path, nclass) ((struct geom_##nclass *)geom_fpath_append_new(path, GEOM_CLASS(nclass)))
#define GEOM_FPATH_APPEND_NEW_CREATE(pinstance, path, nclass, constructor) \
	do{ pinstance = GEOM_FPATH_APPEND_NEW(path, nclass); constructor((void *)pinstance); }while(0)
#define GEOM_FPATH_APPEND_NEW_CREATE2(pinstance, path, nclass, constructor, ...) \
	do{ pinstance = GEOM_FPATH_APPEND_NEW(path, nclass); constructor((void *)pinstance, __VA_ARGS__); }while(0)


/**
 * Should be called after fpath tail-construction and before other path functions.
 * Reserved for planned optimalizations.
 *
 * @ai  path  Path pointer.
 */
static inline void geom_fpath_finish(struct geom_fpath UNUSED *path) {}



#endif


