#ifndef _VRR_GEOM_MATRIX_H_
#define _VRR_GEOM_MATRIX_H_

/**
	\file geomlib/matrix.h

	\brief Matrix manipulation functions.
*/

#include "lib/lib.h"
#include <math.h>
#include "geomlib/types.h"


/**
	\brief Maximum number of matrix rows and colums.

	\see geom_matrix
*/
#define GEOM_MATRIX_MAX_SIZE 1024


/**
	\brief Macro returns coefficient in given position of a matrix.

	\param  a  Pointer to geom_matrix structure.
	\param  i  Row index. Must be of uns type and i < a->m.
	\param  j  Columns index. Must be of uns type and j < a->n.

	\return    Matrix coefficient (double).
*/
#define GEOM_MATRIX_DATA(a, i, j) ((a)->data[(j) + ((i) << (a)->w)])


/**
	\brief Macro returns size of buffer for #geom_matrix_init_buf.

	\param  m  Number of rows (1 <= m <= #GEOM_MATRIX_MAX_SIZE).
	\param  n  Number of columns (1 <= n <= #GEOM_MATRIX_MAX_SIZE).

	\return    Buffer size.
*/
#define GEOM_MATRIX_BUF_SIZE(m, n) (2 * ((m) - 1) * (n) + (n))


/**
	\def GEOM_MATRIX_ASSERT(a)

	\brief Makes assertions to test validity of matrix pointers and sizes.
		Use both #GEOM_MATRIX_ASSERT and #GEOM_MATRIX_DATA_ASSERT
		to assert the whole structure.

	\param  a  Pointer to geom_matrix structre.
*/
#ifdef GEOM_DEBUG
#define GEOM_MATRIX_ASSERT(a) do{ \
		GEOM_BASSERT((a) != NULL); \
		GEOM_BASSERT((a)->m > 0); \
		GEOM_BASSERT((a)->m <= GEOM_MATRIX_MAX_SIZE); \
		GEOM_BASSERT((a)->n > 0); \
		GEOM_BASSERT((a)->n <= GEOM_MATRIX_MAX_SIZE); \
		GEOM_BASSERT((uns)(1 << (a)->w) >= (a)->n); \
		GEOM_BASSERT((a)->data != NULL); \
	}while(0)
#else
#define GEOM_MATRIX_ASSERT(a) do{}while(0)
#endif


/**
	\def GEOM_MATRIX_DATA_ASSERT(a)

	\brief Makes assertions to test validity of matrix coefficients.
		Use both #GEOM_MATRIX_ASSERT and #GEOM_MATRIX_DATA_ASSERT
		to assert the whole structure.

	\param  a  Pointer to geom_matrix structure
*/
#ifdef GEOM_DEBUG
#define GEOM_MATRIX_DATA_ASSERT(a) do{ \
	for (uns i = 0; i < (a)->m; i++) \
		for (uns j = 0; j < (a)->n; j++) \
			GEOM_BASSERT(isfinite(GEOM_MATRIX_DATA(a, i, j))); \
	}while(0)
#else
#define GEOM_MATRIX_DATA_ASSERT(a) do{}while(0)
#endif


/**
	\brief Dumps geom_matrix structure data.

	\param  a  Pointer to geom_matrix structure (could have invalid data).
*/
void geom_matrix_dump(const struct geom_matrix *a);


/**
	\brief Initialize a matrix to given size.
		Data buffer is dynamically allocated and should be cleaned up
		by #geom_matrix_destroy. New matrix has undefined coefficients.

	\param  m  Number of rows (1 <= m <= #GEOM_MATRIX_MAX_SIZE).
	\param  n  Number of columns (1 <= n <= #GEOM_MATRIX_MAX_SIZE).
	\param  a  Resulting matrix.
*/
void geom_matrix_init(const uns m, const uns n, struct geom_matrix *a);


/**
	\brief Cleans up matrix buffer allocated by #geom_matrix_init.

	\param  a  Matrix to clean up.
*/
void geom_matrix_destroy(struct geom_matrix *a);


/**
	\brief Initialize matrix structure to given size.
		Matrix data are stored to given preallocated buffer so
		no structure cleanup is necessary. New matrix has undefined coefficients.

	\param  buf       Pointer to preallocated buffer.
	\param  buf_size  Number of buffer elements (must be large enough).
	\param  m         Number of rows (1 <= m <= #GEOM_MATRIX_MAX_SIZE).
	\param  n         Number of columns (1 <= n <= #GEOM_MATRIX_MAX_SIZE).

	\see #GEOM_MATRIX_BUF_SIZE
*/
void geom_matrix_init_buf(double *buf, const uns buf_size, const uns m, const uns n, struct geom_matrix *a);


/**
	\brief Initialize matrix structure to given size.
		Matrix data points to rectangular area of another matrix.
		Any change to new matrix affets original matrix.
		No structure cleanup is necessary.

	\param  a  Pointer to original matrix (must be initialized).
	\param  i  Submatrix top row index in original matrix (i < a->m).
	\param  j  Submatrix left column index in original matrix (j < a->n).
	\param  m  Number of submatrix rows (1 <= m <= a->m - i).
	\param  n  Number of submatrix colums (1 <= n <= a->n - j).
	\param  r  Resulting matrix.

*/
void geom_matrix_init_submatrix(const struct geom_matrix *a, const uns i, const uns j, const uns m, const uns n, struct geom_matrix *r);


/**
	\brief Function fills given matrix with zeros.

	\param  a  Matrix to fill.
*/
void geom_matrix_set_zero(struct geom_matrix *a);


/**
	\brief Function sets given square matrix to identity.

	\param  a  Matrix to set.
*/
void geom_matrix_set_identity(struct geom_matrix *a);


/**
	\brief Makes a copy of matrix.

	\param  a  Matrix to copy from.
	\param  r  Matrix to copy to (possibly r = a).
*/
void geom_matrix_copy(const struct geom_matrix *a, struct geom_matrix *r);


/**
	\brief Compute R = A * B.
		Matrix R must have as many rows as A and as many colums as B.
		Number of A colums and B rows must be equal.

	\param  a    Matrix A.
	\param  b    Matrix B (possibly a = b).
	\param  r    Resulting matrix R (must differ from both a and b).

	\retval  0   Success.
	\retval  #GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_matrix_mul(const struct geom_matrix *a, const struct geom_matrix *b, struct geom_matrix *r);


/**
	\brief PLU factorization of a matrix.

		Uses Gaussian elimination for rectangular matrix A
		to compute a permutation matrix P,
		a lower triangular matrix L with 1's on diagonal and
		an upper trapezoidal matrix U so that A = P * L * U.

		U is same size as A. P and L are square, with as many rows as A.

		P is returned in one-dimensional array p,
		where p[i] defines matrix P_i swapping rows i and p[i]
		and P = P_0 * P_1 * ... * P_k, where k = Rank(A) - 1.

		If l = u, resulting matrix contains nonzero areas of both L and U matrices
		(ones on L diagonal are not stored).

	\param  a     Rectangular matrix A.
	\param  p     Buffer to store resulting permutation (possibly NULL).
		          On exit, p has Rank(A) items.
	\param  c     Buffer to store diagonal of matrix U (possibly NULL).
		          On exit, c has rank(A) items and pairs (i, c[i]) form diagonal of matrix U.
	\param  l     Resulting matrix L (possibly l = NULL or l = u).
		          If l = u, matrix is not necessary square, but must be same size as A.
	\param  u     Resulting matrix U (possibly u = a).

	\retval  >=0  Rank of matrix A.
	\retval  #GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_matrix_plu(const struct geom_matrix *a, uns *p, uns *c, struct geom_matrix *l, struct geom_matrix *u);


/**
	\brief Compute rank of rectangular matrix A.

	\param  a     Rectangular matrix A.
	\param  temp  Temporary matrix (same size as A, possibly temp = a).

	\retval  >=0  Rank(A).
	\retval  #GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_matrix_rank(const struct geom_matrix *a, struct geom_matrix *temp);


/**
	\brief Compute determinant of square matrix A.

	\param  a     Any square matrix A.
	\param  temp  Temporary matrix (same size as A, possibly temp = a).

	\return       Matrix determinant, NAN for numeric error.
*/
double geom_matrix_determinant(const struct geom_matrix *a, struct geom_matrix *temp);


/**
	\brief Finds particular solution to a system of
		linear equations A * x = b. The general solution
		is x + Ker(A), where Ker(A) is kernel of matrix A.

	\param  a     Rectangular matrix A with m rows and n columns.
	\param  b     m-dimensional vector b.
	\param  temp  Temporary matrix (same size as A, possibly temp = a).
	\param  x     n-dimensional resulting vector (possibly b = x).
	\param  ker   Resulting kernel matrix (n rows, at least Dim(Ker(A)) columns, possibly NULL).

	\retval  >=0  Dim(Ker(A)) (0 for exactly one solution).
	\retval  #GEOM_ERR_NUMERIC             Numeric error.
	\retval  #GEOM_ERR_MATRIX_NO_SOLUTION  System has no solution.
*/
int geom_matrix_solve(const struct geom_matrix *a, const double *b, struct geom_matrix *temp, double *x, struct geom_matrix *ker);


/**
	\brief Finds particular solutions to more systems of
		linear equations A * x_i = b_i with same matrix A.
		The general solution is x_i + Ker(A), where Ker(A) is kernel of matrix A.

		For i-th system of equations,
		vector b_i is i-th column of matrix B and
		vector x_i is i-th column of resulting matrix X.

	\param  a     Rectangular matrix A with m rows and n columns.
	\param  b     Matrix B (m rows, k columns).
	\param  temp  Temporary matrix (same size as A, possibly temp = a).
	\param  x     Resulting matrix X (n rows, k columns, possibly b = x).
	\param  ker   Resulting kernel matrix Ker(A) (n rows, at least Dim(Ker(A)) columns, possibly NULL).

	\retval  >=0  Dim(Ker(A)) (0 for exactly one solution).
	\retval  #GEOM_ERR_NUMERIC             Numeric error.
	\retval  #GEOM_ERR_MATRIX_NO_SOLUTION  No solution to one or more systems.
*/
int geom_matrix_solve_systems(const struct geom_matrix *a, const struct geom_matrix *b, struct geom_matrix *temp, struct geom_matrix *x, struct geom_matrix *ker);


/**
	\brief Uses pivoting LU factorization and the resulting
		reduced row echelon form to find matrix K
		whose columns are a basis for the kernel of A.

	\param  a     Rectangular matrix A with m rows and n columns.
	\param  temp  Temporary matrix (same size as A, possibly a = temp).
	\param  ker   Resulting matrix Ker(A) (n rows, at least Dim(Ker(A)) columns, possibly a = ker or temp = ker).

	\retval  >=0  Dim(Ker(A)).
	\retval  #GEOM_ERR_NUMERIC  Numeric error.
*/
int geom_matrix_kernel(const struct geom_matrix *a, struct geom_matrix *temp, struct geom_matrix *ker);


/**
	\brief Finds inverse matrix Inv(A) to square matrix A.

	\param  a     Square matrix A.
	\param  temp  Temporary matrix (same size as A, possibly a = temp).
	\param  inv   Resulting matrix Inv(A) (same size as A, must be temp != inv).

	\retval  0    Success.
	\retval  #GEOM_ERR_NUMERIC  Numeric error.
	\retval  #GEOM_ERR_MATRIX_SINGULAR  Inverse matrix does not exist (A singular).
*/
int geom_matrix_inverse(const struct geom_matrix *a, struct geom_matrix *temp, struct geom_matrix *inv);


#endif

