#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/polynomial.h"

/* local functions */

/* closed form solvers */
static int geom_solve_quadric(double coef[3], int flags, double result[2]);
static int geom_solve_cubic(double coef[4], int flags, double result[3]);
static int geom_solve_quartic(double coef[5], int flags, double result[4]);

/* generic solver */
static int geom_solve_high_degree(int degree, double coef[], int flags, double result[]);
static int geom_rpoly(int degree, double coef[], double result_r[], double result_i[]);
static void geom_rpoly_quad(double a, double b1, double c, double *sr, double *si, double *lr, double *li);
static void geom_rpoly_fxshfr(int l2, int *nz);
static void geom_rpoly_quadit(double *uu, double *vv, int *nz);
static void geom_rpoly_realit(double *sss, int *nz, int *iflag);
static void geom_rpoly_calcsc(int *type);
static void geom_rpoly_nextk(int *type);
static void geom_rpoly_newest(int type, double *uu, double *vv);
static void geom_rpoly_quadsd(int n, double *u, double *v, double *p, double *q, double *a, double *b);


/* geom_polynomial_valid */

int geom_polynomial_valid(int degree, double coef[])
{
	int i;

	if (degree < 0 || degree > GEOM_MAX_DEGREE || !coef)
		return 0;

	for (i = 0; i <= degree; i++)
		if (!isfinite(coef[i]))
			return 0;

	return 1;
}


/* geom_polynomial_evaluate */

double geom_polynomial_evaluate(int degree, double coef[], double t)
{
	int i;
	double x;

	BASSERT(geom_polynomial_valid(degree, coef) && isfinite(t));

	x = 0;
	for (i = degree; i >= 0; i--)
		x = t * x + coef[i];

	if (!isfinite(x))
		return NAN;

	return x;
}


/* geom_polynomial_mul */

int geom_polynomial_mul(int degree1, double coef1[], int degree2, double coef2[], double result[])
{
	int n, i, j;

	BASSERT(degree1 + degree2 <= GEOM_MAX_DEGREE && result &&
	       geom_polynomial_valid(degree1, coef1) && geom_polynomial_valid(degree2, coef2));

	n = degree1 + degree2;

	for (i = 0; i <= n; i++)
		result[i] = 0;

	for (i = 0; i <= degree1; i++)
		for (j = 0; j <= degree2; j++)
			result[i + j] += coef1[i] * coef2[j];

	for (i = 0; i <= n; i++)
		if (!isfinite(result[i]))
			return -1;

	return n;
}


/* geom_polynomial_solve */

int geom_polynomial_solve(int degree, double coef[], int flags, double result[])
{
	int n, m, i;
	double temp[GEOM_MAX_DEGREE], *p;

	BASSERT(geom_polynomial_valid(degree, coef) && result);

	if (flags & GEOM_SOLVE_LEFT_ONLY)
		p = temp;
	else
		p = result;

	/* Remove leading zeros. */
	while (degree > 0 && IS_ZERO(coef[degree]))
		degree--;

	switch (degree) {
	case 0:
		return 0;

	case 1:
		p[0] = -coef[0] / coef[1];
		if (!isfinite(p[0]))
			return -1;
		n = 1;
		break;

	case 2:
		n = geom_solve_quadric(coef, flags, p);
		break;

	case 3:
		n = geom_solve_cubic(coef, flags, p);
		break;

	case 4:
		n = geom_solve_quartic(coef, flags, p);
		break;

	default:
		if (degree <= GEOM_MAX_DEGREE) {
			n = geom_solve_high_degree(degree, coef, flags, p);
			break;
		} else
			return -1;
	}

	if (n < 0)
		return -1;

	if (flags & GEOM_SOLVE_UNIT_INTERVAL) {
		m = 0;
		for (i = 0; i < n; i++)
			if (p[i] >= 0 - EQ_EPSILON && p[i] <= 1 + EQ_EPSILON) {
				if (p[i] <= 0)
					p[m] = 0;
				else if (p[i] >= 1)
					p[m] = 1;
				else
					p[m] = p[i];
				if (flags & GEOM_SOLVE_MULTIPLICITY || m == 0 || p[m] != p[m - 1])
					m++;
			}
		n = m;
	}

	if (flags & GEOM_SOLVE_LEFT_ONLY) {
		if (n == 0)
			return 0;
		else {
			result[0] = p[0];
			return 1;
		}
	}

	if (!(flags & GEOM_SOLVE_MULTIPLICITY)) {
		m = 0;
		for (i = 0; i < n; i++)
			if (!m || p[m - 1] != p[i])
				p[m++] = p[i];
		n = m;
	}

	return n;
}


/* geom_solve_quadric */

static int geom_solve_quadric(double coef[3], int flags UNUSED, double result[2])
{
	double p, q, D;

	BASSERT(geom_polynomial_valid(2, coef) && result);

	p = -coef[1] / (2 * coef[2]);
	q = coef[0] / coef[2];

	D = p * p - q;

	if (!isfinite(D))
		return -1;

	if (IS_ZERO(D)) {	/* one double solution */
		result[0] = p;
		result[1] = p;

		return 2;
	} else if (D < 0) {	/* no solution */
		return 0;
	} else {		/* two single solutions */

		D = sqrt(D);

		result[0] = -D + p;
		result[1] = D + p;

		if (!isfinite(result[0]) || !isfinite(result[1]))
			return -1;

		return 2;
	}
}

/* geom_solve_cubic */

static int geom_solve_cubic(double coef[4], int flags UNUSED, double result[3])
{
	double a, b, c, sub, aa, p, q, D, t, phi;

	BASSERT(geom_polynomial_valid(3, coef) && result);

	/* normal form: x^3 + ax^2 + px + q = 0 */

	a = coef[2] / coef[3];
	b = coef[1] / coef[3];
	c = coef[0] / coef[3];

	/* substitute x = y - A/3 = y + sub */

	sub = -1.0 / 3 * a;

	if (!isfinite(sub))
		return -1;

	/* eliminate quadratic term: y^3 + py + q = 0 */

	aa = a * a;
	p = 1.0 / 3 * (-1.0 / 3 * aa + b);
	q = 1.0 / 2 * (2.0 / 27 * a * aa - 1.0 / 3 * a * b + c);


	/* use Cardano's formula */

	D = q * q + p * p * p;

	if (!isfinite(D))
		return -1;

	if (IS_ZERO(D)) {
		if (IS_ZERO(q)) {	/* one triple solution */
			result[0] = sub;
			result[1] = sub;
			result[2] = sub;
			return 3;
		} else {	/* one single and one double solution */

			q = cbrt(-q);
			if (q > 0) {
				result[0] = sub - q;
				result[1] = sub + 2 * q;
			} else {
				result[1] = sub - q;
				result[0] = sub + 2 * q;
			}

			if (!isfinite(result[0]) || !isfinite(result[1]))
				return -1;
			return 2;
		}
	} else if (D < 0) {	/* Casus irreducibilis: three real solutions */
		p = sqrt(-p);
		phi = 1.0 / 3 * acos(-q / (p * p * p));
		t = 2 * p;
		result[0] = sub + t * cos(phi);
		result[1] = sub - t * cos(phi + M_PI / 3);
		result[2] = sub - t * cos(phi - M_PI / 3);
		geom_doubles_sort(3, result);

		if (!isfinite(result[0]) || !isfinite(result[1]) || !isfinite(result[2]))
			return -1;
		return 3;
	} else {		/* one real solution */
		D = sqrt(D);
		result[0] = sub + cbrt(D - q) - cbrt(D + q);

		if (!isfinite(result[0]))
			return -1;
		return 1;
	}
}

/* geom_solve_quartic */

static int geom_solve_quartic(double coef[5], int flags, double result[4])
{
	int num, i;
	double a, b, c, d, sub, aa, p, q, r, tcoef[4], z, u, v;

	BASSERT(geom_polynomial_valid(4, coef) && result);

	/* normal form: x^4 + Ax^3 + Bx^2 + Cx + D = 0 */

	a = coef[3] / coef[4];
	b = coef[2] / coef[4];
	c = coef[1] / coef[4];
	d = coef[0] / coef[4];

	/* substitute x = y - A / 4 = y + sub */

	sub = -1.0 / 4 * a;

	if (!isfinite(sub))
		return -1;

	/* eliminate cubic term: x^4 + px^2 + qx + s = 0 */

	aa = a * a;
	p = -3.0 / 8 * aa + b;
	q = 1.0 / 8 * aa * a - 1.0 / 2 * a * b + c;
	r = -3.0 / 256 * aa * aa + 1.0 / 16 * aa * b - 1.0 / 4 * a * c + d;

	if (!isfinite(r))
		return -1;

	if (IS_ZERO(r)) {
		/* no absolute term: y(y^3 + py + q) = 0 */

		tcoef[0] = q;
		tcoef[1] = p;
		tcoef[2] = 0;
		tcoef[3] = 1;

		num = geom_solve_cubic(tcoef, flags, result);
		if (num < 0)
			return -1;

		result[num++] = 0;
	} else {
		/* solve the resolvent cubic ... */

		tcoef[0] = 1.0 / 2 * r * p - 1.0 / 8 * q * q;
		tcoef[1] = -r;
		tcoef[2] = -1.0 / 2 * p;
		tcoef[3] = 1;

		if (geom_solve_cubic(tcoef, flags, result) < 0)
			return -1;

		/* ... and take the one real solution ... */

		z = result[0];

		/* ... to build two quadric equations */

		u = z * z - r;
		v = 2 * z - p;

		if (!isfinite(u) || !isfinite(v))
			return -1;

		if (IS_ZERO(u))
			u = 0;
		else if (u > 0)
			u = sqrt(u);
		else
			return 0;

		if (IS_ZERO(v))
			v = 0;
		else if (v > 0)
			v = sqrt(v);
		else
			return 0;

		tcoef[0] = z - u;
		tcoef[1] = q < 0 ? -v : v;
		tcoef[2] = 1;

		num = geom_solve_quadric(tcoef, flags, result);

		if (num < 0)
			return -1;

		tcoef[0] = z + u;
		tcoef[1] = -tcoef[1];

		i = geom_solve_quadric(tcoef, flags, result + num);

		if (i < 0)
			return -1;

		num += i;
	}

	/* resubstitute */

	for (i = 0; i < num; i++) {
		result[i] += sub;
		if (!isfinite(result[i]))
			return -1;
	}

	geom_doubles_sort(num, result);

	return num;
}

/* geom_solve_high_degree */

static int geom_solve_high_degree(int degree, double coef[], int flags UNUSED, double result[])
{
	double u[GEOM_MAX_DEGREE], v[GEOM_MAX_DEGREE];
	int i, n, m;

	m = geom_rpoly(degree, coef, u, v);

	n = 0;
	for (i = 0; i < m; i++) {
		if (isfinite(v[i]) && isfinite(u[i]) && v[i] == 0)
			result[n++] = u[i];
	}

	geom_doubles_sort(n, result);

	return n;
}

/* RPOLY
 *     Jenkins-Traub generic real polynomial root finder.
 *     Translation of TOMS493 FORTRAN code.
 *
 * Input:
 *     degree - integer degree of polynomial
 *     coef - double precision vector of coefficients in order of increasing powers,
 *            leading coefficient must be non-zero
 *
 * Output:
 *     result_r, result_i - output double precision vectors of the real and imaginary parts of the zeros
 *
 * Return:
 *     returnval - -1 if leading coefficient is zero, otherwise number of roots found.
 */


/* global varriables */

static int geom_rpoly_n;
static double *geom_rpoly_p, *geom_rpoly_qp, *geom_rpoly_k, *geom_rpoly_qk;
static double geom_rpoly_sr, geom_rpoly_si, geom_rpoly_u, geom_rpoly_v;
static double geom_rpoly_a, geom_rpoly_b, geom_rpoly_c, geom_rpoly_d;
static double geom_rpoly_a1, geom_rpoly_a3, geom_rpoly_a7;
static double geom_rpoly_e, geom_rpoly_f, geom_rpoly_g, geom_rpoly_h;
static double geom_rpoly_szr, geom_rpoly_szi, geom_rpoly_lzr, geom_rpoly_lzi;

/* Machine constants. */

#define GEOM_RPOLY_BASE FLT_RADIX
#define GEOM_RPOLY_ETA DBL_EPSILON
#define GEOM_RPOLY_INFIN DBL_MAX
#define GEOM_RPOLY_SMALLNO DBL_MIN
#define GEOM_RPOLY_ARE GEOM_RPOLY_ETA
#define GEOM_RPOLY_MRE GEOM_RPOLY_ETA
#define GEOM_RPOLY_LO (GEOM_RPOLY_SMALLNO / GEOM_RPOLY_ETA)

/* geom_rpoly */

int geom_rpoly(int degree, double coef[], double result_r[], double result_i[])
{
	double p[GEOM_MAX_DEGREE + 1], qp[GEOM_MAX_DEGREE + 1];
	double k[GEOM_MAX_DEGREE + 1], qk[GEOM_MAX_DEGREE + 1];
	double pt[GEOM_MAX_DEGREE + 1], temp[GEOM_MAX_DEGREE + 1];
	double t, aa, bb, cc, factor, rot;
	double max, min, xx, yy, cosr, sinr, xxx, x, sc, bnd;
	double xm, ff, df, dx;
	int cnt, nz, i, j, jj, l, nm1, zerok;

	/* Initialization of constants for shift rotation. */
	xx = sqrt(0.5);
	yy = -xx;
	rot = 94.0 * 0.017453293;
	cosr = cos(rot);
	sinr = sin(rot);
	geom_rpoly_n = degree;

	/* Algorithm fails of the leading coefficient is zero. */
	if (coef[geom_rpoly_n] == 0.0)
		return -1;

	/* Remove the zeros at the origin, if any. */
	j = 0;
	while (coef[j] == 0.0) {
		result_r[j] = 0.0;
		result_i[j] = 0.0;
		j++;
	}
	if (geom_rpoly_n <= j)
		return -1;

	/* Allocate global arrays. */
	geom_rpoly_p = p;
	geom_rpoly_qp = qp;
	geom_rpoly_k = k;
	geom_rpoly_qk = qk;

	/* Make a copy of the coefficients. */
	geom_rpoly_n -= j;
	for (i = 0; i <= geom_rpoly_n; i++)
		geom_rpoly_p[i] = coef[geom_rpoly_n - i];

	/* Start the algorithm for one zero. */
      _40:
	if (geom_rpoly_n == 1) {
		result_r[degree - 1] = -geom_rpoly_p[1] / geom_rpoly_p[0];
		result_i[degree - 1] = 0.0;
		geom_rpoly_n -= 1;
		goto _99;
	}

	/* Calculate the final zero or pair of zeros. */
	if (geom_rpoly_n == 2) {
		geom_rpoly_quad(geom_rpoly_p[0], geom_rpoly_p[1], geom_rpoly_p[2], &result_r[degree - 2],
				&result_i[degree - 2], &result_r[degree - 1], &result_i[degree - 1]);
		geom_rpoly_n -= 2;
		goto _99;
	}

	/* Find largest and smallest moduli of coefficients. */
	max = 0.0;
	min = GEOM_RPOLY_INFIN;
	for (i = 0; i <= geom_rpoly_n; i++) {
		x = fabs(geom_rpoly_p[i]);
		if (x > max)
			max = x;
		if (x != 0.0 && x < min)
			min = x;
	}

	/* Scale if there are large or very small coefficients.
	 * Computes a scale factor to multiply the coefficients of the
	 * polynomial. The scaling si done to avoid overflow and to
	 * avoid undetected underflow interfering with the convergence
	 * criterion. The factor is a power of the base. */
	sc = GEOM_RPOLY_LO / min;
	if (sc > 1.0 && GEOM_RPOLY_INFIN / sc < max)
		goto _110;
	if (sc <= 1.0) {
		if (max < 10.0)
			goto _110;
		if (sc == 0.0)
			sc = GEOM_RPOLY_SMALLNO;
	}
	l = (int) (log(sc) / log(GEOM_RPOLY_BASE) + 0.5);
	factor = pow(GEOM_RPOLY_BASE * 1.0, l);
	if (factor != 1.0)
		for (i = 0; i <= geom_rpoly_n; i++)
			geom_rpoly_p[i] = factor * geom_rpoly_p[i];
      _110:

	/* Compute lower bound on moduli of roots. */
	for (i = 0; i <= geom_rpoly_n; i++)
		pt[i] = (fabs(geom_rpoly_p[i]));
	pt[geom_rpoly_n] = -pt[geom_rpoly_n];

	/* Compute upper estimate of bound. */
	x = exp((log(-pt[geom_rpoly_n]) - log(pt[0])) / (double) geom_rpoly_n);

	/* If Newton step at the origin is better, use it. */
	if (pt[geom_rpoly_n - 1] != 0.0) {
		xm = -pt[geom_rpoly_n] / pt[geom_rpoly_n - 1];
		if (xm < x)
			x = xm;
	}

	/* Chop the interval (0,x) until ff <= 0 */
	while (1) {
		xm = x * 0.1;
		ff = pt[0];
		for (i = 1; i <= geom_rpoly_n; i++)
			ff = ff * xm + pt[i];
		if (ff <= 0.0)
			break;
		x = xm;
	}
	dx = x;

	/* Do Newton interation until x converges to two decimal places. */
	while (fabs(dx / x) > 0.005) {
		ff = pt[0];
		df = ff;
		for (i = 1; i < geom_rpoly_n; i++) {
			ff = ff * x + pt[i];
			df = df * x + ff;
		}
		ff = ff * x + pt[geom_rpoly_n];
		dx = ff / df;
		x -= dx;
	}
	bnd = x;

	/* Compute the derivative as the initial k polynomial and do 5 steps with no shift. */
	nm1 = geom_rpoly_n - 1;
	for (i = 1; i < geom_rpoly_n; i++)
		geom_rpoly_k[i] = (double) (geom_rpoly_n - i) * geom_rpoly_p[i] / (double) geom_rpoly_n;
	geom_rpoly_k[0] = geom_rpoly_p[0];
	aa = geom_rpoly_p[geom_rpoly_n];
	bb = geom_rpoly_p[geom_rpoly_n - 1];
	zerok = (geom_rpoly_k[geom_rpoly_n - 1] == 0);
	for (jj = 0; jj < 5; jj++) {
		cc = geom_rpoly_k[geom_rpoly_n - 1];
		if (!zerok) {
			/* Use a scaled form of recurrence if value of k at 0 is nonzero. */
			t = -aa / cc;
			for (i = 0; i < nm1; i++) {
				j = geom_rpoly_n - i - 1;
				geom_rpoly_k[j] = t * geom_rpoly_k[j - 1] + geom_rpoly_p[j];
			}
			geom_rpoly_k[0] = geom_rpoly_p[0];
			zerok = (fabs(geom_rpoly_k[geom_rpoly_n - 1]) <= fabs(bb) * GEOM_RPOLY_ETA * 10.0);
		} else {
			/* Use unscaled form of recurrence. */
			for (i = 0; i < nm1; i++) {
				j = geom_rpoly_n - i - 1;
				geom_rpoly_k[j] = geom_rpoly_k[j - 1];
			}
			geom_rpoly_k[0] = 0.0;
			zerok = (geom_rpoly_k[geom_rpoly_n - 1] == 0.0);
		}
	}

	/* Save k for restarts with new shifts. */
	for (i = 0; i < geom_rpoly_n; i++)
		temp[i] = geom_rpoly_k[i];

	/* Loop to select the quadratic corresponding to each new shift. */
	for (cnt = 0; cnt < 20; cnt++) {
		/* Quadratic corresponds to a double shift to
		 * a non-real point and its complex conjugate. The point
		 * has modulus bnd and amplitude rotated by 94 degrees
		 * from the previous shift. */
		xxx = cosr * xx - sinr * yy;
		yy = sinr * xx + cosr * yy;
		xx = xxx;
		geom_rpoly_sr = bnd * xx;
		geom_rpoly_si = bnd * yy;
		geom_rpoly_u = -2.0 * geom_rpoly_sr;
		geom_rpoly_v = bnd;
		geom_rpoly_fxshfr(20 * (cnt + 1), &nz);
		if (nz != 0) {
			/* The second stage jumps directly to one of the third
			 * stage iterations and returns here if successful.
			 * Deflate the polynomial, store the zero or zeros and
			 * return to the main algorithm. */
			j = degree - geom_rpoly_n;
			result_r[j] = geom_rpoly_szr;
			result_i[j] = geom_rpoly_szi;
			geom_rpoly_n -= nz;
			for (i = 0; i <= geom_rpoly_n; i++)
				geom_rpoly_p[i] = geom_rpoly_qp[i];
			if (nz != 1) {
				result_r[j + 1] = geom_rpoly_lzr;
				result_i[j + 1] = geom_rpoly_lzi;
			}
			goto _40;
		}

		/* If the iteration is unsuccessful another quadratic is chosen after restoring k. */
		for (i = 0; i < geom_rpoly_n; i++)
			geom_rpoly_k[i] = temp[i];
	}

	/* Return with failure if no convergence with 20 shifts. */
      _99:
	return degree - geom_rpoly_n;
}


/* Computes up to L2 fixed shift k-polynomials, testing for convergence in the linear or quadratic
 * case. Initiates one of the variable shift iterations and returns with the number of zeros found. */

void geom_rpoly_fxshfr(int l2, int *nz)
{
	double svu, svv, ui, vi, s;
	double betas, betav, oss, ovv, ss, vv, ts, tv;
	double ots, otv, tvv, tss;
	int type, i, j, iflag, vpass, spass, vtry, stry;
	double svk[GEOM_MAX_DEGREE + 1];

	*nz = 0;
	betav = 0.25;
	betas = 0.25;
	oss = geom_rpoly_sr;
	ovv = geom_rpoly_v;
	ots = otv = 0;

	/* Evaluate polynomial by synthetic division. */
	geom_rpoly_quadsd(geom_rpoly_n, &geom_rpoly_u, &geom_rpoly_v,
			  geom_rpoly_p, geom_rpoly_qp, &geom_rpoly_a, &geom_rpoly_b);
	geom_rpoly_calcsc(&type);

	for (j = 0; j < l2; j++) {
		/* Calculate next k polynomial and estimate v. */
		geom_rpoly_nextk(&type);
		geom_rpoly_calcsc(&type);
		geom_rpoly_newest(type, &ui, &vi);
		vv = vi;
		/* Estimate s. */
		ss = 0.0;
		if (geom_rpoly_k[geom_rpoly_n - 1] != 0.0)
			ss = -geom_rpoly_p[geom_rpoly_n] / geom_rpoly_k[geom_rpoly_n - 1];
		tv = 1.0;
		ts = 1.0;
		if (j == 0 || type == 3)
			goto _70;

		/* Compute relative measures of convergence of s and v sequences. */
		if (vv != 0.0)
			tv = fabs((vv - ovv) / vv);
		if (ss != 0.0)
			ts = fabs((ss - oss) / ss);

		/* If decreasing, multiply two most recent convergence measures. */
		tvv = 1.0;
		if (tv < otv)
			tvv = tv * otv;
		tss = 1.0;
		if (ts < ots)
			tss = ts * ots;

		/* Compare with convergence criteria. */
		vpass = (tvv < betav);
		spass = (tss < betas);
		if (!(spass || vpass))
			goto _70;

		/* At least one sequence has passed the convergence test. Store variables before iterating. */
		svu = geom_rpoly_u;
		svv = geom_rpoly_v;
		for (i = 0; i < geom_rpoly_n; i++)
			svk[i] = geom_rpoly_k[i];
		s = ss;

		/* Choose iteration according to the fastest converging sequence. */
		vtry = 0;
		stry = 0;
		if ((spass && (!vpass)) || tss < tvv)
			goto _40;
	      _20:
		geom_rpoly_quadit(&ui, &vi, nz);
		if (*nz > 0)
			return;

		/* Quadratic iteration has failed. Flag that it has been tried and decrease the convergence criterion. */
		vtry = 1;
		betav *= 0.25;

		/* Try linear iteration if it has not been tried and the S sequence is converging. */
		if (stry || !spass)
			goto _50;
		for (i = 0; i < geom_rpoly_n; i++)
			geom_rpoly_k[i] = svk[i];
	      _40:
		geom_rpoly_realit(&s, nz, &iflag);
		if (*nz > 0)
			return;

		/* Linear iteration has failed. Flag that it has been tried and decrease the convergence criterion. */
		stry = 1;
		betas *= 0.25;
		if (iflag == 0)
			goto _50;

		/* If linear iteration signals an almost double real zero attempt quadratic iteration. */
		ui = -(s + s);
		vi = s * s;
		goto _20;

		/* Restore variables. */
	      _50:
		geom_rpoly_u = svu;
		geom_rpoly_v = svv;
		for (i = 0; i < geom_rpoly_n; i++)
			geom_rpoly_k[i] = svk[i];

		/* Try quadratic iteration if it has not been tried and the V sequence is convergin. */
		if (vpass && !vtry)
			goto _20;

		/* Recompute QP and scalar values to continue the second stage. */
		geom_rpoly_quadsd(geom_rpoly_n, &geom_rpoly_u, &geom_rpoly_v,
				  geom_rpoly_p, geom_rpoly_qp, &geom_rpoly_a, &geom_rpoly_b);
		geom_rpoly_calcsc(&type);
	      _70:
		ovv = vv;
		oss = ss;
		otv = tv;
		ots = ts;
	}
}


/* Variable-shift k-polynomial iteration for a quadratic factor converges only
 * if the zeros are equimodular or nearly so.
 * uu, vv - coefficients of starting quadratic.
 * nz - number of zeros found. */

void geom_rpoly_quadit(double *uu, double *vv, int *nz)
{
	double ui, vi;
	double mp, omp, ee, relstp, t, zm;
	int type, i, j, tried;

	*nz = 0;
	tried = 0;
	geom_rpoly_u = *uu;
	geom_rpoly_v = *vv;
	j = 0;
	relstp = omp = 0;

	/* Main loop. */
      _10:
	geom_rpoly_quad(1.0, geom_rpoly_u, geom_rpoly_v, &geom_rpoly_szr, &geom_rpoly_szi, &geom_rpoly_lzr,
			&geom_rpoly_lzi);

	/* Return if roots of the quadratic are real and not
	 * close to multiple or nearly equal and of opposite sign. */
	if (fabs(fabs(geom_rpoly_szr) - fabs(geom_rpoly_lzr)) > 0.01 * fabs(geom_rpoly_lzr))
		return;

	/* Evaluate polynomial by quadratic synthetic division. */
	geom_rpoly_quadsd(geom_rpoly_n, &geom_rpoly_u, &geom_rpoly_v,
			  geom_rpoly_p, geom_rpoly_qp, &geom_rpoly_a, &geom_rpoly_b);
	mp = fabs(geom_rpoly_a - geom_rpoly_szr * geom_rpoly_b) + fabs(geom_rpoly_szi * geom_rpoly_b);

	/* Compute a rigorous bound on the rounding error in evaluating p. */
	zm = sqrt(fabs(geom_rpoly_v));
	ee = 2.0 * fabs(geom_rpoly_qp[0]);
	t = -geom_rpoly_szr * geom_rpoly_b;
	for (i = 1; i < geom_rpoly_n; i++)
		ee = ee * zm + fabs(geom_rpoly_qp[i]);
	ee = ee * zm + fabs(geom_rpoly_a + t);
	ee *= (5.0 * GEOM_RPOLY_MRE + 4.0 * GEOM_RPOLY_ARE);
	ee = ee - (5.0 * GEOM_RPOLY_MRE + 2.0 * GEOM_RPOLY_ARE) *
	    (fabs(geom_rpoly_a + t) + fabs(geom_rpoly_b) * zm) + 2.0 * GEOM_RPOLY_ARE * fabs(t);

	/* Iteration has converged sufficiently if the polynomial value is less than 20 times this bound. */
	if (mp <= 20.0 * ee) {
		*nz = 2;
		return;
	}
	j++;

	/* Stop iteration after 20 steps. */
	if (j > 20)
		return;
	if (j < 2)
		goto _50;
	if (relstp > 0.01 || mp < omp || tried)
		goto _50;

	/* A cluster appears to be stalling the convergence.
	 * Five fixed shift steps are taken with a u,v close to the cluster. */
	if (relstp < GEOM_RPOLY_ETA)
		relstp = GEOM_RPOLY_ETA;
	relstp = sqrt(relstp);
	geom_rpoly_u = geom_rpoly_u - geom_rpoly_u * relstp;
	geom_rpoly_v = geom_rpoly_v + geom_rpoly_v * relstp;
	geom_rpoly_quadsd(geom_rpoly_n, &geom_rpoly_u, &geom_rpoly_v,
			  geom_rpoly_p, geom_rpoly_qp, &geom_rpoly_a, &geom_rpoly_b);
	for (i = 0; i < 5; i++) {
		geom_rpoly_calcsc(&type);
		geom_rpoly_nextk(&type);
	}
	tried = 1;
	j = 0;
      _50:
	omp = mp;

	/* Calculate next k polynomial and new u and v. */
	geom_rpoly_calcsc(&type);
	geom_rpoly_nextk(&type);
	geom_rpoly_calcsc(&type);
	geom_rpoly_newest(type, &ui, &vi);

	/* If vi is zero the iteration is not converging. */
	if (vi == 0.0)
		return;
	relstp = fabs((vi - geom_rpoly_v) / vi);
	geom_rpoly_u = ui;
	geom_rpoly_v = vi;
	goto _10;
}


/* Variable-shift H polynomial iteration for a real zero.
 * sss - starting iterate
 * nz  - number of zeros found
 * iflag - flag to indicate a pair of zeros near real axis. */

void geom_rpoly_realit(double *sss, int *nz, int *iflag)
{
	double pv, kv, t, s;
	double ms, mp, omp, ee;
	int i, j;

	*nz = 0;
	s = *sss;
	*iflag = 0;
	j = 0;
	omp = t = 0;

	/* Main loop */
	while (1) {
		pv = geom_rpoly_p[0];

		/* Evaluate p at s. */
		geom_rpoly_qp[0] = pv;
		for (i = 1; i <= geom_rpoly_n; i++) {
			pv = pv * s + geom_rpoly_p[i];
			geom_rpoly_qp[i] = pv;
		}
		mp = fabs(pv);

		/* Compute a rigorous bound on the error in evaluating p. */
		ms = fabs(s);
		ee = (GEOM_RPOLY_MRE / (GEOM_RPOLY_ARE + GEOM_RPOLY_MRE)) * fabs(geom_rpoly_qp[0]);
		for (i = 1; i <= geom_rpoly_n; i++)
			ee = ee * ms + fabs(geom_rpoly_qp[i]);

		/* Iteration has converged sufficiently if the polynomial value is less than 20 times this bound. */
		if (mp <= 20.0 * ((GEOM_RPOLY_ARE + GEOM_RPOLY_MRE) * ee - GEOM_RPOLY_MRE * mp)) {
			*nz = 1;
			geom_rpoly_szr = s;
			geom_rpoly_szi = 0.0;
			return;
		}
		j++;

		/* Stop iteration after 10 steps. */
		if (j > 10)
			return;
		if (j < 2)
			goto _50;
		if (fabs(t) > 0.001 * fabs(s - t) || mp < omp)
			goto _50;

		/* A cluster of zeros near the real axis has been encountered.
		 * Return with iflag set to initiate a quadratic iteration. */
		*iflag = 1;
		*sss = s;
		return;

		/* Return if the polynomial value has increased significantly. */
_50:
		omp = mp;

		/* Compute t, the next polynomial, and the new iterate. */
		kv = geom_rpoly_k[0];
		geom_rpoly_qk[0] = kv;
		for (i = 1; i < geom_rpoly_n; i++) {
			kv = kv * s + geom_rpoly_k[i];
			geom_rpoly_qk[i] = kv;
		}
		if (fabs(kv) <= fabs(geom_rpoly_k[geom_rpoly_n - 1]) * 10.0 * GEOM_RPOLY_ETA) {
			/* Use unscaled form. */
			geom_rpoly_k[0] = 0.0;
			for (i = 1; i < geom_rpoly_n; i++)
				geom_rpoly_k[i] = geom_rpoly_qk[i - 1];
		} else {
			/* Use the scaled form of the recurrence if the value of k at s is nonzero. */
			t = -pv / kv;
			geom_rpoly_k[0] = geom_rpoly_qp[0];
			for (i = 1; i < geom_rpoly_n; i++)
				geom_rpoly_k[i] = t * geom_rpoly_qk[i - 1] + geom_rpoly_qp[i];
		}
		kv = geom_rpoly_k[0];
		for (i = 1; i < geom_rpoly_n; i++)
			kv = kv * s + geom_rpoly_k[i];
		t = 0.0;
		if (fabs(kv) > fabs(geom_rpoly_k[geom_rpoly_n - 1] * 10.0 * GEOM_RPOLY_ETA))
			t = -pv / kv;
		s += t;
	}
}


/* This routine calculates scalar quantities used to compute the next
 * k polynomial and new estimates of the quadratic coefficients.
 * type - integer variable set here indicating how the
 * calculations are normalized to avoid overflow. */

void geom_rpoly_calcsc(int *type)
{
	/* Synthetic division of k by the quadratic 1,u,v */
	geom_rpoly_quadsd(geom_rpoly_n - 1, &geom_rpoly_u, &geom_rpoly_v, geom_rpoly_k,
			  geom_rpoly_qk, &geom_rpoly_c, &geom_rpoly_d);

	if (fabs(geom_rpoly_c) <= fabs(geom_rpoly_k[geom_rpoly_n - 1] * 100.0 * GEOM_RPOLY_ETA) &&
	    fabs(geom_rpoly_d) <= fabs(geom_rpoly_k[geom_rpoly_n - 2] * 100.0 * GEOM_RPOLY_ETA)) {
		*type = 3;
		/* Type=3 indicates the quadratic is almost a factor of k. */
	} else if (fabs(geom_rpoly_d) < fabs(geom_rpoly_c)) {
		*type = 1;
		/* Type=1 indicates that all formulas are divided by c. */
		geom_rpoly_e = geom_rpoly_a / geom_rpoly_c;
		geom_rpoly_f = geom_rpoly_d / geom_rpoly_c;
		geom_rpoly_g = geom_rpoly_u * geom_rpoly_e;
		geom_rpoly_h = geom_rpoly_v * geom_rpoly_b;
		geom_rpoly_a3 =
		    geom_rpoly_a * geom_rpoly_e + (geom_rpoly_h / geom_rpoly_c + geom_rpoly_g) * geom_rpoly_b;
		geom_rpoly_a1 = geom_rpoly_b - geom_rpoly_a * (geom_rpoly_d / geom_rpoly_c);
		geom_rpoly_a7 = geom_rpoly_a + geom_rpoly_g * geom_rpoly_d + geom_rpoly_h * geom_rpoly_f;
	} else {
		*type = 2;
		/* Type=2 indicates that all formulas are divided by d. */
		geom_rpoly_e = geom_rpoly_a / geom_rpoly_d;
		geom_rpoly_f = geom_rpoly_c / geom_rpoly_d;
		geom_rpoly_g = geom_rpoly_u * geom_rpoly_b;
		geom_rpoly_h = geom_rpoly_v * geom_rpoly_b;
		geom_rpoly_a3 =
		    (geom_rpoly_a + geom_rpoly_g) * geom_rpoly_e +
		    geom_rpoly_h * (geom_rpoly_b / geom_rpoly_d);
		geom_rpoly_a1 = geom_rpoly_b * geom_rpoly_f - geom_rpoly_a;
		geom_rpoly_a7 = (geom_rpoly_f + geom_rpoly_u) * geom_rpoly_a + geom_rpoly_h;
	}
}


/* Computes the next k polynomials using scalars computed in calcsc. */

void geom_rpoly_nextk(int *type)
{
	double temp;
	int i;

	if (*type == 3) {
		/* Use unscaled form of the recurrence if type is 3. */
		geom_rpoly_k[0] = 0.0;
		geom_rpoly_k[1] = 0.0;
		for (i = 2; i < geom_rpoly_n; i++)
			geom_rpoly_k[i] = geom_rpoly_qk[i - 2];
		return;
	}
	temp = geom_rpoly_a;
	if (*type == 1)
		temp = geom_rpoly_b;
	if (fabs(geom_rpoly_a1) <= fabs(temp) * GEOM_RPOLY_ETA * 10.0) {
		/* If a1 is nearly zero then use a special form of the recurrence. */
		geom_rpoly_k[0] = 0.0;
		geom_rpoly_k[1] = -geom_rpoly_a7 * geom_rpoly_qp[0];
		for (i = 2; i < geom_rpoly_n; i++)
			geom_rpoly_k[i] =
			    geom_rpoly_a3 * geom_rpoly_qk[i - 2] - geom_rpoly_a7 * geom_rpoly_qp[i - 1];
		return;
	}

	/* Use scaled form of the recurrence. */
	geom_rpoly_a7 /= geom_rpoly_a1;
	geom_rpoly_a3 /= geom_rpoly_a1;
	geom_rpoly_k[0] = geom_rpoly_qp[0];
	geom_rpoly_k[1] = geom_rpoly_qp[1] - geom_rpoly_a7 * geom_rpoly_qp[0];
	for (i = 2; i < geom_rpoly_n; i++)
		geom_rpoly_k[i] =
		    geom_rpoly_a3 * geom_rpoly_qk[i - 2] - geom_rpoly_a7 * geom_rpoly_qp[i - 1] +
		    geom_rpoly_qp[i];
}


/* Compute new estimates of the quadratic coefficients using the scalars computed in calcsc. */

void geom_rpoly_newest(int type, double *uu, double *vv)
{
	double a4, a5, b1, b2, c1, c2, c3, c4, temp;

	/* Use formulas appropriate to setting of type. */

	if (type == 3) {
		/* If type=3 the quadratic is zeroed. */
		*uu = 0.0;
		*vv = 0.0;
		return;
	}
	if (type == 2) {
		a4 = (geom_rpoly_a + geom_rpoly_g) * geom_rpoly_f + geom_rpoly_h;
		a5 = (geom_rpoly_f + geom_rpoly_u) * geom_rpoly_c + geom_rpoly_v * geom_rpoly_d;
	} else {
		a4 = geom_rpoly_a + geom_rpoly_u * geom_rpoly_b + geom_rpoly_h * geom_rpoly_f;
		a5 = geom_rpoly_c + (geom_rpoly_u + geom_rpoly_v * geom_rpoly_f) * geom_rpoly_d;
	}

	/* Evaluate new quadratic coefficients. */
	b1 = -geom_rpoly_k[geom_rpoly_n - 1] / geom_rpoly_p[geom_rpoly_n];
	b2 = -(geom_rpoly_k[geom_rpoly_n - 2] +
	       b1 * geom_rpoly_p[geom_rpoly_n - 1]) / geom_rpoly_p[geom_rpoly_n];
	c1 = geom_rpoly_v * b2 * geom_rpoly_a1;
	c2 = b1 * geom_rpoly_a7;
	c3 = b1 * b1 * geom_rpoly_a3;
	c4 = c1 - c2 - c3;
	temp = a5 + b1 * a4 - c4;
	if (temp == 0.0) {
		*uu = 0.0;
		*vv = 0.0;
	} else {
		*uu =
		    geom_rpoly_u - (geom_rpoly_u * (c3 + c2) +
				    geom_rpoly_v * (b1 * geom_rpoly_a1 + b2 * geom_rpoly_a7)) / temp;
		*vv = geom_rpoly_v * (1.0 + c4 / temp);
	}
}


/* Divides p by the quadratic 1,u,v placing the quotient in q and the remainder in a,b. */

void geom_rpoly_quadsd(int nn, double *u, double *v, double *p, double *q, double *a, double *b)
{
	double c;
	int i;

	*b = p[0];
	q[0] = *b;
	*a = p[1] - (*b) * (*u);
	q[1] = *a;
	for (i = 2; i <= nn; i++) {
		c = p[i] - (*a) * (*u) - (*b) * (*v);
		q[i] = c;
		*b = *a;
		*a = c;
	}
}


/* Calculate the zeros of the quadratic a * z^2 + b1 * z + c.
 * The quadratic formula, modified to avoid overflow, is used
 * to find the larger zero if the zeros are real and both
 * are complex. The smaller real zero is found directly from
 * the product of the zeros c / a. */

void geom_rpoly_quad(double a, double b1, double c, double *sr, double *si, double *lr, double *li)
{
	double b, d, e;

	/* less than two roots */
	if (a == 0.0) {
		if (b1 != 0.0)
			*sr = -c / b1;
		else
			*sr = 0.0;
		*lr = 0.0;
		*si = 0.0;
		*li = 0.0;
		return;
	}

	/* one real root, one zero root */
	if (c == 0.0) {
		*sr = 0.0;
		*lr = -b1 / a;
		*si = 0.0;
		*li = 0.0;
		return;
	}

	/* Compute discriminant avoiding overflow. */
	b = b1 / 2.0;
	if (fabs(b) < fabs(c)) {
		if (c < 0.0)
			e = -a;
		else
			e = a;
		e = b * (b / fabs(c)) - e;
		d = sqrt(fabs(e)) * sqrt(fabs(c));
	} else {
		e = 1.0 - (a / b) * (c / b);
		d = sqrt(fabs(e)) * fabs(b);
	}

	/* complex conjugate zeros */
	if (e < 0.0) {
		*sr = -b / a;
		*lr = *sr;
		*si = fabs(d / a);
		*li = -(*si);
	} else {
		/* real zeros */
		if (b >= 0.0)
			d = -d;
		*lr = (-b + d) / a;
		*sr = 0.0;
		if (*lr != 0.0)
			*sr = (c / *lr) / a;
		*si = 0.0;
		*li = 0.0;
	}
}
