#ifndef _VRR_GEOMLIB_SEGMENT_H_
#define _VRR_GEOMLIB_SEGMENT_H_


/**
 * @file geomlib/segment.h
 *
 * Geomlib - Segments.
 */


#include "geomlib/curve.h"
#include "geomlib/vector.h"


#define geom_segment_VMT    geom_curve_VMT
#define geom_segment_INIT   geom_curve_INIT \
	\
	GEOM_CURVE_TIME_EQUALS_BTIME \
	\
	.alength = (real (*)(void *)) &geom_segment_alength, \
	.time_to_atime = (real (*)(void *, real)) &geom_segment_time_to_atime, \
	.atime_to_time = (real (*)(void *, real)) &geom_segment_atime_to_time, \
	\
	.expansion_append = (int (*)(void *, struct geom_fpath *)) &geom_segment_expansion_append, \
	.to_bezier = (int (*)(void *, struct geom_bezier *)) &geom_segment_to_bezier, \


GEOM_CLASS_HEAD(segment, curve);


/*----------------------------------------------*
 *                                              *
 *              INITIALIZATION                  *
 *                                              *
 *----------------------------------------------*/


static inline void geom_segment_create(struct geom_segment *segment)
{ geom_curve_create(&segment->curve); }

static inline void geom_segment_destroy(struct geom_segment *segment)
{ geom_curve_destroy(&segment->curve); }


/**
 * Sets geom_segment from two given endpoints.
 * Can be used as class constructor.
 *
 * @ao  segment  Resulting segment.
 * @ai  a        Starting point.
 * @ai  b        Ending point.
 *
 * @rv  zero  Always.
 */
static inline void geom_segment_set(struct geom_segment *segment, struct geom_point *a, struct geom_point *b)
{
	GEOM_INSTANCE_IS_ASSERT(segment, segment);
	segment->a = *GEOM_ASSERTX(point, a);
	segment->b = *GEOM_ASSERTX(point, b);
	geom_curve_after_change(&segment->curve);
}


/*----------------------------------------------*
 *                                              *
 *           OVERRIDED VIRTUAL METHODS          *
 *                                              *
 *----------------------------------------------*/


/* Parametrizations */

real geom_segment_alength(struct geom_segment *segment);
real geom_segment_time_to_atime(struct geom_segment *segment, real time);
real geom_segment_atime_to_time(struct geom_segment *segment, real atime);


/* Conversions */

int geom_segment_expansion_append(struct geom_segment *segment, struct geom_fpath *expansion);
int geom_segment_to_bezier(struct geom_segment *segment, struct geom_bezier *bezier);


/* Transformation */

int geom_segment_transform(struct geom_segment *segment, struct geom_transform *transform, struct geom_segment *result);


/* Geometric properties */

static inline int geom_segment_point_start(struct geom_segment *segment, struct geom_point *result)
{ *result = segment->a; return 0; }


/**
 * Returns ending point of given segment.
 *
 * @ai  segment  Segment.
 * @ao  result   Resulting ending point.
 *
 * @rv  zero  Always.
 */
static inline int geom_segment_point_end(struct geom_segment *segment, struct geom_point *result)
{ *result = segment->b; return 0; }


/**
 * Returns point of given segment.
 *
 * @ai  segment  Segment.
 * @ai  time     TIME parameter.
 * @ao  result   Resulting point.
 *
 * @rv  zero              Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_segment_point_at_time(struct geom_segment *segment, real time, struct geom_point *result);


/**
 * Returns first derivation of given segment.
 *
 * @ai  segment  Segment.
 * @ao  result   Resulting derivation vector.
 *
 * @rv  zero                 Success.
 * @rv  GEOM_ERR_NUMERIC     Numeric error.
 */
int geom_segment_derivation(struct geom_segment *segment, struct geom_vector *result);


/**
 * Returns bounding box of given segment.
 *
 * @ai  segment  Segment.
 * @ao  bbox     Resulting bounding box.
 *
 * @rv  zero  Always.
 */
int geom_segment_bbox(struct geom_segment *segment, struct geom_rectangle *bbox);




#endif

