#ifndef _VRR_GEOMLIB_TRANSFORM_H_
#define _VRR_GEOMLIB_TRANSFORM_H_


/**
 * @file
 *
 * Geomlib - Affine transformations.
 */


#include "geomlib/base.h"


/* Transformation flags */

/** identity (coef[i][i]=1, other 0) */
#define GEOM_TRANSFORM_IDENTITY  0x00000001

/** similar transformation */
#define GEOM_TRANSFORM_SIMILAR   0x00000002


/* Predefined transformations. */
extern const struct geom_transform geom_transform_identity; /**< identity transformation */
extern const struct geom_transform geom_transform_hflip;    /**< flip around y axis */
extern const struct geom_transform geom_transform_vflip;    /**< flip around x axis */
extern const struct geom_transform geom_transform_rot90;    /**< rotation by 90 degrees */
extern const struct geom_transform geom_transform_rot180;    /**< rotation by 180 degrees */
extern const struct geom_transform geom_transform_rot270;    /**< rotation by 270 degrees */



int geom_transform_valid(const struct geom_transform *t);

const struct geom_transform *geom_transform_assert(GEOM_ASSERT_DEF, const struct geom_transform *t);

void geom_transform_dump(const struct geom_transform *t, uns flags);




/**
 * Initialize identity transformation matrix (T(x, y) = (x, y)).
 *
 * @a  t  Resulting transformation.
 *
 * @r  0 always.
 */
static inline void geom_transform_set_identity(struct geom_transform *t)
{ *t = geom_transform_identity; }


/**
 * Initialize general transformation matrix (T(x, y) = (ax + by + c, dx + ey + f)).
 *
 * @a  t  Resulting transformation.
 * @a  a  Transformation coefficient.
 * @a  b  Transformation coefficient.
 * @a  c  Transformation coefficient.
 * @a  d  Transformation coefficient.
 * @a  e  Transformation coefficient.
 * @a  f  Transformation coefficient.
 *
 * @r  0 always.
 */
int geom_transform_set(struct geom_transform *t, real a, real b, real c, real d, real e, real f);


/**
 * Initialize move transformation matrix (T(x, y) = (x + mx, y + my)).
 *
 * @a  t   Resulting transformation.
 * @a  mx  Move X coordinate.
 * @a  my  Move Y coordinate.
 *
 * @r  0 always
 */
int geom_transform_set_move(struct geom_transform *t, const real mx, const real my);



/**
 * Initialize scale transformation matrix (T(x, y) = (x * sx, y * sy)).
 *
 * @a  t   Resulting transformation.
 * @a  sx  Scale X coefficient.
 * @a  sy  Scale Y coefficient.
 *
 * @r  0 always
 */
int geom_transform_set_scale(struct geom_transform *t, const real sx, const real sy);


/**
 * Initialize transformation matrix of scale around given center point.
 *
 * @a  t  Resulting transformation.
 * @a  center  Center point.
 * @a  sx      Scale X coefficient.
 * @a  sy      Scale Y coefficient.
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_transform_set_center_scale(struct geom_transform *t, const struct geom_point *center, const real sx, const real sy);


/**
 * Initialize rotation transformation matrix (CCW around origin).
 *
 * @a  t      Resulting transformation.
 * @a  angle  Rotation angle in radians.
 *
 * @r  0 always
 */
int geom_transform_set_rotation(struct geom_transform *t, const real angle);


/**
 * Initialize rotation transformation matrix (CCW around given center point).
 *
 * @a  t       Resulting transformation.
 * @a  center  Center point.
 * @a  angle   Rotation angle in radians.
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_transform_set_center_rotation(struct geom_transform *t, const struct geom_point *center, const real angle);


/**
 * Initilize slant transformation matrix.
 *
 * @a  t  Resulting transformation.
 * @a  a  First coefficint.
 * @a  b  Second coefficient.
 *
 * @r  0 always
 */
int geom_transform_set_slant(struct geom_transform *t, const real a, const real b);


/**
 * Merges two transformations to one (T = T2 * T1).
 *
 * @a  t1  Transformation T1.
 * @a  t2  Transformation T2.
 * @a  t   Resulting transformation T2 * T1 (possibly T = T1 or T2).
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error (resulting structure remains unchanged).
 */
int geom_transform_merge(const struct geom_transform *t1, const struct geom_transform *t2, struct geom_transform *t);


/**
 * Compute inversion of matrix (R = T^-1)
 *
 * @a  t  Transformation T.
 * @a  r  Resulting transformation R = T^-1.
 *
 * @rv  0                            Success.
 * @rv  GEOM_ERR_TRANSFORM_SINGULAR  T is near-singular transformation (resulting structure remains unchanged).
 * @rv  GEOM_ERR_NUMERIC             Numeric error (resulting structure remains unchanged).
 */
int geom_transform_invert(const struct geom_transform *t, struct geom_transform *r);


/**
 * Returns whether given affine transformation is linear transformation (T(0, 0) ?= (0, 9)).
 *
 * @a  t  Transformation.
 *
 * @rv  1  Image of zero is zero (linear transformation)/
 * @rv  0  Image of zero is not zero (non-linear transformation).
 */
int geom_transform_is_linear(const struct geom_transform *t);


/**
 * Returns whether given transformation is regular.
 *
 * @a  t  Transformation.
 *
 * @rv  1                 Transformation is regular.
 * @rv  0                 Transformation is singular.
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_transform_is_regular(const struct geom_transform *t);


/**
 * Returns rank of given transformation.
 *
 * @a  t  Transformation.
 *
 * @rv  2                 Image of R^2 is R^2 (regular transformation).
 * @rv  1                 Image of R^2 is a line (singular transformation).
 * @rv  0                 Image of R^2 is a point (singular transformation).
 * @rv  GEOM_ERR_NUMERIC  Numeric error.
 */
int geom_transform_rank(const struct geom_transform *t);


/**
 * Function transforms geom_point coordinates by transformation matrix.
 *
 * @a  p       Point to transform.
 * @a  t       Transformation.
 * @a  result  Resulting point (possibly result = p).
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error (resulting structure remains unchanged).
 */
int geom_point_transform(const struct geom_point *p, const struct geom_transform *t, struct geom_point *result);


/**
 * Function transforms geom_point_w coordinates by transformation matrix.
 *
 * @a  p       Point to transform.
 * @a  t       Transformation.
 * @a  result  Resulting point (possibly result = p).
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error (resulting structure remains unchanged).
 */
int geom_point_w_transform(struct geom_point_w *p, struct geom_transform *t, struct geom_point_w *result);


/**
 * Function transforms a vector by transformation matrix (no translation is applied).
 *
 * @a  p       Vector to transform.
 * @a  t       Transformation.
 * @a  result  Resulting vector (possibly result = p).
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error (resulting structure remains unchanged).
 */
int geom_vector_transform(const struct geom_vector *p, const struct geom_transform *t, struct geom_vector *result);


/**
 * Function finds bounding box of transformed rectangle.
 *
 * @a  r       Rectangle to transform.
 * @a  t       Transformation.
 * @a  result  Resulting bounding box (possibly result = r).
 *
 * @rv  0                 Success.
 * @rv  GEOM_ERR_NUMERIC  Numeric error (resulting structure remains unchanged).
 */
int geom_rectangle_transform (const struct geom_rectangle *r, const struct geom_transform *t, struct geom_rectangle *result);


static inline int geom_transform_is_similar(struct geom_transform *t)
{ return t->flags & GEOM_TRANSFORM_SIMILAR; }

int geom_transform_is_point_fixed(struct geom_point *p, struct geom_transform *t);


#endif

