#ifndef _VRR_GEOMLIB_TYPES_H_
#define _VRR_GEOMLIB_TYPES_H_


/**
 * @file geomlib/types.h
 *
 * Types and some constants defined in Geomlib.
 */


#include "geomlib/base.h"


/*------------------------------------------------------------------------*
 *                                                                        *
 *                           BASIC DEFINITIONS                            *
 *                                                                        *
 *------------------------------------------------------------------------*/


/** 2D point structure */
struct geom_point {
	real x;                   /**< X coordinate */
	real y;                   /**< Y coordinate */
};


/** 2D point structure with weight */
struct geom_point_w {
	real x;                   /**< X coordinate */
	real y;                   /**< Y coordinate */
	real w;                   /**< weight */
};


/** 2D vector structure */
struct geom_vector {
	real dx;                  /**< X coordinate */
	real dy;                  /**< Y coordinate */
};


/** 2D rectangle structure */
struct geom_rectangle {
	struct geom_point ll;     /**< lower left corner point */
	struct geom_point ur;     /**< upper right corner point */
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                          AFFINE TRANSFORMATIONS                        *
 *                                                                        *
 *------------------------------------------------------------------------*/


/** 2D affine transform matrix */
struct geom_transform {
	uns flags;                      /**< special flags */
	real coef[2][3];          /**< transformation coefficients */
};


/** 2D affine bidirectional transformator */
struct geom_transform2 {
	struct geom_transform primary;  /**< primary matrix */
	struct geom_transform inverted; /**< inverted matrix */
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                            ITEMS AND GROUPS                            *
 *                                                                        *
 *------------------------------------------------------------------------*/


/** header of groupable geomlib structures */
// FIXME: rewrite to balanced tree
struct geom_item {
	struct geom_o o;          /**< Instance header. */
	ccnode node;              /**< node for geom_group.list */
	struct geom_group *owner; /**< owner group */
};


/** ordered group of geom_node structures */
struct geom_group {
	struct geom_item item;    /**< grouping recursion */
	cclist list;              /**< list of group items */
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                                CURVE TYPES                             *
 *                                                                        *
 *------------------------------------------------------------------------*/


/** Curve header structure. */
struct geom_curve {
	struct geom_item item;
};


/** Segment structure. */
struct geom_segment {
	struct geom_curve curve;  /**< Curve header. */
	struct geom_point a;      /**< Starting point. */
	struct geom_point b;      /**< Ending point. */
};


/** Rational Bezier curve structure. */
struct geom_bezier {
	struct geom_curve curve;    /**< Curve header. */
	struct geom_point_w pt[4];  /**< Control points. */
	struct geom_rectangle bbox; /**< Cached bounding box. */
	real alength;               /**< Cached arc length. */
};


/**
 * Parabolic arc structure.
 *
 * Structure represents parabola in standard form: x^2 = 4 * distance * y,
 * moved to stored center point and rotated CCW around this point.
 *
 * Parametric form of parabola:
 *   x = t,
 *   y = t^2 / (4 * distance).
 *
 * Section of parabola (parabolic arc) is given by parameters subinterval.
 */
struct geom_parabolic_arc {
	struct geom_curve curve;    /**< Curve header. */
	struct geom_point center;   /**< Center point. */
	real rotation;              /**< Rotation around the center point (CCW in radians). */
	real distance;              /**< Distance of center point to parabola focus and directrix (distance >= 0). */
	real start;                 /**< Parameter of parabolic arc starting point. */
	real end;                   /**< Parameterof parabolic arc ending point. */
};


/**
 * Circle or circular arc.
 *
 * Structure represents circle in standard form: x^2 + y^2 = radius^2, moved to stored center point.
 *
 * Parametric form of circle:
 *   x = radius * cos t,
 *   y = radius * sin t.
 *
 * Section of circle (arc) is given by parameters subinterval.
 */
struct geom_arc {
	struct geom_curve curve;    /**< Curve header. */
	struct geom_point center;   /**< Center point. */
	real radius;                /**< Circle radius (radius >= 0). */
	real start;                 /**< Parameter of arc starting point (0 for entire circle). */
	real end;                   /**< Parameter of arc ending point (2 * PI for entire circle). */
};


/**
 * Ellipse or elliptic arc.
 *
 * Structure represents ellipse in standard form: x^2 / a_radius^2 + y^2 / b_radius^2 = 1,
 * moved to stored center point and rotated CCW around this point.
 * Special cases with zero radius (radii) are also supported (closure to affine transformations).
 *
 * Parametric form of ellipse:
 *  x = a_radius * cos t,
 *  y = b_radius * sin t.
 *
 * Section of ellipse (elliptic arc) is given by parameters subinterval.
 */
struct geom_elliptic_arc {
	struct geom_curve curve;
	struct geom_point center;   /**< Center point. */
	real rotation;              /**< Rotation around the center point (CCW in radians). */
	real a_radius;              /**< X-axis radius (a_radius >= 0). */
	real b_radius;              /**< Y-axis radius (b_radius >= 0). */
	real start;                 /**<  */
	real dif;                   /**<  */
};


/**
 * Hyperbola or hyperbolic arc.
 *
 * Structure represents hyperbola in standard form: x^2 / major^2 - y^2 / minor^2 = 1,
 * moved to stored center point and rotated CCW around this point.
 *
 * FIXME: domyslet vhodnou parametrizaci
 *
 * Parametric form (interval (0, PI) for right branch, (-PI, 0) for left branch):
 *   x = a / sin t,
 *   y = b * cos t / sin t.
 *
 * Section of hyperbola (hyperbolic arc) is given by parameters subinterval (cannot contain zero).
 */
struct geom_hyperbolic_arc {
	struct geom_curve curve;
	struct geom_point center;   /**< Center point. */
	real rotation;              /**< Rotation around the center point (CCW in radians). */
	real major;                 /**< Semimajor axis (major >= 0). */
	real minor;                 /**< Semiminor axis (minor >= 0). */
	real start;                 /**< Parameter of hyperbolic arc starting point (-infinity for whole hyperbola). */
	real end;                   /**< Parameter of hyperbolic arc ending point (+infinity for whole hyperbola). */
};


/** Nonuniform rational B-splines. */
struct geom_nurbs {
	struct geom_curve curve;
	u8 degree;
	cclist knots;
};

struct geom_nurbs_knot {
	ccnode node;
	real time;
	struct geom_point_w point;
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                   CONTINOUS GROUP OF CURVES - PATH                     *
 *                                                                        *
 *------------------------------------------------------------------------*/


struct geom_path {
	struct geom_group group;
	real alength;                    /**< Cached arc length of path. */
};

/* Path with fast greedy curve allocator. */
struct geom_fpath {
	struct geom_path path;
	struct geom_fpath_block *block_last;
	byte *block_ptr;
	uns block_free;
	uns total_size;
};

struct geom_fpath_block {
	struct geom_fpath_block *prev;
	byte start;
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                               SPECIAL ITEMS                            *
 *                                                                        *
 *------------------------------------------------------------------------*/

 struct geom_point_item {
 	struct geom_item item;
	struct geom_point point;
 };

 struct geom_callback_item {
 	struct geom_item item;
	int (*func)(struct geom_callback_item *, struct geom_fpath *);
 };
 typedef int (geom_callback_item_func)(struct geom_callback_item *, struct geom_fpath *);


/*------------------------------------------------------------------------*
 *                                                                        *
 *                                  PATHSET                               *
 *                                                                        *
 *------------------------------------------------------------------------*/

 struct geom_pathset {
 	struct geom_group group;
 };




/** list of subdivided parts of bezier curve */
struct geom_bezier_subdiv {
	uns count;                      /**< number of items */
	uns size;                       /**< number of allocated items */
	struct geom_bezier_subdiv_item *items; /**< items buffer */
};

/** single item in geom_bezier_subdiv */
struct geom_bezier_subdiv_item {
	double left;                    /**< starting parameter in original curve */
	double right;                   /**< ending parameter in original curve */
	struct geom_bezier bezier;      /**< subdivided part data */
};


#define GEOM_INTERSECTION_POINT 0x00000001
#define GEOM_INTERSECTION_TIME1 0x00000002
#define GEOM_INTERSECTION_TIME2 0x00000004
#define GEOM_INTERSECTION_ALL   (GEOM_INTERSECTION_POINT | GEOM_INTERSECTION_TIME1 | GEOM_INTERSECTION_TIME2)

struct geom_intersection {
	struct geom_point point;
	real time1;
	real time2;
};

#define GEOM_NEAREST_POINT 0x00000001
#define GEOM_NEAREST_TIME  0x00000002
#define GEOM_NEAREST_DIST  0x00000004
#define GEOM_NEAREST_ALL   (GEOM_NEAREST_POINT | GEOM_NEAREST_TIME | GEOM_NEAREST_DIST)

struct geom_nearest {
	struct geom_point point;
	real time;
	real dist;
};

#define GEOM_CURVE_POINT_POINT 0x00000001
#define GEOM_CURVE_POINT_TIME  0x00000002
#define GEOM_CURVE_POINT_ALL   (GEOM_CURVE_POINT_POINT | GEOM_CURVE_POINT_TIME)

struct geom_curve_point {
	struct geom_point point;
	real time;
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                            RTREE DEFINITIONS                           *
 *                                                                        *
 *------------------------------------------------------------------------*/

/*
	Limitations:
		2 <= GEOM_RTREE_MIN <= (GEOM_RTREE_MAX + 1) / 2
		0 <= GEOM_RTREE_REINSERT <= GEOM_RTREE_MAX - GEOM_RTREE_MIN + 1
*/
#define GEOM_RTREE_MIN           3  /**< min number of children in non-root internal node,
                                         root has at least one child */
#define GEOM_RTREE_MAX           5  /**< max number of children internal node */
#define GEOM_RTREE_REINSERT      3  /**< number of forced reinserts in overflow treatment */

/** RTree main data */
struct geom_rtree {
	struct geom_rtree_node *root;   /**< root node (NULL in empty tree) */
	clist lquery;                   /**< list of active dynamic queries */
};

/** RTree object data */
struct geom_rtree_obj {
	struct geom_rectangle bbox;     /**< object bounding box */
	struct geom_rtree_node *parent; /**< parent node */
};

/** RTree internal node data */
struct geom_rtree_node {
	struct geom_rectangle bbox;     /**< subtree bounding box */
	byte count;                     /**< number of children */
	byte height;                    /**< subtree height (0 in nodes containing objects) */
	struct geom_rtree_node *parent; /**< parent node (NULL in root node) */
	struct geom_rtree_node *child[GEOM_RTREE_MAX + 1]; /**< childpointers */
};

/** RTree dynamic query structure */
struct geom_rtree_query {
	struct geom_rtree *rtree;       /**< associated RTree pointer */
	struct geom_rectangle box;      /**< query area */
	cnode nquery;                   /**< node in rtree->lquery */
	void *call_show;                /**< object show callback */
	void *call_hide;                /**< object hide callback */
	void *call_update;              /**< object update callback */
};

/** query callback definition */
typedef void geom_rtree_query_callback(struct geom_rtree_query *query, struct geom_rtree_obj *obj, uns flags);


/*------------------------------------------------------------------------*
 *                                                                        *
 *                          CURVES APPROXIMATION                          *
 *                                                                        *
 *------------------------------------------------------------------------*/

/** curve approximation item */
struct geom_approx_item {
	uns type;                          /**< curve type (GEOM_APPROX_...) */
	double left;                       /**< starting parameter */
	double right;                      /**< ending parameter */
	union {
		/* substructure for any of GEOM_APPROX_BEZIER types */
		struct {
			struct geom_bezier bezier; /**< bezier curve data */
		} bezier;

		/* substructure for GEOM_APPROX_POINT type */
		struct {
			struct geom_point point;   /**< point data */
		} point;

		/* substructure for GEOM_APPROX_ARC type */
		struct {
			struct geom_point center;  /**< center point */
			double radius;             /**< circle radius */
			double angle_start;        /**< arc starting angle */
			double angle_end;          /**< arc ending angle */
		} arc;

		/* substructure for GEOM_APPROX_ELLIPTIC_ARC type */
		struct {
			struct geom_point center;  /**< center point */
			double a_radius;           /**< first radius */
			double b_radius;           /**< second radius */
			double angle_start;        /**< arc starting angle */
			double angle_end;          /**< arc ending angle */
			double rotation;           /**< axis angle */
		} el_arc;
	} data;
};

/** curve approximation structure */
struct geom_approx {
	uns size;                          /**< items buffer size */
	uns count;                         /**< number of items */
	double right;                      /**< ending parameter */
	struct geom_approx_item *items;    /**< items buffer */
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                            MATRIX DEFINITION                           *
 *                                                                        *
 *------------------------------------------------------------------------*/

/** matrix structure */
struct geom_matrix {
	uns m;           /**< number of rows */
	uns n;           /**< number of columns */
	uns w;           /**< data row log size */
	double *data;    /**< data buffer */
};


/*------------------------------------------------------------------------*
 *                                                                        *
 *                              CONIC SECTIONS                            *
 *                                                                        *
 *------------------------------------------------------------------------*/

/** Type of conic. */
enum geom_conic_type {
	GEOM_CONIC_PARABOLIC_ARC,   /**< Parabola or parabolic arc. */
	GEOM_CONIC_ARC,             /**< Circle or arc. */
	GEOM_CONIC_ELLIPTIC_ARC,    /**< Ellipse or elliptic arc. */
	GEOM_CONIC_HYPERBOLIC_ARC   /**< Hyperbola or hyperbolic arc. */
};


/**  Conic or conic section. */
struct geom_conic {
	enum geom_conic_type type;                     /**< Type of conic. */
	union {
		struct geom_parabolic_arc parabolic_arc;   /**< Parabola / parabolic arc data. */
		struct geom_arc arc;                       /**< Circle / arc data. */
		struct geom_elliptic_arc elliptic_arc;     /**< Ellipse / elliptic arc data. */
		struct geom_hyperbolic_arc hyperbolic_arc; /**< Hyperbola / hyperbolic arc data. */
	} data;                                        /**< Conic data. */
};


#endif /* _VRR_GEOM_TYPES_H_ */

