#define GEOM_SOURCE
#include "geomlib/base.h"
#include "geomlib/vector.h"


/* predefined points */
struct geom_point geom_point_zero = {0, 0};
struct geom_point geom_point_right = {1, 0};
struct geom_point geom_point_left = {-1, 0};
struct geom_point geom_point_up = {0, 1};
struct geom_point geom_point_down = {0, -1};

/* predefined vectors */
struct geom_vector geom_vector_zero = {0, 0};
struct geom_vector geom_vector_right = {1, 0};
struct geom_vector geom_vector_left = {-1, 0};
struct geom_vector geom_vector_up = {0, 1};
struct geom_vector geom_vector_down = {0, -1};


/* geom_point_assert */

const struct geom_point *geom_point_assert(GEOM_ASSERT_DEF, const struct geom_point *p)
{
	geom_p_assert(GEOM_ASSERT_COPY, p);
	if (!(RTEST(p->x) && RTEST(p->y)))
		die(GEOM_ASSERT_FORMAT ": point (%f, %f) at %p is invalid\n", GEOM_ASSERT_COPY, p->x, p->y, (void *)p);	
	return p;
}


/* geom_point_w_assert */

const struct geom_point_w *geom_point_w_assert(GEOM_ASSERT_DEF, const struct geom_point_w *p)
{
	geom_p_assert(GEOM_ASSERT_COPY, p);
	if (!(RTEST(p->x) && RTEST(p->y) && RTEST(p->w)))
		die(GEOM_ASSERT_FORMAT ": weighed point (%f, %f, %f) at %p is invalid\n", GEOM_ASSERT_COPY, p->x, p->y, p->w, (void *)p);
	return p;
}


/* geom_vector_assert */

const struct geom_vector *geom_vector_assert(GEOM_ASSERT_DEF, const struct geom_vector *v)
{
	geom_p_assert(GEOM_ASSERT_COPY, v);
	if (!(RTEST(v->dx) && RTEST(v->dy)))
		die(GEOM_ASSERT_FORMAT ": vector (%f, %f) at %p is invalid\n", GEOM_ASSERT_COPY, v->dx, v->dy, (void *)v);
	return v;
}


/* geom_point_dump */

void geom_point_dump(const struct geom_point *p, uns UNUSED flags)
{
	DUMP_ENTER(1, "Dumping geom_point %p: ", (void *)p);
	DUMP_PTR(p) {
		SDUMP("%f", p->x);
		SDUMP("%f", p->y);
	}
	DUMP_LEAVE;
}


/* geom_point_w_dump */

void geom_point_w_dump(const struct geom_point_w *p, uns UNUSED flags)
{
	DUMP_ENTER(1, "Dumping geom_point_w %p: ", (void *)p);
	DUMP_PTR(p) {
		SDUMP("%f", p->x);
		SDUMP("%f", p->y);
		SDUMP("%f", p->w);
	}
	DUMP_LEAVE;
}


/* geom_vector_dump */

void geom_vector_dump(const struct geom_vector *v, uns UNUSED flags)
{
	DUMP_ENTER(1, "Dumping geom_vector %p: ", (void *)v);
	DUMP_PTR(v) {
		SDUMP("%f", v->dx);
		SDUMP("%f", v->dy);
	}
	DUMP_LEAVE;
}


/* geom_vector_init_two_points */

int geom_vector_init_two_points(const struct geom_point *p1, const struct geom_point *p2, struct geom_vector *v)
{
	real l;

	GASSERT(point, p1);
	GASSERT(point, p2);
	PASSERT(v);

	v->dx = p2->x - p1->x;
	v->dy = p2->y - p1->y;
	l = sqrtf(v->dx * v->dx + v->dy * v->dy);

	if (l == 0)
		return -1;

	v->dx /= l;
	v->dy /= l;
	return ((isfinite(v->dx) && isfinite (v->dy)) ? 0 : -1);
}


/* geom_vector_intersection */

int geom_vector_intersection(const struct geom_point *p1, const struct geom_vector *v1, const struct geom_point *p2, const struct geom_vector *v2, struct geom_point *result)
{
	double divisor, t;

	GASSERT(point, p1);
	GASSERT(vector, v1);
	GASSERT(point, p2);
	GASSERT(vector, v2);
	PASSERT(result);

	divisor = v1->dx * (-v2->dy) - v1->dy * (-v2->dx);

	if (divisor == 0)
		return -1;

	t = (p2->x - p1->x) * (-v2->dy) - (p2->y - p1->y) * (-v2->dx);
	t /= divisor;

	result->x = p1->x + v1->dx * t;
	result->y = p1->y + v1->dy * t;

	return ((isfinite(result->x) && isfinite(result->y)) ? 0 : -1);
}


