#ifndef __VRR_CMD_MGR_H__
#define __VRR_CMD_MGR_H__

/** \file cmdmgr.h
 *  \brief The command manager.
 *  Structures, types, enums, functions etc. for commads and menu items
 *  management.
 */

/* bit masks for states that a menu item can be in */
#define VCS_VISIBLE 0x1
#define VCS_ENABLED 0x2
#define VCS_VIS_EN  ( VCS_VISIBLE | VCS_ENABLED )
#define VCS_CHECKED 0x4

#define VCL_MAIN_TBAR 0x1
#define VCL_MAIN_MENU 0x2
#define VCL_VIEW_MENU 0x4
#define VCL_UB_MENU 0x8
#define VCL_VIEW_TBAR 0x10
#define VCL_VIEW_TBAR2 0x20
#define VCL_WUH_TBAR 0x40
#define VCL_LWUH_TBAR 0x80
#define VCL_VIEW ( VCL_VIEW_MENU | VCL_VIEW_TBAR | VCL_VIEW_TBAR2 )
#define VCL_LOCAL_MENUS ( VCL_VIEW_MENU | VCL_UB_MENU )
#define VCL_ALL_MENUS ( VCL_MAIN_MENU | VCL_VIEW_MENU | VCL_UB_MENU )
#define VCL_EVERYWHERE ( VCL_MAIN_TBAR | VCL_ALL_MENUS )

#define VCC_CHECKABLE 0x1000
#define VCC_STOCK 0x2000
#define VCC_META_SEL 0x4000
#define VCC_LOCAL_OP 0x10000
#define VCC_START_OP 0x20000
#define VCC_INSIDE_TRANS 0x40000
//! Can be called during an opened operation
#define VCC_SAFE 0x80000
#define VCC_COLLAPSED 0x100000
#define VCC_CAN_CLPS  0x200000

#define OT_TO_VTM(x) (uns)(1 << (x))
#define GOT_TO_VTM(x) (uns)(1 << (OT_MAX + (x)))
#define WT_TO_VTM(x) (uns)(1 << (OT_MAX + GOT_MAX + (x)))

struct cmd;
struct cmd_category;

/* typedef for covenience :-) */
typedef int cmd_get_state_func( struct cmd * command, GtkWidget * widget );
typedef void cmd_activate_callback( struct cmd * command, void * subj );

struct cmddata {
	struct window * window;
	struct go_group * group;
};

struct cmd {
	char * title; /**< The short menu item title. */
	char * description; /**< Description that appears in tooltip or status bar. */
	int accel; /**< The key shortcut key. */
	GdkModifierType modifier; /**< The shortcut modifier */
	uns type_mask; /**< Type mask for the current selection. */
	int count_request; /**< The number of objects that must be in the selection or 0 if any number is OK. */
	uns flags; /**< Where the command should be located. */
	unsigned int state; /**< The state the command is in. */
	char * icon_path; /**< The icon path. NULL if no icon is set. */
	GdkPixbuf * icon_pixbuf; /**< The icon pixbuf. */
	cmd_get_state_func * get_state_f; /**< The function that gets the state or NULL. */
	cmd_activate_callback * activate_f; /**< The function called when the command is activated. */
	struct of_state * of_state; /**< The state to enter for factory operations */
	struct cmd * next_cmd; /**< A pointer to next command in the same category. */
};

struct cmd_category {
	char * title; /**< The menu title. */
	unsigned int flags; /**< Where the category should be located. */
	struct cmd_category * son; /**< The son category. */
	struct cmd_category * brother; /**< The next category in the same level. */
	struct cmd * first_cmd; /**< The first command in the category. */
	struct cmd * selected_cmd; /**< The selected cmd shown when the category is collapsed. */
};

/** Initializes the command manager.
 */
void command_manager_init(void);
void command_manager_finish(void);

/** Adds a new command.
 * \param category A pointer to the category in which the command should be inserted.
 * \returns A pointer to the added command.
 */
struct cmd * register_new_command( struct cmd_category * category, struct cmd * cmd );
/** Removes a command.
 * \param command A pointer to the command.
 */
void remove_command( struct cmd * command );

/** Adds a new category into the parent category.
 * \param flags Location etc.
 * \returns A pointer to the added category.
 */
struct cmd_category * add_new_category( struct cmd_category * parent, char * title, unsigned int flags );

/** Creates the main menu.
 * \param A bit mask for location, e.g. VCL_MAIN_MENU.
 * \returns A pointer to the main menubar widget */
GtkWidget * create_menu( unsigned int location, struct cmddata cd );

/** Sets the state of the given menu to the current state
 *  of the appropriate commands. Goes through the contents of the menu
 *  and checks if axisting items are the same as items that should exist.
 *  If so, updates their state and returns 1. If not, returns 0 (indicating
 *  that the menu should be recreated).
 * \param menu A pointer to the widget to be set.
 * \param location The bit mask for menu location.
 */
int sync_menu( GtkWidget * menu, unsigned int location );

/** Creates the toolbar for the given location.
 *  \returns Pointer to the newly created toolbar.
 */
GtkWidget * create_toolbar( unsigned int location, int vertical, struct cmddata cd );

/** Sets the state of the given toolbar to the current state
 *  of the appropriate commands. Creates new subwidgets or removes
 *  some if needed.
 * \param menu A pointer to the widget to be set.
 */
int sync_toolbar( GtkWidget * toolbar, unsigned int location );

void execute_command_item( struct cmd * cmd );

/** Returns a pointer to the accel group.
 */
GtkAccelGroup * get_command_accel_group(void);

void cm_set_selection_invalid(void);

#endif
