#include "lib/lib.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <gdk/gdkkeysyms.h>
#include "gui/main.h"

struct console *console_window = NULL;

static scm_t_bits console_tag;
static int scheme_at_work = 0;

static SCM
guile_lookup (const char * s)
{ return (scm_variable_ref (scm_c_lookup (s))); }

static gunichar
get_input_buffer_char (struct console *c)
{
  GtkTextIter i;

  gtk_text_buffer_get_iter_at_mark (c -> buf, &i, c -> close_mark);

  if (! gtk_text_iter_has_tag (&i, c -> edit_tag))
    if (! gtk_text_iter_forward_to_tag_toggle (&i, c -> edit_tag))
      gtk_text_buffer_get_end_iter (c -> buf, &i);

  if (gtk_text_iter_is_end (&i))
    {
      gtk_text_buffer_move_mark (c -> buf, c -> close_mark, &i);
      return (0);
    }
  else
    {
      ASSERT (gtk_text_iter_has_tag (&i, c -> edit_tag));
      gunichar retval = gtk_text_iter_get_char (&i);
      gtk_text_iter_forward_char (&i);
      gtk_text_buffer_move_mark (c -> buf, c -> close_mark, &i);
      return (retval);
    }
}

static void
flush_input_buffer_difference (struct console *c, int blockread)
{
  GtkTextIter i1, i2, i3;
  gtk_text_buffer_get_iter_at_mark (c -> buf, &i1, c -> open_mark);
  gtk_text_buffer_get_iter_at_mark (c -> buf, &i3, c -> close_mark);

  // Block considered only with visible chars
  if (blockread)
    {
      blockread = 0;
      for (i2 = i1; (gtk_text_iter_compare (&i2, &i3) < 0); gtk_text_iter_forward_char (&i2))
	if (g_unichar_isgraph (gtk_text_iter_get_char (&i2)) &&
	    gtk_text_iter_has_tag (&i2, c -> edit_tag))
	  {
	    blockread = 1;
	    break;
	  }
    }

  if (blockread)
    {
      GtkTextMark *m;
      m = gtk_text_buffer_create_mark (c -> buf, NULL, &i1, TRUE);
      gtk_text_mark_set_visible (m, TRUE);
    }

  while (gtk_text_iter_compare (&i1, &i3) < 0)
    {
      if (! gtk_text_iter_has_tag (&i1, c -> edit_tag))
	if ((! gtk_text_iter_forward_to_tag_toggle (&i1, c -> edit_tag))
	    || (gtk_text_iter_compare (&i1, &i3) >= 0)) break;

      ASSERT (gtk_text_iter_has_tag (&i1, c -> edit_tag));
      i2 = i1;

      if ((! gtk_text_iter_forward_to_tag_toggle (&i2, c -> edit_tag))
	  || (gtk_text_iter_compare (&i2, &i3) > 0)) i2 = i3;

      gtk_text_buffer_remove_tag (c -> buf, c -> edit_tag, &i1, &i2);
      gtk_text_buffer_apply_tag (c -> buf, c -> input_tag, &i1, &i2);
      i1 = i2;
    }

  gtk_text_buffer_move_mark (c -> buf, c -> open_mark, &i3);
}

static void
undo_input_buffer_difference (struct console *c)
{
  GtkTextIter i;
  gtk_text_buffer_get_iter_at_mark (c -> buf, &i, c -> open_mark);
  gtk_text_buffer_move_mark (c -> buf, c -> close_mark, &i);
}

static int
mark_at_pos (const GtkTextIter *pos, const char *name)
{
  GSList *l = gtk_text_iter_get_marks (pos);
  GSList *i = l;
  int retval = 0;

  while (i)
    {
      const char *s = gtk_text_mark_get_name ((GtkTextMark *) (i -> data));

      if ((!s && !name) ||
	  (s && name && (strcmp (s, name) == 0)))
	{
	  retval = 1;
	  break;
	}

      i = i -> next;
    }

  g_slist_free (l);
  return (retval);
}

static void
find_border_marks (const GtkTextIter *pos, GtkTextIter *begin, GtkTextIter *end)
{
  *begin = *end = *pos;

  while ((! mark_at_pos (begin, NULL)) && (! gtk_text_iter_is_start (begin)))
    gtk_text_iter_backward_char (begin);

  gtk_text_iter_forward_char (end);
  while ((! mark_at_pos (end, NULL)) && (! gtk_text_iter_is_end (end)))
    gtk_text_iter_forward_char (end);
}

static void
text_insert (struct console *c, const char *s, int len, GtkTextTag *t)
{
  GtkTextIter i;

  scheme_at_work = 1;
  if (len == -1) len = strlen (s);
  gtk_text_buffer_get_end_iter (c -> buf, &i);
  gtk_text_buffer_insert_with_tags (c -> buf, &i, s, len,
				    (!t) ? c -> output_tag : t, NULL);
  scheme_at_work = 0;
}

static SCM
vrr_console_get_port (SCM console)
{
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-read-char");

  struct console *c = (struct console *) SCM_SMOB_DATA (console);

  return (c -> port);
}

static SCM
vrr_console_end_blockread (SCM console)
{
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-end-blockread");

  struct console *c = (struct console *) SCM_SMOB_DATA (console);

  if (c -> blockread) flush_input_buffer_difference (c, 1);
  c -> blockread = 0;

  return (SCM_UNSPECIFIED);
}

static SCM
vrr_console_begin_blockread (SCM console)
{
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-begin-blockread");

  struct console *c = (struct console *) SCM_SMOB_DATA (console);

  if (c -> blockread) vrr_console_end_blockread (console);
  c -> blockread = 1;

  return (SCM_UNSPECIFIED);
}

static SCM
vrr_console_read_char (SCM console)
{
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-read-char");

  struct console *c = (struct console *) SCM_SMOB_DATA (console);

  gunichar uc = get_input_buffer_char (c);
  if (! c -> blockread) flush_input_buffer_difference (c, 0);

  return ((uc == 0) ? SCM_EOF_VAL : (SCM_MAKE_CHAR (uc)));
}

static SCM
vrr_console_write_char (SCM console, SCM obj)
{
  SCM_ASSERT (SCM_CHARP (obj), obj, SCM_ARG2, "vrr-console-write-char");
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-write-char");

  char cc = SCM_CHAR (obj);
  text_insert ((struct console *) SCM_SMOB_DATA (console), &cc, 1, NULL);
  return (SCM_UNSPECIFIED);
}

static SCM
vrr_console_write_string (SCM console, SCM obj)
{
  SCM_ASSERT (SCM_STRINGP (obj), obj, SCM_ARG2, "vrr-console-write-string");
  SCM_ASSERT (SCM_SMOB_PREDICATE(console_tag, console),
	      console, SCM_ARG1, "vrr-console-write-string");

  text_insert ((struct console *) SCM_SMOB_DATA (console),
	       SCM_STRING_CHARS (obj), SCM_STRING_LENGTH (obj), NULL);
  return (SCM_UNSPECIFIED);
}

/*

static int idle_yielding;
static int timeout_yielding;

static gboolean timeout_yielding_cb (gpointer data);
static gboolean idle_yielding_cb (gpointer data);

static void
timeout_yielding_ensure (void)
{
  if (! timeout_yielding)
    {
      g_timeout_add (100, timeout_yielding_cb, NULL);
      timeout_yielding = 1;
    }
}

static void
idle_yielding_ensure (void)
{
  if (! idle_yielding)
    {
      g_idle_add (idle_yielding_cb, NULL);
      idle_yielding = 1;
    }
}

static gboolean
timeout_yielding_cb (gpointer data UNUSED)
{
  scm_yield ();

  if (! SCM_NFALSEP (scm_single_thread_p ()))
    {
      idle_yielding_ensure ();
      timeout_yielding = 0;
      return FALSE;
    }
  else return TRUE;
}

static gboolean
idle_yielding_cb (gpointer data UNUSED)
{
  scm_yield ();

  if (SCM_NFALSEP (scm_single_thread_p ()))
    {
      timeout_yielding_ensure ();
      idle_yielding = 0;
      return FALSE;
    }
  else return TRUE;
}

*/

static void
realized (GtkWidget *widget UNUSED, struct console *c)
{
  c -> data = scm_call_1 (guile_lookup ("vrr-repl-init"), c -> proxy);
}

static void
text_inserted (GtkTextBuffer *buf, GtkTextIter *i1,
	       UNUSED gchar *text, gint length,
	       struct console *c)
{
      GtkTextIter i2 = *i1;
      if (scheme_at_work) return;

      gtk_text_iter_backward_chars (i1, length);

      gtk_text_buffer_remove_tag (buf, c -> input_tag, i1, &i2);
      gtk_text_buffer_remove_tag (buf, c -> output_tag, i1, &i2);
      gtk_text_buffer_apply_tag (buf, c -> edit_tag, i1, &i2);
}

static void
state_update (struct console *c, int state)
{
  char buff [c -> search_buffer_size + 25];
  c -> state = state;

  switch (c -> state)
    {
    case STATE_NORMAL:
      statusbar_msg(&c->status, NULL);
      break;
    case STATE_FSEARCH_FIRST:
      statusbar_msg(&c->status, "Search:");
      break;
    case STATE_RSEARCH_FIRST:
      statusbar_msg(&c->status, "Search backwards:");
      break;
    case STATE_FSEARCH:
      strcpy (buff, "Search: ");
      strcat (buff, c -> search_buffer);
      statusbar_msg(&c->status, buff);
      break;
    case STATE_RSEARCH:
      strcpy (buff, "Search backwards: ");
      strcat (buff, c -> search_buffer);
      statusbar_msg(&c->status, buff);
      break;
    }
}

static void
search_push (struct console *c)
{
  struct console_search_stack * item = xmalloc (sizeof (* item));
  GtkTextIter pos;

  item -> state = c -> state;
  item -> search_buffer_used = c -> search_buffer_used;
  item -> search_buffer_chars = c -> search_buffer_chars;

  gtk_text_buffer_get_iter_at_mark (c -> buf, &pos, gtk_text_buffer_get_insert (c -> buf));
  item -> mark = gtk_text_buffer_create_mark (c -> buf, NULL, &pos, FALSE);
  item -> next = c -> search_stack;

  c -> search_stack = item;
}

static int
search_pop (struct console *c)
{
  struct console_search_stack *item = c -> search_stack;

  if (!item) return (0);

  c -> search_buffer_used = item -> search_buffer_used;
  c -> search_buffer_chars = item -> search_buffer_chars;
  c -> search_buffer [c -> search_buffer_used] = '\0';
  c -> search_stack = item -> next;

  GtkTextIter i1, i2;
  gtk_text_buffer_get_iter_at_mark (c -> buf, &i1, item -> mark);
  i2 = i1;
  gtk_text_iter_forward_chars (&i2, c -> search_buffer_chars);
  gtk_text_buffer_select_range (c -> buf, &i1, &i2);
  gtk_text_buffer_delete_mark (c -> buf, item -> mark);

  state_update (c, item -> state);
  xfree (item);

  return (1);
}

static void
search_flush_stack (struct console *c)
{
  struct console_search_stack *item = c -> search_stack;

  while (item)
    {
      struct console_search_stack *next = item -> next;
      gtk_text_buffer_delete_mark (c -> buf, item -> mark);
      xfree (item);
      item = next;
    }

  c -> search_stack = NULL;
}

static gboolean
directed_search (struct console *c,
		 const GtkTextIter *iter,
		 GtkTextIter *match_start,
		 GtkTextIter *match_end)
{
  if (c -> state == STATE_FSEARCH)
    {
      return (gtk_text_iter_forward_search (iter, c -> search_buffer, 0,
					    match_start, match_end, NULL));
    }

  if (c -> state == STATE_RSEARCH)
    {
      GtkTextIter pos2 = *iter;
      gtk_text_iter_forward_chars (&pos2, c -> search_buffer_chars);
      return (gtk_text_iter_backward_search (&pos2, c -> search_buffer, 0,
					     match_start, match_end, NULL));
    }

  BUG();
  return (0);
}

static int
search_iter (struct console *c, int next)
{
  GtkTextIter pos, i1, i2;
  int i;

  gtk_text_buffer_get_iter_at_mark (c -> buf, &pos, gtk_text_buffer_get_insert (c -> buf));

  i = directed_search (c, &pos, &i1, &i2);
  if (i && next && (gtk_text_iter_compare (&pos, &i1) == 0))
    {
      if (c -> state == STATE_FSEARCH) gtk_text_iter_forward_char (&pos);
      else if (c -> state == STATE_RSEARCH) gtk_text_iter_backward_char (&pos);
      else BUG ();
      i = directed_search (c, &pos, &i1, &i2);
    }

  if (i) gtk_text_buffer_select_range (c -> buf, &i1, &i2);
  return (i);
}

static void
cb_send_line (struct console *c)
{
  GtkTextIter i;

  text_insert (c, "\n", 1, c -> edit_tag);
  scm_call_1 (guile_lookup ("vrr-repl-iter"), c -> proxy);
  undo_input_buffer_difference (c);
  c -> blockread = 0;

  gtk_text_buffer_get_end_iter (c -> buf, &i);
  gtk_text_buffer_place_cursor (c -> buf, &i);

  GtkTextMark *m = gtk_text_buffer_get_insert (c -> buf);
  gtk_text_view_scroll_mark_onscreen (c -> text, m);
}

static int
cb_resurrect_line (struct console *c)
{
  char *buff = NULL;
  int buff_used = 0;
  int buff_size = 0;
  GtkTextIter i1, i2, i3, pos;
  gtk_text_buffer_get_iter_at_mark (c -> buf, &pos, gtk_text_buffer_get_insert (c -> buf));
  find_border_marks (&pos, &i1, &i3);

  while (gtk_text_iter_compare (&i1, &i3) < 0)
    {
      if (! gtk_text_iter_has_tag (&i1, c -> input_tag))
	if ((! gtk_text_iter_forward_to_tag_toggle (&i1, c -> input_tag))
	    || (gtk_text_iter_compare (&i1, &i3) >= 0)) break;

      i2 = i1;
      if (! gtk_text_iter_forward_to_tag_toggle (&i2, c -> input_tag))
	gtk_text_buffer_get_end_iter (c -> buf, &i2);

      if (gtk_text_iter_compare (&i2, &i3) > 0)
	i2 = i3;

      char *s = gtk_text_buffer_get_text (c -> buf, &i1, &i2, 1);

      if (buff_size == 0)
	{
	  buff = s;
	  buff_size = buff_used = strlen (s);
	  buff_size ++; // one for '\0'

	  ASSERT (buff_used > 0);
	}
      else
	{
	  int sl = strlen (s);
	  if ((buff_used + sl + 1) > buff_size)
	    {
	      while ((buff_used + sl + 1) > buff_size) buff_size *= 2;
	      buff = xrealloc (buff, buff_size);
	    }

	  memcpy (buff + buff_used, s, sl);
	  buff_used += sl;
	  buff [buff_used] = '\0';
	  xfree (s);
	}

      i1 = i2;
    }

  while ((buff_used > 0) && (buff [buff_used - 1] == '\n'))
    buff_used --;

  if (buff_used > 0)
    {
      buff [buff_used] = '\0';
      text_insert (c, buff, buff_used, c -> edit_tag);

      gtk_text_buffer_get_end_iter (c -> buf, &pos);
      gtk_text_buffer_place_cursor (c -> buf, &pos);

      GtkTextMark *m = gtk_text_buffer_get_insert (c -> buf);
      gtk_text_view_scroll_mark_onscreen (c -> text, m);
    }

  if (buff_size > 0) xfree (buff);

  return (buff_used > 0);
}


static int
cb_fsearch (struct console *c)
{
  int s = 1;
  switch (c -> state)
    {
    case STATE_NORMAL:
      state_update (c, STATE_FSEARCH_FIRST);
      break;
    case STATE_FSEARCH_FIRST:
      search_push (c);
      state_update (c, STATE_FSEARCH);
      s = search_iter (c, 0);
      break;
    case STATE_RSEARCH_FIRST:
      search_push (c);
      state_update (c, STATE_FSEARCH_FIRST);
      break;
    case STATE_FSEARCH:
      search_push (c);
      s = search_iter (c, 1);
      break;
    case STATE_RSEARCH:
      search_push (c);
      state_update (c, STATE_FSEARCH);
      s = search_iter (c, 0);
      break;
    }
  return (s);
}

static int
cb_rsearch (struct console *c)
{
  int s = 1;
  switch (c -> state)
    {
    case STATE_NORMAL:
      state_update (c, STATE_RSEARCH_FIRST);
      break;
    case STATE_RSEARCH_FIRST:
      search_push (c);
      state_update (c, STATE_RSEARCH);
      s = search_iter (c, 0);
      break;
    case STATE_FSEARCH_FIRST:
      search_push (c);
      state_update (c, STATE_RSEARCH_FIRST);
      break;
    case STATE_RSEARCH:
      search_push (c);
      s = search_iter (c, 1);
      break;
    case STATE_FSEARCH:
      search_push (c);
      state_update (c, STATE_RSEARCH);
      s = search_iter (c, 0);
      break;
    }
  return (s);
}

static int
search_add (struct console *c, guint32 wc)
{
  gchar buff [8];
  int clength;

  if (c -> search_buffer_used == 1024) return (0);

  if (c -> state == STATE_FSEARCH_FIRST)
    {
      c -> search_buffer_used = 0;
      c -> search_buffer_chars = 0;
      state_update (c, STATE_FSEARCH);
    }

  if (c -> state == STATE_RSEARCH_FIRST)
    {
      c -> search_buffer_used = 0;
      c -> search_buffer_chars = 0;
      state_update (c, STATE_RSEARCH);
    }

  clength = g_unichar_to_utf8 (wc, buff);

  if ((c -> search_buffer_used + clength + 1) > c -> search_buffer_size)
    {
      char *old = c -> search_buffer;
      c -> search_buffer_size *= 2;
      // because buffer_size starts at 32 and one char is max 6, it is now ok
      c -> search_buffer = xmalloc (c -> search_buffer_size);
      memcpy (c -> search_buffer, old, c -> search_buffer_used);
      xfree (old);
    }

  search_push (c);
  memcpy (c -> search_buffer + c -> search_buffer_used, buff, clength);
  c -> search_buffer_used += clength;
  c -> search_buffer_chars ++;
  c -> search_buffer [c -> search_buffer_used] = '\0';

  state_update (c, c -> state);
  return (search_iter (c, 0));
}

static void
cb_search_end (struct console *c)
{
  search_flush_stack (c);
  state_update (c, STATE_NORMAL);
}

	  /*      GtkTextIter i;
		  GtkTextMark *m1 = gtk_text_buffer_get_mark (c -> buf, "insert");
		  GtkTextMark *m2 = gtk_text_buffer_get_mark (c -> buf, "selection_bound");
		  gtk_text_buffer_get_end_iter (c -> buf, &i);
		  gtk_text_buffer_move_mark (c -> buf, m1, &i);
		  gtk_text_buffer_move_mark (c -> buf, m2, &i);
		  gtk_text_view_scroll_mark_onscreen (c -> text, m1);
	  */

static inline int
beep (struct console *c)
{
  gdk_display_beep (gtk_widget_get_display (GTK_WIDGET (c -> window)));
  return (1);
}

static gboolean
key_pressed (GtkTextView *view UNUSED, GdkEventKey *event, struct console *c)
{ /* This procedure looks very ugly */
  switch (c -> state)
    {
    case STATE_NORMAL:
      if ((! (event -> state & GDK_SHIFT_MASK))
	  && (event -> keyval == GDK_Return))
	{
	  GtkTextIter pos;
	  gtk_text_buffer_get_iter_at_mark (c -> buf, &pos,
					    gtk_text_buffer_get_insert (c -> buf));
	  if (gtk_text_iter_has_tag (&pos, c -> edit_tag) ||
	      gtk_text_iter_is_end (&pos))
	  {
            factory_op_break ();
	    cb_send_line (c);
	    refresh_gui ();
	  }
	  else
	    cb_resurrect_line (c) || beep (c);
	  return TRUE;
	}
      else if ((event -> state & GDK_CONTROL_MASK)
	       && (event -> keyval == GDK_s))
	{
	  cb_fsearch (c) || beep (c);

	  return TRUE;
	}
      else if ((event -> state & GDK_CONTROL_MASK) &&
	       (event -> keyval == GDK_r))
	{
	  cb_rsearch (c) || beep (c);
	  return TRUE;
	}
      else return FALSE;
      break;
    case STATE_FSEARCH_FIRST:
    case STATE_RSEARCH_FIRST:
    case STATE_FSEARCH:
    case STATE_RSEARCH:
      if ((event -> state & GDK_CONTROL_MASK)
	       && (event -> keyval == GDK_s))
	{
	  cb_fsearch (c) || beep (c);
	  return TRUE;
	}
      else if ((event -> state & GDK_CONTROL_MASK)
	       && (event -> keyval == GDK_r))
	{
	  cb_rsearch (c) || beep (c);
	  return TRUE;
	}
      else if ((event -> keyval == GDK_Return)
	       || (event -> keyval == GDK_Escape))
	{
	  cb_search_end (c);
	  return TRUE;
	}
      else if (event -> keyval == GDK_BackSpace)
	{
	  search_pop (c) || beep (c);
	  return TRUE;
	}
      else if ((event -> keyval == GDK_Left) ||
	       (event -> keyval == GDK_Right) ||
	       (event -> keyval == GDK_Up) ||
	       (event -> keyval == GDK_Down) ||
	       (event -> keyval == GDK_Page_Up) ||
	       (event -> keyval == GDK_Page_Down) ||
	       (event -> keyval == GDK_Home) ||
	       (event -> keyval == GDK_End) ||
	       ((event -> state & (GDK_CONTROL_MASK | GDK_MOD1_MASK |
				   GDK_MOD2_MASK | GDK_MOD3_MASK |
				   GDK_MOD4_MASK | GDK_MOD5_MASK)) != 0))
	{ /* Silly heuristic for identify non-insertable keystrokes */
	  cb_search_end (c);
	  return FALSE;
	}
      else
	{
	  guint32 ug = gdk_keyval_to_unicode (event -> keyval);

	  if (ug != 0)
	    {
	      search_add (c, ug) || beep (c);
	      return TRUE;
	    }
	}
      break;
    }
  return FALSE;
}

static gboolean on_delete_event (GtkWidget *widget UNUSED, GdkEvent *event UNUSED, gpointer data UNUSED)
{
	gtk_widget_hide (console_window -> window);
	console_window -> visible = 0;
	refresh_gui();
	return TRUE;
}

void
console_create(void)
{
  console_window = xmalloc (sizeof (struct console));
  SCM_NEWSMOB (console_window -> proxy, console_tag, console_window);

  console_window -> state = STATE_NORMAL;
  console_window -> search_buffer_size = 32;
  console_window -> search_buffer_used = 0;
  console_window -> search_buffer_chars = 0;
  console_window -> search_buffer = xmalloc (console_window -> search_buffer_size);
  console_window -> search_buffer [0] = '\0';
  console_window -> search_stack = NULL;
  console_window -> visible = 1;

  console_window -> port = scm_call_1 (guile_lookup ("make-vrr-console-port"), console_window -> proxy);

  console_window -> window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW(console_window -> window), "Scheme console");
  gtk_window_set_default_size (GTK_WINDOW(console_window -> window), 400, 400);

  console_window -> status.statusbar = gtk_statusbar_new ();
  console_window -> status.context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR(console_window -> status.statusbar), "info");

  GtkScrolledWindow *sw = GTK_SCROLLED_WINDOW (gtk_scrolled_window_new (NULL, NULL));
  gtk_scrolled_window_set_policy (sw, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  console_window -> text = GTK_TEXT_VIEW (gtk_text_view_new ());
  gtk_text_view_set_editable (console_window -> text, 1);
  gtk_text_view_set_wrap_mode (console_window -> text, GTK_WRAP_WORD);

  console_window -> buf = gtk_text_view_get_buffer (console_window -> text);

  GtkTextIter i;
  gtk_text_buffer_get_start_iter (console_window -> buf, &i);

  console_window -> open_mark = gtk_text_buffer_create_mark (console_window -> buf, "open", &i, TRUE);
  console_window -> close_mark = gtk_text_buffer_create_mark (console_window -> buf, "close", &i, TRUE);

  console_window -> input_tag = gtk_text_buffer_create_tag (console_window -> buf, NULL,
					       "weight", PANGO_WEIGHT_NORMAL,
					       "family", "Monospace",
					       "family-set", 1,
					       "editable", 0, NULL);

  console_window -> output_tag = gtk_text_buffer_create_tag (console_window -> buf, NULL,
						"weight", PANGO_WEIGHT_NORMAL,
						"family", "Monospace",
						"family-set", 1,
						"editable", 0, NULL);

  console_window -> edit_tag = gtk_text_buffer_create_tag (console_window -> buf, NULL,
					      "weight", PANGO_WEIGHT_BOLD,
					      "family", "Monospace",
					      "family-set", 1,
					      "editable", 1, NULL);


  //  gtk_text_buffer_insert_with_tags (buf, &i, "XXX", -1, console_window -> tag, NULL);



  g_signal_connect (G_OBJECT(console_window -> text), "key-press-event", G_CALLBACK (key_pressed), console_window);
  g_signal_connect (G_OBJECT(console_window -> text), "realize", G_CALLBACK (realized), console_window);
  //  g_signal_connect (G_OBJECT(c -> buf), "changed", G_CALLBACK (input_changed), console_window);
  g_signal_connect_after (G_OBJECT(console_window -> buf), "insert-text", G_CALLBACK (text_inserted), console_window);
  g_signal_connect(G_OBJECT(console_window -> window), "delete_event",
		  G_CALLBACK(on_delete_event), NULL);

  gtk_widget_add_events (GTK_WIDGET (console_window -> text), GDK_KEY_PRESS);

  gtk_container_add (GTK_CONTAINER (sw), GTK_WIDGET (console_window -> text));

  GtkBox *b = GTK_BOX (gtk_vbox_new (FALSE, 0));
  gtk_box_pack_start (b, GTK_WIDGET (sw), 1, 1, 0);
  gtk_box_pack_start (b, GTK_WIDGET (console_window -> status.statusbar), 0, 1, 0);

  gtk_container_add (GTK_CONTAINER (console_window -> window), GTK_WIDGET (b));
  gtk_widget_show_all (console_window -> window);
}

void
console_init (void)
{
  console_tag = scm_make_smob_type ("vrr-console", 0);
  scm_c_define_gsubr("vrr-console-begin-blockread", 1, 0, 0, vrr_console_begin_blockread);
  scm_c_define_gsubr("vrr-console-end-blockread", 1, 0, 0, vrr_console_end_blockread);
  scm_c_define_gsubr("vrr-console-get-input-port", 1, 0, 0, vrr_console_get_port);
  scm_c_define_gsubr("vrr-console-get-output-port", 1, 0, 0, vrr_console_get_port);
  scm_c_define_gsubr("vrr-console-read-char", 1, 0, 0, vrr_console_read_char);
  scm_c_define_gsubr("vrr-console-write-char", 2, 0, 0, vrr_console_write_char);
  scm_c_define_gsubr("vrr-console-write-string", 2, 0, 0, vrr_console_write_string);
}
