#undef LOCAL_DEBUG
#include "lib/lib.h"

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <values.h>
#include <fontconfig/fontconfig.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iconv.h>
#include <errno.h>
#include <langinfo.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "font/font.h"
#include "gui/main.h"
#include "gui/dialogs.h"
#include "gui/properties.h"

#define ENTER_CRITICAL_SECTION { if (text_edit_locked) return; text_edit_locked = 1; }
#define LEAVE_CRITICAL_SECTION { text_edit_locked = 0; return; }
#define OFFSET sizeof (fd->path)

#define TEXT_EDIT_PROP_CNT 5

static int text_edit_locked = 0;
const char *text_edit_string_unknown = "(unknown)";

GList *fonts_list;
GSList * text_edit_list;
static int last_used_font;

static int fill_fonts_list (GList **list);

static void text_edit_destroy( GtkWidget * w, struct text_edit * te );
static void text_edit_go_del( struct o * o, struct o *owner UNUSED, struct o *prev UNUSED, struct go_hook* h );
static void text_edit_obj_del( struct o *o, struct o *owner UNUSED, struct o *prev UNUSED, struct obj_hook* h );
static void text_edit_go_prop_changed( struct o *o, struct go_hook* h );
static void text_edit_refresh_go( GtkButton * b, struct text_edit * te );
static void text_edit_refresh_widgets( struct text_edit * te );

struct font_data
{
	char *path;
	char text[1];
};

static void free_list_item (gpointer data, gpointer user_data UNUSED)
{
	struct font_data *fd = (struct font_data *) ((char *) data - OFFSET);

	xfree (fd->path);
	xfree (fd);
}

void fonts_list_free (void)
{
	int free_fi = 1;
	if (fonts_list)
	{
		g_list_foreach (fonts_list, free_list_item, &free_fi);
		g_list_free (fonts_list);
		free_fi = 0;
	}
}

static int compare_fonts (gconstpointer a, gconstpointer data)
{
	return strcmp ((char *) a, (char *) data);
}

static int get_family_and_style(char *filename, FcChar8 **family, FcChar8 **style)
{
	FcFontSet *set;
	FcPattern *pattern, *font;
	FcObjectSet *os = NULL;

	os = FcObjectSetCreate ();
	if (!os)
		return 0;
	FcObjectSetAdd (os, FC_STYLE);
	FcObjectSetAdd (os, FC_FAMILY);
	pattern = FcPatternCreate();
	if (!pattern)
		return 0;
	FcPatternAddString (pattern, FC_FILE, filename);

	set = FcFontList (0, pattern, os);
	FcPatternDestroy (pattern);
	FcObjectSetDestroy (os);
	if (!set)
		return 0;
	font = set->fonts[0];
	if (FcPatternGetString (font, FC_FAMILY, 0, family) == FcResultNoMatch)
		return 0;
	if (FcPatternGetString (font, FC_STYLE, 0, style) == FcResultNoMatch)
		return 0;
	return 1;
}

static void insert_item(GList **list, const char *path, char *family, char *style)
{
	DBG("Insert item\n");
	int r = 1;
	struct font_data *fd = xmalloc_zero (sizeof (struct font_data) + strlen (family) + (style ? strlen (style) : 0) + 16);
	GList *tmp;

	fd->path = xstrdup(path);
	if (!style)
		sprintf(fd->text, "%s", family);
	else
		sprintf(fd->text, "%s (%s)", family, style);
	tmp = *list;
	while (tmp)
	{
		r = strcmp(fd->text, (char *) tmp->data);
	    if (r <= 0)
			break;
		tmp = tmp->next;
	}
	if (r)
	{
		if (!tmp)
			*list = g_list_append (*list, (char *) fd + OFFSET);
		else if (tmp->prev)
			tmp->prev->next = g_list_prepend (tmp, (char *) fd + OFFSET);
		else
			*list = g_list_prepend (*list, (char *) fd + OFFSET);
	}
	else
		xfree (fd);
}

int fill_fonts_list (GList **list)
{
	int i;
	FcPattern *pattern, *font;
	FcChar8 *family, *style, *fontpath;
	FcObjectSet *os = 0;
	FcFontSet *set;

//	show_progress_bar(te);

	pattern = FcPatternCreate();
	if (!pattern)
		return 0;
	FcPatternAddBool (pattern, FC_SCALABLE, FcTrue);

	os = FcObjectSetCreate ();
	if (!os)
		return 0;
	FcObjectSetAdd (os, FC_FILE);
	FcObjectSetAdd (os, FC_STYLE);
	FcObjectSetAdd (os, FC_FAMILY);

	set = FcFontList (0, pattern, os);
	FcPatternDestroy (pattern);
	FcObjectSetDestroy (os);
	if (!set)
		return 0;
	for (i = 0; i < set->nfont; i++)
	{
		font = set->fonts[i];
		if (FcPatternGetString (font, FC_FILE, 0, &fontpath) == FcResultNoMatch)
			return 0;
		if (FcPatternGetString (font, FC_FAMILY, 0, &family) == FcResultNoMatch )
			return 0;
		if (FcPatternGetString (font, FC_STYLE, 0, &style) == FcResultNoMatch)
			return 0;
		DBG("FILL_FONTS: inserting %s %s\n", fontpath, fontname);
		insert_item(list, fontpath, family, style);
	}

	if (!(*list))
		return 0;
	return 1;
}

int fonts_get_default(void)
{
	if ( last_used_font )
		return last_used_font;

	if (!fonts_list)
		if (!fill_fonts_list(&fonts_list))
			trans_fail("Could not load fonts.");
	//get the first font we have found
	struct font_data * fd = (struct font_data*)((char *)fonts_list->data - OFFSET);
	return font_load_file( fd->path );
}

static char *get_text_from_text_buffer (GtkTextBuffer *text_buf)
{
	GtkTextIter startiter, enditer;

	gtk_text_buffer_get_start_iter (text_buf, &startiter);
	gtk_text_buffer_get_end_iter (text_buf, &enditer);
	return gtk_text_buffer_get_text (text_buf, &startiter, &enditer, 1);
}

static void text_edit_refresh_go( GtkButton * b UNUSED, struct text_edit * te )
{
	volatile int new_font_id = 0;

	ASSERT( te );
	char err[TRANS_ERR_SIZE];

	statusbar_clear (&te->status);
	if(te->go->type == GOT_TEXT)
	{
		GList *tmp;
		struct font_data *fd;

		tmp = g_list_find_custom (fonts_list, gtk_entry_get_text (GTK_ENTRY (te->entry_fonts)), compare_fonts);
		if (tmp)
		{
			fd = (struct font_data *) ((char *) tmp->data - OFFSET);
			last_used_font = new_font_id = font_load_file (fd->path);
		}
		else
		{
			statusbar_msg (&te->status, "Invalid font - select a font from combolist.");
			report_error("Invalid font\n");
			return;
		}
		te->fnt_id = new_font_id;
	}
	char *s = get_text_from_text_buffer (te->text_buf);
	/*if (!(*s))
	{
		statusbar_msg(&te->status, "Enter the text.");
		report_error ("Empty text\n");
		return;
	}*/

	te->changing = 1;
	TRANS_BEGIN_MAIN( te->parent_tlo, "Change text properties", err)

		prop_set_text( go2o(te->go), string_text, s);
		if ( te->go->type == GOT_TEXT )
			prop_set_uns ( go2o(te->go), string_font, (uintptr_t)new_font_id );
	TRANS_FAILED
		report_error_f("%s\n", err);
	TRANS_END
	te->changing = 0;

	g_free(s);
}

static void text_edit_refresh_widgets( struct text_edit * te )
{
	struct o * o = go2o(te->go);
	const char * text  = prop_lookup_text( o, string_text );
	char *fontname, *filename;
	FcChar8 *family, *style;

	if ( text ) gtk_text_buffer_set_text( te->text_buf, text, -1 );

	uns counter = 0;
	if ( te->go->type == GOT_TEXT )
	{
		te->fnt_id = prop_lookup_uns( o, string_font, 0 );
		filename = (char *) font_filename (te->fnt_id);
		if (!get_family_and_style (filename, &family, &style))
			fontname = (char *) text_edit_string_unknown;
		else
		{
			fontname = alloca (strlen (family) + strlen (style) + 16);
			sprintf (fontname, "%s (%s)", family, style);
		}
		insert_item (&fonts_list, filename, fontname, 0);
		gtk_entry_set_text (GTK_ENTRY (te->entry_fonts), fontname);
	
		prop_sync( te->pi + counter++, prop_find( o, string_font_size ) );
	}
	prop_sync( te->pi + counter++, prop_find( o, string_rpos_x ) );
	prop_sync( te->pi + counter++, prop_find( o, string_rpos_y ) );
	prop_sync( te->pi + counter++, prop_find( o, string_apos_x ) );
	prop_sync( te->pi + counter++, prop_find( o, string_apos_y ) );
}

static void text_edit_go_prop_changed( struct o *o, struct go_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( go2o(te->go) == o && !te->changing ) text_edit_refresh_widgets(te);
}

static void text_edit_obj_del( struct o *o, struct o *owner UNUSED, struct o *prev UNUSED, struct obj_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( o_is_ancestor(o, obj2o(&te->parent_tlo->obj))) gtk_widget_destroy( te->window );
}

static void text_edit_go_del( struct o * o, struct o *owner UNUSED, struct o *prev UNUSED, struct go_hook* h )
{
	struct text_edit * te = (struct text_edit*)h->data;
	if ( go2o(te->go) == o ) gtk_widget_destroy( te->window );
}

static void text_edit_destroy( GtkWidget * w UNUSED, struct text_edit * te )
{
	go_unref(te->go);

	go_hook_remove( te->parent_tlo, &te->go_hook );
	obj_hook_remove( &te->obj_hook );
	obj_unref( &te->parent_tlo->obj );

	text_edit_list = g_slist_remove( text_edit_list, (gpointer)te );
	xfree(te);
}

static void entry_fonts_changed (GtkWidget *entry, struct text_edit *te)
{
	char *txt;
	GList *tmp;
	struct font_data *fd;
	GList *list = fonts_list;

	ENTER_CRITICAL_SECTION;

	txt = (char *) gtk_entry_get_text (GTK_ENTRY (entry));
	if (!txt || !(*txt))
		LEAVE_CRITICAL_SECTION;
	tmp = g_list_find_custom (list, txt, compare_fonts);
	if (!tmp)
		LEAVE_CRITICAL_SECTION;

	fd = (struct font_data *) ((char *) tmp->data - OFFSET);
	insert_item (&fonts_list, fd->path, fd->text, 0);

	gtk_entry_set_text (GTK_ENTRY (te->entry_fonts), txt);

	LEAVE_CRITICAL_SECTION;
}

static char *convert_text (char *text, int exporting)
{
	iconv_t conv;
	char *local_charset = nl_langinfo (CODESET);
	if (exporting)
		conv = iconv_open (local_charset, "UTF8");
	else
		conv = iconv_open ("UTF8", local_charset);
	if (conv == (iconv_t) -1) {
		report_error ("Cannot initialize charset conversion.\n");
		return NULL;
	}
	size_t in_len = strlen (text);
   	size_t out_len = 4*in_len + 1;
	char *output = xmalloc (out_len);
	char *outpos = output;
	int res = iconv (conv, &text, &in_len, &outpos, &out_len);
	iconv_close(conv);
	if (res < 0) {
		report_error_f ("Charset conversion failed: %s\n", strerror(errno));
		xfree (output);
		return NULL;
	}
	*outpos = 0;
	return output;
}

static int save_text (char *text, const char *filename)
{
	text = convert_text (text, 1);
	if (!text)
		return 1;
	int fd = open(filename, O_WRONLY | O_TRUNC | O_CREAT, 0666);
	int result = 1;
	if (fd < 0)
		report_error_f ("Cannot open %s for writing: %s\n", filename, strerror (errno));
	else {
		int len = strlen (text);
		if (write (fd, text, len) == len)
			result = 0;
		else
			report_error_f ("Error writing %s\n", filename);
		close (fd);
	}
	xfree (text);
	return result;
}

static char *load_text (const char *filename)
{
	int fd = open(filename, O_RDONLY);
	if (fd < 0) {
		report_error_f ("Cannot open %s for reading: %s\n", filename, strerror (errno));
		return NULL;
	}
	struct stat st;
	char *result = NULL;
	if (stat (filename, &st) < 0)
		report_error_f ("Cannot stat %s: %s\n", filename, strerror (errno));
	else {
		char *buf = xmalloc (st.st_size + 1);
		if (read (fd, buf, st.st_size) != st.st_size)
			report_error_f ("Error reading %s\n", filename);
		else {
			buf[st.st_size] = 0;
			result = convert_text (buf, 0);
		}
		xfree (buf);
	}
	close (fd);
	return result;
}

static int save_text_edit (struct text_edit *te, const char *filename)
{
	char *text = get_text_from_text_buffer (te->text_buf);
	int res = save_text (text, filename);
	g_free (text);
	return res;
}

static int load_text_edit (struct text_edit *te, const char *filename)
{
	char *text = load_text (filename);
	if (!text)
		return 1;
	gtk_text_buffer_set_text (te->text_buf, text, -1);
	xfree (text);
	return 0;
}

static void on_save_text (GtkWidget *button UNUSED, struct text_edit *te)
{
	SAVE_DLG_START("Save text", NULL, NULL)
	{
		if (save_text_edit (te, filename))
			statusbar_msg (&te->status, "couldn't save text into the file.");
		else
			statusbar_msg (&te->status, "file saved successfully.");
	}
	SAVE_DLG_END;
}

static void on_load_text (GtkWidget *button UNUSED, struct text_edit *te)
{
	OPEN_DLG_START( "Open text", NULL, NULL )
	{
		if (load_text_edit (te, filenames[0]))
			statusbar_msg (&te->status, "Couldn't load text from the file.");
		else
			statusbar_msg (&te->status, "File loaded successfully.");
		break; //load only the first selected file
	}
	OPEN_DLG_END;
}

static void open_external_editor (GtkWidget *button UNUSED, struct text_edit *te)
{
	const char * editor = prop_lookup_text( obj2o(obj_universum), string_editor );
	if (!editor) { report_error("Please set the text editor in the Settings first!\n"); return; }
	uns run_in_term = prop_lookup_uns( obj2o(obj_universum), string_editor_in_term, 2 );
	if (run_in_term == 2) { report_error("Please set if to run the text editor in a terminal first!\n"); return; }
	const char * term = NULL;
	if (run_in_term)
	{
		term = prop_lookup_text( obj2o(obj_universum), string_terminal );
		if (!term) { report_error("Please set the terminal in the Settings first!\n"); return; }
	}

	char *tmpdir = getenv ("TMPDIR");
	if (!tmpdir)
		tmpdir = "/tmp";
	char *filename = alloca (strlen (tmpdir) + 11);
	sprintf (filename, "%s/vrrXXXXXX", tmpdir);
	int fd = mkstemp(filename);
	if (fd < 0) { report_error("Cannot create temporary file.\n"); return; }
	close (fd);

	if (save_text_edit (te, filename))
		goto remove_file;

	pid_t pid;
	switch (pid = fork())
	{
	case -1: report_error("Cannot fork!\n"); break;
	case 0:
		msg("About to run editor on file %s\n", filename );
		if (run_in_term)
			execlp( term, term, "-e", editor, filename, NULL );
		else
			execlp( editor, editor, filename, NULL );
		report_error_f("Could not run the editor %s on %s: %s\n", editor, filename, strerror(errno));
		exit(1);

	default:;
		int status; waitpid(pid, &status, 0); // TODO: something better! This makes vrr freeze completely!
		if (!WIFEXITED(status) || WEXITSTATUS(status))
			report_error_f("Editor failed (exited with code %d)\n", status);
		else
			load_text_edit (te, filename);
	}

remove_file:
	unlink (filename);
}

#define TE_ADD_PROP( _left, _top, _key, _desc ) \
	ASSERT(counter < TEXT_EDIT_PROP_CNT); \
	pi = te->pi + counter++; \
	prop_item_init( pi, go2o(te->go), _key ); \
	label = gtk_label_new(_desc); \
	g_object_set( G_OBJECT(label), "xalign", 0.0, NULL ); \
	gtk_table_attach( table, label, _left, _left+1, _top, _top+1, GTK_FILL, GTK_FILL, 5, 0 ); \
	if (pi->unit_edit) \
	{ \
		gtk_table_attach( table, pi->value_edit, _left+1, _left+2, _top, _top+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 ); \
		gtk_table_attach( table, pi->unit_edit, _left+2, _left+3, _top, _top+1, GTK_FILL, GTK_FILL, 0, 0 ); \
	} else \
		gtk_table_attach( table, pi->value_edit, _left+1, _left+3, _top, _top+1, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0 )

void text_edit_create( struct go * go )
{
	ASSERT( go && (go->type == GOT_TEXT || go->type == GOT_TEX_TEXT ) );

	struct text_edit * te = xmalloc_zero(sizeof(struct text_edit));
	te->pi = xmalloc_zero(sizeof(struct prop_item) * TEXT_EDIT_PROP_CNT);
	GtkTable *table;
	GtkWidget * widget, * scw, * vbox, * butbox, *hbox;
	GtkWidget *button_save, *button_load, *button_external_editor;
	GtkWidget *label_text, *text_hbox;
	GtkTextView * tw;
	GtkAccelGroup * ag = gtk_accel_group_new();
	GtkTooltips *tooltips = gtk_tooltips_new();

	te->fnt_id = 0;
	te->changing = 0;
	te->go = go;
	uns tex = go->type == GOT_TEX_TEXT;

	te->window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
	gtk_window_set_title( GTK_WINDOW(te->window), tex ? "TeX Text Editor" : "Text Editor" );
	gtk_window_add_accel_group( GTK_WINDOW(te->window), ag );
	g_signal_connect( G_OBJECT(te->window), "destroy", G_CALLBACK(text_edit_destroy), te );

	gtk_container_add( GTK_CONTAINER(te->window), vbox = gtk_vbox_new( FALSE, 0 ) );

	if ( !tex )
	{
		// Font
		GtkWidget *font_label;

		gtk_box_pack_start( GTK_BOX(vbox), hbox = gtk_hbox_new( FALSE, 0 ), FALSE, TRUE, 3 );
		font_label = gtk_label_new("Font: ");
		gtk_widget_show (font_label);
		gtk_box_pack_start (GTK_BOX (hbox), font_label, FALSE, FALSE, 5);

		te->combo_fonts = gtk_combo_new ();
		gtk_widget_show (te->combo_fonts);
		gtk_box_pack_start (GTK_BOX(hbox), te->combo_fonts, TRUE, TRUE, 3);
		if (!fonts_list)
			if (!fill_fonts_list(&fonts_list))
				report_error_f("Couldn't load fonts.\n");

		if (te->go)
			te->fnt_id = prop_lookup_uns( go2o(te->go), string_font, 0 );
		if ( font_id_valid(te->fnt_id) )
		{
			char *filename = (char *) font_filename (te->fnt_id);
			char *fontname;
			FcChar8 *family, *style;

			if (!get_family_and_style (filename, &family, &style))
				fontname = (char *) text_edit_string_unknown;
			else
			{
				fontname = alloca (strlen (family) + strlen (style) + 16);
				sprintf (fontname, "%s (%s)", family, style);
			}
			insert_item(&fonts_list, filename, fontname, 0);
		}

		te->entry_fonts = GTK_COMBO (te->combo_fonts)->entry;
		gtk_tooltips_set_tip (tooltips, te->entry_fonts, "Select the font from this combo.", NULL);
		gtk_widget_show (te->entry_fonts);
		gtk_entry_set_max_length (GTK_ENTRY (te->entry_fonts), 60);

		gtk_combo_set_popdown_strings (GTK_COMBO (te->combo_fonts), fonts_list);
		gtk_combo_disable_activate (GTK_COMBO (te->combo_fonts));
		gtk_entry_set_editable (GTK_ENTRY (te->entry_fonts), FALSE);

		gtk_box_pack_start( GTK_BOX(vbox), gtk_hseparator_new(), FALSE, TRUE, 0 );

		g_signal_connect(G_OBJECT (te->entry_fonts), "changed",
				G_CALLBACK (entry_fonts_changed), te);
	}
	
	table = GTK_TABLE(gtk_table_new( 6, 3, FALSE ));
	gtk_box_pack_start( GTK_BOX(vbox), GTK_WIDGET(table), FALSE, TRUE, 0 );

	uns counter = 0;
	GtkWidget * label;
	struct prop_item * pi;
	if (te->go->type == GOT_TEXT) { TE_ADD_PROP( 0, 0, string_font_size, "Font size:" ); }

	TE_ADD_PROP( 0, 1, string_rpos_x, "Rel-x:" );
	TE_ADD_PROP( 3, 1, string_rpos_y, "Rel-y:" );
	TE_ADD_PROP( 0, 2, string_apos_x, "Abs-x:" );
	TE_ADD_PROP( 3, 2, string_apos_y, "Abs-y:" );

	gtk_box_pack_start( GTK_BOX(vbox), gtk_hseparator_new(), FALSE, TRUE, 0 );

	// label text and buttons Save & Load
	text_hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), text_hbox, FALSE, TRUE, 0);
	label_text = gtk_label_new ("Text:");
	gtk_widget_show (label_text);
	gtk_box_pack_start (GTK_BOX (text_hbox), label_text, FALSE, TRUE, 0);
	button_save = gtk_button_new_with_label ("Save ...");
	gtk_tooltips_set_tip (tooltips, button_save, "Save text into the file.", NULL);
	g_signal_connect( G_OBJECT(button_save), "clicked", G_CALLBACK(on_save_text), te );
	gtk_widget_show (button_save);
	gtk_box_pack_end (GTK_BOX (text_hbox), button_save, FALSE, TRUE, 0);
	button_load = gtk_button_new_with_label ("Load ...");
	gtk_tooltips_set_tip (tooltips, button_load, "Load text from the file.", NULL);
	g_signal_connect( G_OBJECT(button_load), "clicked", G_CALLBACK(on_load_text), te );
	gtk_widget_show (button_load);
	gtk_box_pack_end (GTK_BOX (text_hbox), button_load, FALSE, TRUE, 3);
	button_external_editor = gtk_button_new_with_label ("Edit with external editor ...");
	gtk_tooltips_set_tip (tooltips, button_external_editor, "Use extern editor to edit text.", NULL);
	g_signal_connect (G_OBJECT (button_external_editor), "clicked", G_CALLBACK(open_external_editor), te);
	gtk_widget_show (button_external_editor);
	gtk_box_pack_end (GTK_BOX (text_hbox), button_external_editor, FALSE, TRUE, 0);

	// Text view
	scw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
	gtk_box_pack_start( GTK_BOX(vbox), scw, TRUE, TRUE, 0 );

	tw = GTK_TEXT_VIEW( gtk_text_view_new() );
	gtk_text_view_set_editable( tw, 1 );
	gtk_text_view_set_wrap_mode( tw, GTK_WRAP_WORD );
	gtk_container_add( GTK_CONTAINER(scw), GTK_WIDGET(tw) );
	te->text_buf = gtk_text_view_get_buffer(tw);

	// Buttons
	gtk_box_pack_start( GTK_BOX(vbox), butbox = gtk_hbox_new( FALSE, 5 ), FALSE, TRUE, 0 );
	widget = gtk_button_new_with_label( "Close" );
	gtk_widget_set_size_request (widget, 84, -1);
	gtk_tooltips_set_tip (tooltips, widget, "Close the window without refreshing text's properties.", NULL);
	g_signal_connect_swapped( G_OBJECT(widget), "clicked", G_CALLBACK(gtk_widget_destroy), te->window );
	gtk_box_pack_start( GTK_BOX(butbox), widget, FALSE, TRUE, 0 );
	widget = gtk_button_new_with_label( "Refresh" );
	gtk_widget_set_size_request (widget, 84, -1);
	gtk_tooltips_set_tip (tooltips, widget, "Apply all made changes and save it into the text's properties", NULL);
	g_signal_connect( G_OBJECT(widget), "clicked", G_CALLBACK(text_edit_refresh_go), te );
	gtk_widget_add_accelerator( widget, "clicked", ag, GDK_KP_Enter, GDK_CONTROL_MASK, 0 );
	gtk_widget_add_accelerator( widget, "clicked", ag, GDK_Return, GDK_CONTROL_MASK, 0 );
	gtk_box_pack_end( GTK_BOX(butbox), widget, FALSE, TRUE, 0 );

	te->status.statusbar = gtk_statusbar_new();
	te->status.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(te->status.statusbar), "Text editor");
	gtk_box_pack_start( GTK_BOX(vbox), te->status.statusbar, FALSE, TRUE, 0 );

/*	te->progress_bar = gtk_progress_bar_new();
	gtk_box_pack_start( GTK_BOX(vbox), te->progress_bar, FALSE, TRUE, 0 );
	g_object_set( G_OBJECT(te->progress_bar), "text", "Loading fonts, please wait...", "pulse-step", 0.1, NULL ); */

	gtk_widget_show_all( te->window );
//	gtk_widget_hide( te->progress_bar );
	gtk_widget_grab_focus( GTK_WIDGET(tw) );

	text_edit_list = g_slist_prepend( text_edit_list, (gpointer)te );

	go_ref(te->go);
	te->parent_tlo = go_get_tlo(go);
	text_edit_refresh_widgets(te);
	obj_ref( &te->parent_tlo->obj );

	te->go_hook.data = te;
	te->go_hook.go_unlink = text_edit_go_del;
	te->go_hook.go_prop_change = text_edit_go_prop_changed;
	go_hook_add( te->parent_tlo, &te->go_hook );
	te->obj_hook.data = te;
	te->obj_hook.obj_unlink = text_edit_obj_del;
	obj_hook_add( &te->obj_hook );
}
