#ifndef __VRR_MAIN_H__
#define __VRR_MAIN_H__

/** \file gui/main.h
 * The main gui structures, windows, and functions.
 */
#include "lib/lib.h"

#include <gtk/gtk.h>
#include <libguile.h>
#include "kernel/kernel.h"
#include "gui/visualisation.h"

/** Maximal length of error mssage (includes also additional arguments).  */
#define MAX_ERR_MSG_LENGTH		255
/** URL of the VRR homepage */
#define VRR_HOMEPAGE			"http://vrr.ucw.cz/"

/** The snap tolerance in pixels. */
#define TOLERANCE_PIX			10
/** Text editor, which can be used by vrr to modify text files */
#define EXTERNAL_TEXT_EDITOR		"vim"
/** Do we use a text editor which must be run in a terminal? */
#define TEXT_EDITOR_RUN_IN_TERM		1
/** Which command should be use to open web browser for help?  */
#define WWW_BROWSER			"links"
/** Do we use a browser which must be run in a terminal? */
#define WWW_BROWSER_RUN_IN_TERM		1
/** The terminal to run browser and/or text editor, if needed */
#define FAVOURITE_TERMINAL		"rxvt"
/** The x-coordinate of the first grid vector */
#define GRID_X1				10.0
/** The y-coordinate of the first grid vector */
#define GRID_Y1				0.0
/** The x-coordinate of the second grid vector */
#define GRID_X2				0.0
/** The y-coordinate of the second grid vector */
#define GRID_Y2				10.0
/** The resolution of the horizontal ruler */
#define RULERX_RES			10.0
/** The resolution of the vertical ruler */
#define RULERY_RES			10.0
/** Whether to drag the image, not the view */
#define PANNING_DRAG_IMAGE		0

enum
{
	SNAP_HANGERS		= 1,
	SNAP_GRID		= 2,
	SNAP_LINES		= 4,
	SNAP_INTERS		= 8,
	SNAP_DEPENDENCY		= 16
};

extern GtkWidget * view_menu;		//!< The popup menu of all views.
extern GSList * view_list;		//!< The list of all existing views.
extern GtkWidget * main_window; 	//!< The one and the only main window instance.
extern struct console *console_window;	//!< Console (the one and the only one :))

/** The number of open local windows + main window. If open_windows_cnt == 0,
 program will be closed.  */
extern int open_windows_cnt;
/** Is main window open?  */
extern int main_window_open;
/** Global list of units and multipliers.  */
extern GList *global_units_glist;
extern GList *global_multiplier_glist;

enum window_type
{
	WT_VIEW,
	WT_UNDO_HISTORY,
	WT_UNIVERSE_BROWSER,
	WT_PROPERTIES,
	WT_PLUGINS,
};

#define WIN_O_MAGIC u8 type; SCM proxy

struct window { WIN_O_MAGIC; };

struct status_data
{
	GtkWidget *statusbar;
	GtkWidget *zoom_bar;
	gint context_id;
};

struct context {
	struct window * w;
	struct obj_tlo * tlo;	//! The TLO for nonmeta transactions, ancestor of the context group
	struct go_group * group; //! the group inside the context.tlo for which go selection is calculated
	struct o * o;
	int sel_cnt[GOT_MAX];	//! The array of counts of selected gos in the context.group; per each go type
	int sel_cnt_all;	//! The total count of selected gos in the context.group
	int msel_cnt_all;	//! The count of all selected meta objects
	int sel_valid;
	int msel_valid;
	uns sel_mask;		//! The bitmask of selected gos in the context.group
	uns msel_mask;		//! The bitmask of selected objects
};
extern struct context context;

void change_context( struct window * window, struct go_group * group, struct o * o );

/* ***** MAIN WINDOW *****/

void main_window_create(void);
void main_window_show(void);

void main_window_refresh_menu(void);
void main_window_refresh_toolbar(void);

/* ***** UNIVERSE BROWSER *****/

struct GtkVrrStore;
struct wnd_univ_browser{
	WIN_O_MAGIC;
	struct GtkVrrStore * tree_store;
	GtkWidget * tree_view;
	GtkWidget * menu;
	GtkWidget * window;
	struct status_data status;
	int visible;
};

extern struct wnd_univ_browser wnd_univ_browser;

void create_universe_browser(void);
void ub_refresh_menu(void);

/* ***** OPTIONS WINDOW ******/
struct wnd_settings{
	GtkWidget *window;
	struct prop_item * pi;
	struct obj_hook oh;
};

extern struct wnd_settings wnd_settings;

/* ****** ERROR WINDOW ******/
/**
 Creates modal window to show up error messages (or warning, information...).
 \param widget Parent widget (modal window will be set transient for this widget)
 \param label Message, formated string (like printf)
 \param button_label Label of the button (which closes the window)
 \param ... optional additional arguments used with label
*/
void create_err_window(GtkWidget *widget, const char *title, const char *label, const char *button_label, ...) FORMAT(3,5);

/* ***** PLUGIN WINDOW *****/
struct wnd_plugins
{
	WIN_O_MAGIC;
	GtkWidget * window;
	int visible;
	GtkListStore * plugin_list;
	GtkTreeStore * function_list;
	GtkWidget * plugin_view;
	GtkWidget * function_view;
	GtkWidget * unload_button;
};
extern struct wnd_plugins wnd_plugins;

/* ***** THE VIEW ******/

enum moving_mode
{
	MOVE_NONE,
	MOVE_FACTORY,
	MOVE_TRANSFORM,
	MOVE_TRANSFORM_2,
	MOVE_PANNING,
	MOVE_PANNING_NAV,
	MOVE_PANNING_IMG,
	MOVE_SELECT,
	MOVE_CNT
};

struct view_navigator;
struct visualisation;

struct view{
	WIN_O_MAGIC;
	struct go_group * group;	/**< the top level group */
	struct obj_tlo *tlo;		/**< the parent tlo of the top level group */
	struct obj_hook obj_hook;
	struct go_hook go_hook;
	struct visualisation vis;
	int state;
	int tf_drag_on;
	struct geom_point transform_origin;
	struct geom_point axis;
	uns transform_state;
	real tolerance;
	struct geom_transform2 grid;			//! The grid transformation matrix.

	GtkWidget *table;		/**< the main widget table */
 	GtkWidget *vruler;	/**< vertical ruler */
	GtkWidget *hruler;		/**< horizontal rulers */
	GtkWidget *toolbar;		/**< toolbar */
	GtkWidget *toolbar2;		/**< toolbar 2 */
	GtkWidget *window;		/**< pointer to local window */
	GtkAdjustment *vadj;	/**< adjustment (vertical scrollbar) */
	GtkAdjustment *hadj;	/**< adjustment (horizontal scrollbar) */
	GtkWidget *vscrollbar; /**< pointer to the vertical scrollbar */
	GtkWidget *hscrollbar; /**< pointer to the horizontal scrollbar */
	GtkWidget *image_on;
	GtkWidget *image_modified;
	struct view_navigator * navigator;	/**< the navigator (of orientation in view) window struct */

	struct status_data status;	/** status bar for messages to user */
};

struct view_navigator
{
	GtkWidget * window;
	GtkSpinButton * zoom_entry, * rot_entry;
	real zoom, rotation;
	struct visualisation vis;
	struct vcl_path * box;
	struct view * parent_view;
};

struct view * create_view(struct go_group * group);
void view_refresh_toolbar(struct view * vw);
void view_refresh_menu(void);
void view_update_grid(struct view * vw);

void transform_update( struct view *vw );

//MOVING FUNCTIONS
void view_moving_start(struct geom_point pix, struct view * vw, enum moving_mode mode, GdkCursorType cursor_type);
void view_moving_end(void);
extern enum moving_mode moving_mode;

/* ****** UNDO HISTORY WINDOW *****/

struct VrrUndoStore;
struct wnd_undo_history{
	WIN_O_MAGIC;
	GtkWidget * window;
	GtkWidget * vbox;
	GtkWidget * undo_view;
	struct VrrUndoStore * undo_model;
	GtkWidget * toolbar;
	struct obj_tlo * tlo;
	struct obj_hook obj_hook;
	struct trans_hook trans_hook;
};
extern GSList * wuh_list; //! The list of opened undo history windows.

/** Creates the undo history window.
 * \returns A pointer to the allocated window structure.
 */
struct wnd_undo_history * create_wnd_undo_history(struct obj_tlo *tlo);
void wuh_refresh_toolbar( struct wnd_undo_history * wuh );

/* ****** TEXT EDITOR *******/
struct text_edit
{
	struct go * go;
	struct obj_tlo * parent_tlo;
	struct prop_item * pi;
	int fnt_id;
	GtkWidget * window;
	GtkWidget * entry_fonts;
	GtkWidget * combo_fonts;
	GtkTextBuffer * text_buf;
	struct status_data status;
	struct obj_hook obj_hook;
	struct go_hook go_hook;
	int changing;
};

extern GSList * text_edit_list;
extern GList *fonts_list;

void fonts_list_free (void);
int fonts_get_default(void);
void text_edit_create( struct go * go );

/* ****** SCHEME CONSOLE *******/

enum console_state
{
        STATE_NORMAL,
	STATE_FSEARCH_FIRST,
	STATE_RSEARCH_FIRST,
	STATE_FSEARCH,
	STATE_RSEARCH
};

struct console_string {
        struct snode list;
        char *data;
};

struct console_search_stack
{
        enum console_state state;
	s16 search_buffer_used;
        s16 search_buffer_chars;
        GtkTextMark *mark;
        struct console_search_stack *next;
};

struct console {
	GtkWidget *window;
	GtkTextView *text;
	GtkTextBuffer *buf;
	GtkTextTag *input_tag, *output_tag,  *edit_tag;
	GtkTextMark *open_mark, *close_mark;

	int visible;
        int blockread;
        enum console_state state;

        char *search_buffer;
        s16 search_buffer_size;
        s16 search_buffer_used;
        s16 search_buffer_chars;
        struct console_search_stack *search_stack;

        SCM data;
	SCM port;
	SCM proxy;

	struct status_data status;
};

void console_create (void);
void console_init (void);

/* **** OPERATION FACTORY ******/

enum of_input_type
{
	OFIT_GO = DIST_PASS_GO,
	OFIT_HANGER = DIST_PASS_HANGER,
	OFIT_ANCHOR = DIST_PASS_ANCHOR,
	OFIT_POINT = DIST_PASS_POINT,
	OFIT_FUNC,
	OFIT_SEL,
	OFIT_TRANSFORM,
	OFIT_TF,
};

struct of_state;

enum of_go_role { OFR_DISPLAY, OFR_HANGER_OWNER, OFR_CNT };

struct of_state {
	enum of_input_type input_type;
	struct dist_pass_result value;
	char * description;		//! A sententious description of what is wanted from the user.
	void (* action_func)(struct of_state *);	//! A function to be called when leaving this state.
	void (* cleanup_func)(struct of_state *);	//! Cleanup actions made in this step.
	void (* input_func)(struct of_state *);	//! Custom function for nonstandard choosing.
	struct of_state * prev;         //! The previous state in case of a step back.
	struct of_state * next;         //! The next state in case of success.
	struct undo_gui * prev_undo_item;	//! Pointer to the last undo item before any action in this state.
	struct go * tmp_go[OFR_CNT];	//! Temporarily created gos
};

extern struct factory {
	int operation_active;
	int inside_trans;
	struct obj_tlo * tlo;
	struct go_group * group;
	struct of_state * state;
	struct of_state * start_state; //! Start of the particular operation
	string desc;
	real snap_tolerance;		//!< The current snap tolerance in coords (for the current view)
	uns snap_set;			//!< Bitmask of snap settings
	uns snap_tol_pix; 		//!< The maximum permitted snap distance in pixels
	struct hanger * hanger_cursor;
} factory;

//PUBLIC Operation Factory states
extern struct of_state ofs_select;
extern struct of_state ofs_transform_x;
extern struct of_state ofs_point;
extern struct of_state ofs_decorator_pt;
extern struct of_state ofs_point_inters;
extern struct of_state ofs_segment_1;
extern struct of_state ofs_bez_cub_1;
extern struct of_state ofs_bez_quad_1;
extern struct of_state ofs_arc_3p_1;
extern struct of_state ofs_arc_center_radius_1;
extern struct of_state ofs_circ_3p_1;
extern struct of_state ofs_circ_center_pt_1;
extern struct of_state ofs_circ_center_radius_1;
extern struct of_state ofs_ellipse_3p_smallest_1;
extern struct of_state ofs_ellipse_3p_ecc_1;
extern struct of_state ofs_ellarc_center_radii_1;
extern struct of_state ofs_ellipse_center_radii_1;
extern struct of_state ofs_ellipse_foci_pt_1;
extern struct of_state ofs_ellipse_xy1ecc_1;
extern struct of_state ofs_transform_1;
extern struct of_state ofs_rehang_1, ofs_rehang_2;
extern struct of_state ofs_text_1;
extern struct of_state ofs_tex_text_1;

//OPERATION FACTORY FUNCTIONS
void factory_init(void);
void factory_finish(void);
void factory_op_start( struct go_group * group, struct of_state * state, char * desc, int inside_trans );
void factory_op_break( void );
void factory_moving_step(struct geom_point po);
void factory_op_step( void );
int factory_merge_trans(void);
void factory_op_step_back(void);

struct go * factory_get_created_go(void);

void snap_point( struct geom_point * pos, struct dist_pass_result * dpr, struct geom_transform2 * grid, struct obj_tlo * tlo, real maxdist );
void snap_to_go( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist );
void snap_to_anchor( struct geom_point * pos, struct dist_pass_result * dpr, struct obj_tlo * tlo, real maxdist );

/* ***** MISC ******/
/** Get a new name for an untitled document.
 * \returns The pointer to a static buffer (should not be freed).
 */
char * get_new_noname(void);
/** Get the name for a new untitled page in the document.
 * \returns The pointer to a static buffer (should not be freed).
 */
char * get_new_page_noname(struct obj_doc * doc);

void refresh_gui(void);

/** Closes all opened documents and does all neccessary cleaning
 * before quitting. After that, terminates the program.
 */
void main_quit(void);

//RESOURCES
extern GdkPixbuf * icon_arrow_down, * icon_arrow_up, * icon_arrow_left, * icon_arrow_right;
extern GdkPixbuf * icon_pencil_on, * icon_modified;
extern GdkPixbuf * icon_rotation_left, * icon_rotation_right, * icon_zoom_in, * icon_zoom_out;

//FUNCTIONS
GdkPixbuf * icon_load( const char * filename );
GtkWidget * image_get_from_icon( GdkPixbuf * icon );

void statusbar_clear(struct status_data *status);
void statusbar_msg(struct status_data *status, const char * fmt, ... ) FORMAT(2,3);
void report_error(const char * s);
void report_error_f(const char * fmt, ... ) FORMAT(1,2);

void modify_selection( struct go * go, struct obj_tlo * tlo, GdkEventButton * event, struct geom_rectangle * rect, struct go_group * group );
void modify_meta_selection( struct obj * obj, GdkEventButton * event );

/** Allocates and fills memory with val.  */
float *new_float(float val);

/** Cuts white chars from beginning and end of the str.  */
char *str_trim(char *str);


/* ****** CALLBACKS ******/
struct cmd;

void on_file_new( struct cmd * cmd, void * subj );
void on_file_open( struct cmd * cmd, void * subj );
void on_svg_file_import( struct cmd * cmd, void * subj );
void on_import_ipe5( struct cmd * cmd, void * subj );
void on_file_save( struct cmd * cmd, void * subj );
void on_file_save_as( struct cmd * cmd, void * subj );
void on_file_close( struct cmd * cmd, void * subj );
void on_file_new_page( struct cmd * cmd, void * subj );
void on_file_close_page( struct cmd * cmd, void * subj );
void on_group_new_view( struct cmd * cmd, void * subj );

void on_view_fit_selected( struct cmd * cmd, void * subj );
void on_view_fit_all( struct cmd * cmd, void * subj );
void on_view_zoom_in( struct cmd * cmd, void * subj );
void on_view_zoom_out( struct cmd * cmd, void * subj );
void on_view_zoom_reset( struct cmd * cmd, void * subj );

void on_export_pdf( struct cmd * cmd, void * subj );
void on_export_ps( struct cmd * cmd, void * subj );
void on_export_eps( struct cmd * cmd, void * subj );
void on_export_svg( struct cmd * cmd, void * subj );

void on_set_vrr_options(struct cmd *cmd, void * subj );
void on_help(struct cmd *cmd, void * subj);
void on_open_www(struct cmd *cmd, void * subj);
void on_file_exit( struct cmd * cmd, void * subj );
void on_show_property_window( struct cmd * cmd, void * subj );
void on_show_console_window( struct cmd * cmd, void * subj );
void on_show_universe_browser( struct cmd * cmd, void * subj );
void on_show_main_window( struct cmd * cmd, void * subj );
void on_show_undo_history( struct cmd * cmd, void * subj );
void on_view_clipboard( struct cmd * cmd, void * subj );
void on_show_plugin_window( struct cmd * cmd, void * subj );
void on_show_settings_window( struct cmd * cmd, void * subj );

void on_copy( struct cmd * cmd, void * subj );
void on_cut( struct cmd * cmd, void * subj );
void on_paste( struct cmd * cmd, void * subj );

void on_group_selected( struct cmd * cmd, void * subj );
void on_group_ungroup( struct cmd * cmd, void * subj );

void on_group_select_all( struct cmd * cmd, void * subj );
void on_group_select_none( struct cmd * cmd, void * subj );

void on_move_up( struct cmd * cmd, void * subj );
void on_move_down( struct cmd * cmd, void * subj );
void on_move_front( struct cmd * cmd, void * subj );
void on_move_back( struct cmd * cmd, void * subj );

/** UNDO and REDO */
void on_obj_undo( struct cmd * cmd, void * subj );
void on_obj_redo( struct cmd * cmd, void * subj );
void on_tlo_undo( struct cmd * cmd, void * subj );
void on_tlo_redo( struct cmd * cmd, void * subj );
void on_hist_jump( struct cmd * cmd, void * subj );

void on_go_delete( struct cmd * cmd, void * subj );
void on_go_enclose_sel( struct cmd * cmd, void * subj );

void on_rotate90( struct cmd * cmd, void * subj );
void on_rotate180( struct cmd * cmd, void * subj );
void on_rotate270( struct cmd * cmd, void * subj );
void on_flip_vertical( struct cmd * cmd, void * subj );
void on_flip_horizontal( struct cmd * cmd, void * subj );

/** Change properties */
void on_change_obj_prop( struct cmd *cmd, void * subj );
void on_change_go_prop( struct cmd *cmd, void * subj );
void on_edit_text( struct cmd * cmd, void * subj );

void on_undo_history( struct cmd *cmd, void * subj );
void on_dump( struct cmd * cmd, void * subj );
void on_kerntest( struct cmd * cmd, void * subj );

/* GO creating callbacks. Implemented in creatego.c */
void on_random_point( struct cmd * cmd, void * subj );
void on_random_segment( struct cmd * cmd, void * subj );
void on_random_bezier_quad( struct cmd * cmd, void * subj );
void on_random_bezier_cubic( struct cmd * cmd, void * subj );

void on_step_back( struct cmd * cmd, void * subj );
void on_break_op( struct cmd * cmd, void * subj );

/* Snap modes */
void on_snap_hangers( struct cmd * cmd, void * subj );
void on_snap_grid( struct cmd * cmd, void * subj );
void on_snap_lines( struct cmd * cmd, void * subj );
void on_snap_inters( struct cmd * cmd, void * subj );
void on_snap_depend( struct cmd * cmd, void * subj );

/* Ruler */
void on_ruler_resolution( struct cmd * cmd, void * subj );

#endif

