#include "lib/lib.h"
#include <gtk/gtk.h>
#include "gui/ruler.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

enum {
  PROP_0,
  PROP_POSITION,
  PROP_OFFSET,
  PROP_RESOLUTION
};

enum {
	PRESSED,
	LAST_SIGNAL
};

static void ruler_class_init    (RulerClass  *klass);
static void ruler_init          (Ruler       *ruler);
static void ruler_realize       (GtkWidget      *widget);
static void ruler_unrealize     (GtkWidget      *widget);
static void ruler_size_allocate (GtkWidget      *widget,
				     GtkAllocation  *allocation);
static void ruler_make_pixmap   (Ruler       *ruler);
static void ruler_set_property  (GObject        *object,
				     guint            prop_id,
				     const GValue   *value,
				     GParamSpec     *pspec);
static void ruler_get_property  (GObject        *object,
				     guint           prop_id,
				     GValue         *value,
				     GParamSpec     *pspec);
gint ruler_expose        (GtkWidget      *widget,
				     GdkEventExpose *event);

static GtkWidgetClass *parent_class;
static guint ruler_signals[LAST_SIGNAL] = { 0 };
double scale[3] = {2, 2.5, 2};

GtkType ruler_get_type (void)
{
  static GtkType ruler_type = 0;

  if (!ruler_type)
    {
      static const GtkTypeInfo ruler_info =
      {
	"Ruler",
	sizeof(Ruler),
	sizeof(RulerClass),
	(GtkClassInitFunc) ruler_class_init,
	(GtkObjectInitFunc) ruler_init,
	0, 0, 0 };

      ruler_type = gtk_type_unique (gtk_widget_get_type(), &ruler_info);
    }

  return ruler_type;
}

static void ruler_class_init (RulerClass *class)
{
  GObjectClass   *gobject_class;
  GtkWidgetClass *widget_class;

  gobject_class = G_OBJECT_CLASS (class);
  widget_class = (GtkWidgetClass*) class;

  parent_class = g_type_class_peek_parent (class);

  gobject_class->set_property = ruler_set_property;
  gobject_class->get_property = ruler_get_property;

  widget_class->realize = ruler_realize;
  widget_class->unrealize = ruler_unrealize;
  widget_class->size_allocate = ruler_size_allocate;
  widget_class->expose_event = ruler_expose;

  class->draw_ticks = NULL;
  class->draw_pos = NULL;

  g_object_class_install_property (gobject_class,
                                   PROP_POSITION,
                                   g_param_spec_double ("position",
							"Position",
							"Position of mark on the ruler",
							-G_MAXDOUBLE,
							G_MAXDOUBLE,
							0.0,
							G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_OFFSET,
                                   g_param_spec_double ("offset",
							"Offset",
							"Pixel coordinate of 0",
							-G_MAXDOUBLE,
							G_MAXDOUBLE,
							0.0,
							G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class,
                                   PROP_RESOLUTION,
                                   g_param_spec_double ("resolution",
							"Resolution",
							"Pixels per unit",
							-G_MAXDOUBLE,
							G_MAXDOUBLE,
							0.0,
							G_PARAM_READWRITE));

  ruler_signals[PRESSED] = g_signal_new ("pressed",
		  G_OBJECT_CLASS_TYPE (gobject_class),
		  G_SIGNAL_RUN_FIRST,
		  G_STRUCT_OFFSET (RulerClass, pressed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
}

static void ruler_init (Ruler *ruler)
{
  gtk_widget_set_events (GTK_WIDGET (ruler), GDK_EXPOSURE_MASK |
		  GDK_BUTTON_PRESS_MASK );

  ruler->backing_store = NULL;
  ruler->non_gr_exp_gc = NULL;
  ruler->xsrc = 0;
  ruler->ysrc = 0;
  ruler->slider_size = 0;
  ruler->position = 0;
  ruler->offset = 0;
  ruler->resolution = 1.0;
}

static void ruler_set_property (GObject      *object,
 			guint         prop_id,
			const GValue *value,
			GParamSpec   *pspec UNUSED)
{
  Ruler *ruler = RULER (object);

  switch (prop_id)
    {
    case PROP_POSITION:
      ruler_set_position (ruler, g_value_get_double (value));
      break;
    case PROP_OFFSET:
      ruler_set_offset (ruler, g_value_get_double (value));
      break;
    case PROP_RESOLUTION:
      ruler_set_resolution (ruler, g_value_get_double (value));
      break;
    }
}

static void ruler_get_property (GObject      *object,
			guint         prop_id,
			GValue       *value,
			GParamSpec   *pspec)
{
  Ruler *ruler = RULER (object);

  switch (prop_id)
    {
    case PROP_POSITION:
      g_value_set_double (value, ruler->position);
      break;
    case PROP_OFFSET:
      g_value_set_double (value, ruler->offset);
      break;
    case PROP_RESOLUTION:
      g_value_set_double (value, ruler->resolution);
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

void ruler_set_position (Ruler *ruler, gdouble position)
{
  if (ruler->position != position)
    {
      ruler->position = position;
      g_object_notify (G_OBJECT (ruler), "position");
      if (GTK_WIDGET_DRAWABLE (ruler))
	gtk_widget_queue_draw (GTK_WIDGET (ruler));
    }
}

void ruler_set_offset (Ruler *ruler, gdouble offset)
{
  if (ruler->offset != offset)
    {
      ruler->offset = offset;
      g_object_notify (G_OBJECT (ruler), "offset");
      if (GTK_WIDGET_DRAWABLE (ruler))
	gtk_widget_queue_draw (GTK_WIDGET (ruler));
    }
}

void ruler_set_resolution (Ruler *ruler, gdouble resolution)
{
  if (ruler->resolution != resolution)
    {
      ruler->resolution = resolution;
      g_object_notify (G_OBJECT (ruler), "resolution");
      if (GTK_WIDGET_DRAWABLE (ruler))
	gtk_widget_queue_draw (GTK_WIDGET (ruler));
    }
}

void ruler_draw_ticks (Ruler *ruler)
{
  g_return_if_fail (IS_RULER (ruler));

  if (RULER_GET_CLASS (ruler)->draw_ticks)
    RULER_GET_CLASS (ruler)->draw_ticks (ruler);
}

void ruler_draw_pos (Ruler *ruler)
{
  g_return_if_fail (IS_RULER (ruler));

  if (RULER_GET_CLASS (ruler)->draw_pos)
     RULER_GET_CLASS (ruler)->draw_pos (ruler);
}


static void ruler_realize (GtkWidget *widget)
{
  Ruler *ruler;
  GdkWindowAttr attributes;
  gint attributes_mask;

  ruler = RULER (widget);
  GTK_WIDGET_SET_FLAGS (ruler, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_POINTER_MOTION_MASK |
			    GDK_POINTER_MOTION_HINT_MASK);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, ruler);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);

  ruler_make_pixmap (ruler);
}

static void ruler_unrealize (GtkWidget *widget)
{
  Ruler *ruler = RULER (widget);

  if (ruler->backing_store)
    g_object_unref (ruler->backing_store);
  if (ruler->non_gr_exp_gc)
    g_object_unref (ruler->non_gr_exp_gc);

  ruler->backing_store = NULL;
  ruler->non_gr_exp_gc = NULL;

  if (GTK_WIDGET_CLASS (parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}

static void ruler_size_allocate (GtkWidget     *widget,
			 GtkAllocation *allocation)
{
  Ruler *ruler = RULER (widget);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      ruler_make_pixmap (ruler);
    }
}

gint ruler_expose (GtkWidget      *widget,
		  GdkEventExpose *event UNUSED)
{
  Ruler *ruler;

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      ruler = RULER (widget);

      ruler_draw_ticks (ruler);

      gdk_draw_drawable (widget->window,
			 ruler->non_gr_exp_gc,
			 ruler->backing_store,
			 0, 0, 0, 0,
			 widget->allocation.width,
			 widget->allocation.height);

      ruler_draw_pos (ruler);
    }

  return FALSE;
}

static void ruler_make_pixmap (Ruler *ruler)
{
  GtkWidget *widget;
  gint width;
  gint height;

  widget = GTK_WIDGET (ruler);

  if (ruler->backing_store)
    {
      gdk_drawable_get_size (ruler->backing_store, &width, &height);
      if ((width == widget->allocation.width) &&
	  (height == widget->allocation.height))
	return;

      g_object_unref (ruler->backing_store);
    }

  ruler->backing_store = gdk_pixmap_new (widget->window,
					 widget->allocation.width,
					 widget->allocation.height,
					 -1);

  ruler->xsrc = 0;
  ruler->ysrc = 0;

  if (!ruler->non_gr_exp_gc)
    {
      ruler->non_gr_exp_gc = gdk_gc_new (widget->window);
      gdk_gc_set_exposures (ruler->non_gr_exp_gc, FALSE);
    }
}

void compute_and_draw_ticks (Ruler *ruler, PangoLayout *layout,
		int digit_width, int width, int height,
		draw_line_fn draw_line, print_label_fn print_label,
		int digit_offset)
{
	double lower, upper, increment;
	double dtemp, difference_step, difference_grade, grade, delta;
	double h, cur, pos;
	int i, digit_cnt, possible_cnt, max_cnt, text_width;
	char label[128];
	GtkWidget *widget = GTK_WIDGET (ruler);
	GdkGC *gc;

	gc = widget->style->fg_gc[GTK_STATE_NORMAL];

	lower = ruler_pixels_to_coords (ruler, 0);
	upper = ruler_pixels_to_coords (ruler, width);
	if (lower > upper)
	{
		h = lower;
		lower = upper;
		upper = h;
	}

	increment = fabs (ruler->resolution);
	if ((upper - lower) == 0)
		return;
	dtemp = MAX (fabs (lower), fabs (upper));
	difference_step = upper - lower;
	grade = MAX (trunc (log10 (dtemp)), 0);
	difference_grade = trunc (log10 (difference_step));
	if (difference_grade <= 0)
		digit_cnt = grade - difference_grade + 4;
	else
		digit_cnt = grade + 3;
	difference_step = pow (10, difference_grade);
#define TR(x) trunc (x / difference_step)
#define CL(x) ceil (x / difference_step)
	possible_cnt = CL (upper) - TR (lower);

	text_width = digit_cnt * digit_width + 3;
	max_cnt = width / text_width;

	i = 0;
	if (possible_cnt < max_cnt)
	{
		while ((possible_cnt * scale[i] <= max_cnt)
				&& (difference_step / scale[i] * increment > text_width))
		{
			if (i == 0)
			{
				difference_grade--;
				if (difference_grade < 0)
					digit_cnt++;
			}
			if (difference_grade < 0)
				digit_cnt++;

			difference_step /= scale[i];
			possible_cnt = CL(upper) - TR(lower);
			i = (i + 1) % 3;
		}
	}
	else
	{
		do
		{
			difference_step *= scale[i++];
			possible_cnt = CL(upper) - TR(lower);

			if (i == 3)
			{
				i = 0;
				difference_grade++;
			}
		}
		while (possible_cnt > max_cnt);
	}

	for (cur = TR(lower) * difference_step; cur < upper; cur += difference_step)
	{
		pos = draw_line (ruler, gc, cur, height);
		sprintf (label, "%.*lf", (difference_grade < 0) ? abs (difference_grade) : 0, cur);
		print_label (ruler, layout, pos, digit_width, label, digit_offset);
	}
	delta = difference_step * increment;
	difference_step /= 2.0;
	cur = trunc (lower / difference_step) * difference_step;
	h = height / 3;
	for (i = 0; i < 2 * (possible_cnt + 1); i++, cur += difference_step)
		draw_line (ruler, gc, cur, h);

	if (delta / 10.0 > MINIMUM_INCR)
	{
		difference_step /= 5.0;
		h /= 2.0;
		cur = trunc (lower / difference_step) * difference_step;
		for (i = 0; i < 10 * (possible_cnt + 1); i++, cur += difference_step)
			draw_line (ruler, gc, cur, h);
	}
}
