#undef LOCAL_DEBUG

#include "lib/lib.h"
#include <string.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include "kernel/kernel.h"
#include "gui/main.h"
#include "gui/cmdmgr.h"
#include "gui/glt_gui.h"

/** The column type identifiers. */
enum undo_column_types{
	UCT_NAME,
	UCT_STYLE,
	UCT_COUNT
};

#define VRR_TYPE_UNDO_STORE		(vrr_undo_store_get_type())
#define VRR_UNDO_STORE(obj)		(G_TYPE_CHECK_INSTANCE_CAST ((obj), VRR_TYPE_UNDO_STORE, VrrUndoStore))
#define VRR_UNDO_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST((klass), VRR_TYPE_UNDO_STORE, VrrUndoStoreClass))
#define VRR_IS_UNDO_STORE(obj)		(G_TYPE_CHECK_INSTANCE_TYPE((obj), VRR_TYPE_UNDO_STORE))
#define VRR_IS_UNDO_STORE_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass), VRR_TYPE_UNDO_STORE))
#define VRR_UNDO_STORE_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS((obj), VRR_TYPE_UNDO_STORE, VrrUndoStoreClass))

typedef struct VrrUndoStore 		VrrUndoStore;
typedef struct VrrUndoStoreClass	VrrUndoStoreClass;

const char *global_undo = "Global Undo History";

/** The structure providing the mediator between GtkTreeView and our internal kernel structures.*/
struct VrrUndoStore
{
	GObject parent;		/**< The parent GTK object. */
	gint stamp;		/**< A random naumber used to identify iters. */
	struct obj_tlo * tlo;	/**< The owner tlo of the undo history */
};

struct VrrUndoStoreClass
{
	GObjectClass parent_class;
};

GType vrr_undo_store_get_type(void);
static struct VrrUndoStore * vrr_undo_store_new( struct obj_tlo * tlo );

/** Lets the view know (by emitting a signal) that the particular undo item has changed. */
static void undo_store_item_changed( struct VrrUndoStore * store, struct undo_gui * ug );
/** Lets the view know (by emitting a signal) that a new undo item has been added. */
static void undo_store_item_new( struct VrrUndoStore * store, struct undo_gui * ug );
/** Lets the view know (by emitting a signal) that the undo item was deleted. */
static void undo_store_item_deleted( struct VrrUndoStore * store, struct undo_gui * ug, struct undo_gui * prev_item );

static void wuh_window_set_title( GtkWindow *wnd, struct obj_tlo * tlo );

/** Destroys the window and unrefs all referenced widgets.
 * \param wuh A pointer to the window structure.
 */
static void destroy_wnd_undo_history( GtkWidget * widget UNUSED, struct wnd_undo_history * wuh );

/** Handler of unlinking of either the TLO or his father doc */
static void wuh_on_source_unlink( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook );
/** Rebuild the window title when one of the names has been changed. */
static void wuh_on_names_changed( struct o * o, struct obj_hook *hook );

/* TRANSACTION HOOKS */
static void uh_item_new( struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h );
static void uh_item_delete( struct obj_tlo *tlo, struct undo_gui* ug, struct undo_gui* prev_item, struct trans_hook* h );
static void uh_item_change( struct obj_tlo *tlo, struct undo_gui* ug, struct trans_hook* h );
static void uh_undo( struct obj_tlo* tlo, struct trans_hook* h );
static void uh_redo( struct obj_tlo* tlo, struct trans_hook* h );

struct wnd_undo_history * create_wnd_undo_history(struct obj_tlo * tlo)
{
	struct wnd_undo_history * wuh = xmalloc_zero( sizeof(struct wnd_undo_history)) ;
	GtkWidget * wnd;
	GtkCellRenderer * renderer;
	GtkTreeViewColumn * column;

	wuh->type = WT_UNDO_HISTORY;
	wuh->proxy = SCM_BOOL_F;
	wuh_list = g_slist_prepend( wuh_list, (gpointer)wuh );
	wuh->tlo = tlo;
	obj_ref(&tlo->obj);

	wuh->obj_hook.data = wuh;
	wuh->obj_hook.obj_unlink = wuh_on_source_unlink;
	wuh->obj_hook.obj_prop_change = wuh_on_names_changed;
	obj_hook_add( &wuh->obj_hook );

	trans_hook_add(tlo, trans_hook_fill(&wuh->trans_hook, wuh, uh_item_new, uh_item_delete,
		uh_item_change, uh_undo, uh_redo ));

	wuh->window = wnd = gtk_window_new( GTK_WINDOW_TOPLEVEL );
	wuh_window_set_title( GTK_WINDOW( wnd ), tlo );
	gtk_window_set_default_size( GTK_WINDOW( wnd ), 200, 300 );

	wuh->vbox = gtk_vbox_new( FALSE, 0 );
	GtkWidget * sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start( GTK_BOX( wuh->vbox ), sw, TRUE, TRUE, 0 );

	wuh->undo_model = vrr_undo_store_new( tlo );
	/* create the view */
	wuh->undo_view = gtk_tree_view_new_with_model( GTK_TREE_MODEL( wuh->undo_model) );
	g_object_set( G_OBJECT( wuh->undo_view ), "headers-visible", FALSE, NULL );

	/* Set alternating row colors */
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( wuh->undo_view ), TRUE );

	/* TYPE column */
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes( "", renderer, "text", UCT_NAME,
			"strikethrough", UCT_STYLE, NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW(wuh->undo_view), column );

	/* Selection handling */
	gtk_tree_selection_set_mode( gtk_tree_view_get_selection(GTK_TREE_VIEW( wuh->undo_view )), GTK_SELECTION_BROWSE );
	gtk_container_add( GTK_CONTAINER( sw ), wuh->undo_view );

	gtk_window_add_accel_group( GTK_WINDOW( wuh->window ), get_command_accel_group());
	wuh->toolbar = NULL;
	wuh_refresh_toolbar( wuh );

	g_signal_connect( G_OBJECT( wuh->window ),
			"delete_event", G_CALLBACK( gtk_widget_destroy ), NULL );
	/* Destroying */
	g_signal_connect( G_OBJECT( wuh->window ),
			"destroy", G_CALLBACK( destroy_wnd_undo_history ), (gpointer)wuh );

	gtk_container_add( GTK_CONTAINER( wnd ), wuh->vbox );
	gtk_widget_show_all( GTK_WIDGET(wnd) );

	return wuh;
}

void wuh_refresh_toolbar( struct wnd_undo_history * wuh )
{
	ASSERT( wuh );
	if ( !wuh->toolbar || !sync_toolbar( wuh->toolbar, (wuh->tlo == tlo_universum) ? VCL_WUH_TBAR : VCL_LWUH_TBAR ) ){
		struct cmddata cd = { (struct window*)wuh, tlo_get_tlg(wuh->tlo) };
		DBG("WUH toolbar needs rebuilding!\n");
		if ( wuh->toolbar )
			gtk_container_remove( GTK_CONTAINER(wuh->vbox), wuh->toolbar );
		gtk_widget_show_all( wuh->toolbar = create_toolbar(
					(wuh->tlo == tlo_universum) ? VCL_WUH_TBAR : VCL_LWUH_TBAR, 0, cd) );
		gtk_box_pack_start( GTK_BOX(wuh->vbox), wuh->toolbar, FALSE, FALSE, 0 );
		DBG("Rebuild OK\n");
	}
}

static void wuh_window_set_title( GtkWindow *wnd, struct obj_tlo * tlo )
{
	char *buf;
	const char *objname, *fathername;

	if ( tlo != tlo_universum )
	{
		objname = obj_lookup_name (&tlo->obj);
		fathername = obj_lookup_name (tlo->obj.father);
		buf = alloca (strlen(objname) + strlen(fathername) + 20);
		sprintf( buf, "Undo History (%s of %s)", objname, fathername);
	}
	else
		buf = (char *) global_undo;

	gtk_window_set_title( wnd, buf );
}

static void wuh_on_source_unlink( struct o * o, struct o * owner UNUSED, struct o * prev UNUSED, struct obj_hook *hook )
{
	struct wnd_undo_history * wuh = (struct wnd_undo_history*)hook->data;
	if (o_is_ancestor(o, obj2o(&wuh->tlo->obj))) gtk_widget_destroy( wuh->window );
}

static void wuh_on_names_changed( struct o * o, struct obj_hook *hook )
{
	struct wnd_undo_history * wuh = (struct wnd_undo_history*)hook->data;
	if (o_is_ancestor(o, obj2o(&wuh->tlo->obj))) wuh_window_set_title( GTK_WINDOW( wuh->window ), wuh->tlo );
}

static void uh_item_new( struct obj_tlo *tlo UNUSED, struct undo_gui* ug, struct trans_hook* h )
{ undo_store_item_new( ((struct wnd_undo_history*)h->data)->undo_model, ug ); }

static void uh_item_delete( struct obj_tlo *tlo UNUSED, struct undo_gui* ug, struct undo_gui *prev_item, struct trans_hook* h )
{ undo_store_item_deleted( ((struct wnd_undo_history*)h->data)->undo_model, ug, prev_item ); }

static void uh_item_change( struct obj_tlo *tlo UNUSED, struct undo_gui* ug, struct trans_hook* h )
{ undo_store_item_changed( ((struct wnd_undo_history*)h->data)->undo_model, ug ); }

static void uh_undo( struct obj_tlo* tlo, struct trans_hook* h UNUSED )
{
	undo_store_item_changed( ((struct wnd_undo_history*)h->data)->undo_model,
		tlo->undo_present ? trans_get_next_gui(tlo, tlo->undo_present) : trans_get_first_gui(tlo) );
}

static void uh_redo( struct obj_tlo* tlo, struct trans_hook* h UNUSED )
{ undo_store_item_changed( ((struct wnd_undo_history*)h->data)->undo_model, tlo->undo_present ); }

void on_hist_jump( struct cmd * cmd UNUSED, void * subj )
{
	ASSERT(subj && ((struct window*)subj)->type == WT_UNDO_HISTORY);
	struct wnd_undo_history * wuh = (struct wnd_undo_history*)subj;

	GtkTreeSelection * sel = gtk_tree_view_get_selection( GTK_TREE_VIEW(wuh->undo_view) );
	GtkTreeIter iter;
	if (!gtk_tree_selection_get_selected( sel, NULL, &iter )) return;

	// now find out if the action is in the undo or redo part of history
	struct undo_gui * ug = (struct undo_gui*)iter.user_data;
	struct undo_gui * tmp, * present = wuh->tlo->undo_present;
	char s[TRANS_ERR_SIZE];
	int undone = 0;
	if (ug == present) return; //nothing else is needed to be done

	if ( !present ) undone = 1;
	else for (tmp = trans_get_next_gui(wuh->tlo, present); tmp; tmp = trans_get_next_gui(wuh->tlo, tmp))
			if (tmp == ug) undone = 1;
	if (undone)
		while( wuh->tlo->undo_present != ug )
		{ if (!trans_redo(wuh->tlo, s)) { report_error_f("%s\n", s); break; } }
	else
		while( wuh->tlo->undo_present != ug )
		{ if (!trans_undo(wuh->tlo, s)) { report_error_f("%s\n", s); break; } }
}

static void destroy_wnd_undo_history( GtkWidget * widget UNUSED, struct wnd_undo_history * wuh )
{
	DBG("Closing undo history for %p\n", wuh->tlo );
	obj_hook_remove( &wuh->obj_hook );
	trans_hook_remove( wuh->tlo, &wuh->trans_hook );
	obj_unref( &wuh->tlo->obj );

	wuh_list = g_slist_remove( wuh_list, (gpointer)wuh );
	window_proxy_clean ((struct window *) wuh);
	xfree(wuh);
}

/**** IMPLEMENTATION OF THE WIDGET FOR DISPLAYING UNDO HISTORY ****/

#define VALID_ITER(iter, undo_store) (iter && iter->user_data)

GType undo_column_types[] = { G_TYPE_STRING, G_TYPE_BOOLEAN };

static GObjectClass *parent_class = NULL;
/* initializations and finalization */
static void	vrr_undo_store_init( VrrUndoStore *undo_store );
static void	vrr_undo_store_class_init( VrrUndoStoreClass *undo_store_class );
static void	vrr_undo_store_tree_model_init( GtkTreeModelIface *iface );
static void	vrr_undo_store_finalize( GObject *object );

/* implementation of GtkTreeModelIface */
static GtkTreeModelFlags vrr_undo_store_get_flags( GtkTreeModel *tree_model );
static gint vrr_undo_store_get_n_columns( GtkTreeModel *tree_model);
static GType vrr_undo_store_get_column_type( GtkTreeModel *tree_model, gint index);
static gboolean	vrr_undo_store_get_iter( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath  *path);
static GtkTreePath *vrr_undo_store_get_path( GtkTreeModel *tree_model, GtkTreeIter *iter);
static void vrr_undo_store_get_value( GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value);
static gboolean vrr_undo_store_iter_next( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean vrr_undo_store_iter_children( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter  *parent);
static gboolean vrr_undo_store_iter_has_child( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gint vrr_undo_store_iter_n_children( GtkTreeModel *tree_model, GtkTreeIter *iter);
static gboolean vrr_undo_store_iter_nth_child( GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n);
static gboolean	vrr_undo_store_iter_parent( GtkTreeModel *tree_model, GtkTreeIter  *iter, GtkTreeIter  *child);


GType vrr_undo_store_get_type(void)
{
	static GType undo_store_type = 0;

	if (!undo_store_type){
		static const GTypeInfo undo_store_info =
		{ sizeof (VrrUndoStoreClass), NULL, NULL, (GClassInitFunc) vrr_undo_store_class_init,
		NULL, NULL, sizeof (VrrUndoStore), 0, (GInstanceInitFunc) vrr_undo_store_init, NULL };

		static const GInterfaceInfo tree_model_info =
			{ (GInterfaceInitFunc) vrr_undo_store_tree_model_init, NULL, NULL };

		undo_store_type = g_type_register_static (G_TYPE_OBJECT, "VrrUndoStore", &undo_store_info, 0);
		g_type_add_interface_static (undo_store_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
	}
	return undo_store_type;
}

static void vrr_undo_store_class_init (VrrUndoStoreClass *class)
{
	GObjectClass *object_class;
	parent_class = g_type_class_peek_parent (class);
	object_class = (GObjectClass *) class;
	object_class->finalize = vrr_undo_store_finalize;
}

static void vrr_undo_store_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = vrr_undo_store_get_flags;
	iface->get_n_columns = vrr_undo_store_get_n_columns;
	iface->get_column_type = vrr_undo_store_get_column_type;
	iface->get_iter = vrr_undo_store_get_iter;
	iface->get_path = vrr_undo_store_get_path;
	iface->get_value = vrr_undo_store_get_value;
	iface->iter_next = vrr_undo_store_iter_next;
	iface->iter_children = vrr_undo_store_iter_children;
	iface->iter_has_child = vrr_undo_store_iter_has_child;
	iface->iter_n_children = vrr_undo_store_iter_n_children;
	iface->iter_nth_child = vrr_undo_store_iter_nth_child;
	iface->iter_parent = vrr_undo_store_iter_parent;
}

static void vrr_undo_store_init (VrrUndoStore *undo_store)
{
	do undo_store->stamp = g_random_int ();
	while (undo_store->stamp == 0);
}

VrrUndoStore * vrr_undo_store_new( struct obj_tlo * tlo )
{
	VrrUndoStore * vus = VRR_UNDO_STORE (g_object_new (VRR_TYPE_UNDO_STORE, NULL));
	vus->tlo = tlo;
	return vus;
}

/***********************************************************************/

static void vrr_undo_store_finalize (GObject *object)
{ (* parent_class->finalize) (object); }

static GtkTreeModelFlags vrr_undo_store_get_flags (GtkTreeModel *tree_model)
{
	g_return_val_if_fail (VRR_IS_UNDO_STORE (tree_model), 0);
	return GTK_TREE_MODEL_LIST_ONLY; /* it means that the structure is not branching */
}

static gint vrr_undo_store_get_n_columns (GtkTreeModel *tree_model)
{
	g_return_val_if_fail (VRR_IS_UNDO_STORE (tree_model), 0);
	return UCT_COUNT;
}

static GType vrr_undo_store_get_column_type (GtkTreeModel *tree_model, gint index)
{
	g_return_val_if_fail (VRR_IS_UNDO_STORE (tree_model), G_TYPE_INVALID);
	g_return_val_if_fail (index <= UCT_COUNT && index >= 0, G_TYPE_INVALID);
	return undo_column_types[index];
}

static gboolean vrr_undo_store_get_iter(GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path)
{
	VrrUndoStore *undo_store = (VrrUndoStore *) tree_model;
	gint *indices;
	int i, depth;
	struct undo_gui *ug;
	g_return_val_if_fail (VRR_IS_UNDO_STORE (undo_store), FALSE);

	indices = gtk_tree_path_get_indices (path);
	depth = gtk_tree_path_get_depth(path);
	g_return_val_if_fail (depth > 0, FALSE);

	//DBG("Path is: %s\n", gtk_tree_path_to_string(path) );

	for( ug = trans_get_first_gui(undo_store->tlo), i=0; i < indices[0] && ug;
		ug = trans_get_next_gui( undo_store->tlo, ug )) i++;

	if ( !ug ) return FALSE;
	else {
		iter->stamp = undo_store->stamp;
		iter->user_data = (gpointer)ug;
	}
	return TRUE;
}

static GtkTreePath * vrr_undo_store_get_path(GtkTreeModel *tree_model, GtkTreeIter  *iter)
{
	GtkTreePath *path;
	VrrUndoStore * undo_store = VRR_UNDO_STORE(tree_model);
	struct undo_gui * ug;

	g_return_val_if_fail (VRR_IS_UNDO_STORE (tree_model), NULL);
	g_return_val_if_fail (iter, NULL);
	g_return_val_if_fail (iter->user_data, NULL);
	g_return_val_if_fail (iter->stamp == VRR_UNDO_STORE (tree_model)->stamp, NULL);

	path = gtk_tree_path_new();
	ug = (struct undo_gui*)(iter->user_data);
	gtk_tree_path_prepend_index( path, trans_get_gui_rank( undo_store->tlo, ug ));
	return path;
}

/* Get the value at given position */
static void vrr_undo_store_get_value(GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value)
{
	struct undo_gui * ug;
	g_return_if_fail (VRR_IS_UNDO_STORE (tree_model));
	VrrUndoStore * vus = VRR_UNDO_STORE(tree_model);
	g_return_if_fail (iter);
	g_return_if_fail (iter->stamp == VRR_UNDO_STORE (tree_model)->stamp);
	g_return_if_fail (column < UCT_COUNT );

	ug = (struct undo_gui*)(iter->user_data);
	g_value_init( value, undo_column_types[column] );
	switch ( column ){
	case UCT_NAME:
		if ( ug && ug->name_operation ) g_value_set_string( value, ug->name_operation->text );
		else g_value_reset( value );
		break;
	case UCT_STYLE:
		if ( vus->tlo->undo_present ){
			//FIXME: this is terribly ineffective!
			struct undo_gui * ug2 = trans_get_next_gui( vus->tlo, vus->tlo->undo_present);
			for ( ; ug2 && ug2 != ug; ug2 = trans_get_next_gui(vus->tlo, ug2) );
			g_value_set_boolean( value, ug2 != NULL );
		} else g_value_set_boolean( value, TRUE ); //no done actions, so it must be redone
		break;
	default: BUG();
	}
}

/* Move the iter to next node */
static gboolean vrr_undo_store_iter_next (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	g_return_val_if_fail (iter, FALSE);
	g_return_val_if_fail (iter->user_data, FALSE);
	g_return_val_if_fail (iter->stamp == VRR_UNDO_STORE (tree_model)->stamp, FALSE);

	struct undo_gui * ug = (struct undo_gui*)iter->user_data;
	ug = trans_get_next_gui(VRR_UNDO_STORE(tree_model)->tlo, ug);
	iter->user_data = (gpointer)ug;

	return ug != NULL;
}

/* Move the iter to child */
static gboolean vrr_undo_store_iter_children (GtkTreeModel *tree_model UNUSED, GtkTreeIter *iter UNUSED, GtkTreeIter *parent UNUSED)
{ return FALSE; }

/* If the iter has a child */
static gboolean vrr_undo_store_iter_has_child (GtkTreeModel *tree_model UNUSED, GtkTreeIter  *iter UNUSED)
{ return FALSE; }

/* Probably the number of childen of given node */
static gint vrr_undo_store_iter_n_children (GtkTreeModel *tree_model UNUSED, GtkTreeIter  *iter UNUSED)
{ return 0; }

/* Move to the nth child */
static gboolean vrr_undo_store_iter_nth_child (GtkTreeModel *tree_model UNUSED,
		GtkTreeIter *iter UNUSED, GtkTreeIter *parent UNUSED, gint n UNUSED)
{ return FALSE; }

/* Move to parent */
static gboolean vrr_undo_store_iter_parent (GtkTreeModel *tree_model UNUSED, GtkTreeIter *iter UNUSED, GtkTreeIter *child UNUSED )
{ return FALSE; }

static void undo_store_item_changed( VrrUndoStore * store, struct undo_gui * ug )
{
	GtkTreeIter iter;
	DBG("undo_store_item_changed\n");
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)ug;
	g_signal_emit_by_name( (gpointer)store, "row-changed",  NULL, &iter, NULL );
}

static void undo_store_item_new( VrrUndoStore * store, struct undo_gui * ug )
{
	GtkTreeIter iter;
	DBG("undo_store_item_new\n");
	iter.stamp = store->stamp;
	iter.user_data = (gpointer)ug;
	g_signal_emit_by_name( (gpointer)store, "row-inserted",  NULL, &iter, NULL );
}

static void undo_store_item_deleted( VrrUndoStore * store, struct undo_gui * ug UNUSED, struct undo_gui * prev_item )
{
	DBG("undo_store_item_deleted\n");
	GtkTreePath * path = gtk_tree_path_new();
	gtk_tree_path_prepend_index( path, prev_item ? (trans_get_gui_rank(store->tlo, prev_item) + 1) : 0 );
	g_signal_emit_by_name( (gpointer)store, "row-deleted", path, NULL );
	gtk_tree_path_free( path );
}
