#include "kernel/internal.h"
#include "lib/hashfunc.h"

#ifdef DEBUG_EXPAND
#define DBG_EXPAND(...) msg(__VA_ARGS__)
#else
#define DBG_EXPAND(...) do { } while(0)
#endif

#define HASH_NODE struct expand_rtree
#define HASH_PREFIX(x) expand_sht_ ## x
#define HASH_KEY_ATOMIC go
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_INSERT
#define HASH_WANT_REMOVE
#define HASH_GIVE_ALLOC
static inline void expand_sht_free(struct expand_sht_table *table UNUSED, void *p UNUSED) { }
#define HASH_GIVE_HASHFN
static inline uns expand_sht_hash(struct expand_sht_table *table UNUSED, struct go *go) { return hash_int((intptr_t)go); }
#define HASH_CONSERVE_SPACE
#define HASH_ATOMIC_TYPE struct go*
#define HASH_TABLE_DYNAMIC
#include "lib/sht.h"

struct expand_path
{
	struct expand_sht_bucket bucket;
	struct geom_path path;
};

static inline struct expand_path* rtree2path(struct expand_rtree *b)
	{ ASSERT(b->usage != 0); return SKIP_BACK(struct expand_path, bucket.n, b); }
static inline struct expand_sht_bucket* rtree2bucket(struct expand_rtree *b)
	{ return SKIP_BACK(struct expand_sht_bucket, n, b); }

/*
 * Compute the box enclosing all visible parts of GO, hangers and anchors.
 */

#define EXPAND_BBOX(_id) case GO_TYPE(_id): ret = GLUEE(GO_NAME(_id), _bbox) ((GO_STRUCT(_id)*)go, bbox); break;

static int expand_get_bbox(struct go *go, struct geom_rectangle *bbox)
{
	int volatile ret = 0;
	DBG_EXPAND("expand_get_bbox: go %s at %p\n", go_lookup_name(go), go);
	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS(buf)
		switch(go->type)
		{
			EXPAND_BBOX(POINT)
			EXPAND_BBOX(SEGMENT)
			EXPAND_BBOX(BEZIER)
			EXPAND_BBOX(PARPOINT)
			EXPAND_BBOX(INTERSECTION)
			EXPAND_BBOX(TEXT)
			EXPAND_BBOX(TEX)

			default:
				die("Function expand_get_bbox hasn't label for the go type %d - %s.\n", go->type, string_go_type[go->type]->text);
		}
	TRANS_FAILED
		DBG_EXPAND("Cannot compute bbox of go %s at %p: %s\n", go_lookup_name(go), go, buf);
	TRANS_END
	return ret;
}

#define EXPAND_PATH(_id) case GO_TYPE(_id): ret = GLUEE(GO_NAME(_id), _path_and_bbox) ((GO_STRUCT(_id)*)go, path, bbox); break;

static int expand_get_path_and_bbox(struct go *go, struct geom_path *path, struct geom_rectangle *bbox)
{
	int volatile ret = 0;
	DBG_EXPAND("expand_get_path_and_bbox: go %s at %p\n", go_lookup_name(go), go);
	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS(buf)
		GEOM_ITEM_CREATE(path, path);
		switch(go->type)
		{
			EXPAND_PATH(ELARC)
			EXPAND_PATH(DECPOINT)
			EXPAND_PATH(DIRECTION)

			default:
			  die("Function expand_get_path_and_bbox hasn't label for the go type %d - %s.\n", go->type, string_go_type[go->type]->text);
		}
	TRANS_FAILED
		DBG_EXPAND("Cannot compute expansion of go %s at %p: %s\n", go_lookup_name(go), go, buf);
	TRANS_END
	return ret;
}

const struct geom_rectangle* expand_insert(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo);
	struct expand_sht_bucket *b;
	struct geom_rectangle bbox;
	int result;
	if(!tlo->expand || go->type == GOT_GROUP)
		return NULL;
	DBG_EXPAND("expand_insert: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
	ASSERT(!expand_sht_find(tlo->expand, go));
	if(go_get_info(go)->expand == ET_PATH)
	{
		struct expand_path *e = xmalloc_zero(sizeof(struct expand_path));
		result = expand_get_path_and_bbox(go, &e->path, &bbox);
		b = &e->bucket;
		b->n.usage = -1; // FIXME: time stamp
	}
	else
	{
		b = xmalloc_zero(sizeof(struct expand_sht_bucket));
		result = expand_get_bbox(go, &bbox);
		b->n.usage = 0;
	}
	if(!result)
	{
		xfree(b);
		return NULL;
	}
	geom_rtree_insert(&tlo->rtree, &b->n.rtree, &bbox);
	expand_sht_insert(tlo->expand, go, b);
	return &b->n.rtree.bbox;
}

int expand_remove(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo || go_get_tlo(go) == tlo_zombie);
	struct expand_rtree *b;
	if(!tlo->expand || !(b = expand_sht_find(tlo->expand, go)))
		return 0;
	DBG_EXPAND("expand_remove: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
//	if(go->rtree.parent)  // When transaction fails before inserting this go into rtree then parent is NULL.
	geom_rtree_delete(&tlo->rtree, &b->rtree);
	expand_sht_remove(tlo->expand, b);
	if(b->usage <= 0)
		xfree(rtree2bucket(b));
	return 1;
}

const struct geom_rectangle* expand_update(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo);
	struct expand_rtree *b;
	struct geom_rectangle bbox;
	struct expand_path *e;
	if(!tlo->expand)
		return NULL;
	if(!(b = expand_sht_find(tlo->expand, go)))
		return expand_insert(tlo, go);
	DBG_EXPAND("expand_update: tlo %s at %p, go %s at %p\n", obj_lookup_name(&tlo->obj), tlo, go_lookup_name(go), go);
	if(b->usage > 0)
	{
		geom_rtree_delete(&tlo->rtree, &b->rtree);
		expand_sht_remove(tlo->expand, b);
		struct expand_path *e = xmalloc_zero(sizeof(struct expand_path));
		if(!expand_get_path_and_bbox(go, &e->path, &bbox))
		{
			xfree(e);
			return NULL;
		}
		b = &e->bucket.n;
		b->usage = -1; // FIXME: time stamp
		geom_rtree_insert(&tlo->rtree, &b->rtree, &bbox);
		expand_sht_insert(tlo->expand, go, &e->bucket);
	}
	else if(b->usage == 0)
	{
		if(!expand_get_bbox(go, &bbox))
		{
			geom_rtree_delete(&tlo->rtree, &b->rtree);
			expand_sht_remove(tlo->expand, b);
			xfree(rtree2bucket(b));
			return NULL;
		}
		geom_rtree_update(&tlo->rtree, &b->rtree, 0, &bbox);
	}
	else
	{
		e = rtree2path(b);
		if(!expand_get_path_and_bbox(go, &e->path, &bbox))
		{
			geom_rtree_delete(&tlo->rtree, &b->rtree);
			expand_sht_remove(tlo->expand, b);
			xfree(e);
			return NULL;
		}
		geom_rtree_update(&tlo->rtree, &b->rtree, 0, &bbox);
		b->usage = -1; // FIXME: time stamp
	}
	return &b->rtree.bbox;
}

void expand_ref(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	tlo->expand_ref_count++;
	DBG_EXPAND("  Increase expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 1)
		return;

	ASSERT(!tlo->expand);
	struct go *go;
	struct go_group *tlg = &tlo->tlg->group, *owner;
	geom_rtree_init(&tlo->rtree);
	tlo->expand = xmalloc(sizeof(struct expand_sht_table));
	expand_sht_init(tlo->expand);

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_ANONYMOUS(buf)
		WALK_GROUP_SMART(tlg, 1, go, owner, , expand_insert(tlo, go), expand_insert(tlo, &owner->go));
	TRANS_FAILED
		die(buf);
	TRANS_END
}

void expand_unref(struct obj_tlo *tlo)
{
	ASSERT_TLO(tlo);
	ASSERT(tlo->expand_ref_count);
	tlo->expand_ref_count--;
	DBG_EXPAND("  Decrease expand reference: tlo %s at %p, new ref_count %d\n", obj_lookup_name(&tlo->obj), tlo, tlo->expand_ref_count);
	if(tlo->expand_ref_count > 0)
		return;

	HASH_FOR_ALL_DYNAMIC(expand_sht, tlo->expand, b)
		DANGEROUS(b->usage <= 0, "Somebody has %u reference to expansion %p of go %s at %p.\n", b->usage, b, go_lookup_name(b->go), b->go);
//		msg("free: %p, go %s at %p\n", b, go_lookup_name(b->go), b->go);
//		expand_sht_remove(tlo->expand, b);
		if(b->usage)
			geom_path_destroy(&rtree2path(b)->path);
//		xfree(rtree2bucket(b));
	HASH_END_FOR;

	geom_rtree_clear(&tlo->rtree);
	expand_sht_cleanup(tlo->expand);
	xfree(tlo->expand);
	tlo->expand = NULL;
}

struct geom_path* go_path_expand_ref(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo && tlo->expand);
	struct expand_rtree *b = expand_sht_find(tlo->expand, go);
	if(!b || b->usage == 0)
		return NULL;
	if(b->usage > 0)
		b->usage++;
	else if(b->usage < 0)
		return NULL; // FIXME: time stamp
	return &rtree2path(b)->path;
}

void go_path_expand_unref(struct obj_tlo *tlo, struct geom_path *path)
{
	struct expand_path *e = SKIP_BACK(struct expand_path, path, path);
	ASSERT(tlo->expand && e->bucket.n.usage > 0);
	if(e->bucket.n.usage > 1)
		e->bucket.n.usage--;
	else
	{
		if(expand_sht_find(tlo->expand, e->bucket.n.go) == &e->bucket.n)
			xfree(e);
		else
			e->bucket.n.usage = -1; // FIXME: time stamp
	}
}


/**
 * Returns expansion box of given GO.
 *
 * @ai tlo  You must have expand reference of this TLO to use this function.
 * @ai go   GO which belong into TLO.
 *
 * @rv nonzero Box enclosing all visible parts of GO, hangers and anchors. Read only.
 * @rv NULL    Numeric error.
 */

const struct geom_rectangle* go_get_expand_bbox(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo && tlo->expand);
	struct expand_rtree *b = expand_sht_find(tlo->expand, go);
//	o_dump(go2o(go)); msg("\t"); geom_rectangle_msg(b ? &b->rtree.bbox : NULL); msg("\n");
	return b ? &b->rtree.bbox : NULL;
}

/**
 * Returns expansion box of given TLO.
 *
 * @ai tlo  You must have expand reference of this TLO to use this function.
 *
 * @rv Box enclosing all visible parts in TLO, hangers and anchors. Read only.
 */

const struct geom_rectangle* tlo_get_expand_bbox(struct obj_tlo *tlo)
{
	ASSERT(tlo->expand);
	return &tlo->rtree.root->bbox;
}

/**
 * Returns path expansion of given GO.
 *
 * @ai tlo  You must have expand reference of this TLO to use this function.
 * @ai go   GO which belong into TLO.
 *
 * @rv nonzero Path expansion of GO. Read only.
 * @rv NULL    Numeric error.
 */

const struct geom_path* go_get_expand_path(struct obj_tlo *tlo, struct go *go)
{
	ASSERT(go_get_tlo(go) == tlo && tlo->expand);
	struct expand_rtree *b = expand_sht_find(tlo->expand, go);
	// FIXME: time stamp
	if(!b || b->usage == 0)
		return NULL;
	return &rtree2path(b)->path;
}

