#ifndef _KERNEL_FUNCTIONS_
#define _KERNEL_FUNCTIONS_

/** @file
 *
 * 	Kernels' functions
 *
 */


/*
 *  Map of this header file.
 *
 * x OBJ and GO handling
 *   - Verify macros and type-cast functions
 *   - Reference
 *   - Creating functions
 *   - Linking
 *   - Selectings
 *   - Transformation
 *
 * x Movement functions and walk macros
 *   - Movement in OBJ hierarchy
 *   - Movement in GROUP
 *   - Movement in O (OBJ and GO)
 *   - Movement in hanger/anchor dependence
 *   - Movement in selections
 *
 * x OBJ and GO hooks handling
 *
 * x Clipboard
 *
 * x Dumping functions
 *
 * x Save and load function
 *
 * x Miscellaneous
 *   - Kernel initialization and cleanup
 *
 */



/***********************************
 *                                 *
 *      OBJ handling functions     *
 *                                 *
 ***********************************/

/*
 * Verify macros and type-cast functions
 */

#define ASSERT_O(_o) ASSERT((_o) && (((struct o*)(_o))->kind == T_OBJ || ((struct o*)(_o))->kind == T_GO))
#define ASSERT_OBJ(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->subtype == 0)
#define ASSERT_DOC(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->type == OT_DOCUMENT)
#define ASSERT_TLO(_o) ASSERT((_o) != NULL && ((struct o*)(_o))->kind == T_OBJ && ((struct o*)(_o))->type == OT_TLO)
#define ASSERT_GO(_o) ASSERT((_o) && ((struct o*)(_o))->kind == T_GO)
#define ASSERT_GROUP(_group) ASSERT((_group) && ((struct o*)(_group))->kind == T_GO && ((struct o*)(_group))->type == GOT_GROUP)

#ifdef DEBUG_KERNEL_SLOW
static inline struct go* assert_go_in_tlo(struct go *go, struct obj_tlo *tlo)
	{ ASSERT_GO(go); ASSERT_TLO(tlo); ASSERT(go_get_tlo(go) == tlo); return go; }
#else
static inline struct go* assert_go_in_tlo(struct go *go, struct obj_tlo *tlo UNUSED)
	{ return go; }
#endif

#define ASSERT_GO_IN_TLO(_go, _tlo) ASSERT(go_get_tlo(_go) == _tlo)
#define ASSERT_GO_IN_TLO_EXPAND(_go, _tlo) ASSERT(go_get_tlo(_go) == _tlo && _tlo->expand)

static inline struct obj_tlo* obj2tlo(struct obj* obj)
	{ ASSERT_TLO(obj); return (struct obj_tlo*)obj; }
static inline struct obj_doc* obj2doc(struct obj* obj)
	{ ASSERT_DOC(obj); return (struct obj_doc*)obj; }

static inline struct obj* o2obj(struct o* o)
	{ ASSERT_OBJ(o); return (struct obj*)o; }
static inline struct o* obj2o(struct obj* o)
	{ ASSERT_OBJ(o); return (struct o*)o; }

static inline struct go* o2go(struct o* o)
	{ ASSERT_GO(o); return (struct go*)o; }
static inline struct o* go2o(struct go* g)
	{ ASSERT_GO(g); return (struct o*)g; }

static inline struct go_group* go2group(struct go* go)
	{ ASSERT_GROUP(go); return (struct go_group*)go; }

static inline struct o* void2o(void *o)
	{ ASSERT_O(o); return o; }
static inline struct obj* void2obj(void *o)
	{ ASSERT_OBJ(o); return o; }
static inline struct go* void2go(void *o)
	{ ASSERT_GO(o); return o; }

static inline struct geom_item* go2geom(struct go *go)
	{ ASSERT_GO(go); return SKIP_OFFSET(struct geom_item, go, OFFSETOF(struct go_group, geom)); }
static inline struct go* geom2go(struct geom_item *geom)
	{ struct go *go = SKIP_OFFSET(struct go, geom, -OFFSETOF(struct go_group, geom)); ASSERT_GO(go); return go; }



/*
 * Reference functions
 */

/* Macro DEBUG_REF_O_ARGS is used for debugging. You can ignore it. */
struct obj* obj_ref(struct obj *obj DEBUG_REF_O_ARGS);
void obj_unref(struct obj *obj DEBUG_REF_O_ARGS);

struct go* go_ref(struct go *go DEBUG_REF_O_ARGS);
void go_unref(struct go *go DEBUG_REF_O_ARGS);

struct o* o_ref(struct o *o DEBUG_REF_O_ARGS);
void o_unref(struct o *o DEBUG_REF_O_ARGS);

struct hanger* hanger_ref(struct hanger *hanger DEBUG_REF_O_ARGS);
void hanger_unref(struct hanger *hanger DEBUG_REF_O_ARGS);



/*
 * Creating functions
 */

SCHEME SO_obj_doc doc_new(void) SMETATRANS ;
SCHEME SO_obj_doc doc_new_named(SSTRINGC name) SMETATRANS ;

SCHEME SO_obj_tlo tlo_new(void) SMETATRANS ;
SCHEME SO_obj_tlo tlo_new_named(SSTRINGC name) SMETATRANS ;


struct hanger* hanger_new_mc(struct geom_point *gp);
static inline struct hanger* hanger_new_mc_xy(real x, real y)
	{ struct geom_point gp = (struct geom_point) { x, y }; return hanger_new_mc(&gp); }
SCHEME SHANGER hanger_new_mc_curve(void);


/*
 * Linking functions
 */

SCHEME SO_obj obj_link(SO_obj obj, SO_obj father) SMETATRANS ;
SCHEME SO_obj obj_unlink(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unlink_selected(void) SMETATRANS ;

SCHEME static inline SO_obj_doc doc_new_named_link(SSTRINGC name) SMETATRANS
	{ return obj2doc(obj_link(&doc_new_named(name)->obj, obj_universum)); }
SCHEME static inline SO_obj_tlo tlo_new_named_link(SO_obj_doc doc, SSTRINGC name) SMETATRANS
	{ return obj2tlo(obj_link(&tlo_new_named(name)->obj, &doc->obj)); }

/* If after if NULL then go is inserted in the begin of group. */
SCHEME SO_go go_link(SO_go go, SO_go_group group, SO_go after) STRANS ;
SCHEME SO_go go_unlink(SO_go go) STRANS ;
SCHEME SO_go go_relink(SO_go go, SO_go_group group, SO_go after) STRANS ;

SCHEME SVOID group_relink_selected_go(SO_go_group source, SO_go_group target, SO_go after) STRANS ;
SCHEME SVOID group_relink_all_go(SO_go_group source, SO_go_group target, SO_go after) STRANS ;

SCHEME SVOID tlo_unlink_all_go(void) STRANS ;
SCHEME SVOID group_unlink_selected_go(SO_go_group group) STRANS ;
SCHEME SVOID go_remove_rehang(SO_go go) STRANS ;
SCHEME SVOID group_remove_rehang_selected_go(SO_go_group group); STRANS
uns go_remove_carefully(struct go *go);

SCHEME SVOID anchor_rehang(SANCHOR a, SHANGER h); STRANS
int hanger_mc_move(struct hanger* h, struct geom_point* gp);


/*
 * Selecting functions
 */

SCHEME SVOID obj_select(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unselect(SO_obj obj) SMETATRANS ;
SCHEME SVOID obj_unselect_all(void) SMETATRANS ;

SCHEME SO_go go_select(SO_go g) STRANS ;
SCHEME SO_go go_unselect(SO_go g) STRANS ;
SCHEME SVOID tlo_unselect_all_go(void) STRANS ;
SCHEME SVOID group_select_all_go(SO_go_group group) STRANS ;
SCHEME SVOID group_unselect_all_go(SO_go_group group) STRANS ;

void o_select(struct o* o);
void o_unselect(struct o* o);
void o_unselect_all(void);

SCHEME static inline SUNS obj_is_selected(SO_obj o)
	{ return (o->flags & OF_SELECTED); }
SCHEME static inline SUNS go_is_selected(SO_go o)
	{ return (o->flags & GOF_SELECTED); }
SCHEME static inline SUNS o_is_selected(SO_o o)
	{ return o->kind == T_OBJ ? obj_is_selected(o2obj(o)) : go_is_selected(o2go(o)); }



/**********************************************
 *                                            *
 *    Movement functions and walk macros      *
 *                                            *
 **********************************************/

/*
 * Movement in OBJ hierarchy
 */

/* Basic OBJ movement */
SCHEME static inline SO_obj obj_get_first_son(SO_obj obj)
	{ cnode *n = clist_head(&obj->lson); return n ? SKIP_BACK(struct obj, nson, n) : NULL; }
SCHEME static inline SO_obj obj_get_next_son(SO_obj obj)
	{ cnode *n = clist_next(&obj->father->lson, &obj->nson); return n ? SKIP_BACK(struct obj, nson, n) : NULL; }
SCHEME static inline SO_obj obj_get_last_son(SO_obj obj)
	{ cnode *n = clist_tail(&obj->lson); return n ? SKIP_BACK(struct obj, nson, n) : NULL; }
SCHEME static inline SO_obj obj_get_prev_son(SO_obj obj)
	{ cnode *n = clist_prev(&obj->father->lson, &obj->nson); return n ?  SKIP_BACK(struct obj, nson, n) : NULL; }
#define WALK_OBJ_SON(_father, _walk) for(_walk = obj_get_first_son(_father); _walk; _walk = obj_get_next_son(_walk))

/* Type-specified OBJ movement */
#define WALK_UNIVERSUM_DOC(_walk) for(cnode *_n = obj_universum->lson.head.next; _n != &obj_universum->lson.head && (_walk = obj2doc(SKIP_BACK(struct obj, nson, _n)), 1); _n = _n->next)
#define WALK_DOC_TLO(_doc, _walk) for(_walk = (struct obj_tlo*)(obj_get_first_son(&_doc->obj)); _walk; _walk = (struct obj_tlo*)(obj_get_next_son(&_walk->obj)))

/* Recursive OBJ movement */
SCHEME SO_obj obj_recursive_next(const SO_obj root, SO_obj walk);
#define WALK_OBJ_RECURSIVE(_root, _walk) for(_walk = _root; _walk; _walk = obj_recursive_next(_root, _walk))

/* Miscellaneous OBJ movement */
SCHEME static inline SO_obj obj_get_nth_first_son(SO_obj obj, SUNS n)
	{ for(obj = obj_get_first_son(obj); n && obj; obj = obj_get_next_son(obj), n--); return obj; }
SCHEME static inline SO_obj obj_get_nth_last_son(SO_obj obj, SUNS n)
	{ for(obj = obj_get_last_son(obj); n && obj; obj = obj_get_prev_son(obj), n--); return obj; }
SCHEME static inline SUNS obj_get_rank_son(SO_obj obj)
	{ uns n = -1; for(; obj; obj = obj_get_prev_son(obj)) n++; return n; }
static inline struct obj* obj_get_root(struct obj *obj)
	{ while(obj->father) obj = obj->father; return obj; }
SCHEME static inline SO_go_group tlo_get_tlg(SO_obj_tlo tlo)
	{ return &tlo->tlg->group; }

static inline int obj_is_ancestor(struct obj *ancestor, struct obj *descendant)
	{ for(; descendant; descendant = descendant->father) if(descendant == ancestor) return 1; return 0; }

/*
 * Movement in GROUP
 */

/* Basic GROUP movement */
#ifdef STUPID_GROUP
static inline struct go* group_get_first_go(struct go_group *group)
	{ cnode *n = clist_head(&group->lmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
static inline struct go* group_get_next_go(struct go_group *group, struct go *go)
	{ cnode *n = clist_next(&group->lmember, &go->nmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
static inline struct go* group_get_last_go(struct go_group *group)
	{ cnode *n = clist_tail(&group->lmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
static inline struct go* group_get_prev_go(struct go_group *group, struct go *go)
	{ cnode *n = clist_prev(&group->lmember, &go->nmember); return n ? SKIP_BACK(struct go, nmember, n) : NULL; }
static inline struct go_group* go_get_group(struct go *go)
	{ return go->owner; }
static inline struct go_tlg* go_get_root(struct go* go)
	{ while(go->owner) go = &go->owner->go; return (struct go_tlg*)go; }
#else
static inline struct go* group_get_first_go(struct go_group *group UNUSED) { return NULL; }
static inline struct go* group_get_next_go(struct go_group *group UNUSED, struct go *go UNUSED) { return NULL; }
static inline struct go* group_get_last_go(struct go_group *group UNUSED) { return NULL; }
static inline struct go* group_get_prev_go(struct go_group *group UNUSED, struct go *go UNUSED) { return NULL; }
static inline struct go_group* go_get_group(struct go *go UNUSED) { return NULL; }
static inline struct go_tlg* go_get_root(struct go* go UNUSED) { return NULL; }
#endif
#define WALK_GROUP_GO(_group, _walk) for(_walk = group_get_first_go(_group); _walk; _walk = group_get_next_go(_group, _walk))

/* Recursive OBJ movement */
struct go* go_recursive_next(const struct go *root, struct go *walk);
#define WALK_GO_RECURSIVE(_root, _walk) for(_walk = _root; _walk; _walk = go_recursive_next(_root, _walk))
#define WALK_GO_RECURSIVE_WITHOUT_ROOT(_root, _walk) for(_walk = go_recursive_next(_root, _root); _walk; _walk = go_recursive_next(_root, _walk))

/* Miscellaneous GROUP movement */
static inline uns group_get_rank_go(struct go_group *group, struct go *go)
	{ uns n = -1; for(; go; go = group_get_prev_go(group, go)) n++; return n; }
static inline int group_is_go_descentant(struct go_group *group, struct go *go)
	{ do { if(&group->go == go) return 1; } while((go = &go_get_group(go)->go)); return 0; }
static inline int go_is_linked(struct go *go)
	{ return go_get_tlo(go)->obj.type == OT_TLO; }
SCHEME static inline SUNS group_get_count_go(SO_go_group group)
	{ struct go *go; uns u = 0; WALK_GROUP_GO(group, go) u++; return u; }


/*
 * Movement in O (OBJ and GO)
 */

struct o* o_get_ancestor(struct o* o);

/* Type of _o must be a point to structure which is derived from o. Type of _walk must be struct o*. */
#define WALK_O_ANCESTOR(_o, _walk) for(_walk = void2o(_o); _walk; _walk = o_get_ancestor(_walk))

static inline int o_is_ancestor(struct o *ancestor, struct o *descendant)
	{ struct o *walk; WALK_O_ANCESTOR(descendant, walk) if(walk == ancestor) return 1; return 0; }

SCHEME static inline SINT o_is_linked(SO_o o)
	{ return o->kind == T_OBJ ? obj_is_ancestor(obj_universum, o2obj(o)) : go_is_linked(o2go(o)); }

/*
 * Dependencies movement
 */

/* Hanger - GO movement */
#define WALK_GO_HANGER_BEGIN(_go, _h) do { \
	struct go_info *_gi = go_get_info(_go); \
	clist *_l = go_get_lhanger(_go);	\
	int _i = _gi->h_count; \
	for(_h = _i ? SKIP_OFFSET(struct hanger, _go, _gi->h_offset) : (_l ? &go_lhanger_get_first_hanger_smart(_l)->h : NULL); \
			_h; \
			_i--, _h = _i > 0 ? _h + 1 : (_i < 0 ? &go_lhanger_get_next_hanger_smart(_l, hanger2smart(_h))->h : \
				(_l ? &go_lhanger_get_first_hanger_smart(_l)->h : NULL))) \
					do {
#define WALK_GO_HANGER_END } while(0); } while(0);

SCHEME static inline SO_go hanger_get_go(SHANGER h)
	{ return void2go((hanger_is_smart(h) ? hanger2smart(h)->go : SKIP_OFFSET(struct go, h, -h->info->go_offset))); }

SCHEME static inline SHANGER go_get_hanger(SO_go go, SUNS n)
	{ struct go_info *gi = go_get_info(go); return (gi->h_count > n) ? SKIP_OFFSET(struct hanger, go, gi->h_offset) + n : NULL; }
SCHEME static inline SHANGER go_get_hanger_curve(SO_go go)
	{ int i = go_get_info(go)->h_curve; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_start(SO_go go)
	{ int i = go_get_info(go)->h_start; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_end(SO_go go)
	{ int i = go_get_info(go)->h_end; return i == -1 ? NULL : go_get_hanger(go, i); }
SCHEME static inline SHANGER go_get_hanger_center(SO_go go)
	{ int i = go_get_info(go)->h_center; return i == -1 ? NULL : go_get_hanger(go, i); }

SCHEME static inline SHANGER go_get_hanger_by_name(SO_go go, SSYMBOLS name)
	{ struct hanger *h; WALK_GO_HANGER_BEGIN(go, h) if(h->info->name == name) return h; WALK_GO_HANGER_END return NULL; }

/* Anchor - hanger movement */
SCHEME static inline SANCHOR hanger_get_first_anchor(SHANGER h)
	{ cnode *n = clist_head(&h->lanchor); return n ? SKIP_BACK(struct anchor, nanchor, n) : NULL; }
SCHEME static inline SANCHOR hanger_get_next_anchor(SANCHOR a)
	{ cnode *n = clist_next(&a->h->lanchor, &a->nanchor); return n ? SKIP_BACK(struct anchor, nanchor, n) : NULL; }
#define WALK_HANGER_ANCHOR(_hanger, _walk) for(_walk = hanger_get_first_anchor(_hanger); _walk; _walk = hanger_get_next_anchor(_walk))


/* Anchor - GO movement */
SCHEME static inline SO_go anchor_get_go(SANCHOR a)
	{ return void2go(SKIP_OFFSET(struct go, a, -a->info->offset)); }
SCHEME static inline SANCHOR go_get_anchor(SO_go go, SUNS anchor)
	{ struct go_info *gi = go_get_info(go); return anchor < gi->a_count ? SKIP_OFFSET(struct anchor, go, gi->a_offset) + anchor : NULL; }
static inline struct anchor* go_get_first_anchor(struct go *go)
	{ return SKIP_OFFSET(struct anchor, go, go_get_info(go)->a_offset); }
#define WALK_GO_ANCHOR(_go, _walk, _i) for(_i = -go_get_info(_go)->a_count, _walk = go_get_first_anchor(_go); _i; _i++, _walk++)

SCHEME static inline SANCHOR go_get_anchor_by_name(SO_go go, SSYMBOLS name)
	{ struct anchor *a; int i; WALK_GO_ANCHOR(go, a, i) if(a->info->name == name) return a; return NULL; }

/*
 * Movement in selections
 */

SCHEME static inline SO_obj obj_get_first_selected(void)
	{ cnode *n = obj_lsel.head.next; return n != &obj_lsel.head ? SKIP_BACK(struct obj, nsel, n) : NULL; }
SCHEME static inline SO_obj obj_get_next_selected(SO_obj o)
	{ cnode *n = o->nsel.next; return n != &obj_lsel.head ? SKIP_BACK(struct obj, nsel, n) : NULL; }
#define WALK_OBJ_SELECTED(_walk) for(_walk = obj_get_first_selected(); _walk; _walk = obj_get_next_selected(_walk))

SCHEME static inline SO_go group_get_first_selected_go(SO_go_group group)
	{ cnode *n = clist_head(&group->lselected); return n ? SKIP_BACK(struct go, nselected, n) : NULL; }
SCHEME static inline SO_go group_get_next_selected_go(SO_go_group group, SO_go go)
	{ cnode *n = clist_next(&group->lselected, &go->nselected); return n ? SKIP_BACK(struct go, nselected, n) : NULL; }
#define WALK_GROUP_SELECTED_GO(_group, _walk) for(_walk = group_get_first_selected_go(_group); _walk; _walk = group_get_next_selected_go(_group, _walk))



/**********************************************
 *                                            *
 *     OBJ and GO hooks handling functions    *
 *                                            *
 **********************************************/

struct obj_hook* obj_hook_add(struct obj_hook *hook);
static inline struct go_hook* go_hook_add(struct obj_tlo *tlo, struct go_hook *hook)
	{ slist_add(&tlo->go_hooks, &hook->n); return hook; }

struct obj_hook* obj_hook_remove(struct obj_hook *hook);
static inline struct go_hook* go_hook_remove(struct obj_tlo *tlo, struct go_hook *hook)
	{ slist_remove(&tlo->go_hooks, &hook->n); return hook; }

struct obj_hook* obj_hook_remove_by_data(void *data);
struct go_hook* go_hook_remove_by_data(struct obj_tlo *tlo, void * data);

static inline struct obj_hook* obj_hook_fill(
		struct obj_hook *h,
		void* data,
		void (*o_link) (struct o *obj, struct obj_hook* h),
		void (*o_unlink) (struct o *obj, struct o *father, struct o *prev, struct obj_hook* h),
		void (*obj_selected) (struct o *obj, struct obj_hook* h),
		void (*obj_unselected) (struct o *obj, struct obj_hook* h),
		void (*obj_prop_create) (struct o *obj, struct prop* p, struct obj_hook* h),
		void (*obj_prop_change) (struct o *obj, struct obj_hook* h),
		void (*obj_prop_delete) (struct o *obj, struct prop* p, struct obj_hook* h))
	{ h->data = data; h->obj_link = o_link; h->obj_unlink = o_unlink; h->obj_selected = obj_selected; h->obj_unselected = obj_unselected; h->obj_prop_create = obj_prop_create; h->obj_prop_change = obj_prop_change; h->obj_prop_delete = obj_prop_delete; return h; }

static inline struct go_hook* go_hook_fill(
		struct go_hook *h,
		void* data,
		void (*g_link) (struct o *go, struct go_hook* h),
		void (*g_unlink) (struct o *go, struct o *owner, struct o *prev, struct go_hook* h),
		void (*g_relink) (struct o *go, struct go_group *old_group, struct go *old_prev, struct go_hook* h),
		void (*go_selected) (struct o *go, struct go_hook* h),
		void (*go_unselected) (struct o *go, struct go_hook* h),
		void (*go_prop_create) (struct o *go, struct prop* p, struct go_hook* h),
		void (*go_prop_change) (struct o *go, struct go_hook* h),
		void (*go_prop_delete) (struct o *go, struct prop* p, struct go_hook* h),
		void (*go_changed) (struct go *go, const struct changed_data_generic *data, struct go_hook* h))
	{ h->data = data; h->go_link = g_link; h->go_unlink = g_unlink; h->go_relink = g_relink; h->go_selected = go_selected; h->go_unselected = go_unselected; h->go_prop_create = go_prop_create; h->go_prop_change = go_prop_change; h->go_prop_delete = go_prop_delete; h->go_changed = go_changed; return h; }


/*
 * Miscellaneous
 */

SCHEME static inline SBOOL hanger_is_mc(SHANGER h)
	{ struct go *go = hanger_get_go(h); return go->type == GOT_GROUP && (go->subtype == GOST_GROUP_TLG || go->subtype == GOST_GROUP_TEMP); }
SCHEME static inline SBOOL hanger_is_position(SHANGER h)
	{ return h->info->quantity == QT_POINT; }
SCHEME static inline SBOOL hanger_is_curve(SHANGER h)
	{ return h->info->quantity == QT_CURVE; }
static inline quantity hanger_get_quantity(struct hanger *h)
	{ ASSERT(!hanger_is_curve(h)); return h->info->q_offset ? *SKIP_OFFSET(quantity, h, h->info->q_offset) : h->info->get_quantity(h); }
static inline quantity* hanger_get_quantity_ptr(struct hanger *h, quantity *q)
	{ *q = hanger_get_quantity(h); return q; }
static inline struct geom_point hanger_get_position(struct hanger *h)
	{ ASSERT(hanger_is_position(h)); return hanger_get_quantity(h).p; }
static inline struct geom_point* hanger_get_position_ptr(struct hanger *h, struct geom_point *gp)
	{ *gp = hanger_get_position(h); return gp; }
uns go_can_anchor_rehang(struct anchor* a, struct hanger* h, struct obj_tlo *tlo);

SCHEME SBOOL go_has_dependent(SO_go go);
SCHEME SUNS go_get_count_dependents(SO_go go);
SCHEME SO_go go_get_owner_and_member(SO_go go, SO_go_group group);


// RTREE & expansion

SCHEME SVOID expand_ref(SO_obj_tlo tlo);
SCHEME SVOID expand_unref(SO_obj_tlo tlo);

struct geom_path* go_path_expand_ref(struct obj_tlo *tlo, struct go *go);
void go_path_expand_unref(struct obj_tlo *tlo, struct geom_path *path_expand);

const struct geom_rectangle* go_get_expand_bbox(struct obj_tlo *tlo, struct go *go);
const struct geom_path* go_get_expand_path(struct obj_tlo *tlo, struct go *go);

const struct geom_rectangle* tlo_get_expand_bbox(struct obj_tlo *tlo);

// FIXME: Lepsi implementaci pomoci rtree a stromu go-cek

#define WALK_GROUP_SMART(_group, _down, _go, _owner, _pre, _in, _post) \
	do { \
		 _owner = _group; cnode *_n = &_owner->lmember.head; \
NEXT: \
		while((_n = clist_next(&_owner->lmember, _n))) { \
			_go = SKIP_BACK(struct go, nmember, _n); \
			if(_go->type == GOT_GROUP) \
		  		{ if(_down) { _owner = go2group(_go); goto DOWN; } } \
			else do { _in; } while(0); \
		} \
		if(_owner == _group) break; \
		do { _post; } while(0); \
		_n = &_owner->go.nmember; _owner = _owner->go.owner; goto NEXT;  \
DOWN: \
		do { _pre; } while(0); \
		_n = &_owner->lmember.head; goto NEXT; \
	} while(0)

#define WGSB(_r, _tlo, _go, _bbox) ((_r = go_get_expand_bbox(_tlo, _go)) && geom_rectangle_have_intersection(_r, _bbox))

#define WALK_GROUP_SMART_BBOX(_tlo, _group, _bbox, _go, _owner, _pre, _in, _post) \
	do { const struct geom_rectangle *_r; \
		WALK_GROUP_SMART(_group, WGSB(_r,_tlo,_go,_bbox), _go, _owner, _pre, if(WGSB(_r,_tlo,_go,_bbox)) do { _in; } while(0), _post); \
	} while(0)



/**********************
 *                    *
 *    Clipboard       *
 *                    *
 **********************/


void clipboard_copy_selected_go(struct go_group *source, struct go_group *target, struct go *after);
uns clipboard_copy(struct go_group *source, const char* name_trans, char* error_string);
void clipboard_paste(struct go_group *target, struct go *after);
void clipboard_cut(struct go_group *source);



/*****************************
 *                           *
 *    Dumping functions      *
 *                           *
 *****************************/

SCHEME SVOID o_dump(SO_o o);
SCHEME SVOID o_dump_prop(SO_o o);

SCHEME SVOID obj_dump_father(SO_obj obj);
SCHEME SVOID obj_dump_son(SO_obj obj);
SCHEME SVOID obj_dump_subtree(SO_obj root);
SCHEME SVOID obj_dump_flag(SO_obj root);
SCHEME SVOID obj_dump_hook(SO_obj obj);
SCHEME SVOID obj_dump_selected(void);
SCHEME SVOID obj_dump_all(SO_obj obj);

SCHEME SVOID tlo_dump_bbox(SO_obj_tlo tlo);
SCHEME SVOID tlo_dump_tsort(SO_obj_tlo tlo);

SCHEME SVOID go_dump_flag(SO_go go);
SCHEME SVOID go_dump_hanger(SO_go go);
SCHEME SVOID go_dump_anchor(SO_go go);
SCHEME SVOID go_dump_bbox(SO_go go);
SCHEME SVOID go_dump_type(SO_go go);
SCHEME SVOID go_dump_all(SO_go go);

SCHEME SVOID group_dump_go(SO_go_group group);
SCHEME SVOID group_dump_subtree(SO_go_group root);
SCHEME SVOID group_dump_selected_go(SO_go_group group);

SCHEME SVOID hanger_dump(SHANGER h);
SCHEME SVOID hanger_dump_anchor(SHANGER h);
SCHEME SVOID anchor_dump(SANCHOR a);

 SCHEME SVOID kernel_dump_memory(void);



/*************************
 *                       *
 *     Save and load     *
 *                       *
 *************************/

SCHEME SVOID doc_save(SO_obj_doc doc, SSTRINGC filename);
SCHEME SO_obj_doc doc_load(SSTRINGC filename);

SCHEME SVOID settings_save(SSTRINGC filename) SMETATRANS ;
SCHEME SVOID settings_load(SSTRINGC filename) SMETATRANS ;

SCHEME static inline SUNS tlo_is_saved(SO_obj_tlo tlo)
	{ return tlo->undo_saved == tlo->undo_present; }
SCHEME static inline SVOID tlo_set_saved(SO_obj_tlo tlo)
	{ tlo->undo_saved = tlo->undo_present; }

SCHEME static inline SUNS doc_is_saved(SO_obj_doc doc)
	{ return doc->saved; }
SCHEME SVOID doc_set_saved(SO_obj_doc doc);

SCHEME SUNS doc_is_all_saved(SO_obj_doc doc);
SCHEME SVOID doc_set_all_saved(SO_obj_doc doc);



/*************************
 *                       *
 *     Miscellaneous     *
 *                       *
 *************************/


void kernel_init(int argc, char **argv, uns gui);
void kernel_finish(void);

#endif
