/** \file kernel/glt_kernel.c
  \brief GUILE-link - interface functions for kernel data types. */


#include "lib/lib.h"
#include <stdio.h>
#include <string.h>
#include <guile/gh.h>
#include "kernel/kernel.h"
#include "build/path.h"

#include "kernel/gl_trans.h"
#include "kernel/glt_common.h"
#include "kernel/glt_kernel.h"

SCM obj_type_numbers;
SCM go_type_numbers;
SCM go_subtype_numbers;

SCM *obj_type_symbols;
SCM *go_type_symbols;
SCM **go_subtype_symbols;

SCM symbol_obj    = SCM_BOOL_F;
SCM symbol_go     = SCM_BOOL_F;
SCM symbol_hanger = SCM_BOOL_F;
SCM symbol_anchor = SCM_BOOL_F;
SCM symbol_transaction_failed = SCM_BOOL_F;
SCM symbol_condition_not_satisfied = SCM_BOOL_F;

typedef struct rst_item
{
  struct rst_item* next;
  SCM proxy;
  void* p;
} rst_item;


#define HT_PREFIX(x) rst_ ## x
#define HT_KEY void*
#define HT_KEY_POINTER p
#define HT_TYPE SCM
#define HT_STRUCT rst_item
#define HT_ALLOC(_struct, _key, _type) \
	(_struct) = xmalloc (sizeof (rst_item)); \
	(_struct)->p = _key; \
	(_struct)->proxy = _type;
#define HT_ONLY_ONE rst_oo
#include "lib/hashtable.h"

#define WALK_RST(_walk) for(rst_item **_index = rst_oo.table, *_walk = *_index ? *_index : rst_next(&_index); _index < rst_oo.table + rst_oo.size && _walk; _walk = _walk->next ? _walk->next : rst_next(&_index))

scm_t_bits o_tag;
scm_t_bits hanger_tag;
scm_t_bits anchor_tag;

int is_o_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(o_tag, proxy); }

int is_hanger_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(hanger_tag, proxy); }

int is_anchor_proxy(SCM proxy)
{ return SCM_SMOB_PREDICATE(anchor_tag, proxy); }


int is_specific_proxy(SCM proxy, int kind, int type, int subtype)
{
  if ((kind == T_HANGER) && (is_hanger_proxy (proxy)) && (type == -1))
    return (1);
  if ((kind == T_ANCHOR) && (is_anchor_proxy (proxy)) && (type == -1))
    return (1);
  else if (is_o_proxy (proxy))
    {
      struct o * obj = optr (proxy);
      return (((kind == -1) || (obj -> kind == kind)) &&
	      ((type == -1) || (obj -> type == type)) &&
	      ((subtype == -1) || (obj -> subtype == subtype)));
    }
  else return (0);
}

SCM vrr_exception_handler;

static size_t o_proxy_free (SCM proxy)
{
  rst_delete (optr (proxy));
  o_unref (optr (proxy));
  return (0);
}

static size_t hanger_proxy_free (SCM proxy)
{
  rst_delete( hptr (proxy));
  hanger_unref (hptr (proxy));
  return (0);
}

static size_t anchor_proxy_free (SCM proxy)
{
  rst_delete( aptr (proxy));
  go_unref (anchor_get_go (aptr (proxy)));
  return (0);
}

SCM o2scm(struct o *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      o_ref (p);
      SCM proxy;
      SCM_NEWSMOB (proxy, o_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

SCM hanger2scm(struct hanger *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      hanger_ref (p);
      SCM proxy;
      SCM_NEWSMOB (proxy, hanger_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

SCM anchor2scm(struct anchor *p)
{
  if (p == NULL) return (SCM_BOOL_F);

  struct rst_item *it = rst_get(p);
  if (it == NULL)
    {
      go_ref ( anchor_get_go(p));
      SCM proxy;
      SCM_NEWSMOB (proxy, anchor_tag, p);
      rst_insert(p, proxy);
      return proxy;
    }
  else return it->proxy;
}

void condition_not_satisfied(const char *name, const char *dsc)
{ scm_error_scm (symbol_condition_not_satisfied, scm_makfrom0str (name), scm_makfrom0str (dsc), SCM_EOL, SCM_BOOL_F); }

static SCM gl_proxy_get_kind (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (symbol_obj);
      else if (o -> kind == T_GO) return (symbol_go);
      else BUG();
    }
  else if (is_hanger_proxy (proxy)) return (symbol_hanger);
  else if (is_anchor_proxy (proxy)) return (symbol_anchor);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-kind");
}

static SCM gl_proxy_get_type (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (obj_type_symbols [o -> type]);
      else if (o -> kind == T_GO) return (go_type_symbols [o -> type]);
      else BUG ();
    }
  else if (is_hanger_proxy (proxy)) return (SCM_BOOL_F);
  else if (is_anchor_proxy (proxy)) return (SCM_BOOL_F);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-type");
}

static SCM gl_proxy_get_subtype (SCM proxy)
{
  if (is_o_proxy (proxy))
    {
      struct o * o = optr (proxy);
      if (o -> kind == T_OBJ) return (SCM_BOOL_F);
      else if (o -> kind == T_GO) return (out_SSYMBOLS (go_get_info ((struct go *) o) -> name));
      else BUG ();
    }
  else if (is_hanger_proxy (proxy)) return (SCM_BOOL_F);
  else if (is_anchor_proxy (proxy)) return (SCM_BOOL_F);
  else SCM_ASSERT (0, proxy, SCM_ARG1, "proxy-get-subtype");
}

static SCM gl_proxy_p (SCM proxy, SCM kind, SCM type, SCM subt)
{
  SCM_ASSERT (SCM_EQ_P (kind, SCM_UNDEFINED) || SCM_SYMBOLP (kind), kind,
	      SCM_ARG2, "proxy?");
  SCM_ASSERT (SCM_EQ_P (type, SCM_UNDEFINED) || SCM_SYMBOLP (type), type,
	      SCM_ARG3, "proxy?");
  SCM_ASSERT (SCM_EQ_P (subt, SCM_UNDEFINED) || SCM_SYMBOLP (subt), subt,
	      SCM_ARG4, "proxy?");

  return SCM_BOOL ((is_o_proxy (proxy) ||
		    is_hanger_proxy (proxy) ||
		    is_anchor_proxy (proxy)) &&
		   (SCM_EQ_P (kind, SCM_UNDEFINED) ||
		    SCM_EQ_P (kind, gl_proxy_get_kind (proxy))) &&
		   (SCM_EQ_P (type, SCM_UNDEFINED) ||
		    SCM_EQ_P (type, gl_proxy_get_type (proxy))) &&
		   (SCM_EQ_P (subt, SCM_UNDEFINED) ||
		    SCM_EQ_P (subt, gl_proxy_get_subtype (proxy))));
}

static SCM
gl_group_get_children (SCM ar0)
{
  SCM retval = SCM_EOL;

  assert_SO (ar0, SCM_ARG1, "go-link", T_GO, GOT_GROUP);
  struct go_group *grp = (struct go_group *) scm2o (ar0);
  struct go* child;

  for (child = group_get_last_go (grp); child != NULL; child = group_get_prev_go (grp, child))
    retval = scm_cons (o2scm ((struct o*) child), retval);

  return retval;
}



static void
string_array_convert (string * src, uns size, SCM ** dst, SCM * rev,
		     const char *dst_name, const char *rev_name)
{
    SCM sdst = scm_c_make_vector (size, SCM_BOOL_F);
    *rev = scm_c_make_hash_table (size * 2);
    *dst = xmalloc (size * sizeof (SCM));

    for (uns i = 0; i < size; i ++)
    {
	SCM symbol = (*dst) [i] = scm_str2symbol (src[i] -> text);
	scm_vector_set_x (sdst, SCM_MAKINUM (i), symbol);
	scm_hashq_set_x (*rev, symbol, SCM_MAKINUM (i));
    }

    scm_c_define (dst_name, sdst);
    scm_c_define (rev_name, *rev);
}

UNUSED static void
string_array_line (string * src, uns size, uns ln,
		   SCM ** dst, SCM sdst, SCM rev)
{
    dst [ln] = xmalloc (size * sizeof (SCM));
    SCM line = scm_c_make_vector (size, SCM_BOOL_F);
    scm_vector_set_x (sdst, SCM_MAKINUM (ln), line);

    for (uns i = 0; i < size; i ++)
    {
	SCM symbol = dst [ln][i] = scm_str2symbol (src[i] -> text);
	scm_vector_set_x (line, SCM_MAKINUM (i), symbol);
	scm_hashq_set_x (rev, symbol, scm_cons (SCM_MAKINUM (ln), SCM_MAKINUM (i)));
    }
}


void
glt_kernel_finish (void)
{
  WALK_RST(walk)
    {
      if (is_o_proxy (walk -> proxy)) o_unref (optr (walk -> proxy));
      else if (is_hanger_proxy (walk -> proxy)) hanger_unref (hptr (walk -> proxy));
      else if (is_anchor_proxy (walk -> proxy)) go_unref ( anchor_get_go( aptr (walk -> proxy)));
      else BUG ();
    }

  xfree (obj_type_symbols);
  xfree (go_type_symbols);
  xfree (go_subtype_symbols);

  rst_finish();
}

void
glt_kernel_init(void)
{
  rst_init(31);

  // Scheme smob definitions
  o_tag = scm_make_smob_type ("vrr-o", 0);
  scm_set_smob_free (o_tag, o_proxy_free);

  hanger_tag = scm_make_smob_type ("vrr-hanger", 0);
  scm_set_smob_free (hanger_tag, hanger_proxy_free);

  anchor_tag = scm_make_smob_type ("vrr-anchor", 0);
  scm_set_smob_free (anchor_tag, anchor_proxy_free);

  // string-symbol-number tables
  string_array_convert (string_obj_type, OT_MAX,  &obj_type_symbols, &obj_type_numbers, "obj-type-symbols", "obj-type-numbers");
  string_array_convert (string_go_type,	 GOT_MAX, &go_type_symbols,  &go_type_numbers,  "go-type-symbols",  "go-type-numbers");

  go_subtype_symbols = xmalloc_zero (GOT_MAX * sizeof (SCM *));
  for (int i = 0; i < GOT_MAX; i ++) go_subtype_symbols [i] = go_type_symbols + i;
  SCM subtype_symbols_scm = scm_c_make_vector (GOT_MAX, SCM_BOOL_F);
  go_subtype_numbers = scm_c_make_hash_table (GOT_MAX * 8); // heuristic
  scm_c_define ("go-subtype-symbols", subtype_symbols_scm);
  scm_c_define ("go-subtype-numbers", go_subtype_numbers);

//  string_array_line (string_elarc_subtype,  GOST_ELARC_MAX,  GOT_ELARC,  go_subtype_symbols, subtype_symbols_scm, go_subtype_numbers);
//  string_array_line (string_bezier_subtype, GOST_BEZIER_MAX, GOT_BEZIER, go_subtype_symbols, subtype_symbols_scm, go_subtype_numbers);

  // misc symbols
  SCM symbols = scm_c_make_vector (6, SCM_BOOL_F);
  scm_c_define("symbols-gc-protect", symbols);
  #define DEF(N, VAR, STR) scm_vector_set_x (symbols, SCM_MAKINUM (N), VAR = scm_str2symbol (STR))
  DEF (0, symbol_obj,		"obj");
  DEF (1, symbol_go,		"go");
  DEF (2, symbol_hanger,	"hanger");
  DEF (3, symbol_anchor,	"anchor");
  DEF (4, symbol_transaction_failed,	  "transaction-failed");
  DEF (5, symbol_condition_not_satisfied, "condition-not-satisfied");
  #undef DEF

  scm_c_define_gsubr ("proxy?", 1, 3, 0, gl_proxy_p);
  scm_c_define_gsubr ("proxy-get-kind", 1, 0, 0, gl_proxy_get_kind);
  scm_c_define_gsubr ("proxy-get-type", 1, 0, 0, gl_proxy_get_type);
  scm_c_define_gsubr ("proxy-get-subtype", 1, 0, 0, gl_proxy_get_subtype);

  scm_c_define_gsubr ("group-get-children", 1, 0, 0, gl_group_get_children);
}
