

;;; miscellaneous functions


(define (make-thunk fn . args)
  (lambda () (apply fn args)))

(define (curry fn arg1)
  (lambda args (apply fn (cons arg1 args))))

(define (safe-force p)
  (if (promise? p) (force p) p))

(define (natural? n)
  (and (integer? n) (exact? n) (>= n 0)))

(define (every predicate list)
  (or (null? list)
      (and (predicate (car list))
	   (every predicate (cdr list)))))

(define (flatten data)
  (let rec ((node data)
	    (tail '()))
    (cond
     ((pair? node)
      (rec (car node) (rec (cdr node) tail)))
     ((null? node)
      tail)
     (else
      (cons node tail)))))

(define (true . x) #t)
(define (id x) x)

(define (flag? number flag) (positive? (logand number flag)))


(define (sav f) (save-doc (o-lookup-global "Untitled 1") f))


;;; Vrr REPL and console support


(define (typearg-error fn arg val)
  (scm-error 'wrong-type-arg fn "Wrong type argument in position ~S: ~S" (list arg val) #f))

(define implicit-tlo 'active)

(define (set-implicit-tlo! tlo)
  (if (or (not tlo)
	  (eq? tlo 'active)
	  (proxy? tlo 'obj 'tlo)
	  (proxy? tlo 'obj 'universum))
      (set! implicit-tlo tlo)
      (typearg-error "set-implicit-tlo!" 1 tlo)))

(define (get-implicit-tlo)
  (if (eq? implicit-tlo 'active)
      (get-active-tlo)
      implicit-tlo))

(define (make-vrr-console-port console)
  (make-soft-port
   (vector (lambda (c) (vrr-console-write-char console c))
	   (lambda (s) (vrr-console-write-string console s))
	   #f
	   (lambda () (vrr-console-read-char console))
	   #f)
   "rw"))

(define vrr-repl-prompt-1 "> ")
(define vrr-repl-prompt-2 "... ")

(define (vrr-repl-init console)
  (display vrr-repl-prompt-1 (vrr-console-get-output-port console))
  #f)

(define (vrr-repl-iter console)
  (call-with-io-ports
   (vrr-console-get-input-port console)
   (vrr-console-get-output-port console)
   (lambda () (vrr-repl-iter-int console))))

(define (display-exception . e)
  (if (= (length e) 5)
      (apply display-error (cons* #f (current-output-port) (cdr e)))
      (format #t "Error - exception: ~s\n" e))
  *unspecified*)


(define (vrr-repl-eval call)
  (if (get-implicit-tlo)
      (trans-main (get-implicit-tlo)
		  (lambda ()
		    (with-fluids ((autolink-state user-autolink-state))
				 (primitive-eval call)))
		  display-exception
		  "Console command")
      (catch #t
	     (lambda () (with-fluids ((autolink-state user-autolink-state))
				     (primitive-eval call)))
	     display-exception)))

(define (vrr-repl-read console)
  (vrr-console-begin-blockread console)
  (let ((data (read)))
    (vrr-console-end-blockread console)
    data))

(define (vrr-repl-iter-int console)
  (catch 'misc-error
	 (lambda ()
	   (let loop ((data (vrr-repl-read console)))
	     (if (not (eof-object? data))
		 (let ((val (vrr-repl-eval data)))
		   (if (not (unspecified? val))
		       (begin
			 (write val)
			 (newline)))
		   (loop (vrr-repl-read console)))
		 (display vrr-repl-prompt-1))))
	 (lambda err
	   (if (and (> (length err) 2) (string=? (caddr err) "end of file"))
	       (display vrr-repl-prompt-2)
	       (begin
	         (vrr-console-end-blockread console)
		 (display "ERROR ")
		 (write err)
		 (newline)
		 (vrr-repl-iter-int console))))))

(define (call-with-io-ports input output thunk)
  (let ((ci (current-input-port))
	(co (current-output-port)))
    (set-current-input-port input)
    (set-current-output-port output)
    (thunk)
    (set-current-input-port ci)
    (set-current-output-port co)))


;;; Support functions for common user functions

(define autolink-state (make-fluid))
(fluid-set! autolink-state #t)

(define user-autolink-state #t)

(define (set-autolink-state! value)
  (if (boolean? value)
      (begin
	(set! user-autolink-state value)
	(fluid-set! autolink-state value))
      (typearg-error "set-autolink-state!" 1 value)))

(define (get-autolink-state)
  (fluid-ref autolink-state))

(define (call-with-autolink-state value thunk)
  (with-fluids ((autolink-state value))
	       (thunk)))

(define (autolink go)
  (if (get-autolink-state)
      (go-link go (tlo-get-tlg (get-trans-tlo)) #f)
      go))

(define (make-idtable . items)
  (let* ((l (length items))
	 (t1 (make-vector l))
	 (t2 (make-hash-table (* l 2))))
    (for-each (lambda (key value)
		(vector-set! t1 value key)
		(hashq-set! t2 key value))
	      items (iota l))
    (cons t1 t2)))

(define (make-idtable-lookup-1 idt)
  (lambda (n) (vector-ref (car idt) n)))

(define (make-idtable-lookup-2 idt)
  (lambda (k) (hashq-ref (cdr idt) k)))

(define conic-table (make-idtable
		     'start-entire 'start-dif 'point-entire 'point-dif
		     'ccw 'cw 'smaller 'bigger 'middle 'opposite 'entire))

(define code-conic (make-idtable-lookup-2 conic-table))
(define decode-conic (make-idtable-lookup-1 conic-table))

(define cap-style-table (make-idtable 'butt 'round 'projecting))

(define code-cap-style (make-idtable-lookup-2 cap-style-table))
(define decode-cap-style (make-idtable-lookup-1 cap-style-table))


(define (prepare fn fargs largs)
  (lambda args
    (autolink (apply fn (append fargs args largs)))))

(define a-vtables '(#t))
(define h-vtables '(#t))

(define rh-table (make-hash-table 31))
(define src-table (make-hash-table 31))
(define cons-table (make-hash-table 31))

(define (get-vtable vtlist op)
  (or (assq-ref (cdr vtlist) op)
      (begin
	(set-cdr! vtlist (assq-set! (cdr vtlist) op (make-hash-table 31)))
	(assq-ref (cdr vtlist) op))))

(define (prepare-aget op)
  (let ((vt (get-vtable a-vtables op)))
    (lambda (go)
      (let ((num (hashq-ref vt (proxy-get-subtype go))))
	(if num (go-get-anchor go num) (typearg-error #f 1 go))))))

(define (prepare-hget op)
  (let ((vt (get-vtable h-vtables op)))
    (lambda (go)
      (let ((num (hashq-ref vt (proxy-get-type go))))
	(if num (go-get-hanger go num) (typearg-error #f 1 go))))))

(define (define-aslot subtype op val)
  (let ((vt (get-vtable a-vtables op)))
    (hashq-set! vt subtype val)))

(define (define-aslots subtype . ops)
  (for-each (lambda (a b) (define-aslot subtype a b))
	    ops (iota (length ops))))

(define (define-hslot type op val)
  (let ((vt (get-vtable h-vtables op)))
    (let ((v (hashq-ref rh-table type)))
      (if (vector? v) (vector-set! v val op)))
    (hashq-set! vt type val)))

(define (define-hcount type count)
  (hashq-set! rh-table type (make-vector count #f)))

(define (define-h-all type val)
  (hashq-set! rh-table type val))

(define (define-source subtype srclist)
  (hashq-set! src-table subtype srclist))

(define (define-constructor subtype make-fn)
  (hashq-set! cons-table subtype make-fn))

(define (make-go subtype args)
  (apply (hashq-ref cons-table subtype) args))

(define prop-table (make-hash-table 31))

(define prop-table2 (make-vector CC_PT_MAX #f))
(vector-set! prop-table2 CC_PT_UNS    (make-vector CC_PTU_MAX #f))
(vector-set! prop-table2 CC_PT_POINTER (make-vector CC_PTP_MAX #f))
(vector-set! prop-table2 CC_PT_STRING (make-vector CC_PTS_MAX #f))
(vector-set! prop-table2 CC_PT_REAL   (make-vector CC_PTR_MAX #f))

(define (register-prop-subtype symbol type-number subtype-number decoder)
  (let ((obj (vector symbol type-number subtype-number decoder)))
    (hashq-set! prop-table symbol obj)
    (vector-set! (vector-ref prop-table2 type-number) subtype-number obj)))

(define (property-type-symbol t s)
  (pt-symbol (vector-ref (vector-ref prop-table2 t) s)))

(define (pt-symbol x) (vector-ref x 0))
(define (pt-type-number x) (vector-ref x 1))
(define (pt-subtype-number x) (vector-ref x 2))
(define (pt-decode x y) ((vector-ref x 3) y))


(define (prop-decode tn stn value)
  (pt-decode (vector-ref (vector-ref prop-table2 tn) stn) value))

(define (real-between? a b c)
  (and (real? a) (real? b) (real? c) (<= a b) (<= b c)))

(define (natural-between? a b c)
  (and (natural? a) (natural? b) (natural? c) (<= a b) (<= b c)))


(define prop-definition-table (make-hash-table 31))

(define (register-prop-definition name fn)
  (hashq-set! prop-definition-table name fn))

(define (property-specific-set obj prop name data)
  (let ((fn (hashq-ref prop-definition-table name)))
    (if fn
	(fn obj prop data)
	(error "Bad property value definition"))))


(define (rgb-list? a)
  (and (list? a) (= (length a) 3) (every (lambda (x) (natural-between? 0 x 255)) a)))

(define (rgba-list? a)
  (and (list? a) (= (length a) 4) (every (lambda (x) (natural-between? 0 x 255)) a)))

(define unpack-r uns2r-uns)
(define unpack-g uns2g-uns)
(define unpack-b uns2b-uns)
(define unpack-a uns2a-uns)

(define pack-rgba rgba-uns2uns)
(define (pack-rgb r g b)
  (pack-rgba r g b 255))


(define (natural-list? data)
  (and (pair? data) (natural? (car data)) (null? (cdr data))))

(define (property-natural-set obj prop data)
  (or (natural-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_UNSPECIFIED #t (car data)))

(define (property-rgb-set obj prop data)
  (or (rgb-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_COLOR #t (apply pack-rgb data)))

(define (property-rgba-set obj prop data)
  (or (rgba-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_COLOR #t (apply pack-rgba data)))

(define (font-list? data)
  (and (pair? data) (string? (car data)) (null? (cdr data))))

(define (property-font-set obj prop data)
  (or (font-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_FONT #t (font-load (car data))))

(define (conic-type-list? data)
  (and (pair? data) (code-conic (car data)) (null? (cdr data))))

(define (property-conic-type-0-set obj prop data)
  (or (conic-type-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_CONIC_TYPE_0P #t (code-conic (car data))))

(define (property-conic-type-1-set obj prop data)
  (or (conic-type-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_CONIC_TYPE_1P #t (code-conic (car data))))

(define (property-conic-type-2-set obj prop data)
  (or (conic-type-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_CONIC_TYPE_2P #t (code-conic (car data))))

(define (property-conic-type-3-set obj prop data)
  (or (conic-type-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_CONIC_TYPE_3P #t (code-conic (car data))))

(define (property-boolean-set obj prop data)
  (or (boolean? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_BOOLEAN #t (if data 1 0)))

(define (cap-style-list? data)
  (and (pair? data) (code-cap-style (car data)) (null? (cdr data))))

(define (property-cap-style-set obj prop data)
  (or (cap-style-list? data)
      (error "Bad property value definition"))
  (prop-uns-set-forte obj prop CC_PT_UNS CC_PTU_CAP_STYLE #t (code-cap-style (car data))))



(register-prop-definition 'natural property-natural-set)
(register-prop-definition 'rgb property-rgb-set)
(register-prop-definition 'rgba property-rgba-set)
(register-prop-definition 'font property-font-set)
(register-prop-definition 'conic-type-0 property-conic-type-0-set)
(register-prop-definition 'conic-type-1 property-conic-type-1-set)
(register-prop-definition 'conic-type-2 property-conic-type-2-set)
(register-prop-definition 'conic-type-3 property-conic-type-3-set)
(register-prop-definition 'cap-style property-cap-style-set)
(register-prop-definition 'transform property-transform-set)


(define (decode-prop-natural n)
  (list 'natural n))

(define (decode-prop-color n)
  (if (= (unpack-a n) 255)
      (list 'rgb (unpack-r n) (unpack-g n) (unpack-b n))
      (list 'rgba (unpack-r n) (unpack-g n) (unpack-b n) (unpack-a n))))

(define (decode-prop-font n)
  (list 'font (font-file-name n)))

(define (decode-prop-boolean n)
  (not (zero? n)))

(define (decode-prop-cap-style n)
  (list 'cap-style (decode-cap-style n)))

(define (decode-prop-conic-0 n)
  (list 'conic-type-0 (decode-conic n)))

(define (decode-prop-conic-1 n)
  (list 'conic-type-1 (decode-conic n)))

(define (decode-prop-conic-2 n)
  (list 'conic-type-2 (decode-conic n)))

(define (decode-prop-conic-3 n)
  (list 'conic-type-3 (decode-conic n)))

(register-prop-subtype 'natural  CC_PT_UNS CC_PTU_UNSPECIFIED decode-prop-natural)
(register-prop-subtype 'color    CC_PT_UNS CC_PTU_COLOR decode-prop-color)
(register-prop-subtype 'font     CC_PT_UNS CC_PTU_FONT decode-prop-font)
(register-prop-subtype 'conic-type-0 CC_PT_UNS CC_PTU_CONIC_TYPE_0P decode-prop-conic-0)
(register-prop-subtype 'conic-type-1 CC_PT_UNS CC_PTU_CONIC_TYPE_1P decode-prop-conic-1)
(register-prop-subtype 'conic-type-2 CC_PT_UNS CC_PTU_CONIC_TYPE_2P decode-prop-conic-2)
(register-prop-subtype 'conic-type-3 CC_PT_UNS CC_PTU_CONIC_TYPE_3P decode-prop-conic-3)
(register-prop-subtype 'boolean  CC_PT_UNS CC_PTU_BOOLEAN decode-prop-boolean)
(register-prop-subtype 'cap-style CC_PT_UNS CC_PTU_CAP_STYLE decode-prop-cap-style)

(register-prop-subtype 'string   CC_PT_STRING CC_PTS_UNSPECIFIED id)
(register-prop-subtype 'filename CC_PT_STRING CC_PTS_FILE_NAME id)
(register-prop-subtype 'text     CC_PT_STRING CC_PTS_LARGE_TEXT id)

(register-prop-subtype 'transform CC_PT_POINTER CC_PTP_TRANSFORM decode-prop-transform)

(register-prop-subtype 'real CC_PT_REAL CC_PTR_UNSPECIFIED id)
(register-prop-subtype 'coordinate CC_PT_REAL CC_PTR_COORDINATE id)
(register-prop-subtype 'nonnegative CC_PT_REAL CC_PTR_NON_NEGATIVE id)
(register-prop-subtype 'angle-1pi CC_PT_REAL CC_PTR_ANGLE_PI id)
(register-prop-subtype 'angle-2pi CC_PT_REAL CC_PTR_ANGLE_2PI id)
(register-prop-subtype 'angle-4pi CC_PT_REAL CC_PTR_ANGLE_4PI id)
(register-prop-subtype 'reference CC_PT_REAL CC_PTR_REFERENCE id)


;;; Common user functions


(define (property-get-subtype obj name)
  (let ((x (property-get-subtype-i obj name)))
    (and x (property-type-symbol (car x) (cdr x)))))

(define (property-get obj name . missing)
  (let ((st (property-get-int obj name)))
    (if st
	(prop-decode (cadr st) (caddr st) (car st))
	(if (null? missing) #f (car missing)))))

(define (property-set! obj name value)
  (cond
   ((real? value)
    (property-real-set obj name value))
   ((string? value)
    (property-string-set obj name value))
   ((boolean? value)
    (property-boolean-set obj name value))
   ((pair? value)
    (property-specific-set obj name (car value) (cdr value)))
   (else
    (typearg-error "property-set!" 3 value))))

(define (skip-head f)
  (if (pair? f)
      (cdr f)
      f))

(define (go-get-sources go)
  (let ((v (hashq-ref src-table (proxy-get-subtype go))))
    (and v (flatten (map (lambda (x) ; in v should be only numbers, getters or property names
			   (cond
			    ((number? x)
			     (if (negative? x)
				 (go-get-anchor go (- -1 x))
				 (go-get-anchor go x)))
			    ((symbol? x) (skip-head (property-get go x)))
			    ((procedure? x) (x go))
			    (else (error "BUG in go-get-sources"))))
			 (if (number? v) (iota v) v))))))

(define (go-get-source-dsc subtype)
  (let ((v (hashq-ref src-table subtype)))
    (and v (map (lambda (x) ; in v should be only numbers, getters or property names
		  (if (number? x)
		      (if (negative? x) 'hanger-c 'hanger)
		      'data))
		(if (number? v) (iota v) v)))))

(define coords mc-new)
(define (coords-c p) (coords (real-part p) (imag-part p)))

(define make-point (prepare go-point-new '() '() ))
(define make-segment (prepare go-segment-new '() '()))

(define make-ellipse-by-center-2-radii-rotation
  (lambda (c r1 r2 rot) (autolink (go-elarc-new-xyr c rot r1 r2 0 7 (code-conic 'start-entire)))))

(define make-ellipse-by-2-foci-point (prepare go-elarc-new-foci '() (list (code-conic 'point-entire))))
(define make-ellipse-by-3-points-smallest (prepare go-elarc-new-3small '() (list (code-conic 'point-entire))))
(define make-ellipse-by-3-points-rotation-eccentricity (prepare go-elarc-new-3ecc '() (list (code-conic 'point-entire))))
(define make-ellipse-by-center-point-rotation-eccentricity (prepare go-elarc-new-xy1ecc '() (list (code-conic 'point-entire))))

(define make-ellipse/c2rr make-ellipse-by-center-2-radii-rotation)
(define make-ellipse/2fp make-ellipse-by-2-foci-point)
(define make-ellipse/3ps make-ellipse-by-3-points-smallest)
(define make-ellipse/3pre make-ellipse-by-3-points-rotation-eccentricity)
(define make-ellipse/cpre make-ellipse-by-center-point-rotation-eccentricity)
(define make-ellipse make-ellipse-by-center-2-radii-rotation)

(define make-elarc-by-center-2-radii-rotation-2-angles
  (lambda (c r1 r2 rot a1 a2 int) (autolink (go-elarc-new-xyr c rot r1 r2 a1 a2 (code-conic int)))))

(define make-elarc-by-3-points-smallest
  (lambda (p1 p2 p3 int) (autolink (go-elarc-new-foci p1 p2 p3 (code-conic int)))))

(define make-elarc/c2rr2a make-elarc-by-center-2-radii-rotation-2-angles)
(define make-elarc/3ps make-elarc-by-3-points-smallest)

(define make-quadratic-bezier
  (lambda (p1 p2 p3) (autolink (go-bezier-new-quadratic p1 1 p2 1 p3 1))))
(define make-quadratic-rational-bezier (prepare go-bezier-new-quadratic '() '()))
(define make-cubic-bezier
  (lambda (p1 p2 p3 p4) (autolink (go-bezier-new-cubic p1 1 p2 1 p3 1 p4 1))))
(define make-cubic-rational-bezier (prepare go-bezier-new-cubic '() '()))
(define make-bezier/2 make-quadratic-bezier)
(define make-bezier/2r make-quadratic-rational-bezier)
(define make-bezier/3 make-cubic-bezier)
(define make-bezier/3r make-cubic-rational-bezier)

(define (make-tex-text h str)
  (autolink (go-tex-text-new h 0 0 0 0 str)))

(define (make-text h str fontfile fontsize)
  (autolink (go-text-new h 0 0 0 0 str (font-load fontfile) fontsize)))

					;  (lambda (p rx1 ry mx my str) (autolink (go-text-new 0 p1 1 p2 1 p3 1)))

(define make-parametric-point (prepare go-parametric-point-new '() '()))
(define make-parametric-point-on-hanger (prepare go-parametric-point-new-on-hanger '() '(#f 0 0)))

(define make-intersection (prepare go-intersection-new '() '()))
(define make-intersection-on-hanger (prepare go-intersection-new-on-hanger '() '(#f 0 0)))

(define make-decorator-point (prepare go-decorator-point-new '() '()))

(define make-group (prepare go-group-new '() '()))

(define-constructor 'point make-point)
(define-constructor 'segment make-segment)
					;(define-constructor 'circle-by-center-radius make-circle-by-center-radius)
					;(define-constructor 'circle-by-center-point make-circle-by-center-point)
					;(define-constructor 'circle-by-3-points make-circle-by-3-points)

					;(define-constructor 'arc-by-center-radius-2-angles make-arc-by-center-radius-2-angles)
					;(define-constructor 'arc-by-center-point-angle make-arc-by-center-point-angle)
					;(define-constructor 'arc-by-3-points make-arc-by-3-points)

; (define-constructor 'ellipse-by-center-2-radii-rotation make-ellipse-by-center-2-radii-rotation)
(define-constructor 'ellipse-by-center-2-radii-rotation-2-angles make-elarc-by-center-2-radii-rotation-2-angles)
(define-constructor 'ellipse-by-2-foci-point make-ellipse-by-2-foci-point)
(define-constructor 'ellipse-by-3-points-smallest make-ellipse-by-3-points-smallest)
(define-constructor 'ellipse-by-3-points-rotation-eccentricity make-ellipse-by-3-points-rotation-eccentricity)
(define-constructor 'ellipse-by-center-point-rotation-eccentricity make-ellipse-by-center-point-rotation-eccentricity)

(define-constructor 'quadratic-bezier make-quadratic-rational-bezier)
(define-constructor 'cubic-bezier make-cubic-rational-bezier)

(define-constructor 'text make-text)
(define-constructor 'tex-text make-tex-text)

(define-constructor 'parametric-point make-parametric-point-on-hanger)
(define-constructor 'intersection make-intersection-on-hanger)
(define-constructor 'decorator-point make-decorator-point)

(define-constructor 'group make-group)

(define hget-center (prepare-hget 'center))
(define hget-endpoint-1 (prepare-hget 'endpoint-1))
(define hget-endpoint-2 (prepare-hget 'endpoint-2))

(define-hcount 'point 1)
(define-hcount 'segment 3)
(define-hcount 'elliptic-arc 4)
(define-hcount 'bezier 3)
(define-hcount 'parametric-point 1)
(define-hcount 'intersection 1)
(define-hcount 'decorator 0)

(define-h-all 'mouse-click 'coords)

(define-hslot 'point 'center 0)
(define-hslot 'segment 'endpoint-1 1)
(define-hslot 'segment 'endpoint-2 2)
(define-hslot 'elliptic-arc 'endpoint-1 1)
(define-hslot 'elliptic-arc 'endpoint-2 2)
(define-hslot 'elliptic-arc 'center 3)
(define-hslot 'bezier 'endpoint-1 1)
(define-hslot 'bezier 'endpoint-2 2)
(define-hslot 'parametric-point 'center 0)
(define-hslot 'intersection 'center 0)

(define center (prepare-aget 'center))
(define endpoint-1 (prepare-aget 'endpoint-1))
(define endpoint-2 (prepare-aget 'endpoint-2))
(define controlpoint (prepare-aget 'controlpoint))
(define controlpoint-1 (prepare-aget 'controlpoint-1))
(define controlpoint-2 (prepare-aget 'controlpoint-2))
(define controlpoint-3 (prepare-aget 'controlpoint-3))
(define focus-1 (prepare-aget 'focus-1))
(define focus-2 (prepare-aget 'focus-2))

(define (aget-center go) (anchor-get-hanger (center go)))
(define (aget-endpoint-1 go) (anchor-get-hanger (endpoint-1 go)))
(define (aget-endpoint-2 go) (anchor-get-hanger (endpoint-2 go)))
(define (aget-controlpoint go) (anchor-get-hanger (controlpoint go)))
(define (aget-controlpoint-1 go) (anchor-get-hanger (controlpoint-1 go)))
(define (aget-controlpoint-2 go) (anchor-get-hanger (controlpoint-2 go)))
(define (aget-controlpoint-2 go) (anchor-get-hanger (controlpoint-3 go)))
(define (aget-focus-1 go) (anchor-get-hanger (focus-1 go)))
(define (aget-focus-2 go) (anchor-get-hanger (focus-2 go)))

(define (set-center go h) (anchor-rehang (center go) h))
(define (set-endpoint-1 go h) (anchor-rehang (endpoint-1 go) h))
(define (set-endpoint-2 go h) (anchor-rehang (endpoint-2 go) h))
(define (set-controlpoint go h) (anchor-rehang (controlpoint go) h))
(define (set-controlpoint-1 go h) (anchor-rehang (controlpoint-1 go) h))
(define (set-controlpoint-2 go h) (anchor-rehang (controlpoint-2 go) h))
(define (set-controlpoint-3 go h) (anchor-rehang (controlpoint-3 go) h))
(define (set-focus-1 go h) (anchor-rehang (focus-1 go) h))
(define (set-focus-2 go h) (anchor-rehang (focus-2 go) h))

(define-aslots 'point 'center)
(define-aslots 'segment 'endpoint-1 'endpoint-2)
					;(define-aslots 'circle-by-center-radius 'center)
					;(define-aslots 'circle-by-center-point 'center 'controlpoint)
					;(define-aslots 'circle-by-3-points 'controlpoint-1 'controlpoint-2 'controlpoint-3)

					;(define-aslots 'arc-by-center-radius 'center)
					;(define-aslots 'arc-by-center-point 'center 'conrolpoint)
					;(define-aslots 'arc-by-3-points 'endpoint-1 'controlpoint 'endpoint-2)

(define-aslots 'ellipse-by-center-2-radii-rotation-2-angles 'center)
(define-aslots 'ellipse-by-2-foci-point 'focus-1 'focus-2 'controlpoint)
(define-aslots 'ellipse-by-3-points-smallest 'controlpoint-1 'controlpoint-2 'controlpoint-3)
(define-aslots 'ellipse-by-3-points-rotation-eccentricity 'controlpoint-1 'controlpoint-2 'controlpoint-3)
(define-aslots 'ellipse-by-center-point-rotation-eccentricity 'center 'controlpoint)



(define-aslots 'quadratic-bezier 'endpoint-1 'controlpoint 'endpoint-2)
(define-aslots 'cubic-bezier 'endpoint-1 'controlpoint-1 'controlpoint-2 'endpoint-2)

(define-aslots 'text 'controlpoint)
(define-aslots 'tex-text 'controlpoint)

(define-aslots 'parametric-point 'curve)
(define-aslots 'intersection 'curve-1 'curve-2)
(define-aslots 'decorator-point 'decorator-point)

(define-aslots 'group)

(define (go-text-get-transform-11 g) (go-text-get-transform g 0 0))
(define (go-text-get-transform-12 g) (go-text-get-transform g 0 1))
(define (go-text-get-transform-21 g) (go-text-get-transform g 1 0))
(define (go-text-get-transform-22 g) (go-text-get-transform g 1 1))

(define (go-tex-text-get-transform-11 g) (go-tex-text-get-transform g 0 0))
(define (go-tex-text-get-transform-12 g) (go-tex-text-get-transform g 0 1))
(define (go-tex-text-get-transform-21 g) (go-tex-text-get-transform g 1 0))
(define (go-tex-text-get-transform-22 g) (go-tex-text-get-transform g 1 1))

					;(define (go-arc-get-angle-length a)
					;  (- (go-arc-get-angle-end a) (go-arc-get-angle-start a)))

(define-source 'point 1)
(define-source 'segment 2)

; (define-source 'ellipse-by-center-2-radii-rotation (list 0 'a-radius 'b-radius 'rotation))
(define-source 'ellipse-by-center-2-radii-rotation-2-angles (list 0 'a-radius 'b-radius 'rotation 'start 'dif 'conic))
(define-source 'ellipse-by-2-foci-point 3)
(define-source 'ellipse-by-3-points-smallest 3)
(define-source 'ellipse-by-3-points-rotation-eccentricity (list 0 1 2 'rotation 'ecc))
(define-source 'ellipse-by-center-point-rotation-eccentricity (list 0 1 'rotation 'ecc))

(define-source 'quadratic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2))
(define-source 'cubic-bezier (list 0 'weight-0 1 'weight-1 2 'weight-2 3 'weight-3))

(define-source 'text (list 0 'text 'font 'size))
(define-source 'tex-text (list 0 'text))

(define-source 'parametric-point (list -1 'parameter))
(define-source 'intersection (list -1 -2 'parameter))
(define-source 'decorator-point (list 0 'vertices 'radius 'rotation))

(define-source 'group '())


;;; Save


;; FIXME - check whether id property is different for every go
;; in one tlo and be string or undefined

(define (direct-write? go)
  (not (or (proxy? go 'go 'group 'top-level-group)
	   (go-has-dependent go))))

(define (save-internal doc port)
  (define (write-space x) (write-char #\space port) (write x port))
  (define (write-open-list) (write-char #\space port) (write-char #\( port))
  (define (write-close-list) (write-char #\) port))

  (define (write-properties obj next)
    (write-open-list)
    (for-each (lambda (x) (write-space x)) next)
    (property-for-each obj (lambda (name value type subtype flags)
			     (if (flag? flags CC_PTF_SAVE)
				 (write-space (list name (prop-decode type subtype value))))))
					;				 (if (or (flag? flags CC_PTF_VIRTUAL)
					;					 (= subtype 0))
					;			     (write-space (list name value (property-type-symbol type subtype)))))))
    (write-close-list))

  (define (save-hanger h)
    (cond
     ((and (hanger-is-mc h) (hanger-is-position h))
      (write-open-list)
      (write 'coords port)
      (write-space (hanger-get-pos-x h))
      (write-space (hanger-get-pos-y h))
      (write-close-list))
     ((and (hanger-is-mc h) (hanger-is-curve h))
      (write-space #f))
     ((hanger-is-curve h)
      (link-go (hanger-get-go h)))
     (else
      (write-open-list)
      (let ((tp (hanger-get-class-name h)))
	(write tp port)
	(link-go (hanger-get-go h)))
      (write-close-list))))

  (define (link-go go)
    (if (direct-write? go)
	(write-go go #f)
	(let ((id (property-get go 'id)))
	  (if id
	      (write-space id)
	      (write-space (proxy2ptr go))))))

  (define (save-go go)
    (if (not (direct-write? go))
	(write-go go #t)))

  (define (write-go go want-id)
    (newline port)
    (write-open-list)
    (write (proxy-get-subtype go) port)

    (if (not (proxy? go 'go 'group))
	(for-each (lambda (src)
		    (cond
		     ((proxy? src 'anchor)
		      (save-hanger (anchor-get-hanger src)))
					;		 ((proxy? src 'go)
					;		  (link-go src))
		     (else
		      (write-space src))))
		  (go-get-sources go)))

    (write-properties go (if (and want-id
				  (not (proxy? go 'go 'group 'top-level-group))
				  (not (property-get go 'id)))
			     (list (list 'id (proxy2ptr go)))
			     '()))

    (if (proxy? go 'go 'group)
	(begin
	  (write-open-list)
	  (for-each link-go (group-get-children go))
	  (write-close-list)))

    (write-close-list))


  (define (save-tlo-content tlo)
    (trans-main tlo (lambda ()
		      (tsort-tlo tlo save-go)
		      (save-go (tlo-get-tlg tlo)))))

  ;; body of save-internal
  (let rec ((obj doc))
    (if obj
	(begin
	  (newline port)
	  (write-open-list)
	  (write (proxy-get-type obj) port)
	  (if (eq? obj doc)
	      (write-space 0.6))
	  (write-properties obj '())
	  (if (not (eq? obj obj-universum))
	      (begin
		(write-open-list)
		(if (proxy? obj 'obj 'tlo)
		    (save-tlo-content obj)
		    (rec (obj-get-first-son obj)))
		(write-close-list)))
	  (write-close-list)
	  (if (not (eq? obj doc))
	      (rec (obj-get-next-son obj)))))))

(define (save-doc doc filename)
  (trans-main obj-universum
	      (lambda ()
		(call-with-output-file filename
		  (lambda (port) (save-internal doc port))))))


;;; Tokenize routines for load (warning: extreme perversion)


(define (convert-to-tokens struct)
  (cond
   ((null? struct)
    (list (cons 'open-list #f) (cons 'close-list #f)))
   ((list? struct)
    (append! (list (cons 'open-list (source-properties struct)))
	     (apply append! (map convert-to-tokens struct))
	     (list (cons 'close-list #f))))
   ((pair? struct)
    (parse-error-int "Invalid token - improper list: ~s" #f (list (last-pair struct))))
   ((eof-object? struct)
    (list (cons 'end-of-file #f)))
   (else
    (list (cons 'data struct)))))

(define (make-token-stream port)
  (vector port '() #f))

(define (fill-token-buffer! stream)
  (if (null? (vector-ref stream 1))
      (vector-set! stream 1 (convert-to-tokens (read (vector-ref stream 0))))))

(define (read-next-token stream)
  (fill-token-buffer! stream)
  (let ((old-data (vector-ref stream 1)))
    (vector-set! stream 1 (cdr old-data))
    (if (and (eq? (caar old-data) 'open-list) (cdar old-data))
	(vector-set! stream 2 (cdar old-data)))
    (car old-data)))

(define (look-next-token stream)
  (fill-token-buffer! stream)
  (car (vector-ref stream 1)))

(define (get-last-source-properties stream) (vector-ref stream 2))


;;; Load

(read-enable 'positions)

(define (parse-error-int string sp args)
  (if sp
      (scm-error 'parse-error #f (string-append "Parse error after ~s:~s:~s - " string)
		 (cons* (assq-ref sp 'filename) (assq-ref sp 'line)
			(assq-ref sp 'column) args) #f)
      (scm-error 'parse-error #f (string-append "Parse error - " string) args #f)))


(define (load-internal token-stream)
  (define (parse-error string . args)
    (parse-error-int  string (get-last-source-properties token-stream) args))

  (define (want-specific type pred dsc)
    (let ((data (read-next-token token-stream)))
      (if (and (eq? (car data) type) (pred (cdr data)))
	  (if (eq? (car data) 'data)
	      (cdr data)
	      *unspecified*)
	  (parse-error "invalid token ~s (expected ~s)"
		       (if (eq? (car data) 'data)
			   (cdr data)
			   (car data)) dsc))))

  (define (look-specific type pred)
    (let ((data (look-next-token token-stream)))
      (and (eq? (car data) type) (pred (cdr data)))))

  (define (want-data) (want-specific 'data true 'data))
  (define (want-open-list) (want-specific 'open-list true 'open-list))
  (define (want-close-list) (want-specific 'close-list true 'close-list))
  (define (want-symbol) (want-specific 'data symbol? 'symbol))
  (define (want-number) (want-specific 'data number? 'number))

  (define (look-open-list) (look-specific 'open-list true))
  (define (look-close-list) (look-specific 'close-list true))

  (define (want-sexp)
    (if (look-open-list)
	(begin
	  (want-open-list)
	  (do ((acc '() (cons (want-sexp) acc)))
	      ((look-close-list)
	       (want-close-list)
	       (reverse acc))))
	(want-data)))

  (define (go-get-hanger-by-name-or-fail a b)
    (or (go-get-hanger-by-name a b)
	(parse-error-int "unsupported go hanger '~s' at '~s'" #f (list b (proxy-get-subtype a)))))

  (define (go-get-hanger-curve-or-fail a)
    (or (go-get-hanger-curve a)
	(parse-error-int "unsupported go curve hanger at '~s'" #f (list (proxy-get-subtype a)))))

  (define (read-properties obj)
    (define id #f)
    (want-open-list)
    (do () ((look-close-list))
      (want-open-list)
      (let* ((prop-name (want-symbol))
	     (prop-value (want-sexp)))
	(cond
	 ((not (eq? prop-name 'id))
	  (property-set! obj prop-name prop-value))
	 ((natural? prop-value)
	  (set! id prop-value))
	 ((string? prop-value)
	  (set! id prop-value)
	  (property-set! obj 'id prop-value))
	 (else
	  (parse-error "invalid object identifier definition"))))
      (want-close-list))
    (want-close-list)
    id)

  (define (read-tlo-content tlo)
    (define ht (make-hash-table 255)) ; larger value?

    (define (table-lookup id)
      (define item (hash-ref ht id))
      (if (proxy? item) item #f))

    (define (table-register-todo id todo)
      (hash-set! ht id (cons todo (hash-ref ht id '()))))

					;    (define (table-register-patch id h h-type)
					;    (hash-set! ht id (cons (cons h hname) (hash-ref ht id '()))))

    (define (table-register-go id go)
      (let ((patches (hash-ref ht id '())))
	(if (not (list? patches)) (parse-error "more objects with one ID: ~S ~S" patches go))
	(for-each (lambda (todo) (todo go)) patches)
	(hash-set! ht id go)))

    (define (read-go-link none)
      (if (look-open-list)
	  (read-go)
	  (let ((id (want-data)))
	    (if (not (or (string? id) (number? id) (and none (not id))))
		(parse-error "invalid referenced ID: ~S" id))
	    id)))

    (define (know-link go)
      (if (proxy? go) go (table-lookup go)))

    (define (make-rehang-todo fake-hanger hanger-type)
      (lambda (go)
	(anchor-rehang (hanger-get-first-anchor fake-hanger)
		       (go-get-hanger-by-name-or-fail go hanger-type))))

    (define (register-rehang id fake-hanger hanger-type)
      (table-register-todo id (make-rehang-todo fake-hanger hanger-type))
      fake-hanger)

    ;; read-hanger-inside doesn't return hanger, but promise of hanger
    (define (read-pos-hanger-in)
      (let ((hanger-type (want-symbol)))
	(if (eq? hanger-type 'coords)
	    (let* ((x (want-number)) (y (want-number))) (coords x y))
	    (let ((link (read-go-link #t)))
	      (delay
		(let ((go (know-link link)))
		  (if go
		      (go-get-hanger-by-name-or-fail go hanger-type)
		      (register-rehang link (coords 0 0) hanger-type))))))))

    (define (read-pos-hanger)
      (want-open-list)
      (let ((retval (read-pos-hanger-in)))
	(want-close-list)
	retval))


    (define (read-cur-hanger)
      (let ((link (read-go-link #t)))
	(if link
	    (delay
	      (let ((go (know-link link)))
		(if go
		    (go-get-hanger-curve-or-fail go)
		    (register-rehang link (hanger-new-mc-curve) 'curv))))
	    (hanger-new-mc-curve))))

    ;; warning - #t (not #f) in after slot to link at front.
    (define (make-link-package grp what after)
      (lambda (input-what input-after)
	(if input-what  (set! what  input-what))
	(if input-after (set! after input-after))
	(if (and what after)
	    (go-link what grp (if (boolean? after) #f after)))))

    (define (make-what-todo pkg)
      (lambda (go) (pkg go #f)))

    (define (make-after-todo pkg)
      (lambda (go) (pkg #f go)))

    (define (read-go-args type)
      (map safe-force ;; read-hanger return promise of hanger
	   (map (lambda (x)
		  (case x
		    ((hanger) (read-pos-hanger))
		    ((hanger-c) (read-cur-hanger))
		    ((data) (want-data))
		    ;; ((go) (read-hanger)) ((number) (want-number)) ((string) (want-string))
		    (else (error "BUG in read-go-header"))))
		(or (go-get-source-dsc type)
		    (parse-error "unsupported go subtype (~s)" type)))))

    (define (read-group-content-raw)
      (want-open-list)
      (do ((acc '() (cons (read-go-link #f) acc)))
	  ((look-close-list)
	   (want-close-list)
	   (reverse acc))))

    (define (read-group-content grp)
      (let ((list (read-group-content-raw)))
	(if (not (null? list))
	    (let rec ((act (car list))
		      (old #t)
		      (known-act (know-link (car list)))
		      (known-old #t)
		      (rest (cdr list)))
	      (let ((pkg (make-link-package grp known-act known-old)))
		(pkg #f #f) ; in case that everything is ok
		(if (not known-act) (table-register-todo act (make-what-todo pkg)))
		(if (not known-old) (table-register-todo old (make-after-todo pkg))))
	      (if (not (null? rest))
		  (rec (car rest) act (know-link (car rest)) known-act (cdr rest)))))))

    (define (read-go-header type)
      (if (eq? type 'top-level-group)
	  (tlo-get-tlg (get-trans-tlo))
	  (make-go type (read-go-args type))))

    (define (read-go)
      (want-open-list)
      (let* ((go (read-go-header (want-symbol)))
	     (id (read-properties go)))
	(if id (table-register-go id go))
	(if (proxy? go 'go 'group)
	    (read-group-content go))
	(want-close-list)
	go))

    ;; body of read-tlo-content
    (want-open-list)
    (do () ((look-close-list))
      (read-go))
    (want-close-list)
    tlo)

  (define (read-obj-children obj)
    (want-open-list)
    (do () ((look-close-list))
      (obj-link (read-obj) obj))
    (want-close-list))

  (define (read-tlo-rest)
    (let ((tlo (tlo-new)))
      (read-properties tlo)
      (trans-main tlo
		  (lambda ()
		    (with-fluids ((autolink-state #f))
				(read-tlo-content tlo))))
      (want-close-list)
      tlo))

  (define (read-doc-rest)
    (let ((doc (doc-new)))
      (let ((version (want-number)))
	(if (not (eqv? version 0.6))
	    (parse-error "unsupported file version ~s" version)))
      (read-properties doc)
      (read-obj-children doc)
      (want-close-list)
      doc))

  (define (read-univ-rest)
    (let ((version (want-number)))
      (if (not (eqv? version 0.6))
	  (parse-error "unsupported file version ~s" version)))
    (read-properties obj-universum)
    (want-close-list)
    #f)

  (define (read-obj)
    (want-open-list)
    (let ((type (want-symbol)))
      (case type
	((tlo) (read-tlo-rest))
	((document) (read-doc-rest))
	((universum) (read-univ-rest))
	(else (parse-error "unsupported obj type (~s)" type)))))

  ;; body of load-internal
  (let ((doc (read-obj)))
    (want-specific 'end-of-file true 'end-of-file)
    (if doc (obj-link doc obj-universum))
    doc))

(define (load-doc filename)
  (trans-main obj-universum
	      (lambda ()
		(call-with-input-file filename
		  (lambda (port) (load-internal (make-token-stream port)))))
	      (lambda x  (apply throw x))
	      "Document load"))

;;; Trash

(define (make-doc name)
  (trans-main obj-universum
	      (lambda ()
		(let ((doc (doc-new-named name)))
		  (obj-link doc obj-universum)
		  doc))))

(define (make-tlo doc name)
  (trans-main obj-universum
	      (lambda ()
		(let ((tlo (tlo-new-named name)))
		  (obj-link tlo doc)
		  tlo))
	      (lambda x x)
	      "Making tlo"))

(define pi (angle -1))

(define (cochline p1 p2 dpt)
  (if (= dpt 0)
      (make-segment (coords-c p1) (coords-c p2))
      (let* ((delta (/ (- p2 p1) 3))
	     (pn1 (+ p1 delta))
	     (pn2 (+ pn1 (make-polar (magnitude delta) (+ (angle delta) (/ pi 3)))))
	     (pn3 (+ pn1 delta))
	     (nd (- dpt 1)))
	(cochline p1 pn1 nd)
	(cochline pn1 pn2 nd)
	(cochline pn2 pn3 nd)
	(cochline pn3 p2 nd))))

(define (coch x1 y1 x2 y2 x3 y3 dpt)
  (cochline (make-rectangular x1 y1) (make-rectangular x2 y2) dpt)
  (cochline (make-rectangular x2 y2) (make-rectangular x3 y3) dpt)
  (cochline (make-rectangular x3 y3) (make-rectangular x1 y1) dpt))

(define (avg a b) (/ (+ a b) 2))
(define (ravg a b) (let ((p (+ 0.25 (/ (random 20) 40)))) (+ (* p a) (* (- 1 p) b))))

(define (tree x y d d1 d2 l b i)
  (define (treeline p d l i)
    (make-segment (coords-c p) (coords-c (+ p (make-polar l d))))
    (if (> i 0)
	(let ((i2 (- i 1)))
	  (treeline (+ p (make-polar (* (+ 0.6 (/ (random 10) 50)) l) d))
		    (ravg d d1) (* l (+ (/ (random 10) 30) 0.5)) i2)
	  (treeline (+ p (make-polar (* (+ 0.6 (/ (random 10) 50)) l) d))
		    (ravg d d2) (* l (+ (/ (random 10) 30) 0.5)) i2))))
  (for-each (lambda (none)
	      (treeline (make-rectangular x y) d l i))
	    (iota b)))

					; example (trans-main (lookup "Page 1") (lambda () (tree-final 0 0 20 6 2)))
(define (tree-final x y length iterations density)
  (tree x y (/ pi 2) 0 pi length density iterations))

					;(trans-main tlo (lambda () (coch -2 -2 0 2 2 -2 3)))
					;
					;(trans-main tlo (lambda () (let ((circ (go_arc_new_3points 0 "0x8260778" "0x825f5a0" "0x825b8f0"))) (go-link circ))))

					;(trans-main (lookup "Page 1") (lambda () (coch -40 -40 0 40 40 -40 4)))
					;(tst-quadratic -2 -2 1 0 2 1  2 -1 1)

(define (go-remove-all tlo)
  (trans-main tlo
	      (lambda () (tlo-remove-relink-selected-go))
	      (lambda () ())
	      "Unlinking all go"))

(define (go-remove go)
  (trans-main (go-get-tlo go)
	      (lambda () (go-remove-relink go))
	      (lambda () ())
	      "Unlinking go"))

(define (make-mc tlo x y)
  (trans-main tlo
	      (lambda () (mc-new x y))
	      (lambda () ())
	      "Making mouse click"))

(define (rnd) (- (/ (random 8000) 100)  40))

(define (click x y)
  (trans-main (lookup "Page 1") (lambda () (mc-new x y))))

(define (make func . args)
  (trans-main (lookup "Page 1") (lambda ()
				  (go-link (apply func args)))))

(define (vlocka)
  (trans-main (lookup "Page 1") (lambda () (coch -20 -20 0 20 50 -20 3))))

(define (clear-undo)
  (define (cu obj)
    (if (proxy? obj (string->symbol "obj") (string->symbol "tlo")) (trans-clear-undo obj 0))
    (if (obj-recurse-next obj-universum obj) (cu (obj-recurse-next obj-universum obj))))
  (trans-clear-undo obj-universum 0)
  (cu obj-universum))

(define (clipboard-paste tlo) (trans-main tlo (lambda () (tlo-clipboard-paste))))

(define (clipboard-cut tlo) (trans-main tlo (lambda () (tlo-clipboard-cut))))

(define (copy-paste-test)
  (Kn2 (lookup "Page 1") 5)
  (trans-main (lookup "Page 1") (lambda ()
				  (go-select (lookup "segment-0-1"))
				  (go-select (lookup "segment-1-2"))
				  (go-select (lookup "segment-2-3"))
				  (go-select (lookup "segment-3-4"))
				  ))
  (clipboard-copy (lookup "Page 1"))
  (clipboard-paste (lookup "Page 2"))
  )

(define (cut-paste-test)
  (Kn2 (lookup "Page 1") 5)
  (trans-main (lookup "Page 1") (lambda ()
				  (go-select (lookup "segment-0-1"))
				  (go-select (lookup "segment-1-2"))
				  (go-select (lookup "segment-2-3"))
				  (go-select (lookup "segment-3-4"))
				  ))
  (clipboard-cut (lookup "Page 1"))
  (clipboard-paste (lookup "Page 2"))
  )

(define (lookup name) (o-named-lookup name))

; LocalWords:  functions
