#include "kernel/internal.h"
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <pwd.h>
#include <guile/gh.h>
#include "font/font.h"
#include "kernel/plugin.h"
#include "lib/cache.h"

string NAMES_STRINGS,
	string_stroke_cap_type[PSC_MAX],
	string_go_type[GOT_MAX], string_obj_type[OT_MAX],
	string_prop_type[PT_MAX],
	string_quantity_type[QT_MAX],
	string_conic_type[GEOM_CONIC_ARC_MAX];

static struct obj_tlo tlo_univ;
struct obj *obj_universum = &tlo_univ.obj;
struct obj_tlo *tlo_universum = &tlo_univ;

static struct obj_tlo tlo_zomb;
struct obj *obj_zombie = &tlo_zomb.obj;
struct obj_tlo *tlo_zombie = &tlo_zomb;

static struct obj obj_tem;
struct obj *obj_temp = &obj_tem;
struct obj_tlo *tlo_temp;

struct obj_tlo *tlo_clipboard;

struct trans *trans_present;
struct obj_tlo *trans_tlo;
struct trans* trans_present;

clist obj_lsel;
uns obj_sel_count[OT_MAX];

struct obj_info obj_info[OT_MAX];

#define INFO(_id) [GO_TYPE(_id)] = GLUEE(GO_NAME(_id), _info),
struct go_info *go_info[GOT_MAX] = { GO_FOR_ALL_ID(INFO) };

uns vrr_with_gui;


static void default_settings(void)
{
	struct o * ou = obj2o(obj_universum);
	char buf[TRANS_ERR_SIZE];

	TRANS_BEGIN_MAIN(tlo_universum, "Set default settings", buf )
		prop_create(ou, string_maxhist, PT_UNS, PTU_UNSPECIFIED, PROP_UNIT_MAX, (prop_value){ .u = MAX_UNDO_HISTORY_VOLUME_KB });
	TRANS_FAILED
		BUG();
	TRANS_END

}

static void load_settings(void)
{
	struct stat st;
	struct passwd * pasw = getpwuid( getuid() );
	if (!pasw)
	{ err("Cannot getpwuid:%s\n", strerror(errno)); return; }

	char filename[strlen(pasw->pw_dir) + strlen(VRR_DIR GLOBAL_SETTINGS_FILE) + 1];
	strcpy( filename, pasw->pw_dir );
        strcat( filename, VRR_DIR GLOBAL_SETTINGS_FILE );

	if (stat( filename, &st ))
	{
		if ( errno != ENOENT ) err("Cannot stat file %s: %s\n", filename, strerror(errno));
		return;
	}

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Load settings", buf )
		settings_load(filename);
	TRANS_FAILED
		err("Cannot read settings file %s: %s\n", filename, buf);
	TRANS_END
}

static void save_settings(void)
{
	struct stat st;
	struct passwd * pasw = getpwuid( getuid() );
	if (!pasw)
	{ err("Cannot getpwuid:%s\n", strerror(errno)); return; }

	char filename[strlen(pasw->pw_dir) + strlen(VRR_DIR GLOBAL_SETTINGS_FILE) + 1];
	strcpy( filename, pasw->pw_dir );
	strcat( filename, VRR_DIR );

	int res = stat(filename, &st);
	if (res && errno == ENOENT)
	{
		if (mkdir(filename, S_IRWXU))
		{ err("Cannot create dir %s: %s\n", filename, strerror(errno)); return; }
	} else {
		if (res == 0 && !S_ISDIR(st.st_mode))
		{ err("%s is not a directory\n", filename); return; }
	}
	strcat( filename, GLOBAL_SETTINGS_FILE );

	char buf[TRANS_ERR_SIZE];
	TRANS_BEGIN_MAIN( tlo_universum, "Save settings", buf )
		settings_save( filename );
	TRANS_FAILED
		err("Cannot write settings file %s: %s\n", filename, buf);
	TRANS_END
}

void kernel_init(int argc UNUSED, char **argv UNUSED, uns gui)
{
	char buf[TRANS_ERR_SIZE];
	int opt;
	vrr_with_gui = gui;

	smalloc_init();

	cache_init(&global_cache);	
	geomlib_init();

	string_init();
	property_init();
	object_init();
	hook_init();

#define INIT(_id) GLUEE(GO_NAME(_id), _init) ();
	GO_FOR_ALL_ID(INIT)

	clipboard_init();

	font_init(argv[0]);
	plugin_init();

	gl_first_init();
   glt_kernel_init();
   gl_trans_init();
   gl_misc_init();
   gl_last_init();

	default_settings();
	load_settings();

	while((opt = getopt(argc, argv, "s:c:h")) != -1)
		switch(opt)
		{
			case 's':
				if(!gh_eval_file(optarg))
					exit(1);
				break;

			case 'c':
				if(!gh_eval_str(optarg))
					exit(1);
				break;

			case '?':
				err("usage: %s [-s script] [-c expresion] [<vrr-file> ...]\n", argv[0]);
				exit(1);

			case 'h':
				err("usage: %s [-s script] [-e expresion] [<vrr-file> ...]\n", argv[0]);
				err("  -s SCRIPT      load Scheme source code from SCRIPT\n"
					 "  -c EXPR        evalute Scheme expression EXPR\n"
					 "  -h             print usage information\n");
				exit(1);

			default:;
				BUG();
		}

	TRANS_BEGIN_ANONYMOUS(buf)
		for(; optind < argc; optind++)
			doc_load(argv[optind]);
	TRANS_FAILED
		err("Error: Couldn\'t open file \'%s\'\n", argv[optind]);
		DBG("%s\n", buf);
		exit(1);
	TRANS_END
}

void kernel_finish(void)
{
	save_settings();
	
	glt_kernel_finish();
	plugin_finish();
	font_finish();

	clipboard_finish();

#define FINISH(_id) GLUEE(GO_NAME(_id), _finish) ();
	GO_FOR_ALL_ID(FINISH)

	hook_finish();
	object_finish();
	property_finish();
	string_finish();

	geomlib_finish();
	cache_destroy(&global_cache);
	
	smalloc_finish();
}

