/** @file
 *
 * 	VRR plugin mechanism
 *
 */

#ifndef _KERNEL_PLUGIN_H_
#define _KERNEL_PLUGIN_H_

/**
 * Flags, which may be returned by plugin_start()
 */
enum plugin_flags {
	PLUGIN_UNLOADABLE = 1,		///< Plugin unloading is possible (default is impossible)
};

#include "kernel/scheme_def.h"

/**
 * Initialize the plugin mechanism
 */
void plugin_init(void);

/**
 * Cleanup the plugin mechanism
 */
void plugin_finish(void);

/**
 * Load given plugin into VRR.
 * Plugin must be shared library satisfying certain conditions,
 * see the definition of valid VRR plugin.
 * Function searches in plugin for function void plugin_start(void *plugin)
 * and executes it if found.
 *
 * @ai filename path to plugin file
 * @r 0 on success, -1 on error
 */
SCHEME SINT plugin_load(SSTRINGC filename);

/**
 * Try to unload given plugin.
 * Not all plugin can be unloaded. Only possible if the PLUGIN_UNLOADABLE flag
 * was returned by the plugin_start() function of the plugin.
 * If unloading is possible, function searches plugin for function void plugin_stop(void)
 * and executes it if found.
 *
 * @ai plugin plugin internal handler
 * @r 0 on success, -1 on error
 */
int plugin_unload(void *plugin);

struct garr;

/**
 * Return the list of all loaded plugins.
 *
 * @ao list growing array holding void* elements, the internal plugin handlers
 * @r number of loaded plugins
 */
int plugin_listing(struct garr *list);

/**
 * Return the list of all registered functions of one given plugin.
 *
 * @ai plugin internal plugin handler
 * @ao list growing array holding struct plugin_function* elements
 * @r number of registered functions
 */
int plugin_function_listing(void *plugin, struct garr *list);

/**
 * Return name of a given plugin.
 *
 * @ai plugin internal plugin handler
 * @r plugin name string pointer
 */
const char *plugin_name(void *plugin);

/**
 * Maximum number of arguments an exported function can have
 */
#define PLUGIN_MAX_ARGS 10

enum plugin_prototypes {
	/* Basic C types */
	PLUGIN_T_VOID,
	PLUGIN_T_INT,
	PLUGIN_T_UNS,
	PLUGIN_T_REAL,
	PLUGIN_T_BOOL,

	/* Special structures */
	PLUGIN_T_CCHAR,
	PLUGIN_T_SYMBOL_C,
	PLUGIN_T_STRING_C,
	PLUGIN_T_SYMBOL_S,
	PLUGIN_T_STRING_S,
	PLUGIN_T_HANGER,
	PLUGIN_T_ANCHOR,

	/* Kernel structures */
	PLUGIN_T_O = 0x10,
	PLUGIN_T_OBJ,
	PLUGIN_T_OBJ_TLO,
	PLUGIN_T_OBJ_DOC,
	PLUGIN_T_GO,
	PLUGIN_T_GO_BEZIER,
	PLUGIN_T_GO_ELARC,
	PLUGIN_T_GO_SEGMENT,
	PLUGIN_T_GO_POINT,
	PLUGIN_T_GO_TEXT,
	PLUGIN_T_GO_TEX_TEXT,
	PLUGIN_T_GO_PARAMETRIC_POINT,
	PLUGIN_T_GO_INTERSECTION,
	PLUGIN_T_GO_GROUP,


	/* Flags to be ORed with the type */
	//	PLUGIN_T_PTR = 0x8000,
};

/**
 * Union describing the type of allowed plugin function arguments and return values.
 */
union plugin_arg {
	int t_int;
	uns t_uns;
	real t_real;
	void *ptr;
	const char *t_cs;
};

#include "lib/clists.h"

/**
 * Structure describing one loaded plugin.
 * Read-only for the rest of the world.
 */
struct plugin_rec {
	cnode n;		///< internal, do not touch
	u32 flags;		///< plugin flags
	void *handle;		///< dynamic library plugin handle, do not touch
	uns func_count;		///< number of registered functions
	clist func_list;	///< linked list of functions
	char name[1];		///< plugin name
};

/**
 * Where the plugin linked list begins.
 * Read-only.
 */
extern clist plugin_list;

/**
 * Number of loaded plugins.
 * Read-only.
 */
extern uns plugin_count;

/**
 * Structure describing one registered plugin function.
 * Read-only for the rest of the world.
 */
struct plugin_function {
	cnode n;				///< internal, do not touch
	void *func;				///< executable function pointer
	uns argc;				///< number of function arguments
	uns retval;				///< return value code
	uns prototype[PLUGIN_MAX_ARGS];		///< argc argument codes
	char name[128];				///< function name
	char comment[1];			///< optional comment
};

/**
 * Plugin can register its functions into scheme interface
 * during startup via this call.
 * The function prototype _must_ be sometype func(union plugin_arg *arg).
 * Number of function arguments is limited by PLUGIN_MAX_ARGS and
 * can only of types constructable with plugin_prototypes.
 *
 * @ai funcname		function name to be registered, must satisfy the conditions above
 * @ai comment		optional comment, may be NULL
 * @ai retval		function return value, described by plugin_prototypes
 * @ai argc		number of arguments function has
 * @ai ...		argc uns arguments describing function prototype, all constructed by plugin_prototypes.
 * @r 0 on success, -1 on error
 */
int plugin_function_register(const char *funcname, const char *comment, uns retval, uns argc, ...);

/**
 * Callback for notifications of plugin changes.
 * Is called every time plugin is loaded/unloaded.
 */
extern void (*plugin_change_callback)(void);

/**
 * Fill buffer with the string description of a given function.
 * Function always returns at least something in the output buffer.
 *
 * @ai f function to describe
 * @ao buf output buffer, should be large enough
 * @ai bufsize buffer size
 * @r buf always
 */
char *plugin_function_prototype(struct plugin_function *f, char *buf, uns bufsize);

SCHEME SVOID plugin_dbg(SSTRINGC plugname);

#endif
