/**
 * @file kernel/tlo_rtree.h
 *
 * VRR kernel - 2D search algorithms.
 */

#ifndef _KERNEL_TLO_RTREE_H_
#define _KERNEL_TLO_RTREE_H_

/* From this point, we need to export functions into Scheme */
#include "kernel/scheme_def.h"


/* Rectangular selection */
static inline int go_rect_enclose(struct go *go, struct geom_rectangle *rect)
{ const struct geom_rectangle *bbox = go_get_expand_bbox(go_get_tlo(go), go); return bbox ? geom_rtree_box_enclose(rect, bbox) : 0; }

#define TLO_RECT_ENCLOSE_QUERY_BEGIN(_name, _tlo, _rect, _go) do{ \
	GEOM_RTREE_RECT_INTERSECT_QUERY_BEGIN(_name, &(_tlo)->rtree, _rect, _name##_obj) { \
	struct go *_go = SKIP_BACK(struct expand_rtree, rtree, _name##_obj)->go; if (go_rect_enclose(_go, (_rect))) {

#define TLO_RECT_ENCLOSE_QUERY_END(_name) \
	}} GEOM_RTREE_RECT_INTERSECT_QUERY_END(_name); }while(0)


/* pass global options */
//#define DIST_PASS_STAT            /**< if defined, additional statistics are computed */
#define DIST_PASS_INITIAL_SIZE 64 /**< initial size of pass buffer */

/* dist_pass.flags possible values */
#define DIST_PASS_MAX_DIST      0x00000001 /**< if set, pass is limited with max distance */
#define DIST_PASS_BOX           0x00000002 /**< if set, pass only gos with bbox in given rectangular area */
#define DIST_PASS_RTREE         0x00000010 /**< if set, rtree is used to speed up pass */

#define DIST_PASS_GOS           0x00000020
#define DIST_PASS_HANGERS       0x00000040
#define DIST_PASS_ANCHORS       0x00000080
#define DIST_PASS_INTERSECTIONS 0x00000100

#define DIST_PASS_FLAGS (DIST_PASS_MAX_DIST | DIST_PASS_BOX | DIST_PASS_RTREE)

/** pass result structure */
struct dist_pass_result
{
	uns type;
	union {
		struct { struct go *go; struct geom_nearest geom; } go;
		struct { real dist; struct hanger *hanger; } hanger;
		struct { real dist; struct anchor *anchor; } anchor;
		struct { real dist; struct go *go1, *go2; struct geom_intersection geom; } intersection;
		struct { real dist; struct geom_point point; } point;
	} data;
};

/** pass types */
enum dist_pass_type
{
	DIST_PASS_CENTER_PASS,
	DIST_PASS_DIRECTION_PASS,
	DIST_PASS_TYPE_COUNT
};

/** pass main structure */
struct dist_pass
{
	enum dist_pass_type type;     /**< pass type */
	uns flags;                    /**< pass options */
	real max_dist;                /**< pass max distance (if DIST_PASS_MAX_DIST set) */
	struct geom_rectangle box;    /**< query rectangular area (if DIST_PASS_BOX set) */
	uns size;                     /**< items buffer and heap size */
	uns count;                    /**< number of items in heap */
	uns used;                     /**< number of valid items in buffer (allocated or in free-list) */
	int free;                     /**< free-list head index (-1 if empty) */

	uns temp_index;               /**< temp item index */
	struct dist_pass_item *temp;

	struct dist_pass_result result;

	struct dist_pass_item *items; /**< items buffer */
	uns *heap;                    /**< items heap */
	struct garr intersections;

	/* additional info */
#ifdef DIST_PASS_STAT
	struct obj_tlo *stat_tlo;     /**< tlo pointer */
	uns stat_box_dist;            /**< number of box distance calls */
	uns stat_point_dist;          /**< number of point distance calls */
	uns stat_segment_dist;        /**< number of segment distance calls */
	uns stat_bezier_dist;         /**< number of bezier distance calls */
	uns stat_returned;            /**< number of returned gos */
#endif
};

/** pass item types */
enum dist_pass_item_type
{
	DIST_PASS_GO,
	DIST_PASS_HANGER,
	DIST_PASS_ANCHOR,
	DIST_PASS_INTERSECTION,
	DIST_PASS_POINT,
	
	DIST_PASS_NODE,
	DIST_PASS_GO_BBOX,
	/*DIST_PASS_BEZIER_BBOX,
	DIST_PASS_BEZIER,*/
	DIST_PASS_ITEM_TYPE_COUNT
};

/** pass item structure */
struct dist_pass_item
{
	enum dist_pass_type type;    /**< item type */
	real dist;
	struct geom_point point;
	uns index;                   /**< index in pass heap */
	union {
		struct { struct go *go; struct geom_nearest geom; } go;
		struct { struct hanger *hanger; } hanger;
		struct { struct anchor *anchor; } anchor;
		struct { struct go *go1, *go2; struct geom_intersection geom; } intersection;

		struct { int next; } free;
		struct { struct obj_tlo *tlo; } tlo;
		struct { struct geom_rtree_node *node; } node;
		/*struct { struct go *go; struct expansion *expansion; } go;
		struct { struct geom_bezier *bezier; uns next; uns go; } bezier;*/
	} data;
};

/** center-pass main structure */
struct center_pass
{
	struct dist_pass pass;       /**< pass structure */
	struct geom_point center;    /**< center point of the pass */
};

/** direction-pass main structure */
/*struct direction_pass
{
	struct dist_pass pass;
	struct geom_vector vector;
};*/


/* pass routines */

void dist_pass_dump(struct dist_pass *pass);

void dist_pass_assert(struct dist_pass *pass);

int dist_pass_next(struct dist_pass *pass, struct dist_pass_result *result);

/* center-pass routines */

#define CENTER_PASS_UNLIMITED_FLAGS (DIST_PASS_RTREE)
#define CENTER_PASS_FLAGS (CENTER_PASS_UNLIMITED_FLAGS | DIST_PASS_MAX_DIST)

static inline void center_pass_dump(struct center_pass *cp)
{ dist_pass_dump(&cp->pass); }

void center_pass_init(struct center_pass *cp, uns flags, struct geom_rectangle *box, struct geom_point *center, real radius, struct obj_tlo *tlo);

void center_pass_destroy(struct center_pass *cp);

static inline int center_pass_next(struct center_pass *cp, struct dist_pass_result *result)
{ return dist_pass_next(&cp->pass, result); }

int center_pass_nearest(uns flags, struct geom_rectangle *box, struct geom_point *center, real radius, struct obj_tlo *tlo, struct dist_pass_result *result);

/* direction-pass routines */

/*#define DIRECTION_PASS_FLAGS (DIST_PASS_RTREE)
#define DIRECTION_PASS_ALL ((1<<GOT_POINT)|(1<<GOT_SEGMENT))

void direction_pass_init(struct direction_pass *dp, uns flags, uns filter, struct geom_rectangle *box, struct geom_vector *vector, struct obj_tlo *tlo);

void direction_pass_destroy(struct direction_pass *dp);

static inline struct go *direction_pass_next(struct direction_pass *dp, struct dist_pass_result *result)
{ return dist_pass_next(&dp->pass, result); }

*/


#endif
