#include "kernel/internal.h"

#ifdef DEBUG_REF_STRING
#define DBG_STRING(...) msg(__VA_ARGS__)
#else
#define DBG_STRING(...) do { } while(0)
#endif


/*
 * Strings are stored in following hash table.
 */

#define HASH_NODE struct string_entry
#define HASH_PREFIX(x) hts_##x
#define HASH_KEY_ENDSTRING text
#define HASH_WANT_CLEANUP
#define HASH_WANT_FIND
#define HASH_WANT_NEW
#define HASH_WANT_REMOVE
#define HASH_GIVE_EQ
static inline uns hts_eq(const char* x, const char*  y) { return !strcmp(x, y); }
#define HASH_GIVE_INIT_KEY
static inline void hts_init_key(string s, char *str) { strcpy(s->text, str); s->length = strlen(str); }
#include "lib/sht.h"


/*
 * A string can has no referencies in transaction. These strings are stored in dead list.
 * String aren't removed from dead list during transaction.
 * In the end of transaction all strings are removed from dead list and strings without
 * referencies are removed from hash table and freed.
 * A string is in dead list iff string_entry::dead_next != NULL.
 */

static struct string_entry dead_list = { &dead_list, 0, 0, {'a'}};

#define REF(_name, _string) string_##_name = string_lookup_ref(_string)

void string_init(void)
{
	hts_init();

	REF(name, "name");
	o_prop_name.p.key = string_name;
	REF(noname, "noname");
	REF(filename, "filename");
	REF(file_unsaved, "unsaved");

	REF(obj_type[OT_UNIVERSUM], "universum");
	REF(obj_type[OT_DOCUMENT], "document");
	REF(obj_type[OT_TLO], "tlo");
	REF(obj_type[OT_ZOMBIE], "zombie");
	REF(obj_type[OT_TEMP], "temp");

	REF(quantity_type[QT_POINT], "point");
	REF(quantity_type[QT_CURVE], "curve");
	
	REF(prop_type[PT_UNS], "uns");
	REF(prop_type[PT_POINTER], "pointer");
	REF(prop_type[PT_STRING], "string");
	REF(prop_type[PT_REAL], "real");

	REF(conic_type[GEOM_CONIC_ARC_START_ENTIRE], "start-entire");
	REF(conic_type[GEOM_CONIC_ARC_START_DIF], "start-dif");
	REF(conic_type[GEOM_CONIC_ARC_POINT_ENTIRE], "point-entire");
	REF(conic_type[GEOM_CONIC_ARC_POINT_DIF], "point-dif");
	REF(conic_type[GEOM_CONIC_ARC_CCW], "ccw");
	REF(conic_type[GEOM_CONIC_ARC_CW], "cw");
	REF(conic_type[GEOM_CONIC_ARC_SMALLER], "smaller");
	REF(conic_type[GEOM_CONIC_ARC_BIGGER], "bigger");
	REF(conic_type[GEOM_CONIC_ARC_MIDDLE], "middle");
	REF(conic_type[GEOM_CONIC_ARC_OPPOSITE], "opposite");
	REF(conic_type[GEOM_CONIC_ARC_ENTIRE], "entire");

	REF(stroke_cap_type[PSC_BUTT], "butt");
	REF(stroke_cap_type[PSC_ROUND], "round");
	REF(stroke_cap_type[PSC_PROJECTING], "projecting");
	
	REF(coordinate_x, "coordinate-x");
	REF(coordinate_y, "coordinate-y");

	REF(fill_color, "fill-color");
	REF(stroke_color, "stroke-color");
	REF(stroke_width, "stroke-width");
	REF(stroke_cap, "stroke-cap");
	REF(text, "text");
	REF(font, "font");
	REF(font_size, "size");
	REF(rpos_x, "relative-position-x");
	REF(rpos_y, "relative-position-y");
	REF(apos_x, "absolute-shift-x");
	REF(apos_y, "absolute-shift-y");
	
	REF(maxhist, "max-undo-history-volume");
}

static void string_array_unref(string *array, int size)
{
	for(int i = 0; i < size; i ++)
		string_unref(array[i]);
}

void string_finish(void)
{
	string_array_unref(string_obj_type, OT_MAX);
	string_array_unref(string_go_type, GOT_MAX);
	string_array_unref(string_conic_type, GEOM_CONIC_ARC_MAX);
	string_array_unref(string_stroke_cap_type, PSC_MAX);
	string_array_unref(string_prop_type, PT_MAX);
	string_array_unref(string_quantity_type, QT_MAX);

	string array[] = { NAMES_STRINGS };
	for(int i = sizeof(array) / sizeof(string) - 1; i >= 0; i--)
		string_unref(array[i]);

#ifdef DEBUG
	HASH_FOR_ALL(hts, walk)
	{
		DANGEROUS(0, "There are %d references to the string %s.\n", walk->ref_count, walk->text);
	}
	HASH_END_FOR;
#endif

	hts_cleanup();
}

/*
 * Function string_trans_end is called in the end of transaction.
 * It cleanups the dead list.
 */
void string_trans_end(void)
{
	string s;
	while(s = dead_list.dead_next, s != &dead_list)
	{
		dead_list.dead_next = s->dead_next;
		if(s->ref_count)
			s->dead_next = NULL;
		else
		{
			DBG_STRING("  Remove string %s at %p\n", s->text, s);
			hts_remove(s);
		}
	}
}

#undef string_ref
string string_ref(string s DEBUG_REF_STRING_ARGS)
{
	s->ref_count++;
	DBG_STRING("  Increase reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

#undef string_unref
string string_unref(string s DEBUG_REF_STRING_ARGS)
{
	ASSERT(s->ref_count > 0);
	s->ref_count--;
	DBG_STRING("  Decrease reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);

	if(s->ref_count > 0 || s->dead_next)
		return s;

	if(trans_present)
	{
		s->dead_next = dead_list.dead_next;
		dead_list.dead_next = s;
		return s;
	}

	DBG_STRING("  Remove string %s at %p\n", s->text, s);
	hts_remove(s);
	return NULL;
}

#undef string_lookup
/** Lookup for given string and create it if not exists. No new reference is increased. Must be called in transaction. */
string string_lookup(const char* str DEBUG_REF_STRING_ARGS)
{
	ASSERT(trans_present);
	string s = hts_find((char*)str);
	if(!s)
	{
		s = hts_new((char*)str);
		s->ref_count = 0;
		s->dead_next = dead_list.dead_next;
		dead_list.dead_next = s;
	}
	DBG_STRING("  String lookup: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

#undef string_lookup_ref
/** Lookup for given string and create it if not exists. Increase one new reference. Calling in transaction isn't requested. */
string string_lookup_ref(const char* str DEBUG_REF_STRING_ARGS)
{
	string s = hts_find((char*)str);
	if(s)
		s->ref_count++;
	else
	{
		s = hts_new((char*)str);
		s->ref_count = 1;
		s->dead_next = NULL;
	}
	DBG_STRING("  Increase reference: string %s at %p, new ref_count %d, file %s, line %d, fce %s\n", s->text, s, s->ref_count, file, line, fce);
	return s;
}

/** Lookup for given string and return NULL if not exists. Calling in transaction isn't requested. */
string string_find(const char* str)
{
	return hts_find((char*)str);
}

