#ifndef _KERNEL_TYPES_H_
#define _KERNEL_TYPES_H_

/*
 * Map of this header file.
 *
 * x Kernel's main structuries
 *    - Main objects: o, obj, go, go_group, go_tlg, obj_tlo, obj_doc
 *    - Objects dependence realization: quantity, hanger, hanger_smart, anchor
 *    - String
 *
 * x OBJ and GO hooks
 *    - Structuries for visualization's hook
 *
 * x Enumerations and other constants
 *    - OBJ and GO types
 *    - OBJ and GO flags
 *    - Dependencies' types
 *    - Type of visualization's hook
 *
 * x Description subsystem
 *
 * x Declaration of global variables
 *
 */



/**********************************
 *                                *
 *   Kernel's main structuries    *
 *                                *
 **********************************/


/*
 * Objects dependence realization - hangers and anchors
 */

typedef union quantity
{
	struct geom_point p;
	real r;
} quantity;

struct hanger
{
	struct hanger_info *info;
	clist lanchor;			// The list of anchors which are hanging on this hanger.
};

struct hanger_smart
{
	struct hanger h;
	struct go *go;			// The hanger's owners.
	cnode nhanger;
	uns ref_count;
	quantity q;
};

struct anchor
{
	struct anchor_info *info;
	cnode nanchor;			// Node for hanger.lanchor.
	struct hanger *h;		// This anchor is hanging on the hanger.
};



/*
 * Main structuries follow in this order:
 * o, obj, go, go_group, go_group_main, obj_tlo, obj_doc
 */

/* First we define the greatgrandfather of all name-space objects */
#define STRUCT_O_MAGIC 		u8 kind; u8 type; u8 subtype; u8 _dummy; uns ref_count; struct slist prop; uns flags
struct o
{
	STRUCT_O_MAGIC;
};

/* Abstract non-graphical object */
struct obj
{
	STRUCT_O_MAGIC;
	struct obj* father;		// predecessor in name-space tree
	clist lson;
	cnode nson;			// brothers in NS tree are linked via this node
	cnode nsel;			// selected objects are linked via this node
};

/* Abstract graphical object */
struct go
{
	STRUCT_O_MAGIC;
#ifdef STUPID_GROUP
	cnode nmember;
	struct go_group *owner;
#else
#endif
	cnode nselected;		// Node for obj_tlo.lselected
	snode nfifo;			// Internal: for topological sort
	uintptr_t temp_data;  // Internal: data for topological sort and clipboard
};

struct go_group
{
	struct go go;
	struct geom_group geom;
#ifdef STUPID_GROUP
	clist lmember;
#else
#endif
	clist lselected;
};

struct go_tlg
{
	struct go_group group;
	struct obj_tlo *tlo;
	clist mc;
};

/* OBJ TLO, top-level object, descendant of OBJ */
struct obj_tlo
{
	struct obj obj;
	struct undo_gui undo_start;				// undo history
	struct undo_gui* undo_present;
	struct undo_gui* undo_saved;		// pointer to current gui command in undo history
	slist trans_hooks, go_hooks;
	struct go_tlg *tlg; 		// top-level group
	struct go_tlg *temp;
	
	struct flist tsort;
	
	struct geom_rtree rtree;	// rtree main data
	uns expand_ref_count;
	struct expand_sht_table *expand;
		
	struct fmalloc_head *fname, *fprop, *fuk;
};

/* OBJ DOCUMENT, descendant of OBJ */
struct obj_doc
{
	struct obj obj;
	struct undo_kernel *saved_start, *saved_end;
	int saved;
};



/**********************************
 *                                *
 *        OBJ and GO hooks        *
 *                                *
 **********************************/


/*
 * Structuries for visualization's hook
 */

struct changed_data_generic
{
	uns kind;
};

struct changed_data
{
	uns kind;
	const struct geom_rectangle *old_bbox;
	const struct geom_rectangle *new_bbox;
};

struct altered_data
{
	uns kind;
	const struct geom_rectangle *where;
};

struct transformed_data
{
	uns kind;
	const struct geom_rectangle *old_bbox;
	const struct geom_rectangle *new_bbox;
	const struct geom_transform2 *transformation;
};


struct obj_hook
{
	snode n;
	void* data;
	void (*obj_link) (struct o *o, struct obj_hook* h);		// A obj has been linked.
	void (*obj_unlink) (struct o *o, struct o *father, struct o *prev, struct obj_hook* h);	// The obj is going to be unlinked.
	void (*obj_selected) (struct o *o, struct obj_hook* h);	// The obj has been selected.
	void (*obj_unselected) (struct o *o, struct obj_hook* h);	// The obj has been unselected.
	void (*obj_prop_create) (struct o *o, struct prop* p, struct obj_hook* h);	// The property of obj was created.
	void (*obj_prop_change) (struct o *o, struct obj_hook* h);	// The property of obj was changed.
	void (*obj_prop_delete) (struct o *o, struct prop* p, struct obj_hook* h);	// The property of obj will be deleted.
};

struct go_hook
{
	snode n;
	void* data;
	void (*go_link) (struct o *o, struct go_hook* h);		// The go has been linked.
	void (*go_unlink) (struct o* o, struct o *father, struct o *prev, struct go_hook* h);		// The go is going to be unlinked.
	void (*go_relink) (struct o *o, struct go_group *old_group, struct go *old_prev, struct go_hook* h);		// The go has been relinked.
	void (*go_selected) (struct o *o, struct go_hook* h);		// The go has been selected.
	void (*go_unselected) (struct o *o, struct go_hook* h);	// The go has been unselected.
	void (*go_prop_create) (struct o *o, struct prop* p, struct go_hook* h);	// The property of go was created.
	void (*go_prop_change) (struct o *o, struct go_hook* h);	// The property of go was changed.
	void (*go_prop_delete) (struct o *o, struct prop* p, struct go_hook* h);	// The property of go will be deleted.
	void (*go_changed) (struct go *go, const struct changed_data_generic *d, struct go_hook* h);	// Visualization's hook: The go has been changed.
};



/*****************************************
 *                                       *
 *   Enumerations and other constants    *
 *                                       *
 *****************************************/

/*
 * OBJ and GO types
 */

enum o_kind
{
	T_OBJ = 17,
	T_GO  = 29,
	T_HANGER = 34,	// Ignore this - not used in o.kind
	T_ANCHOR = 76,	// Ignore this - not used in o.kind
};

enum obj_type
{
	OT_UNIVERSUM,
	OT_DOCUMENT,
	OT_TLO,
	OT_TEMP,
	OT_ZOMBIE,
	OT_MAX,
};

enum go_type
{
	GOT_GROUP,
	GOT_POINT,
	GOT_SEGMENT,
	GOT_BEZIER,
	GOT_ELARC,
	GOT_PARAMETRIC_POINT,
	GOT_INTERSECTION,
	GOT_TEXT,
	GOT_TEX_TEXT,
	GOT_DECORATOR_POINT,
	GOT_DIRECTION,
	GOT_MAX,
};

enum go_group_subtype
{
	GOST_GROUP,
	GOST_GROUP_TLG,
	GOST_GROUP_TEMP,
	GOST_GROUP_ZOMBIE,
	GOST_GROUP_MAX,
};


/*
 * OBJ and GO flags
 */

enum obj_flag
{
	OF_SELECTED = 0x1,
	OF_TSORT_ACTIVE = 0x2, // used only in tlo
	OF_TSORT_DIRTY = 0x4, // used only in tlo
	OF_TSORT_PRESORT = 0x8, // used only in tlo
};

enum go_flag
{
	GOF_SELECTED = 0x1,
	GOF_TSORT = 0x2,
	
	// Internal flags
	GOF_PARPOINT_INVALID_X = 0x40, // used only in parametric point
	GOF_PARPOINT_INVALID_Y = 0x80, // used only in parametric point
	GOF_INTERSECTION_INVALID_X = 0x100, // used only in intersection
	GOF_INTERSECTION_INVALID_Y = 0x200, // used only in intersection
	GOF_ELARC_START = 0x400, // used only in elarc
	GOF_ELARC_DIF = 0x800, // used only in elarc
};



/*
 * Objects dependence realization's constants
 */

enum quantity_type
{
	QT_POINT,
	QT_CURVE,
	QT_MAX,
};

#define HANGER_IS_SMART UINT_MAX

enum linearity_type
{
	LT_NONE,
	LT_SIMILARITY,
	LT_AFFINITY,
};


/*
 * Type of visualization's hook
 */

enum changed_kind
{
	CK_CHANGED,
	CK_ALTERED,
	CK_TRANSFORMED
};



/******************************
 *                            *
 *   Description subsystem    *
 *                            *
 ******************************/


struct obj_info
{
	string name;
};

enum expand_type
{
	ET_PATH,
	ET_NONE,
};

struct go_info
{
	string name;
	uns h_count; // Number of hangers in fixed-size array.
	int h_offset; // Offset of the first hanger and go. Zero iff go hasn't stupid hanger.
	int hs_offset; // Offset of clist (which contains smart hangers) and go. Zero iff go hasn't smart hanger.
	struct hanger_info *h_info; // Information of the first hanger.
	int h_curve; // Index of hanger reprezenting whole curve. -1 if go is not parametrizable.
	int h_start; // Index of hanger reprezenting start point of curve. -1 if go hasn't start.
	int h_end; // Index of hanger reprezenting end point of curve. -1 if go hasn't end.
	int h_center; // Index of hanger reprezenting center of curve. -1 if go hasn't center.
	uns a_count; // Number of anchors in fixed-size array.
	int a_offset; // Offset of first anchor and go.
	struct anchor_info *a_info; // Information of the first anchor.
	int expand;	// Expand type.
	struct prop_virtual *prop_virtual_list;
};

struct hanger_info
{
	string name;
	uns quantity;
	int go_offset; // Offset of hanger and go. Zero iff hanger is smart.
	struct go_info *go_info;
	int q_offset; // Offset of hanger and value of quantity. Zero iff following function must be used instead.
	quantity (*get_quantity) (struct hanger *h); // Also zero iff type of hanger is QT_CURVE.
	uns linear;
};

struct anchor_info
{
	string name;
	uns quantity;
	int offset; // Offset of anchor and go.
	struct go_info *go_info;
	uns linear;
};



/***************************************
 *                                     *
 *   Declaration of global variables   *
 *                                     *
 ***************************************/

extern struct obj *obj_universum; 		// Tree of all objects linked to the universum.
extern struct obj_tlo *tlo_universum; 		// Tree of all objects linked to the universum.
extern struct obj_tlo *tlo_clipboard;

extern struct obj *obj_zombie; 			// Forest of unused objects.
extern struct obj_tlo *tlo_zombie;
extern struct obj *obj_temp;			// Forest of temp (working) objects.

extern clist obj_lsel;				// List of selected objects.
extern uns obj_sel_count[OT_MAX];		// Counts of selected objects per each object type.

extern struct obj_info obj_info[];
extern struct go_info *go_info[], go_group_info[];

extern uns vrr_with_gui;

#endif
