/** \file hashtable.h
  \brief Hash table

This is not a normal header file, it's a generator of hash tables. Each time
you include it with parameters set in the corresponding preprocessor macros, it
generates a hash table with the parameters given.

Must be defined: HT_PREFIX && HT_TYPE && HT_STRUCT && HT_ALLOC ||
	( HT_KEY_UINTPTR || HT_KEY_STRING || (HT_KEY && HT_KEY_POINTER) ||
	(HT_KEY && HT_KEY_NAME && HT_MAGIC && HT_CMP))
Can be defined: HT_ONLY_ONE, HT_IS_FULL, HT_EXPAND, HT_DEFAULT_SIZE, HT_MAGIC, HT_CMP, HT_STATIC, HT_DUMP

	\todo Description hashtable.h
*/

#if !defined(HT_PREFIX) || !defined(HT_TYPE) || !defined(HT_STRUCT) || !defined(HT_ALLOC)
#error HT_PREFIX and HT_TYPE and HT_STRUCT and HT_ALLOC must be defined
#endif

#ifdef HT_KEY_UINTPTR
#define HT_KEY uintptr_t
#define HT_KEY_NAME HT_KEY_UINTPTR

#elif defined(HT_KEY_STRING)
#define HT_KEY const char*
#define HT_KEY_NAME HT_KEY_STRING

#elif defined(HT_KEY_POINTER)
#ifndef HT_KEY
#error HT_KEY must be defined
#endif
#define HT_KEY_NAME HT_KEY_POINTER

#elif defined(HT_KEY) && defined(HT_KEY_NAME)
#if !defined(HT_MAGIC) || !defined(HT_CMP)
#error HT_MAGIC and HT_CMP must be defined
#endif
#else
#error One of { HT_KEY_INT, HT_KEY_STRING, {HT_KEY && HT_NAME} } must be defined
#endif

#ifndef HT_HASH
#define HT_HASH HT_PREFIX(ht)
#endif

#ifndef HT_STATIC
#define HT_STATIC static inline
#endif

#ifndef HT_IS_FULL
#define HT_IS_FULL(_size, _count) ((_count) > (_size) * 2)
#endif

#ifndef HT_EXPAND
#define HT_EXPAND(_size) (((_size) * 4) | 1)
#endif

#ifndef HT_DEFAULT_SIZE
#define HT_DEFAULT_SIZE 5
#endif

typedef struct HT_HASH
{
	uns size, count;
	HT_STRUCT** table;
} HT_HASH;

#ifdef HT_ONLY_ONE

HT_HASH HT_ONLY_ONE;
#define HT_C HT_ONLY_ONE.count
#define HT_S HT_ONLY_ONE.size
#define HT_T HT_ONLY_ONE.table
#define HT_ARG_ONE void
#define HT_ARG_TWO
#define HT_PARAM_ONE
#define HT_PARAM_TWO

#define HT_DO_ALL(_walk) for(HT_STRUCT **_index = HT_T, *_walk = HT_T->next ? HT_T->next : HT_PREFIX(next) (HT_PARAM_TWO &_index); _walk; _walk = _walk->next ? _walk->next : HT_PREFIX(next) (HT_PARAM_TWO &_index))

#else // HT_ONLY_ONE

#define HT_C _ht->count
#define HT_S _ht->size
#define HT_T _ht->table
#define HT_ARG_ONE HT_HASH* _ht
#define HT_ARG_TWO HT_HASH* _ht,
#define HT_PARAM_ONE _ht
#define HT_PARAM_TWO _ht,

#define HT_DO_ALL(ht, _walk) for(HT_STRUCT **_index = (ht)->table, *_walk = *_index ? *_index : HT_PREFIX(next) ((ht), &_index); (_index < (ht)->table + (ht)->size) && _walk; _walk = _walk->next ? _walk->next : HT_PREFIX(next) ((ht), &_index))

#endif // HT_ONLY_ONE

HT_STATIC PURE HT_STRUCT* HT_PREFIX(next) (HT_ARG_TWO HT_STRUCT*** index)
{
	for((*index)++; *index < HT_T + HT_S && (**index) == 0; (*index)++);
	return **index;
}


HT_STATIC void HT_PREFIX(init) (HT_ARG_TWO uns size)
{
	if(size < HT_DEFAULT_SIZE)
		size = HT_DEFAULT_SIZE;
	HT_S = size + 1;
	HT_T = xmalloc(sizeof(HT_STRUCT*) * HT_S);
	for(uns i = 0; i < HT_S; i++)
		HT_T[i] = 0;
	HT_C = 0;
}

HT_STATIC void HT_PREFIX(finish) (HT_ARG_ONE)
{
	if(HT_C)
		for(uns i = 0; i < HT_S; i++)
			while(HT_T[i])
			{
				HT_STRUCT* del = HT_T[i];
				HT_T[i] = HT_T[i]->next;
				xfree(del);
				HT_C--;
			}
	ASSERT(HT_C == 0); // All items must be deleted.
	xfree(HT_T);
}

#ifndef HT_MAGIC

#ifdef HT_KEY_STRING

HT_STATIC PURE uns HT_PREFIX(magic) (HT_ARG_TWO const char* str)
{
	uns key = 0;
	const char* pointer = str;
	for(; *pointer; pointer++)
	{
		key += *pointer;
		key <<= 4;
	}
	return key % HT_S;
}

#define HT_MAGIC(_key) HT_PREFIX(magic) (HT_PARAM_TWO _key)

#elif defined(HT_KEY_UINTPTR)
#define HT_MAGIC(_key) ((_key) % HT_S)
#elif defined(HT_KEY_POINTER)
#define HT_MAGIC(_key) (((uintptr_t)(_key)) % HT_S)
#else
#error HT_MAGIC is not defined
#endif
#endif // HT_MAGIC

#ifndef HT_CMP
#ifdef HT_KEY_STRING
#define HT_CMP(_x, _y) ((_x) == (_y) || !strcmp(_x, _y))
#elif defined(HT_KEY_UINTPTR) || defined(HT_KEY_POINTER)
#define HT_CMP(_x, _y) ((_x) == (_y))
#else
#error HT_CMP must be defined
#endif
#endif // HT_CMP

HT_STATIC PURE HT_STRUCT* HT_PREFIX(pos) (HT_ARG_TWO HT_KEY key, uns pos)
{
	HT_STRUCT *walk = HT_T[pos];
	while(walk)
		if(HT_CMP(walk->HT_KEY_NAME, key))
			return walk;
		else
			walk = walk->next;
	return 0;
}

HT_STATIC PURE HT_STRUCT* HT_PREFIX(get) (HT_ARG_TWO HT_KEY key)
{
	return HT_PREFIX(pos) (HT_PARAM_TWO key, HT_MAGIC(key));
}

HT_STATIC HT_STRUCT* HT_PREFIX(insert) (HT_ARG_TWO HT_KEY key, HT_TYPE type);

HT_STATIC void HT_PREFIX(expand) (HT_ARG_ONE)
{
	uns size_old = HT_S;
	HT_S = HT_EXPAND(HT_S);
	HT_STRUCT **ht_old = HT_T;
	HT_T = xmalloc(sizeof(HT_STRUCT*) * HT_S);
	for(uns i = 0; i < HT_S; i++)
		HT_T[i] = 0;
	HT_STRUCT *walk;
	for(uns i = 0; i < size_old; i++)
	{
		while((walk = ht_old[i]))
		{
			ht_old[i] = ht_old[i]->next;
			uns index = HT_MAGIC(walk->HT_KEY_NAME);
			walk->next = HT_T[index];
			HT_T[index] = walk;
		}
	}
	xfree(ht_old);
}

HT_STATIC HT_STRUCT* HT_PREFIX(find) (HT_ARG_TWO HT_KEY key, HT_TYPE type)
{
	uns index = HT_MAGIC(key);
	HT_STRUCT* walk = HT_PREFIX(pos) (HT_PARAM_TWO key, index);
	if(walk)
		return walk;
	HT_C++;
	if(HT_IS_FULL(HT_S, HT_C))
	{
		HT_PREFIX(expand) (HT_PARAM_ONE);
		index = HT_MAGIC(key);
	}
	HT_ALLOC(walk, key, type);
	walk->next = HT_T[index]->next;
	HT_T[index] = walk;
	return walk;
}

HT_STATIC HT_STRUCT* HT_PREFIX(insert) (HT_ARG_TWO HT_KEY key, HT_TYPE type)
{
	uns index = HT_MAGIC(key);
	HT_STRUCT* walk = HT_PREFIX(pos) (HT_PARAM_TWO key, index);
	ASSERT(walk == 0); // This key may not exist.
	HT_C++;
	if(HT_IS_FULL(HT_S, HT_C))
	{
		HT_PREFIX(expand) (HT_PARAM_ONE);
		index = HT_MAGIC(key);
	}
	HT_ALLOC(walk, key, type);
	walk->next = HT_T[index];
	HT_T[index] = walk;
	return walk;
}

#ifdef HT_OVERWRITE

HT_STATIC HT_STRUCT* HT_PREFIX(overwrite) (HT_ARG_TWO HT_KEY key, HT_TYPE type)
{
	uns index = HT_MAGIC(key);
	HT_STRUCT* walk = HT_PREFIX(pos) (HT_PARAM_TWO key, index);
	if(walk)
	{
		HT_OVERWRITE(walk, key, type);
		return walk;
	}
	HT_C++;
	if(HT_IS_FULL(HT_S, HT_C))
	{
		HT_PREFIX(expand) (HT_PARAM_ONE);
		index = HT_MAGIC(key);
	}
	HT_ALLOC(walk, key, type);
	walk->next = HT_T[index]->next;
	HT_T[index] = walk;
	return walk;
}

#endif // HT_OVERWRITE

HT_STATIC void HT_PREFIX(clean) (HT_ARG_TWO HT_KEY key)
{
	for(HT_STRUCT **prev = &HT_T[HT_MAGIC(key)];
			*prev; prev = &((*prev)->next))
	{
		if(HT_CMP((*prev)->HT_KEY_NAME, key))
		{
			HT_STRUCT* del = *prev;
			*prev = del->next;
			xfree(del);
			HT_C--;
			return;
		}
	}
}

HT_STATIC void HT_PREFIX(delete) (HT_ARG_TWO HT_KEY key)
{
	for(HT_STRUCT **prev = &HT_T[HT_MAGIC(key)];
			*prev; prev = &((*prev)->next))
	{
		if(HT_CMP((*prev)->HT_KEY_NAME, key))
		{
			HT_STRUCT* del = *prev;
			*prev = del->next;
			xfree(del);
			HT_C--;
			return;
		}
	}
	die("Item may exist.");
}

#ifdef HT_DUMP

HT_STATIC void HT_PREFIX(dump) (HT_ARG_ONE)
{
	uns max = 0, i, n, access = 0;
	for(i = 0; i < HT_S; i++)
	{
		n = 0;
		msg("%u shelf:\n", i);
		for(HT_STRUCT* walk = HT_T[i]; walk; walk = walk->next)
		{
			n++;
			HT_DUMP(walk);
			access += n;
		}
		if(max < n)
			max = n;
	}
	max++;
	uns* length = alloca(sizeof(uns) * max);
	for(i = 0; i < max; i++)
		length[i] = 0;
	for(i = 0; i < HT_S; i++)
	{
		n = 0;
		for(HT_STRUCT* walk = HT_T[i]; walk; walk = walk->next)
			n++;
		length[n]++;
	}
	for(i = 0; i < max; i++)
		msg("There are %u shelves with %u items.\n", length[i], i);
	msg("Count of shelves is %d. Count of items is %d. Average access time %f.\n", HT_S, HT_C, ((float)access)/HT_C);
}

#endif // HT_DUMP

#undef HT_DUMP
#undef HT_CMP
#undef HT_MAGIC
#undef HT_NO_TYPE
// #undef HT_DO_ALL
#undef HT_ALLOC
// #undef HT_PREFIX
#undef HT_KEY
#undef HT_KEY_NAME
#undef HT_KEY_UINTPTR
#undef HT_KEY_STRING
#undef HT_KEY_POINTER
#undef HT_TYPE
#undef HT_HASH
#undef HT_ONLY_ONE
#undef HT_S
#undef HT_C
#undef HT_T
#undef HT_ARG_ONE
#undef HT_ARG_TWO
#undef HT_PARAM_ONE
#undef HT_PARAM_TWO
#undef HT_EXPAND
#undef HT_IS_FULL
#undef HT_DEFAULT_SIZE
#undef HT_STATIC
// #undef HT_STRUCT
#undef HT_OVERWRITE

