#include "lib/lib.h"
#include <malloc.h>
#include "lib/smalloc.h"

#ifdef FMALLOC_DUMP
#define DBG_FMALLOC(...) msg(__VA_ARGS__)
#else
#define DBG_FMALLOC(...) do {} while(0)
#endif

struct fmalloc_item
{
	struct fmalloc_item *next;
};

struct fmalloc_node
{
	cnode nnode, nempty;
	struct fmalloc_item *empty;
	struct fmalloc_head *head;
	uns full;
	struct fmalloc_item items[];
};

static inline struct fmalloc_head* fmalloc_get_head(void *p)
	{ return ((struct fmalloc_node*)((uintptr_t)p & (~(FMALLOC_ALIGNMENT-1))))->head; }

static inline struct fmalloc_node* get_first_empty(struct fmalloc_head* f)
{ struct node* n = clist_head(&f->lempty); return n ? SKIP_BACK(struct fmalloc_node, nempty, n) : NULL; }
static inline struct fmalloc_node* get_first_node(struct fmalloc_head* f)
{ struct node* n = clist_head(&f->lnode); return n ? SKIP_BACK(struct fmalloc_node, nnode, n) : NULL; }

void fmalloc_init(struct fmalloc_head* f, uns size)
{
	DBG_FMALLOC("fmalloc_init: head %p, size %d\n", f, size);
	clist_init(&f->lnode);
	clist_init(&f->lempty);
	f->size = size;
	f->call_if_empty = NULL;
}

void fmalloc_finish(struct fmalloc_head* f)
{
	DBG_FMALLOC("fmalloc_finish: head %p\n", f);
	struct fmalloc_node* n;
	while((n = get_first_node(f)))
	{
		clist_remove(&n->nnode);
		free(n);
	}
}

#define NEXT_ITEM(_i) ((struct fmalloc_item*)(((char*)(_i))+f->size))

void* fmalloc(struct fmalloc_head* f)
{
	struct fmalloc_node *n = get_first_empty(f);
	struct fmalloc_item *i;
	if(n)
	{
		i = n->empty;
		n->empty = i->next;
		n->full++;
		if(!n->empty)
			clist_remove(&n->nempty);
		DBG_FMALLOC("fmalloc: head %p, memory %p\n", f, i);
		return i;
	}

	n = memalign(FMALLOC_ALIGNMENT, FMALLOC_ALIGNMENT);
	if(!n)
		die("Can't allocate %d bytes of memory", FMALLOC_ALIGNMENT);
	ASSERT(!((uintptr_t)n & (FMALLOC_ALIGNMENT-1)));

	clist_add_tail(&f->lnode, &n->nnode);
	clist_add_head(&f->lempty, &n->nempty);
	i = NEXT_ITEM(n->items);
	n->empty = i;
	n->full = 1;
	n->head = f;

	for(uns u = (FMALLOC_ALIGNMENT - sizeof(struct fmalloc_node)) / f->size - 2; u; u--)
	{
		i->next = NEXT_ITEM(i);
		i = i->next;
	}
	i->next = NULL;

	DBG_FMALLOC("fmalloc: head %p, memory %p\n", f, n->items);
	return n->items;
}

void ffree(void* p)
{
	struct fmalloc_item* const i = p;
	struct fmalloc_node* const n = (struct fmalloc_node*)((uintptr_t)p & (~(FMALLOC_ALIGNMENT-1)));
	struct fmalloc_head* const f = n->head;
	DBG_FMALLOC("ffree: head %p, memory %p\n", f, p);
	if(n->full == 1)
	{
		clist_remove(&n->nnode);
		clist_remove(&n->nempty);
		free(n);
		if(fmalloc_is_empty(f) && f->call_if_empty)
			f->call_if_empty(f);
		return;
	}
	if(!n->empty)
		clist_add_head(&f->lempty, &n->nempty);
	i->next = n->empty;
	n->empty = i;
	n->full--;
}

struct fmalloc_head smalloc_table[SMALLOC_MAX];

void smalloc_init(void)
{
	for(int i = 1; i < SMALLOC_MAX; i++)
		fmalloc_init(smalloc_table + i, i * sizeof(void*));
}

void smalloc_finish(void)
{
	for(int i = 1; i < SMALLOC_MAX; i++)
		fmalloc_finish(smalloc_table + i);
}

