#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_canvas;

static void
recompute_canvas (struct vcl_canvas *obj)
{
  real cx, cy;

  if (! (obj -> flags & VCL_CANVAS_REALIZED)) return;

  if (obj -> flags & VCL_CANVAS_FULL_WIDTH)
    {
      struct geom_rectangle temp;
      vcl_node_get_bbox (obj -> root, &temp);
      if (geom_rectangle_is_empty (&temp)) return;
      cx = - temp . ll . x;
      cy = - temp . ll . y;
    }
  else
    {
      cx = obj -> win_w * obj -> center_x;
      cy = obj -> win_h * obj -> center_y;
    }

  if ((obj -> unit_scale_x == 0) && (obj -> flags & VCL_CANVAS_REAL_UNITS))
    {
      #ifdef GTK_2_0_OR_OLDER
      	real dpmm_x = 3.7;
	real dpmm_y = 3.7;
      #else
      GdkScreen *s = gtk_widget_get_screen (obj -> gtk -> widget);
      real dpmm_x = ((real) gdk_screen_get_width (s))
	/ ((real) gdk_screen_get_width_mm (s));
      real dpmm_y = ((real) gdk_screen_get_height (s))
	/ ((real) gdk_screen_get_height_mm (s));
      #endif
      obj -> unit_scale_x = obj -> unit_size * dpmm_x;
      obj -> unit_scale_y = obj -> unit_size * dpmm_y;
    }

  struct geom_transform2 tmp;
  geom_transform2_set_identity (&tmp);
  geom_transform2_scale
    (&tmp,
     ((obj -> flags & VCL_CANVAS_FLIP_X) ? -1 : 1) *
     ((obj -> flags & VCL_CANVAS_REAL_UNITS) ? obj -> unit_scale_x : 1),
     ((obj -> flags & VCL_CANVAS_FLIP_Y) ? -1 : 1) *
     ((obj -> flags & VCL_CANVAS_REAL_UNITS) ? obj -> unit_scale_y : 1));
  geom_transform2_translate (&tmp, cx, cy);
  vcl_affinity_set (obj -> root, &tmp);
}


static void
update_canvas (struct vcl_canvas *obj, const struct geom_rectangle *where)
{
  GdkRectangle rect;

  if (! (obj -> flags & VCL_CANVAS_REALIZED)) return;

  if (where != NULL)
    {
      rect . x = where -> ll . x - 2;
      rect . y = where -> ll . y - 2;
      rect . width = where -> ur . x - where -> ll . x + 4;
      rect . height = where -> ur . y - where -> ll . y + 4;
    }
  else
    {
      rect . x = 0;
      rect . y = 0;
      rect . width = obj -> win_w;
      rect . height = obj -> win_h;
    }

  DBG ("INVALIDATE PX: %d %d %d %d\n", rect .x , rect. y, rect.width, rect.height);
  gdk_window_invalidate_rect (obj -> gtk -> widget -> window, &rect, 0);
}


// callbacks

static gboolean
canvas_configure (GtkWidget *w UNUSED, GdkEventConfigure *event, gpointer user_data)
{
  struct vcl_canvas * obj = (struct vcl_canvas *) user_data;
  DBG ("CONFIGURE\n");

  obj -> win_w = event -> width;
  obj -> win_h = event -> height;

  recompute_canvas (obj);
  return TRUE;
}

static gboolean
canvas_expose (GtkWidget *widget UNUSED, GdkEventExpose *event, gpointer user_data)
{
  struct vcl_canvas * obj = (struct vcl_canvas *) user_data;

  DBG ("EXPOSE %d %d %d %d\n", event -> area . x, event -> area . y,
       event -> area . width, event -> area . height);

  struct vcl_rectangle r;
  r . lx = event -> area . x;
  r . ly = event -> area . y;
  r . hx = r . lx + event -> area . width;
  r . hy = r . ly + event -> area . height;

  vcl_painter_draw_node (obj -> painter, obj -> root, &r, NULL);
  return TRUE;
}


static void
canvas_realize (GtkWidget *widget UNUSED, gpointer user_data)
{
  struct vcl_canvas * obj = (struct vcl_canvas *) user_data;
  DBG ("REALIZE\n");

  /* FIXME painter */
  obj -> painter = vcl_painter_plainx_new (obj -> gtk -> widget);
  obj -> flags = obj -> flags | VCL_CANVAS_REALIZED;

  recompute_canvas (obj); /* FIXME : needed here ? */
}


// OBJECT interface

static void /* FIXME call this automatically when destroy-event */
vcl_canvas_destroy (struct vcl_canvas * obj)
{
  vcl_node_set_parent (obj -> root, NULL);
  vcl_object_destroy (obj -> root);

  if (obj -> painter != NULL)
    vcl_object_destroy (obj -> painter);

  gtk_widget_destroy (obj -> gtk -> widget);
  g_object_unref (obj -> gtk -> widget);

  xfree (obj -> gtk);
}


// CONTAINER interface

static void
vcl_canvas_child_altered (struct vcl_canvas * obj, void * child UNUSED,
			  const struct geom_rectangle *where)
{ update_canvas (obj, where); }

static void
vcl_canvas_child_changed (struct vcl_canvas * obj, UNUSED void * child,
			  const struct geom_rectangle *old_bbox,
			  const struct geom_rectangle *new_bbox)
{
  struct geom_rectangle temp;
  geom_rectangle_merge (old_bbox, new_bbox, &temp);
  update_canvas (obj, &temp);
}


// specific functions

void
vcl_canvas_setup (struct vcl_canvas *obj, int flags, real center_x,
		  real center_y, real unit_size)
{
  obj -> flags = ((~ VCL_CANVAS_USER_FLAGS) & (obj -> flags)) |
    (VCL_CANVAS_USER_FLAGS & flags);
  obj -> center_x = center_x;
  obj -> center_y = center_y;
  obj -> unit_size = unit_size;

  recompute_canvas (obj);
}

struct vcl_canvas *
vcl_canvas_new (void)
{
  struct vcl_gtk_canvas *cv = xmalloc (sizeof (struct vcl_gtk_canvas));
  struct vcl_canvas *obj = &(cv -> data);

  vcl_object_init (obj, vcl_vt_canvas);

  obj -> painter = NULL;
  obj -> root = vcl_affinity_new ();
  vcl_node_set_parent (obj -> root, obj);

  cv -> widget = gtk_drawing_area_new ();
  g_object_ref (cv -> widget);
  gtk_signal_connect (GTK_OBJECT (cv -> widget), "realize",
                      GTK_SIGNAL_FUNC (canvas_realize), obj);
  gtk_signal_connect (GTK_OBJECT (cv -> widget), "configure_event",
                      GTK_SIGNAL_FUNC (canvas_configure), obj);
  gtk_signal_connect (GTK_OBJECT (cv -> widget), "expose_event",
                      GTK_SIGNAL_FUNC (canvas_expose), obj);

  obj -> gtk = cv;

  obj -> flags = 0;

  obj -> win_w = 0;
  obj -> win_h = 0;
  obj -> unit_scale_x = 0;
  obj -> unit_scale_y = 0;

  obj -> center_x = 0.5;
  obj -> center_y = 0.5;
  obj -> unit_size = 1;

  return (obj);
}

void
vcl_canvas_main (void)
{
  vcl_vt_canvas = vcl_class_new ("canvas");

  vcl_class_add_interface (vcl_vt_canvas, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_canvas_destroy);

  vcl_class_add_interface (vcl_vt_canvas, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_canvas_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_canvas_child_changed);
}
