#include "vcl_internal.h"

#include <stdio.h>
#include <stdlib.h>

vcl_vtable *vcl_vt_property;

struct control_record
{
  int id;
  void *data;
};


// OBJECT interface

static void
vcl_property_destroy (struct vcl_property * obj)
{
  ASSERT (obj -> parent == NULL);

  if (obj -> child != NULL)
  {
    vcl_node_set_parent (obj -> child, NULL);
    vcl_object_destroy (obj -> child);
  }

  vcl_ga_finish (&(obj -> values));
  vcl_ga_finish (&(obj -> control));
  xfree (obj);
}


// NODE interface

static void
vcl_property_set_data (struct vcl_property * obj, int data)
{ obj -> data = data; }

static int
vcl_property_get_data (struct vcl_property * obj)
{ return (obj -> data); }

static void
vcl_property_set_parent (struct vcl_property * obj, void * parent)
{ obj -> parent = parent; }

static void *
vcl_property_get_parent (struct vcl_property * obj)
{ return (obj -> parent); }

static void
vcl_property_get_bbox (struct vcl_property * obj, struct geom_rectangle * bbox)
{
  if (obj -> child)
    vcl_node_get_bbox (obj -> child, bbox);
  else
    geom_rectangle_set_empty (bbox);
}


// CONTAINER interface

static void
vcl_property_child_altered (struct vcl_property * obj, UNUSED void * child,
			 const struct geom_rectangle *where)
{
  if (obj -> parent)
    vcl_container_child_altered (obj -> parent, obj, where);
}

static void
vcl_property_child_changed (struct vcl_property * obj, UNUSED void * child,
			 const struct geom_rectangle *old_bbox,
			 const struct geom_rectangle *new_bbox)
{
  if (obj -> parent)
    vcl_container_child_changed (obj -> parent, obj, old_bbox, new_bbox);
}


// ENCLOSURE interface

static void *
vcl_property_get_child (struct vcl_property * obj)
{ return (obj -> child); }

static void
vcl_property_set_child (struct vcl_property * obj, void * child)
{
  struct geom_rectangle old_bbox, new_bbox;

  vcl_property_get_bbox (obj, &old_bbox);

  if (obj -> child != NULL)
    vcl_node_set_parent (obj -> child, NULL);
  obj -> child = child;
  vcl_node_set_parent (obj -> child, obj);

  vcl_property_get_bbox (obj, &new_bbox);

  if (obj -> parent)
    vcl_container_child_changed (obj -> parent, obj, &old_bbox, &new_bbox);
}


// specific functions

struct vcl_property *
vcl_property_new (void)
{
  struct vcl_property *obj = xmalloc (sizeof (struct vcl_property));
  vcl_object_init (obj, vcl_vt_property);
  vcl_ga_init (&(obj -> values), 80, 1);
  vcl_ga_init (&(obj -> control), 2, sizeof (struct control_record));
  obj -> child = NULL;
  obj -> parent = NULL;
  obj -> data = 0;

  return (obj);
}

static void
property_add (struct vcl_property * obj, int prop_id, void *prop_data)
{
  void *data = vcl_ga_allocate (&(obj -> values), vcl_property_sizes [prop_id]);
  struct control_record *cr = vcl_ga_allocate (&(obj -> control), 1);
  cr -> id = prop_id;
  cr -> data = data;
  memcpy (data, prop_data, vcl_property_sizes [prop_id]);
}

void
vcl_property_set (struct vcl_property * obj, int prop_id, void *prop_data)
{
  int stop = vcl_property_get_count (obj);
  for (int i = 0; i < stop; i ++)
    {
      if (vcl_property_get_nth_type (obj, i) == prop_id)
	{
	  memcpy (vcl_property_get_nth_value (obj, i), prop_data, vcl_property_sizes [prop_id]);
	  return;
	}
    }
  property_add (obj, prop_id, prop_data);
}

int
vcl_property_get_count (struct vcl_property * obj)
{
  return (obj -> control . used);
}


int
vcl_property_get_nth_type (struct vcl_property * obj, int n)
{
  return (VCL_GA_REF (&(obj -> control), n, struct control_record) . id);
}


void *
vcl_property_get_nth_value (struct vcl_property * obj, int n)
{
  return (VCL_GA_REF (&(obj -> control), n, struct control_record) . data);

}

void *
vcl_property_get (struct vcl_property * obj, int prop_id)
{
  int stop = vcl_property_get_count (obj);
  for (int i = 0; i < stop; i ++)
    {
      if (vcl_property_get_nth_type (obj, i) == prop_id)
	return (vcl_property_get_nth_value (obj, i));
    }
  return (NULL);
}


void
vcl_property_main (void)
{
  vcl_vt_property = vcl_class_new ("property");

  vcl_class_add_interface (vcl_vt_property, vcl_if_object);
  vcl_class_add_method_li (vcl_mid_object_destroy, vcl_property_destroy);

  vcl_class_add_interface (vcl_vt_property, vcl_if_node);
  vcl_class_add_method_li (vcl_mid_node_set_data, vcl_property_set_data);
  vcl_class_add_method_li (vcl_mid_node_get_data, vcl_property_get_data);
  vcl_class_add_method_li (vcl_mid_node_set_parent, vcl_property_set_parent);
  vcl_class_add_method_li (vcl_mid_node_get_parent, vcl_property_get_parent);
  vcl_class_add_method_li (vcl_mid_node_get_bbox, vcl_property_get_bbox);

  vcl_class_add_interface (vcl_vt_property, vcl_if_container);
  vcl_class_add_method_li (vcl_mid_container_child_altered, vcl_property_child_altered);
  vcl_class_add_method_li (vcl_mid_container_child_changed, vcl_property_child_changed);

  vcl_class_add_interface (vcl_vt_property, vcl_if_enclosure);
  vcl_class_add_method_li (vcl_mid_enclosure_set_child, vcl_property_set_child);
  vcl_class_add_method_li (vcl_mid_enclosure_get_child, vcl_property_get_child);
}
