#include "vdl_internal.h"

#include <stdio.h>
#include <stdlib.h>

struct vdl_polygon *
vdl_polygon_new (int size)
{
  MALLOCATED (struct vdl_polygon, pg);

  pg -> edges_used = 0;
  pg -> edges_size = (size < 3) ? 16 : size;
  pg -> edges = xmalloc (pg -> edges_size * sizeof (struct vdl_polygon_edge));
  return (pg);
}

void
vdl_polygon_destroy (struct vdl_polygon *pg)
{
  xfree (pg -> edges);
  xfree (pg);
}

void
vdl_polygon_flush (struct vdl_polygon *pg)
{
  pg -> edges_used = 0;
}

void
vdl_polygon_add_edge (struct vdl_polygon *pg, xxxnum_t x1, xxxnum_t y1, xxxnum_t x2, xxxnum_t y2)
{

  //  printf ("C %f %f %f %f\n", x1, y1, x2, y2);

  if (! IS_ZERO (y1 - y2))
    {
      int i;
      if (pg -> edges_used == pg -> edges_size)
	{
	  pg -> edges_size *= 1.5;
	  pg -> edges = (struct vdl_polygon_edge *) xrealloc (pg -> edges, pg -> edges_size *
							      sizeof (struct vdl_polygon_edge));
	}
      i = pg -> edges_used ++;
      if (y1 < y2) /* perhaps optimalization ? */
	{
	  pg -> edges [i] . x1 = x1;
	  pg -> edges [i] . y1 = y1;
	  pg -> edges [i] . x2 = x2;
	  pg -> edges [i] . y2 = y2;
	  pg -> edges [i] . upline = 0;
	}
      else
	{
	  pg -> edges [i] . x1 = x2;
	  pg -> edges [i] . y1 = y2;
	  pg -> edges [i] . x2 = x1;
	  pg -> edges [i] . y2 = y1;
	  pg -> edges [i] . upline = 1;
	}
    }
}

static inline xxxnum_t
compute_x (struct vdl_polygon_edge *e, xxxnum_t y)
{
  return (e -> x1 + ((y - e -> y1) * ((e -> x2 - e -> x1) / (e -> y2 - e -> y1))));
}

static int
compute_intersection_y (struct vdl_polygon_edge *e1,
			struct vdl_polygon_edge *e2,
			xxxnum_t *y)
{
  struct geom_point p1, p2, p3, p4;
  struct geom_vector v1, v2;
  int i;


  p1 . x = e1 -> x1;
  p2 . x = e1 -> x2;
  p3 . x = e2 -> x1;
  p4 . x = e2 -> x2;
  p1 . y = e1 -> y1;
  p2 . y = e1 -> y2;
  p3 . y = e2 -> y1;
  p4 . y = e2 -> y2;

  geom_vector_init_two_points (&p1, &p2, &v1);
  geom_vector_init_two_points (&p3, &p4, &v2);
  i = geom_vector_intersection (&p1, &v1, &p3, &v2, &p4);

  /*  printf ("CRO %f %f %f %f\n x %f %f %f %f\n x %d %f %f\n",
	  e1 -> x1, e1 -> y1, e1 -> x2, e1 -> y2,
	  e2 -> x1, e2 -> y1, e2 -> x2, e2 -> y2,
 i, p4 . x, p4 . y);
  */
  *y = p4 . y;

  //  *y += 0.001; // Small epsilon ?

  return (! i);
}

#define SIGNUM(x) (IS_ZERO ((x)) ? 0 : (((x) < 0) ? -1 : 1))

static int
compare_edges_by_top_y (const void *a, const void *b)
{
  xxxnum_t diff = ((struct vdl_polygon_edge *) a) -> y1
    - ((struct vdl_polygon_edge *) b) -> y1;
  return (SIGNUM (diff));
}

static int
compare_edges_by_temporary_x (const void *a, const void *b)
{

  xxxnum_t diff = ((struct vdl_polygon_edge *) a) -> tmp_x
    - ((struct vdl_polygon_edge *) b) -> tmp_x;
  return (SIGNUM (diff));
}


void
vdl_polygon_tesselate_to_trapezoids (struct vdl_polygon *pg, struct vdl_traps *tr, int fill_rule)
{
  xxxnum_t sweep, sweep_next, intersection;
  int lbound, ubound, in_out, i;
  int edges_used = pg -> edges_used;
  struct vdl_polygon_edge *edges = pg -> edges;

  qsort (edges, edges_used, sizeof (struct vdl_polygon_edge), compare_edges_by_top_y);

  sweep = edges [0] . y1;
  lbound = ubound = 0;

  while (lbound < edges_used)
    {

      while ((ubound < edges_used) && (edges [ubound] . y1 <= sweep)) ubound ++;
      for (i = lbound; i < ubound; i ++)
	edges [i] . tmp_x = compute_x (&(edges [i]), sweep + 0.1); // Small epsilon ?

      qsort (edges + lbound, ubound - lbound, sizeof (struct vdl_polygon_edge),
	     compare_edges_by_temporary_x);

      sweep_next = edges [lbound] . y2;
      for (i = lbound + 1; i < ubound; i ++)
	{
	  if (edges [i] . y2 < sweep_next)
	    sweep_next = edges [i] . y2;

	  if ((edges [i - 1] . tmp_x != edges [i] . tmp_x) &&
	      compute_intersection_y (&(edges [i - 1]), &(edges [i]), &intersection) &&
	      (intersection > sweep) && (intersection < sweep_next))
	    sweep_next = intersection;
	}

      if ((ubound < edges_used) && (edges [ubound] . y1 < sweep_next))
	sweep_next = edges [ubound] . y1;

      printf ("XX %f %f, %d %d\n", sweep, sweep_next, lbound, ubound);

      in_out = 0;
      for (i = lbound + 1; i < ubound; i ++)
	{
	  if (fill_rule == VDL_FILL_RULE_WINDING)
	    {
	      in_out += (edges [i - 1] . upline ? 1 : -1);
	      if (in_out == 0) continue;
	    }
	  else if (fill_rule == VDL_FILL_RULE_EVEN_ODD)
	    {
	      in_out ++;
	      if ((in_out & 1) == 0) continue;
	    }
	  else BUG ();

	  vdl_traps_add (tr, sweep, sweep_next,
			  edges [i - 1] . x1, edges [i - 1] . y1,
			  edges [i - 1] . x2, edges [i - 1] . y2,
			  edges [i] . x1, edges [i] . y1,
			  edges [i] . x2, edges [i] . y2);
	}

      for (i = lbound; i < ubound; i ++)
	{
	  if (edges [i] . y2 <=  sweep_next)
	    {
	      memmove (edges + lbound + 1, edges + lbound,
		       (i - lbound) * sizeof (struct vdl_polygon_edge));
	      lbound ++;
	    }
	}

      sweep = sweep_next;
    }

  vdl_polygon_flush (pg);
}
